unit UnitDebug;

interface

uses Classes, SysUtils, Generics.Collections;

type TDebugLog = class(TObject)
public
    procedure Append(s : string);

    procedure StartTimer( s: string = '');
    procedure EndTimer( s : string );
    procedure CurrentTimer( s : string );
private
    sl : TStringList;
    enabled : boolean;

    tickList : TList<int64>;

    perFreq : int64;

    constructor Create;
    destructor Destroy; override;
end;

var DebugLog : TDebugLog;


implementation

uses UnitUtil, IOUtils, Windows;

constructor TDebugLog.Create;
begin
    sl := TStringList.Create;

    enabled := TDirectory.Exists( TPath.Combine(UnitUtil.GetINIPathOnly, 'debug') );

    tickList := TList<int64>.create;
    QueryPerformanceFrequency(perFreq);
end;
destructor TDebugLog.Destroy;
begin
    if enabled then begin
        sl.SaveToFile(
            TPath.Combine( UnitUtil.GetINIPathOnly, 'debug\debuglog.txt' ), TEncoding.Unicode
        );
    end;
    tickList.Free;

    inherited;
end;

procedure TDebugLog.Append(s: string);
begin
    if enabled then begin
        sl.Add(DateToStr(now) + ': ' + s);
    end;
end;


procedure TDebugLog.StartTimer;
var pc : int64;
begin
    QueryPerformanceCounter(pc);
    tickList.Add(pc);

    if (s <> '') then Append(s);
end;
procedure TDebugLog.EndTimer( s : string );
begin
    CurrentTimer(s);
    tickList.Delete(ticklist.Count-1);
end;
procedure TDebugLog.CurrentTimer( s : string );
var c, c2, c3 : int64;

begin
    c := tickList.Items[ticklist.Count-1];

    QueryPerformanceCounter(c2);
    c3 := trunc(
        (c2 - c) / perFreq * 1000
    );
    self.Append('ms('+ IntToStr(c3)  +'): ' + s);
end;


initialization
begin
    DebugLog := TDebugLog.Create;
end;
finalization
begin
    DebugLog.free;
end;

end.
