unit UnitFolderMonitor;

interface

uses Windows, Classes, SysUtils, Messages, VCL.Dialogs;

type TFolderMonitorThread = class(TThread)
    private
        h : THandle;
        folder : string;
        fOnChange : TNotifyEvent;

    public
        procedure Execute; override;
end;


type TFolderMonitor = class(TObject)
    private
        folder : string;
        thread : TFolderMonitorThread;
        fOnChange : TNotifyEvent;

    public
        procedure SetFolder(f : string);
        procedure Cancel;
    property
        OnChange :  TNotifyEvent read fOnChange write fOnChange;

end;


var FolderMonitor : TFolderMonitor;

implementation

uses Forms{for Application object};

procedure TFolderMonitor.SetFolder(f : string);
begin
    f := IncludeTrailingBackslash(f);
    if (f <> folder) and (f <> '') then begin
        // kill thread and re-create each time

        if thread <> nil then begin
            thread.Terminate;
            thread.Free;
        end;
        thread := TFolderMonitorThread.Create(True);
        thread.folder := f;
        self.folder := f;
        thread.fOnChange := self.fOnChange;

        thread.Start;
    end;

end;
procedure TFolderMonitor.Cancel;
begin
    if thread <> nil then
        self.thread.Terminate;
end;


procedure TFolderMonitorThread.Execute;
var c : Cardinal;
begin
    h := FindFirstChangeNotification(
        pchar(folder),
        false,
        FILE_NOTIFY_CHANGE_FILE_NAME or FILE_NOTIFY_CHANGE_DIR_NAME
    );

    if (h = INVALID_HANDLE_VALUE) then begin
    end else begin
        try
            repeat
                c := WaitForSingleObject(h, 1000);
            until (c <> WAIT_TIMEOUT) or self.Terminated  or (application=nil) or application.terminated ;
            // this rarely causes an error on exit
            FindCloseChangeNotification(h);
        except

        end;

        if not self.Terminated and (not application.terminated) then begin
            if assigned(fOnChange) then begin
                fOnChange(nil);
            end;
        end;
       // h := 0;
    end;
end;

initialization
begin
    FolderMonitor := TFolderMonitor.Create();
end;

end.
