unit UnitFrameBreadcrumb;

interface


// NOTES: The dropdown is being created an destroyed for every time is it shown.
// I've no idea what this is a workaround for.

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms, 
  Dialogs, StdCtrls, Buttons, ExtCtrls, UnitFrameFolderDropdown, UnitFrameCrumb;

type
  TFrameBreadcrumbs = class(TFrame)
    timHideText: TTimer;
    fpCrumbs: TFlowPanel;

    pnlPath: TPanel;
    txtPath: TEdit;
    pnlDown: TPanel;
    btnDown: TSpeedButton;
    timResize: TTimer;
    timClickDelay: TTimer;
    timHideDropDown: TTimer;


    procedure fpCrumbsClick(Sender: TObject);
    procedure timHideTextTimer(Sender: TObject);
    procedure txtPathMouseEnter(Sender: TObject);
    procedure txtPathExit(Sender: TObject);
    procedure fpCrumbsMouseEnter(Sender: TObject);
    procedure fpCrumbsMouseLeave(Sender: TObject);

    procedure CrumbButtonEnter(Sender:Tobject);
    procedure CrumbButtonLeave(Sender:Tobject);
    procedure CrumbButtonMouseUp(Sender:TObject; Button: TMouseButton; Shift: TShiftState;  X, Y: Integer);
    procedure txtPathKeyPress(Sender: TObject; var Key: Char);
    procedure btnDownClick(Sender: TObject);
    procedure FrameFolderDropdown1Exit(Sender: TObject);
    procedure FrameFolderDropdown1lvFoldersClick(Sender: TObject);
    procedure FrameFolderDropdown1lvFoldersMouseMove(Sender: TObject;
      Shift: TShiftState; X, Y: Integer);
    procedure txtPathMouseLeave(Sender: TObject);
    procedure FrameFolderDropdown1lvFoldersExit(Sender: TObject);
    procedure fpCrumbsResize(Sender: TObject);
    procedure timResizeTimer(Sender: TObject);
    procedure timClickDelayTimer(Sender: TObject);
    procedure timHideDropDownTimer(Sender: TObject);
  private
    fdd : TFrameFolderDropdown;
    fStartingColor : TColor;
    lastEnteredCrumb : TFrameCrumb;
    fOnChange,
    fOnCrumbRightClick : TNotifyEvent;
    fHasSubfolders : boolean;
    fHistory : TStringList;
    fMaxHistory : integer;
    fLastPath : string;
    { Private declarations }

    function getDropDown(cmp : TComponent) : TFrameFolderDropdown;
    procedure PathToCrumbs;
    procedure FolderClick(Sender: Tobject);
    procedure MoreClick(Sender: Tobject);
    procedure DrivesClick(Sender: Tobject);
    procedure SetPath(s : string); overload;
    procedure SetPath(s : string; DetectChangeEvent : boolean); overload;
    function GetPath : string;
    procedure UpdateHeight;
    procedure CMShowingChanged(var M: TMessage); message CM_SHOWINGCHANGED;
    procedure HideDropDown;

  public
    constructor Create(AOwner: TComponent) ; override;
    { Public declarations }
    procedure SetFolder(s : string);
    property Path : string read GetPath write SetPath;
    property History : TStringList read fHistory write fHistory;
    
    property OnChange : TNotifyEvent read fOnChange write fOnChange;

    function IsTextMode : boolean;
    function IsCrumbMode : boolean;
    function HasSubfolders : boolean;
    procedure ReportFocusLost;
  published
    property OnCrumbRightClick : TNotifyEvent read  fOnCrumbRightClick write fOnCrumbRightClick;
  end;

implementation

uses  strutils, UnitUtil, UnitToken, DiskTreeView;

{$R *.dfm}


procedure TFrameBreadcrumbs.ReportFocusLost;
begin
    if (fdd <> nil) then begin
        timHideDropDown.Enabled := true;
    end;
end;

//==========================
// A path is clicked
//==========================
procedure TFrameBreadcrumbs.FolderClick(Sender: Tobject);
var s : string;
begin
    with TFrameCrumb(TControl(sender).parent) do begin
        s := GetFolder;
    end;
    SetPath(s, true);
end;
var DelayPath : string;
procedure TFrameBreadcrumbs.FrameFolderDropdown1lvFoldersClick(Sender: TObject);
var s : string;
begin
    fdd.lvFoldersClick(sender);

    DelayPath := fdd.GetFolder;
    if DelayPath = '' then DelayPath := fLastPath;

    timHideDropDown.Enabled := true;

    //self.SetPath(s, true);
    timClickDelay.Enabled := true;
end;
// workaround for an Access Violation caused by the left mouse button
// and some funky timing
procedure TFrameBreadcrumbs.timClickDelayTimer(Sender: TObject);
begin
    timClickDelay.Enabled := false;
    self.SetPath(DelayPath, true);
end;
procedure TFrameBreadcrumbs.txtPathKeyPress(Sender: TObject; var Key: Char);
begin
    if ord(key) = VK_RETURN then begin
        key := #0;
        timHideText.Enabled := true;
    end;
end;

procedure TFrameBreadcrumbs.timHideTextTimer(Sender: TObject);
begin
    timHideText.Enabled := false;
    if pnlPath.Visible  then begin
        txtPath.text := unitutil.ExpandEnvironment(txtPath.text);
        pnlPath.Visible := false;
        fpCrumbs.Visible := true;
        if LongerDirectoryExists(txtPath.text) then begin
            self.SetPath(txtPath.text, true);
        end else begin
            txtPath.text := self.GetPath;
        end;
    end;
end;







procedure TFrameBreadcrumbs.timResizeTimer(Sender: TObject);
begin
	timResize.Enabled := false;
    self.PathToCrumbs; 
end;
procedure TFrameBreadcrumbs.btnDownClick(Sender: TObject);
var //fdd: TFrameFolderDropdown;
    pt : TPoint;
begin
//    btnDown.Down := false;
    if (fdd <> nil) then begin
        fdd.Free;
        fdd := nil;
    end;
    fdd := getDropDown(fpCrumbs);
    pt := btnDown.ClientToScreen(point(-1* (fdd.Width-btnDown.Width),btnDown.height));
    fdd.DropDownStrings(pt, fHistory);
end;
procedure TFrameBreadcrumbs.CMShowingChanged(var M: TMessage);
begin
   inherited;

   //UnitUtil.ShowMessageI(m.msg);
   if Showing then begin
        fpCrumbs.Refresh;
   end;
end;
constructor TFrameBreadcrumbs.Create(AOwner: TComponent);
begin
    inherited;
    fMaxHistory := 10;
    fHistory := TStringList.Create;
    fHistory.CaseSensitive := false;
    fHistory.Sorted := false;
    //fHistory.Duplicates := dupIgnore;

    fStartingColor := fpCrumbs.Color;

    fpCrumbs.Top := 0;
    pnlPath.top := 0;

    fdd := nil;

    

    //FrameFolderDropdown1.Parent := self;
end;
procedure TFrameBreadcrumbs.CrumbButtonEnter(Sender: Tobject);
begin
    fpCrumbsMouseEnter(sender);
    lastEnteredCrumb := TFrameCrumb(TControl(sender).parent);
    with lastEnteredCrumb do begin
        btnFolder.Flat := false;
        btnMore.Flat := false;
     end;
end;
procedure TFrameBreadcrumbs.CrumbButtonLeave(Sender: Tobject);
begin
    fpCrumbsMouseLeave(sender);

    with TFrameCrumb(TControl(sender).parent) do begin
        btnFolder.Flat := true;
        btnMore.Flat := true;
    end;
end;
procedure TFrameBreadcrumbs.CrumbButtonMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState;  X, Y: Integer);
begin
    if (Button = mbRight) and assigned(fOnCrumbRightClick) then begin
        //fOnCrumbRightClick( TButton(Sender).parent);
        fOnCrumbRightClick( lastEnteredCrumb );
    end;
end;




procedure TFrameBreadcrumbs.DrivesClick(Sender: Tobject);
var s : string;
    pt : TPoint;
    btnDrives : TSpeedButton;
    Crumb : TFrameCrumb;
    sl : TStringList;

    DriveLetter : char;
    DriveType : cardinal;
    root : string;
begin
    btnDrives := TSpeedButton(sender);
    crumb := TFrameCrumb(btnDrives.parent);

    pt := btnDrives.ClientToScreen(point(-25,btnDrives.height));


    if (fdd <> nil) then begin
        fdd.Free;
    end;
    fdd := getDropDown(fpCrumbs);

    sl := TStringList.create;
    for DriveLetter := 'A' to 'Z' do begin
        Root := DriveLetter + ':';

        DriveType := GetDriveType(PChar( Root + '/' )) ;
        if (DriveType <> DRIVE_NO_ROOT_DIR) then begin
            SetErrorMode(SEM_FailCriticalErrors);
            //if DirectoryExists(root+'\') then
            //if DiskSize(Ord(DriveLetter) - Ord('A') + 1) <> -1 then
            //sl.Add(root + '\');
            sl.Values[root] := '-1';
            case DriveType of
                DRIVE_REMOVABLE: begin
                    sl.Values[root] := IntToStr(icon_drive_floppy);
                end;
                DRIVE_CDROM: begin
                    sl.Values[root] := IntToStr(icon_drive_cdrom);
                end;
                DRIVE_FIXED: begin
                    sl.Values[root] := IntToStr(icon_drive_fixed);
                end;
                DRIVE_REMOTE: begin
                    sl.Values[root] := IntToStr(icon_drive_network);
                end;
                DRIVE_NO_ROOT_DIR: begin
                end;
                else begin
                    sl.Values[root] := IntToStr(icon_drive_ghost);
                end;
            end;
        end;
    end;



    fdd.DropDownStrings(pt, sl);
    sl.free;
    self.UpdateHeight;
end;

function TFrameBreadcrumbs.getDropDown(cmp : TComponent) : TFrameFolderDropdown;
begin
    result := TFrameFolderDropdown.Create(cmp);
    result.Parent := self;
    result.Name := '';

    result.OnHide := self.FrameFolderDropdown1Exit;
    result.lvFolders.OnClick := self.FrameFolderDropdown1lvFoldersClick;
end;
procedure TFrameBreadcrumbs.MoreClick(Sender: Tobject);
var s : string;
    pt : TPoint;
    btnMore : TSpeedButton;
    Crumb : TFrameCrumb;

begin
    btnMore := TSpeedButton(sender);
    crumb := TFrameCrumb(btnMore.parent);

    pt := btnMore.ClientToScreen(point(-25,btnMore.height));


    if (fdd <> nil) then begin
        fdd.Free;
        fdd := nil;
    end;
    fdd := getDropDown(fpCrumbs );
    fdd.DropDown(pt, crumb.GetFolder );
    //FrameFolderDropdown1.lvFolders.SetFocus;
    self.UpdateHeight;
end;


procedure TFrameBreadcrumbs.PathToCrumbs;
var i : integer;
    s, temps, share : string;
    sl : TStringList;
    fc : TFrameCrumb;
begin
    for i := fpcrumbs.ControlCount - 1 downto 0 do begin
        fpCrumbs.RemoveControl(fpcrumbs.Controls[i]);
    end;

    s :=  ExcludeTrailingPathDelimiter(self.fLastPath);

    s := StringReplace(s,'\\',#10,[rfreplaceall]);
    s := StringReplace(s,#10, '\\',[rfreplaceall]);

    if AnsiStartsText('\\',s) then begin
        temps := s;
        share := TokenString(temps,'\\', false);
        share := '\\' + TokenString(temps,'\', false);
        share := share + '\' + TokenString(temps,'\',false);
        temps := StringReplace(temps,'\',#9,[rfReplaceAll]);
        if temps <> '' then begin
            s := share + #9 + temps;
        end else begin
            s := share;
        end;
    end else begin
        s := StringReplace(s,'\',#9,[rfReplaceAll]);
    end;


    sl := TStringList.Create;
    sl.StrictDelimiter := true;
    sl.Delimiter := #9;
    sl.DelimitedText := s;
    fc := nil;

    for i := 0 to sl.count - 1 do begin
        fc := TFrameCrumb.Create(self);

        fc.Parent := fpCrumbs;
        fc.Visible := true;
        fc.name := '';
        fc.btnFolder.OnClick := self.FolderClick;
        fc.btnFolder.OnMouseEnter := self.CrumbButtonEnter;
        fc.btnFolder.OnMouseLeave := self.CrumbButtonLeave;
        fc.btnFolder.OnMouseUp := self.CrumbButtonMouseUp;
        fc.btnMore.OnClick := self.MoreClick;
        fc.btnMore.OnMouseEnter := self.CrumbButtonEnter;
        fc.btnMore.OnMouseLeave := self.CrumbButtonLeave;

        fc.btnDrives.Visible := false;
        if i = 0 then begin
            fc.btnDrives.Visible := true;;
            fc.btnDrives.OnClick := self.DrivesClick
        end;

        if i=0 then begin
            fc.Left := 0;
            s := sl[i] + '\';
        end else begin
            s := s + sl[i] + '\';
        end;


        if Pos('&', s) > 0 then begin
            s := StringReplace(s,'&&','?',[rfreplaceall]);
            s := StringReplace(s,'&','&&',[rfreplaceall]);
            s := StringReplace(s,'?','&&',[rfreplaceall]);
        end;
        fc.SetFolder(s, i = (sl.count-1));
    end;

    fHasSubfolders := false;
    if fc.btnMore <> nil then begin
    	fHasSubfolders := fc.btnMore.Visible;
    end;

    if fpCrumbs.Visible and fpCrumbs.Showing  then begin
        while ((fc.Left + fc.Width) > (fpCrumbs.Width-10)) and (fpCrumbs.ControlCount > 2) do begin
           fpCrumbs.RemoveControl(fpCrumbs.Controls[0]);
           fpCrumbs.Invalidate;
        end;



        if (fpCrumbs.ControlCount = 2) and ((fc.Left + fc.Width) > (fpCrumbs.Width-10)) then begin
			TFrameCrumb(fpcrumbs.Controls[0]).shrinkto(40);
            //TFrameCrumb(fpcrumbs.Controls[0]).shrinkto(fpcrumbs.Width-10);
            fpcrumbs.Realign;
			if ((fc.Left + fc.Width) > (fpCrumbs.Width-10)) then begin
                fc.ShrinkTo(fc.Width - ((fc.Left + fc.Width)-fpCrumbs.Width));
            end;

        end;

    end;


    fpCrumbs.Refresh;
end;
procedure TFrameBreadcrumbs.SetFolder(s: string);
begin
    self.SetPath(s);
    fpCrumbs.Repaint;
end;
procedure TFrameBreadcrumbs.SetPath(s: string);
begin
    self.setpath(s, FALSE);

    if txtPath.Visible  then begin
       txtPathExit(txtPath);
    end;
end;
procedure TFrameBreadcrumbs.SetPath(s: string; DetectChangeEvent : boolean);
var changed : boolean;
begin
    s := unitutil.ExpandEnvironment(s);

    changed :=     (fLastPath <> IncludeTrailingPathDelimiter(s));
    txtPath.text := IncludeTrailingPathDelimiter(s);
    fLastPath := IncludeTrailingPathDelimiter(s);
    self.PathToCrumbs;
    if changed then begin
        if fHistory.IndexOf(fLastPath) = -1 then begin
            if fHistory.Count = fMaxHistory then begin
                fHistory.Delete(0);
            end;
            fHistory.Add(fLastPath);
        end;
    end;


    //
    // This is SO BAD
    // - I'm changing the click event to a change event
    //
    if DetectChangeEvent then begin
        if changed and Assigned(OnClick) then begin
            //self.OnChange(self);
            self.OnClick(self);
        end;
    end;
end;

procedure TFrameBreadcrumbs.fpCrumbsClick(Sender: TObject);
begin
    fpCrumbs.Visible := false;
    pnlPath.Visible := true;
    txtPath.Visible := true;
    txtPath.SetFocus;
    txtPath.SelStart := length(txtPath.text);
end;
procedure TFrameBreadcrumbs.fpCrumbsMouseEnter(Sender: TObject);
begin
    fpCrumbs.FlowStyle := fsLeftRightTopBottom;

    fpCrumbs.Color := dimcolor(ColorToRGB(clBtnFace), 1.03);
    fpCrumbs.Realign;

    fpCrumbs.Controls[0].Update;
end;
procedure TFrameBreadcrumbs.fpCrumbsMouseLeave(Sender: TObject);
begin
    fpCrumbs.Color := self.fStartingColor;
end;
procedure TFrameBreadcrumbs.fpCrumbsResize(Sender: TObject);
begin
    timResize.Enabled := true;
end;

//
// DropDown callbacks
//
procedure TFrameBreadcrumbs.timHideDropDownTimer(Sender: TObject);
begin
    timHideDropDown.Enabled := false;
    fdd.hide;
    fdd.Free;
    fdd := nil;
end;

procedure TFrameBreadcrumbs.HideDropDown;
begin
    if assigned(fdd) then begin
        timHideDropDown.Enabled := true;
    end;
end;
procedure TFrameBreadcrumbs.FrameFolderDropdown1Exit(Sender: TObject);
begin
    HideDropDown;
    self.UpdateHeight;
end;
procedure TFrameBreadcrumbs.FrameFolderDropdown1lvFoldersExit(Sender: TObject);
begin
//    HideDropDown;
    self.UpdateHeight;
end;
procedure TFrameBreadcrumbs.FrameFolderDropdown1lvFoldersMouseMove(
  Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  fdd.lvFoldersMouseMove(Sender, Shift, X, Y);

end;

function TFrameBreadcrumbs.GetPath: string;
begin
    //result := txtPath.text;
    result := fLastPath;
end;


function TFrameBreadcrumbs.HasSubfolders: boolean;
begin
    result :=  fHasSubfolders;
end;
function TFrameBreadcrumbs.IsCrumbMode: boolean;
begin
    result := not IsTextMode;
end;
function TFrameBreadcrumbs.IsTextMode: boolean;
begin
    result := pnlPath.Visible;
end;
procedure TFrameBreadcrumbs.txtPathExit(Sender: TObject);
begin
    timHideText.Enabled := true;
end;


procedure TFrameBreadcrumbs.txtPathMouseEnter(Sender: TObject);
begin
    //timHideText.Enabled := false;
end;
procedure TFrameBreadcrumbs.txtPathMouseLeave(Sender: TObject);
begin
    //timHideText.Enabled := true;
    // this is annoying if you try to type and the mouse is not over
    // the text field
end;

procedure TFrameBreadcrumbs.UpdateHeight;
begin
    if (fdd <> nil) and (fdd.Visible) then begin
        self.Height := pnlPath.Height + fdd.Height + 20;
    end else begin
        self.Height := pnlPath.Height;
    end;
end;

end.
