unit UnitFrameFolderDropdown;
///
/// This window has the style of a popup window and will dismiss when
/// focus is lost. An OnHide event is generated here when the show state
/// is altered by the OS.
///
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, ImgList, StdCtrls;





type
  TFrameFolderDropdown = class(TFrame)

    ilFolders: TImageList;
    lvFolders: TListView;
    procedure CreateParams(var Params: TCreateParams);  override;
    procedure lvFoldersMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure lvFoldersClick(Sender: TObject);
    procedure lvFoldersChange(Sender: TObject; Item: TListItem;
      Change: TItemChange);
    procedure lvFoldersExit(Sender: TObject);
    procedure lvFoldersAdvancedCustomDrawItem(Sender: TCustomListView;
      Item: TListItem; State: TCustomDrawState; Stage: TCustomDrawStage;
      var DefaultDraw: Boolean);
    procedure lvFoldersAdvancedCustomDraw(Sender: TCustomListView;
      const ARect: TRect; Stage: TCustomDrawStage; var DefaultDraw: Boolean);

  private
    { Private declarations }
    fClicked : string;
    fPath : string;
//    fOnChange : TNotifyEvent;
    fOnHide : TNotifyEvent;
    procedure CMShowingChanged(var M: TMessage); message CM_SHOWINGCHANGED;

  protected

  public
    //cbxFolders : TMyComboboxEx;
    { Public declarations }
    constructor Create(AOwner: TComponent) ; override;
    procedure DropDownStrings(pt : TPoint; sl : TStringList);
    function DropDown(pt : TPoint; folder : string) : string;
    function GetFolder : string;

    property OnHide : TNotifyEvent read fOnHide write fOnHide;

  end;



implementation

//
//  DIRTY TRICKS
// - no Owner Draw
// - use Advanced Draw to disable the Icon and draw it ourselves
// - the ImageList width is used to space the icon and label
// (but the icon is only 16 x 16)
// - a bevel is drawn in the newly created blank space
// - DoubleBuffer is a must to get rid of flicker

uses Math, CommCtrl, UnitUtil, UnitFrmDummyShowPreview, UnitFrmMain,uxTheme, UnitDebug,
    System.UITypes {so Brush color can be used inline}
;
{$R *.dfm}

{ TFrameFolderDropdown }
constructor TFrameFolderDropdown.Create(AOwner: TComponent);
begin
    inherited;

    lvFolders.Parent := self;
    lvFolders.DoubleBuffered := true;
    ShowScrollBar(lvFolders.Handle, SB_HORZ, false);
    FrmDummyShowPreview.RefreshDriveIcons
end;
procedure TFrameFolderDropdown.CreateParams(var Params: TCreateParams);
begin
    inherited;
    Params.Style :=WS_POPUP;
end;
procedure TFrameFolderDropdown.CMShowingChanged(var M: TMessage);
begin
    // we need to report when the DropDown is hidden by lost focus
    // so the container can update its size smaller

    if showing then begin

    end else begin
//        if assigned(fOnHide) then begin
//            fOnHide(self);
//        end;
    end;
end;


function TFrameFolderDropdown.DropDown(pt : TPoint; folder : string) : string;
var hideonce : boolean;
    rc : TREct;
    procedure Add(str : string);
    var
        li : TListItem;
//        c : Char;

    begin

        li := lvFolders.Items.Add;
        li.ImageIndex := 0;
        li.StateIndex := -1;
        li.Caption := str;

        if not hideonce then begin
            hideonce := true;
            ShowWindow(self.Handle, SW_NORMAL);
            ListView_GetItemRect(lvFolders.Handle,li.Index, rc, LVIR_BOUNDS);
        end;
    end;
var rec: TSearchRec;
    res : integer;
    //{s,} temp : string;
    cnt : integer;
    //h, h2 : THandle;
    //ps : array [0..MAX_PATH] of char;
   // b : Boolean;
    needsClose : boolean;
const CBEM_GETEDITCONTROL = 1031;
var fd : _WIN32_FIND_DATAA;

begin
    self.Top := pt.Y;
    self.Left := pt.X;
    lvFolders.Items.clear;
    lvFolders.Items.BeginUpdate;

    self.Height := 60;
    self.Width := 256;
    lvFolders.Columns[0].Width := self.Width - GetSystemMetrics(SM_CYHSCROLL) - 4;
    self.fpath := folder;
    self.fClicked := '';
    //self.top := 0;
    //self.Left := 0;

    cnt := 0;
    hideonce := false;
    {$i-}
    fd.dwFileAttributes := FILE_ATTRIBUTE_DIRECTORY;
    res := findfirst(ExcludeTrailingPathDelimiter(folder) + '\*.*', faDirectory or faArchive, rec);
    needsClose := (res = 0); // no error code
    {$i+}
    while res = 0 do begin
        if ((rec.Attr and FILE_ATTRIBUTE_DIRECTORY) > 0) and
            (rec.name <> '.') and (rec.name <> '..') then begin
            Add(rec.name);
            inc(cnt);
        end;
        res := findnext(rec);
    end;
    if needsClose then
        FindClose(rec);

    if (cnt > 0) then   begin
        lvFolders.align := alNone;

        //
        // Show 16 items at most, minimum size is 20 pixels
        //
        lvFolders.Height := max(
            ((rc.Bottom-rc.top)  * min(cnt,  16)) + 5,
            20
        );
        self.AutoSize := true;

        lvFolders.align := alClient;
        self.AutoSize := false;
    end;

    if lvFolders.Visible  then begin
        lvFolders.SetFocus;
    end;
    lvFolders.HandleNeeded;
    lvFolders.AlphaSort;
    lvFolders.Items.EndUpdate;
end;

function TextHeight(c : TListView; s : string) : integer;
var bm : TBitmap;
begin
    bm := TBitmap.Create;
    bm.Canvas.Font := c.Font;
    result := bm.Canvas.TextHeight(s);
    bm.free;
end;
function TextWidth(c : TListView; s : string) : integer;
var bm : TBitmap;
begin
    bm := TBitmap.Create;
    bm.Canvas.Font := c.Font;
    result := bm.Canvas.TextWidth(s);
    bm.free;
end;


procedure TFrameFolderDropdown.DropDownStrings(pt: TPoint; sl: TStringList);
var hideonce : boolean;
    rc : TRect;
    procedure Add(str : string; iconindex : integer = -1);
    var //ci : TComboExItem;
        li : TListItem;

    begin

        li := lvFolders.Items.Add;
        li.StateIndex := -1;
        if iconindex = -1 then begin
            li.ImageIndex := 0;
        end else begin
            li.ImageIndex := 14+iconindex;
            li.StateIndex := 11+iconindex;
        end;

        li.Caption := str;


        lvFolders.width := min(
            self.Parent.Width,
             max(lvFolders.width, textwidth(lvFolders, li.caption) + GetSystemMetrics(SM_CYHSCROLL) + ilFolders.Width + 20 )
        );

        lvFolders.Width := max(200, lvFolders.width);

        lvFolders.Columns[0].Width := lvFolders.width - 25;
        if not hideonce then begin
            hideonce := true;
            ShowWindow(self.Handle, SW_NORMAL);
            ListView_GetItemRect(lvFolders.Handle,li.Index, rc, LVIR_BOUNDS);
        end;

    end;
var
    cnt, i : integer;
    x : integer;
    name : string;
begin
    self.Top := pt.Y;
    self.Left := pt.X;
    lvFolders.Items.clear;
    self.Height := 60;
    self.Width := 256;
    lvFolders.Columns[0].Width := self.Width - GetSystemMetrics(SM_CYHSCROLL) - 4;
    self.fpath := '';
    self.fClicked := '';
    hideonce := false;
    //self.top := 0;
    //self.Left := 0;

    cnt := sl.count;
    for i := 0 to cnt - 1 do begin
        name := sl.names[i];
        if name <> '' then begin
            x := safeint(sl.Values[name],0);
            Add(name, x+1);
        end else begin
            add(sl[i]);
        end;
    end;


    if (cnt > 0) then   begin
        lvFolders.align := alNone;

        //
        // Show 16 items at most, minimum size is 20 pixels
        //
        lvFolders.Height := max(
            ((rc.Bottom-rc.top)  * min(cnt,  16)) + 5,
            20
        );
        self.AutoSize := true;

        lvFolders.align := alClient;
        self.AutoSize := false;
        lvFolders.align := alNone;
    end;

    if lvFolders.Visible  then begin
        lvFolders.SetFocus;
        //ShowScrollBar(lvFolders.Handle, SB_HORZ, false);
    end;

    // update position for a new width
    self.Left := pt.X - (lvFolders.Width - 256);
end;
function TFrameFolderDropdown.GetFolder: string;
begin
    result := IncludeTrailingPathDelimiter(fPath) +  fClicked;
    if fpath = '' then begin
        result :=  fClicked;
    end;
end;


procedure TFrameFolderDropdown.lvFoldersChange(Sender: TObject; Item: TListItem;
  Change: TItemChange);
begin
    ShowScrollBar(lvFolders.Handle, SB_HORZ, false);
end;
procedure TFrameFolderDropdown.lvFoldersAdvancedCustomDraw(
  Sender: TCustomListView; const ARect: TRect; Stage: TCustomDrawStage;
  var DefaultDraw: Boolean);
begin
    DefaultDraw := True;
end;
procedure TFrameFolderDropdown.lvFoldersAdvancedCustomDrawItem(
  Sender: TCustomListView; Item: TListItem; State: TCustomDrawState;
  Stage: TCustomDrawStage; var DefaultDraw: Boolean);

    procedure SetBrush(c : TColor);
    begin
        sender.Canvas.Brush.color := 0;
        sender.Canvas.Brush.color := c;
    end;
var r : TRect;
    edge : TRect;
    ic : TIcon;
    b : TBrush;
    c, c2 : TCOlor;
begin
    SendMessage(sender.handle,LVM_SETICONSPACING,0,0);

    defaultdraw := true;

//  with the advanced view, it draws the selected text as a black bar
//    uxTheme.SetWindowTheme(sender.handle, nil, nil);

    b := sender.Canvas.Brush;
    c := b.Color;
    c2 := sender.Canvas.Pen.Color;


    if (stage = cdPrePaint) and (
        (cdsSelected in state) or
        (cdsFocused in state)
    ) then begin
        r := Item.DisplayRect(drBounds);
        edge := r;


        // draw icon
        SetBrush(ColorToRGB(clHighlight));
        sender.Canvas.Pen.Color := clBtnText;

        ic := TIcon.create;
        if item.StateIndex = -1 then begin
            //self.ilFolders.GetIcon(0,ic);
            ilFolders.Draw(sender.Canvas, r.Left+2, r.Top, 0);
            //DrawIconEx(sender.Canvas.handle,r.left+2, r.top, ic.Handle,ic.Width,ic.Height,0,0, DI_NORMAL);
            r := item.DisplayRect(drLabel);
            //windows.SetBkColor(sender.Canvas.Handle, ColorToRGB(clBtnFace));
            //DrawText(sender.Canvas.Handle, item.Caption,length(item.caption), r,0);


            //self.Hint := item.Caption;

        end else begin
            FrmMain.DiskTreeView1.images.GetIcon(item.StateIndex+2, ic);
            DrawIconEx(sender.Canvas.handle,r.left+2, r.top, ic.Handle,ilFolders.Height,ilFolders.Height,0,0, DI_NORMAL);
        end;



        ic.free;
        item.ImageIndex := -1 ;

    end else if (stage = cdPostPaint)then  begin
        r := Item.DisplayRect(drBounds);
        edge := r;

        DefaultDraw := true;
        // draw icon
        SetBrush(ColorToRGB(clBtnFace));
        //sender.Canvas.FillRect(edge);
        ic := TIcon.create;
        if item.ImageIndex = -1 then begin
            self.ilFolders.GetIcon(0,ic);
        end else begin
            FrmDummyShowPreview.ilFolderBrowser.GetIcon(item.ImageIndex, ic);
        end;
        //DrawIconEx(sender.Canvas.handle,0, 0, ic.Handle,ilFOlders.Width,ilFolders.Height,0,0, DI_NORMAL);
        ic.free;
        item.ImageIndex := -1 ;



         // draw edge lines
        SetBrush(ColorToRGB(clBtnFace));
        edge.Right := 16 + 7;

        edge.right := edge.right + 1;
        Windows.DrawEdge(sender.Canvas.Handle,edge,BDR_SUNKENINNER,BF_RIGHT);
        edge.right := edge.right + 1;
        Windows.DrawEdge(sender.Canvas.Handle,edge,BDR_SUNKENOUTER,BF_RIGHT);
        edge.right := edge.right - 2;
    end;

    //sender.Canvas.Brush := b;
    setbrush(c);
    sender.Canvas.Pen.Color := c2;

    windows.SetBkMode(sender.canvas.Handle, TRANSPARENT);
//  works around the BLACK RECTANGE bug
end;

procedure TFrameFolderDropdown.lvFoldersExit(Sender: TObject);
begin
    if (assigned(fOnHide)) then begin
        fOnHide(self);
    end;
    //self.Hide;
end;
procedure TFrameFolderDropdown.lvFoldersClick(Sender: TObject);
var li : TListItem;
begin
    li := lvFolders.Selected;
    if li <> nil then begin
        self.fClicked := li.caption;
    end else begin
        self.fClicked := '';
    end;
end;
procedure TFrameFolderDropdown.lvFoldersMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var li : TlistItem;
begin
    li := lvFolders.GetItemAt(x,y);
    if li <> nil then begin
        li.Selected := true;
    end else begin
        lvFolders.Selected := nil;
    end;
end;

end.
