unit UnitFrmDummyShowPreview;
{
    Purpose:
        the code for showing and previewing the file list is pretty detailed
        and needed to be moved into its own module

        This module is responsible for formatting the filelist, extracting icons,
        retreiving "tag data" for each item (like alternate date information),
        and so on

    Updates:

        - Calls to this form are asynchronis, so, without the blocks some code
        on FrmMain would crash when activated repeatedly. Added a poor-mans
        semaphore to ignore concurrent calls

}


interface

uses  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, StdCtrls, Buttons, ExtCtrls,
  Contnrs {for object list},  Menus, UnitColumnManager,
  DiskTreeView, UnitMetaVariables, UnitUtil, StrUtils, UnitFrmConfig,
  UnitFoldersCount, UnitFileIcon, ImgList, UnitWildCards;


type TTagData = class(TObject)
    IsFolder : boolean;
    changed : boolean;
    Modified, Created : integer; {FileDate}
    IsCollision : boolean;
    OriginalCollision : boolean;
    DirectSelected : boolean;
    IsIllegalNewname : boolean;

    ManualNewName : string;
end;

type TVirtualListItem = record
    index : integer;
    selected : boolean;
    focused : boolean;
end;
CONST FOLDER_TYPE_STRING = '<DIR>';

type TFileListData = class(TObject)
    index : integer;
    caption, altname, basefolder : string;
    data : pointer;
    ImageIndex : integer;

    SubItems : TStrings;
    StateIndex : integer;
    Checked : boolean;
    Selected : boolean;
    AltNameUsed : boolean;
    fcustomdata : TStringList;
public
    constructor Create;
    procedure SetName(s : string);
    procedure SetNewName(s : string);
    procedure SetSize(s : string);
    procedure SetType(s : string; isDir : boolean = false);
    procedure SetDate(s : string);
    procedure SetDateCreated(s : string);
    function GetAltName : string;
    function GetName : string;
    function GetNewName : string;
    function GetNewNameWithDir : string;
    function GetSize : string;
    function GetType : string;
    function GetDate : string;
    function GetDateCreated : string;
    function GetBaseFolder : string;

    function IsFolder : boolean;
published
    property CustomData : TStringList read fcustomdata write fcustomdata;
end;


type
  TFrmDummyShowPreview = class(TForm)
    ILDynamic: TImageList;
    ilFolderBrowser: TImageList;
    ilStateX: TImageList;
    ilTabs: TImageList;
    ilState: TImageList;
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
    wc : TWildCards;
    DoCancelRefreshFileList : boolean;
    tagdata : TObjectList;
    IgnoreInterface : boolean;

    //sl : TStringList;
    //li : TListItem;
    containsMP3s : boolean;
    containsJPEGs : boolean;

    DataList : TObjectList;
    SelectedList : TObjectList;
    MyCheckedCount : integer;
    CollisionCount : integer;
    IllegalCount : integer;
    ColSize : array[0 .. 999] of integer;

    replaceNum : TStringList;
    replaceSeq : integer;

    showPreviewRunning : boolean;

    iconsRefreshed : boolean;

    procedure NewListItem(
        subfolder, filepath : string;
        IsFolder, IsChecked : boolean; rec : TSearchRec);
   //function CompareData(Item1, Item2: Pointer): Integer;
    function TitleCase(ins : string): string;
    function RomanNumberals(ins : string): string;
  public

    { Public declarations }
    procedure RefreshFileList(
        var UseRecursiveOnce : boolean;
        var FilesCount : integer;
        var FoldersCount : integer;
        var CheckedCount : integer;
        Unchecked : boolean = false
    );
    procedure CancelRefreshFileList;
    function GetCancelRefresh : boolean;
    procedure RebuildIcons;
    procedure ShowPreviewList(var ChangedCount : integer);
    function GetCollisionCount : integer;
    function GetIllegalCount : integer;
    procedure AssignIcon(li : TListItem);
    function GetTagData(li : TListItem) : TTagData; overload;
    function GetTagData(index : integer) : TTagData; overload;
    function GetTagDataAbsolute(index : integer) : TTagData; 

    function IsFolder(li : TListItem) : boolean;

    function GetFileData(li : TListItem) : TFileListData; overload;
    function GetFileData(i : integer) : TFileListData; overload;

    function GetDateStringFormatted(td : TTagData) : string; overload;
    function GetDateStringFormatted(i : integer) : string; overload;
    procedure UpdateFileData(li : TListItem);
    function GetColSize(index : integer) : integer;
    // selection routines
    procedure SwapItems(index1, index2 : integer);
    procedure ToggleSelected(index : integer);
    function ToggleChecked(index : integer) : integer;
    function SetChecked(index : integer) : integer; overload;
    function GetChecked(index : integer) : boolean;
    function SetChecked(index : integer; value : boolean) : integer; overload;
    function GetCheckedCount : integer;
    function GetSelected(index : integer) : boolean;
    procedure SetSelected(index : integer); overload;
    procedure SetSelected(index : integer; value :boolean); overload;
    function GetItemFromSelected(selindex: integer):TListItem;
    procedure ClearSelected(index : integer);
    procedure ClearAllSelected(includeHovered : boolean = False);
    function SelectedToggleCheckboxes : integer;
    procedure SelectedInitFromControl;
    procedure SelectedRedraw;
    procedure SelectedRedrawFull;
    function SelectedMyCount : integer;

    //
    // Virtual List routines
    //
    procedure PopulateListItem(li : TListItem); overload;
    procedure PopulateListItem(li : TListItem; index : integer); overload;

    procedure DeleteListItem(index : integer);
    procedure FindListItem(
        Find: TItemFind;
      const FindString: String; const FindPosition: TPoint; FindData: Pointer;
      StartIndex: Integer; Direction: TSearchDirection; Wrap: Boolean;
      var Index: Integer
    );
    procedure StateChangeItems(
        StartIndex,
      EndIndex: Integer; OldState, NewState: TItemStates
    );

    function GetVirtualItem(index : integer) : TVirtualListItem;


    procedure GetRecord(var myrec : RenameRecord; index : integer);

    procedure MoveUp;
    procedure MoveDown;
    procedure MoveTop(TopIndex : integer = 0);
    procedure MoveBottom(BottomIndex :integer = 0);
    procedure AfterMove;
    procedure SortData;

    function ListContainsJPEGs : boolean;
    function ListContainsMP3s : boolean;
    function FindByFilename(filename : string) : integer;
    procedure RefreshDriveIcons;
    function GetReplacementSequence(numbers : string) : string;
  end;

var
  FrmDummyShowPreview: TFrmDummyShowPreview;

implementation

uses UnitFrmMain, UnitFrmAbout, UnitFrmCollisionDialog, UnitToken,
  UnitFrmDummyThumbExtract, CommCtrl, Math, UnitSpecialPaths,
  UnitFrmScriptActions, UnitTextFileFormat, System.Character, UnitFrmExclude, UnitDebug,
  IOUtils, System.Types, Generics.Collections;

{$R *.dfm}

type MyAddTime = procedure(li : TListItem; t : integer);
var AddTime : MyAddTime;
    recursive : boolean;
    comparecount : integer;


CONST REPORT_AT = 1000;
procedure AddTimeFull(li : TListItem; t : integer);
begin

    UnitColumnManager.setDate(li,
     DateToStr(FileDateToDateTime(t)) + ' '
    + FormatDateTime('tt', FileDateToDateTime(t)) );

end;
procedure AddTimeShort(li : TListItem; t : integer);
begin
    UnitColumnManager.setDate(li, DateToStr(FileDateToDateTime(t)) );
end;


procedure TFrmDummyShowPreview.NewListItem(
    subfolder, FilePath : string; IsFolder, IsChecked : boolean; rec : TSearchRec);
var s, ext : string;
    w1, w2 : word;
    td : TTagData;
    //li : TListItem;
//    tempdate : TDateTime;
    i : integer;
    fld : TFileListData;
    systemtime : Windows.TSystemTime;

    function IsZero(ft : TFileTime) : boolean;
    begin
    	result := (ft.dwLowDateTime =0) and (ft.dwHighDateTime=0);
    end;

    function DateTimeToFileTime(FileTime: TDateTime): TFileTime;
    var
		LocalFileTime, Ft: TFileTime;
       	SystemTime: TSystemTime;
    begin
    	Result.dwLowDateTime  := 0;
		Result.dwHighDateTime := 0;
       	DateTimeToSystemTime(FileTime, SystemTime);
       	SystemTimeToFileTime(SystemTime, LocalFileTime);
       	LocalFileTimeToFileTime(LocalFileTime, Ft);
       	Result := Ft;
    end;
    function DateTimeFromFileTime(ft : _FILETIME) : integer;
    begin
        FileTimeToSystemTime(ft, systemtime);
        SystemTimeToTzSpecificLocalTime(nil, systemtime, systemtime);
        SystemTimeToFileTime(systemtime, ft);
        FileTimeToDOSDateTime(ft, w1, w2); //1980 is the lowest year
        result := MakeLong(w2,w1);
    end;
begin

    fld := TFileListData.Create;
    fld.index := DataList.Count;
    td := TTagData.Create;
    td.IsCollision := false;
    td.OriginalCollision := false;

    fld.SetName(subfolder + rec.Name);
    {
    if not FileExists(filepath + subfolder + rec.Name) then begin
        fld.AltNameUsed := true;
        s := fld.caption;
        fld.altname := subfolder + rec.FindData.cAlternateFileName;
    end;
    }

    ColSize[0] := max( GetFileList.Canvas.TextWidth(fld.caption) + 50, ColSize[0]);
    //fld.SubItems[0] := '';


    td.IsFolder := false;
    {$WARN SYMBOL_PLATFORM OFF}  // ignore the FindData warning
    If ((rec.FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) <> 0)  then begin
        //UnitColumnManager.setSize(li, '0');
        //UnitColumnManager.setTypeDir(li);

        //fld.SetSize('0');
        fld.SetType('', true); // := '<DIR>';
        td.IsFolder := true;
    end else begin
        // speed - GetSizeOf is a slow ass function, use it sparingly
        // in this case, a negative size represents an overflow of an integer
        if (rec.Size < 0) then begin
            s := Format('%n',[GetSizeOf( FilePath + subfolder + rec.name)]);
        end else begin
            s := Format('%n',[rec.size/1]);
        end;
        if rightstr(s,3) = FormatSEttings.decimalseparator + '00' then begin
            s := leftstr(s, length(s)-3);
        end;
        //UnitColumnManager.setSize(li, s);
        fld.SetSize(s);

        ext := uppercase(ExtractFileExt(rec.name));
        if (length(ext) > 1) then ext := rightstr(ext, length(ext) - 1);


        //UnitColumnManager.setType(li, ext);
        fld.SetType(ext);
    end;

    // Modified or Created date

    if IsZero(rec.FindData.ftCreationTime) then begin
    	if not IsZero(rec.FindData.ftLastWriteTime) then begin
			rec.FindData.ftCreationTime := rec.FindData.ftLastWriteTime;
        end else begin
        	rec.FindData.ftCreationTime := DateTimeToFileTime(now);
        end;
    end;
    if IsZero(rec.FindData.ftLastWriteTime) then begin
    	rec.FindData.ftLastWriteTime := rec.FindData.ftCreationTime;
    end;



    // Workaround for Time being off by One Hour
    td.Created := DateTimeFromFileTime(rec.FindData.ftCreationTime);
    if td.created = 0 then begin
        td.created := DateTimeFromFileTime(DateTimeToFileTime(now));
    end;

    td.Modified := DateTimeFromFileTime(rec.FindData.ftLastWriteTime);
    if td.Modified = 0 then begin
        td.Modified := DateTimeFromFileTime(DateTimeToFileTime(now));
    end;
    {$WARN SYMBOL_PLATFORM ON}

    td.DirectSelected := true;

    fld.SetDate( self.GetDateStringFormatted(td) );
    fld.SetDateCreated( self.GetDateStringFormatted(td.created) );
    fld.Checked := IsChecked;
    if (IsChecked) then inc(MyCheckedCount);
    fld.ImageIndex := ExtractIconCached(
        ILDynamic,
        IncludeTrailingPathDelimiter (filepath )+ fld.caption
    );
    fld.Data := Pointer(TagData.Add(td));
    DataList.Add(fld);

    for i := 0 to fld.SubItems.Count - 1 do begin
        ColSize[1+i] := max( GetFileList.Canvas.TextWidth('[+]'+fld.SubItems[i])+20, ColSize[i+1]);
    end;

    for i := 0 to (fld.CustomData.Count-1) do begin
        ColSize[1+i+fld.SubItems.Count ] := max( GetFileList.Canvas.TextWidth('[+]'+fld.CustomData[i])+20, ColSize[1+i+fld.SubItems.Count]);
    end;



    if not recursive then begin
        if (datalist.count mod REPORT_AT) = 0 then begin
            FrmMain.SetStatusText('Reading Files... ' + IntToStr(datalist.count), false);
        end;
    end;
end;


procedure TFrmDummyShowPreview.GetRecord(var myrec : RenameRecord; index : integer);
var //currentselected : integer;
    curtd : TTagData;
    fld : TFileListData;
    i : integer;
    selcount : integer;
begin
    fld := self.GetFileData(index);
    curtd := FrmDummyShowPreview.GetTagData(index);

    myrec.shortname := fld.altname;

    selcount := 0;
    for i := 0 to index do begin
        if GetChecked(i) then inc(selcount);
        if i = index then BREAK;
    end;
    myrec.listposition := index + 1;
    myrec.selectedposition := selcount;
    myrec.filedate := UnitColumnManager.GetDate(index);
    if (length(myrec.filedate) > 0) and (myrec.filedate[1] = '[')then begin
        myrec.filedate := copy(myrec.filedate,2,length(myrec.filedate) - 2);
    end;
    myrec.filesize := UnitColumnManager.GetSize(index);

    myrec.Created := curtd.Created;
    myrec.Modified := curtd.Modified;

    myrec.columnname := UnitColumnManager.GetName(index);;
    myrec.foldertotal := FoldersCounter.GetCount(
        ExtractFilePath(
            IncludeTrailingPathDelimiter(frmmain.GetPath) + myrec.columnname));
    //
    // collect file data

    myrec.ext := ExtractFileExt(myrec.columnname);
    if (myrec.ext <> '') then begin
        if (myrec.ext[1] = '.') then myrec.ext := rightstr(myrec.ext, length(myrec.ext) - 1);
    end;
    myrec.name := ExtractFileName(myrec.columnname);
    if curtd.IsFolder then begin
        myrec.ext := '';
        myrec.name := ExtractFileName(myrec.columnname);
    end;

    // (bugfix - no extension = truncates name by 1 char)
    if (myrec.ext = '') then begin
        myrec.name := leftstr(myrec.name, length(myrec.name) - (length(myrec.ext) + 1));
    end;
    myrec.FullName := IncludeTrailingPathDelimiter(frmmain.GetPath) + myrec.columnname;



    myrec.listtotal := DataList.Count;
    myrec.selectedtotal := selcount;
    myrec.changedcount := 1; // the startcounting and increment need positions to be 1 based, not zero
    myrec.incrementby := SafeInt(frmmain.txtIncrementBy.Text, 1);
    myrec.selectedtotal := MyCheckedCount;
end;
function TFrmDummyShowPreview.FindByFilename(filename : string) : integer;
var
    i, foundindex : integer;
    found: boolean;
    fld : TFileListData;
begin
    result := -1;
    foundIndex := -1;
    found := false;
    for i := 0 to datalist.Count - 1 do begin
        fld := TFileListData(DataList[i]);
        found := filename = fld.caption;
        if found then begin
            foundIndex := i;
            BREAK;
        end;
    end;

    if Found then result := FoundIndex;
end;

//var kcount : integer;
function CompareData(Item1, Item2: Pointer): Integer;
var f1, f2 : TFileListData;
//    li1, li2 : TListItem;

var c1, c2 : double;
    d1, d2 : tdate;
    s1, s2 : string;
    IgnoreSortDirection : boolean;

    a1, a2 : array[0.. Windows.MAX_PATH] of double;

    function GetFileFlag(li : TFileListData) : string;
    var td : TTagData;
    begin
        td := FrmDummyShowPreview.GetTagDataAbsolute(integer(li.data));
        result := '1';
        if td.IsFolder then result := '0';
    end;
    //
    // Numbers are treated as ASCII 57 (char '9') so they group together
    // This keeps situations like '37', '&', '38' from happening in the sorting
    //
    procedure StringToAlphaArray(s : string; var a : array of double);
    var i,j,k  : integer;
        s2 : string;

    begin
        for i := low(a) to high(a) do begin
            a[i] := -1;
        end;
        j := 0;
        i := 1;
        s2 := '';

        while (i <= length(s)) do begin
            // don't accept digits that can be long enought o overflow an integer
            while (i <= length(s)) and isdigit(s[i]) and (length(s2)<9)  do begin
                s2 := s2 + s[i];
                inc(i);
            end;

            if (s2 = '') then begin
                a[j] := ord(s[i]);
                inc(i);
            end else begin
                try
                    a[j] := StrToFloat(s2+ FormatSettings.DecimalSeparator + '57');  // 57 is '9' in ASCII
                    //a[j] := StrToFloat(s2+ '.57');  // 57 is '9' in ASCII
                except
                    // the numbers in a file can easily max out an integer
                    // just accept the substring as alpha digits
                    for k := 0 to length(s2) - 1 do begin
                        a[j] := ord(s2[k+1]);
                        inc(j)
                    end;
                    //a[j] := MaxInt;
                    s2 := '';
                end;
                s2 := '';
            end;
            inc(j);
        end;
    end;
    function AlphaCompare(a1, a2 : array of double) : integer;
        function DblCompare(c1, c2 : double) : integer;
            function FixNumber(d : double) : integer;
            var s : string;
            begin
                result := trunc(d);
                if trunc(d) <> d then begin
                    s := FloatToStr(d);
                    s := copy(s,1,length(s)-3); // trim .57 off

                    result := StrToInt(s);
                end;
            end;
            function HasMarker(d : double) : boolean;
            var s : string;
            begin
                s := FloatToStr(d);
                result := (rightstr(s,2) = '57') and (not isDigit(rightstr(s,3)[1]));
            end;
            var i1 : double;
                i2 : double;
        begin
            result := 0;

            // detect when two marked numbers are being compared
            i1 := c1;
            i2 := c2;
            if (hasmarker(c1) and hasmarker(c2)) then begin
                i1 := FixNumber(c1);
                i2 := FixNumber(c2);
            end else begin
                if hasmarker(i1) then i1 := 57;
                if hasmarker(i2) then i2 := 57;
            end;



            if i1 <> i2 then begin
                if i1 < i2 then begin
                    result := -1;
                end else begin
                    result := 1;
                end;
            end;
        end;
    var i : integer;
    begin
        result := 0;
        for i:= Low(a1) to high(a1) do begin
            result := DblCompare(a1[i], a2[i]);

            if result <> 0 then EXIT;
            if (a1[i] = -1) or (a2[i] = -1) then Exit;
        end;
    end;
var  clicked : TColumnType;
  //  s3, s4 : string;
    i : integer;
    compare : integer;
    fname1, fname2 : string;
    hadProblem : boolean;

    procedure HandleNameColumn;
        procedure compareNames(fname1, fname2 : string);
        begin
           s1 := lowercase(fname1);
            s2 := lowercase(fname2);
            if (ExtractFileExt(s1)= ExtractFileExt(s2)) then begin
                // chop off the extensions
                i := length(ExtractFileExt(s1));

                s1 := copy(s1,1,length(s1) - i);
                s2 := copy(s2,1,length(s2) - i);
            end;
            if ContainsDigit(s1) or ContainsDigit(s2) then begin
                StringToAlphaArray(s1, a1);
                StringToAlphaArray(s2, a2);

                compare := AlphaCompare(a1, a2);
            end else begin
                compare := CompareText(s1, s2);
            end;
        end;
    begin
        // name
        // show folders before files (both sorted)
        // sort by folder path, filename second
        if recursive then begin
            s1 := lowercase(f1.GetBaseFolder);
            s2 := lowercase(f2.getBaseFolder);
            if (s1 = s2) then begin
                s1 := ExtractFilename( lowercase(fname1) );
                s2 := ExtractFilename( lowercase(fname2) );
                // TODO: What in the hell is this trying to do?
                if (f1.ImageIndex = 6) then s1 := uppercase(s1);
                if (f2.ImageIndex = 6) then s2 := uppercase(s2);
//                compare := comparetext(s1, s2);
                compareNames(s1, s2);
            end else begin
                compare := comparetext(s1, s2);
            end;
        end else begin
            compareNames(fname1, fname2);
        end;
    end;
    procedure HandleSizeColumn;
    begin
             // size
            c1 := strtofloat(
                UnitUtil.StripNumberFormat(fname1)
            );
            c2 := strtofloat(
                UnitUtil.StripNumberFormat(fname2)
            );
            //c2 := strtofloat( self.GetDataByColumn(COL_SIZE, item2.Index));
            if (f1.ImageIndex = 6) then s1 := uppercase(s1);
            if (f2.ImageIndex = 6) then s2 := uppercase(s2);
            if (c1 < c2) then begin
                compare := -1;
            end else if (c1 = c2) then begin
                compare := 0;
            end else begin
                compare := 1;
            end;

    end;
    procedure HandleTypeColumn;
    begin
        compare := CompareText(fname1, fname2);
    end;
    procedure HandleDateColumn;
    begin
            hadProblem := false;
            s1 := fname1;
            if (s1 <> '') and (s1[1] = '[') then begin
                s1 := copy(s1,2,length(s1) - 2);
            end;
            s2 := fname2;
            if (s2 <> '') and (s2[1] = '[') then begin
                s2 := copy(s2,2,length(s2) - 2);
            end;

            //d1 := strtodatetime(s1);
            //d2 := strtodatetime(s2);
            try
                if (s1 = '') then begin
                    hadProblem := true;
                end else begin
                    d1 := VarToDateTime(  ansistring(s1));
                end;
            except
                // not quite sure how to handle this
                hadProblem := true;
            end;
            try
                if (s2 = '') then begin
                    hadProblem := true;
                end else begin
                    d2 := VarToDateTime(ansistring(s2));
                end;
            except
                hadProblem := true;
            end;


            if (HadProblem) then begin
                if (s1=s2) then begin
                    compare := 0;
                end else if (s1 = '') then begin
                    compare := -1;
                end else begin
                    compare := 1;
                end;
            end else begin
                if (d1 < d2) then begin
                    compare := -1;
                end else if (d1 = d2) then begin
                    compare := 0;
                end else begin
                    compare := 1;
                end;
            end;

    end;
    procedure HandleCustomColumn;
    begin

    end;
begin
    comparecount := (comparecount + 1);
    if (comparecount mod 1000) = 0 then begin
        i := comparecount mod 4000;
        i := i div 1000;

        case i of
        1 : FrmMain.SetSubStatusText(' |', true);
        2 : FrmMain.SetSubStatusText('/', true);
        3 : FrmMain.SetSubStatusText('-', true);
        0 : FrmMain.SetSubStatusText('\', true);
        end;
    end;

    f1 := TFileListData(Item1);
    f2 := TFileListData(Item2);

    IgnoreSortDirection := false;

    // ditry trick alert!
    // I'm using an unused integer to flag a file
    /// and to flag a folder
    compare := 0;
    s1 := GetFileFlag(Item1);
    s2 := GetFileFlag(Item2);
    if (s1 <> s2) then begin
        // a folder always comes before a file
        compare := StrComp(Pchar(s1), PChar(s2));
        IgnoreSortDirection := true;
    end else begin
        clicked := UnitColumnManager.GetClickedColumnType;
        case clicked of
            COL_TYPE_NAME: begin
                fname1 := f1.caption;
                fname2 := f2.caption;
                HandleNameColumn;
            end;
            COL_TYPE_SIZE: begin
                fname1 := F1.GetSize;
                fname2 := F2.GetSize;
                HandleSizeColumn;
            end;
            COL_TYPE_TYPE: begin
                fname1 :=f1.GetType;
                fname2 :=f2.GetType;
                HandleTypeColumn;
            end;
            COL_TYPE_DATE: begin
                fname1 := f1.GetDate;
                fname2 := f2.GetDate;
                HandleDateColumn;
            end;
            COL_TYPE_DATECREATED: begin
                fname1 := f1.GetDateCreated;
                fname2 := f2.GetDateCreated;
                HandleDateColumn;
            end;
            COL_TYPE_CUSTOM: begin
                clicked := UnitColumnManager.GetCustomColumnType(UnitColumnManager.GetClickedColumn);

                //
                // Unchecked names will have no preview data
                //
                fname1 := '';
                i := UnitColumnManager.GetCustomColumnIndex(UnitColumnManager.GetClickedColumn);
                try
                    if f1.Checked then begin
                        if i < f1.CustomData.count then
                            fname1 :=  f1.CustomData[ i ];
                    end;
                finally
                end;
                fname2 := '';
                try
                    if f2.checked then begin
                        if i < f2.CustomData.count then
                            fname2 :=  f2.CustomData[ i ];
                    end;
                finally

                end;

                if clicked = COL_TYPE_SIZE then begin
                    if fname1 = '' then fname1 := '0';
                    if fname2 = '' then fname2 := '0';
                end;
                case clicked of
                COL_TYPE_NAME:
                    HandleNameColumn;
                COL_TYPE_SIZE:
                    HandleSizeColumn;
                COL_TYPE_DATE:
                    HandleDateColumn;
                end
            end;
        end;
    end;

    if not IgnoreSortDirection then begin
        if UnitColumnManager.GetClickedColumnDesc  then begin
            compare := compare * -1;
        end;
    end;

    result := compare;
end;


procedure TFrmDummyShowPreview.RefreshDriveIcons;
var driveletter : char;
    sp : TSpecialPaths;
    root, last : string;
    drivetype : cardinal;
    dint  : integer;
    s : string;
    i, icoIdx : integer;
    ico : TIcon;

    drives : TStringDynArray;
    driveTypeList : TDictionary<integer, integer>;
begin
//    if iconsRefreshed then EXIT;
//    iconsRefreshed := true;

    sp := UnitSpecialPaths.SpecialPaths;

    ExtractIconNoCache(ilFolderBrowser, sp.GetStartupPath, Icon_Folder_Network);
    ExtractIconNoCache(ilFolderBrowser, sp.GetSystemRoot, Icon_Folder_Normal, true);
    ExtractIconNoCache(ilFolderBrowser, sp.GetMyPictures, Icon_MyPictures);
    ExtractIconNoCache(ilFolderBrowser, sp.GetDesktop, Icon_MyDesktop, true);

    while ilFolderBrowser.Count > Icon_MyDesktop+1 do begin
        ilFolderBrowser.Delete(Icon_MyDesktop+1);
    end;




    //for DriveLetter := 'A' to 'Z' do begin
    drives := TDirectory.GetLogicalDrives;
    driveTypeList := TDictionary<integer, integer>.create;
    for i:=0 to length(drives)-1 do begin
        Root := drives[i];

        DriveType := GetDriveType(PChar( Root  )) ;
        if (DriveType = DRIVE_NO_ROOT_DIR) then CONTINUE;
        if driveTypeList.TryGetValue(drivetype, icoIdx) then begin
            ico := TIcon.Create;
            ilFolderBrowser.GetIcon(icoIdx, ico);
            ilFolderBrowser.AddIcon(ico);
            CONTINUE;
        end;

        // C drive has a different icon that other volumes, don't use it for the cache
        if Root <> 'C:\' then
            driveTypeList.Add(drivetype, ilFolderBrowser.Count);

        root := IncludeTrailingPathDelimiter(root);
        dint := -1;

        case DriveType of
            DRIVE_REMOVABLE: begin
                last := root;
                ExtractIconNoCache(ilFolderBrowser, root, dint{Icon_Drive_Floppy});
            end;
            DRIVE_CDROM: begin
                last := root;
                ExtractIconNoCache(ilFolderBrowser, root, dint{icon_drive_cdrom});
            end;
            DRIVE_FIXED: begin
                last := root;
                ExtractIconNoCache(ilFolderBrowser, root, dint{icon_drive_fixed});
            end;
            DRIVE_REMOTE: begin
                last := root;
                ExtractIconNoCache(ilFolderBrowser, root, dint{icon_drive_network});
            end;
            DRIVE_NO_ROOT_DIR: begin
            end;
            else begin
                last := root;
                ExtractIconNoCache(ilFolderBrowser, root, dint{icon_drive_ghost});
            end;
        end;

    end;
    driveTypeList.Free;


end;

procedure TFrmDummyShowPreview.RefreshFileList(
    var UseRecursiveOnce : boolean;
    var FilesCount : integer;
    var FoldersCount : integer;
    var CheckedCount : integer;
    Unchecked : boolean = false
);
    // for FindRecursive
var RenameFolders : boolean;
    FolderFilter : string;
    DoCheckFiles, DoCheckFolders : boolean;

    procedure FindRecursive(FilePath : string; subfolder : string);
    var rec : TSearchRec;
        r: cardinal;
        i  : integer;
        j,k : cardinal;
        attr : integer;
        b : boolean;
        ext : string;
//        fd :  _WIN32_FIND_DATA;
    label
        FindNextLabel;

    begin
        FilePath := IncludeTrailingPathDelimiter(FilePath);
        if not DirectoryExists(FilePath) and LongerDirectoryExists(FilePath) then begin
            FilePath := '\\?\\' + FilePath;
        end;


        subfolder := IncludeTrailingPathDelimiter(subfolder);
        if subfolder = '\' then subfolder := '';
        if (UseRecursiveOnce) then
            FrmMain.SetSubStatusText('Scanning ..\' + subfolder, TRUE);

        i := 0;
            {$WARN SYMBOL_PLATFORM OFF}
        attr := faAnyFile or faHidden;

        if frmConfig.cbNoHiddenFiles.Checked then begin
            attr := attr xor faHidden;
        end;
            {$WARN SYMBOL_PLATFORM ON}
        r := 0;

        // send me the error instead
        
        {$i-}
        // engulf any "no disk" exceptions from removable drives
        SetErrorMode(SEM_FAILCRITICALERRORS);
        try
            r := FindFirst(FilePath + subfolder + FolderFilter,attr, rec);
        except
        end;
        {$i+}
        while (r = 0) do begin
            if self.DoCancelRefreshFileList then EXIT;
            // skip non-folders and the Self,Parent special folders
            if ((rec.Attr and fadirectory) = 0) then begin
                //
                // File Found - update total file tally and per-folder count
                //
                wc.SetSearchString(rec.name);
                b := wc.FindFirst(j,k);
                if (b and (j=1) and (k = length(rec.name))) then begin
                    FoldersCounter.IncFolderCount(filepath + subfolder);
                    if (FrmConfig.RenameFiles) then begin

                        self.NewListItem(
                            subfolder, filepath,
                            false,
                            DoCheckFiles and not frmexclude.isexcluded(rec.Name),
                            rec);
                    end;
                    inc(FilesCount);

                    if not containsMP3s then begin
                        ext := lowercase(ExtractFileExt(rec.name));
                        if ext = '.mp3' then
                            containsMP3s := true;
                    end;
                    if not containsJPEGs then begin
                        ext := lowercase(ExtractFileExt(rec.name));
                        if StringIn(ext,['.jpg','.jpeg']) then
                            containsJPEGs := true;
                    end;
                end;
            end else begin
                if (rec.name = '.') or(rec.Name = '..') then begin
                    goto FindNextLabel;
                end;

                //
                // Folder Founder - trigger a recursive scan if needed
                //

                if (RenameFolders) and (not UseRecursiveOnce) then begin
                    self.NewListItem(subfolder, filepath, true, DoCheckFolders, rec);
                    inc(FoldersCount);
                end;

                if (UseRecursiveOnce) then begin
                    FindRecursive(FilePath, subfolder + rec.name);
                end;

            end;
         FindNextLabel:
            r := FindNext(rec);
            inc(i);
            if ((i mod REPORT_AT) = 0 )then begin
                FrmMain.SetSubStatusText(rec.name, false);

            end;

            {if (DataList.Count mod 5000) = 0 then begin
                DataList.Sort(CompareData);
            end;}
            if self.DoCancelRefreshFileList then EXIT;
        end;
        FindClose(rec);

    end;
var i : integer;
begin
    recursive :=  UseRecursiveOnce;
    frmmain.btnCancelScan.Visible := UseRecursiveOnce;
    if self.IgnoreInterface then EXIT;
    self.IgnoreInterface := true;

    try
        // pre scan initialization
        self.DoCancelRefreshFileList := false;
        if (FrmConfig.UseFileIcons) then begin
            GetFileList.SmallImages := nil;
        end else begin
            GetFileList.SmallImages := self.ILDynamic;
        end;

        containsMP3s := false;
        containsJPEGs := false;
        FilesCount := 0;
        FoldersCount := 0;
        CheckedCount := 0;
        MyCheckedCount := 0;
        for i := 0 to getfilelist.Columns.Count -1  do begin
            colsize[i] := GetFileList.Canvas.TextWidth( getfilelist.Column[i].Caption + '[+]    ');
        end;


        RenameFolders := FrmConfig.RenameFolders;
        if (FrmConfig.UseFullDateTime) then begin
            AddTime := AddTimeFull;
        end else begin
            AddTime := @AddTimeShort;
        end;

        if (FrmConfig.UseFolderFilters) then begin
            FolderFilter := frmmain.GetFilter;
        end else begin
            FolderFilter := '*.*';
        end;

        wc.AssignFilemode(true);
        if (FrmMain.GetFilter = '') then begin
            wc.SetSearchPattern('*.*');
        end else begin
            wc.SetSearchPattern(frmMain.GetFilter);
        end;

        DoCheckFiles := not Unchecked;
        DoCheckFolders := not UseRecursiveOnce and frmConfig.SelectFolders;

        //
        // Finally, update the file list and all the associated list
        // (icons, tagdata, folders, etc)
        //

        GetFileList.Items.BeginUpdate;
        {Clear out all these damn lists}
        if not GetFileList.OwnerData then begin
            GetFileList.Clear;
        end;

        DataList.Clear;
        //
        FoldersCounter.Clear;
        UnitFileIcon.ClearCache;
        ILDynamic.Clear;
        ILDynamic.AddImages(ilFolderBrowser);
        TagData.Clear;
        SelectedList.clear;
        GetFileList.ClearSelection;
        Application.ProcessMessages;

        FindRecursive(frmmain.Getpath, '');

        if FrmMain.AutoWidth1.checked then begin
            for i := 0 to (GetFileList.Columns.Count - 1) do begin
                GetFileList.Columns[i].Width := colsize[i];
            end;
        end;
        frmmain.btnCancelScan.Visible := false;
        GetFileList.items.Count := DataList.Count;

        UseRecursiveOnce := false;
        CheckedCount := MyCheckedCount;
        frmmain.setsubstatustext('', false);
        frmmain.SetStatusText('Sorting ...', TRUE);
        comparecount := 0;
        DataList.Sort(CompareData);
        GetFileList.UpdateItems(0, datalist.Count-1);
        frmmain.SetStatusText('Sorting Done.', TRUE);
        frmmain.setsubstatustext('', false);

        //GetFileList.UpdateItems(0,DataList.Count-1);

        GetFileList.Items.EndUpdate;

    finally
        IgnoreInterface := false;
    end;
end;
procedure TFrmDummyShowPreview.ShowPreviewList(var ChangedCount : integer);
var i, j, k, l, appendCount, line, fields, collisionIndex : integer;
    b : boolean;
    startidx, endidx: cardinal;
    selcount: cardinal;
    ColumnNameData, name, s, s2, tempremove: string;
    ext: string;
    path, lastPath: string;
    rec, temprec: RenameRecord;
    sequence, lastsequence : string;

    rwc : TWildcards; // can be replaced with full regular expressions
    newname : string;
    DisableCollisions : boolean;
    charlist : array[#0 .. #255] of boolean;
    c : char;

    td, curTd, miscTD : TTagData;

    replaceStr : string;

    fld : TFileListData;

    procedure CorrectTextNumbers;
        function SafeStr(s : string; default : integer = 0) : string;
        var i : integer;
        begin
            i := SafeInt(s, default);
            result := IntToStr(i);
        end;
    begin
        frmmain.txtIncrementBy.text := SafeStr(frmmain.txtIncrementBy.text, 1);
        frmmain.txtStartCountingAt.Text := SafeStr(frmmain.txtStartCountingAt.Text, 1);
        frmmain.txtANPosition.text := SafeStr(frmmain.txtANPosition.text, 1);
        frmmain.txtAEPosition.text := SafeStr(frmmain.txtAEPosition.text, 1);
        frmmain.txtRP2X.text := SafeStr(frmmain.txtRP2X.text,1);
        frmmain.txtRP2Y.text := SafeStr(frmmain.txtRP2Y.text,1);
        frmmain.txtRMFromNameStart.text := SafeStr(frmmain.txtRMFromNameStart.text, 0);
        frmmain.txtRMFromNameEnd.text := SafeStr(frmmain.txtRMFromNameEnd.text, 0);
        frmmain.txtRMFromExtensionStart.text := SafeStr(frmmain.txtRMFromExtensionStart.text, 0);
        frmmain.txtRMFromExtensionEnd.text := SafeStr(frmmain.txtRMFromExtensionEnd.text, 0);
    end;
    procedure ApplyRenamingOptions;
    var j : integer;
    begin
//
        // removes
        //
        if frmmain.cbRM2Characters.checked then begin
           tempremove := copy(
                name,
                frmmain.udR2CharX.Position,
                frmmain.udR2CharY.Position - frmmain.udR2CharX.Position + 1
           );

           delete(
                name,
                frmmain.udR2CharX.Position,
                frmmain.udR2CharY.Position - frmmain.udR2CharX.Position + 1
            );
            if frmmain.cbRM2AndMove.checked then begin
                {if FrmMain.udR2CharX.Position > frmmain.udR2AndMove.position then begin}
                    insert(tempremove,name, frmmain.udR2AndMove.position);
                {end else begin
                    insert(
                        tempremove,
                        name,
                        frmmain.udR2AndMove.position - (frmmain.udR2CharY.Position - frmmain.udR2CharX.Position + 1)
                    );
                end;}

            end;
        end;

        if frmmain.cbEverything.checked then begin
            name := copy(
                name,
                FrmMain.udRM2X.Position,
                frmmain.udRM2Y.Position - frmmain.udRM2X.Position + 1
            );
        end;
        if (frmmain.cbRM2NumFromName.Checked) then begin
            s := '';
            for j := 1 to length(name) do begin
                if not ((name[j] >= '0') and (name[j] <= '9')) then begin
                    s := s + name[j];
                end;
            end;
            name := s;
        end;
        if (frmmain.cbRM2NumFromExt.Checked) then begin
            s := '';
            for j := 1 to length(ext) do begin
                if not ((ext[j] >= '0') and (ext[j] <= '9')) then begin
                    s := s + ext[j];
                end;
            end;
            ext := s;
        end;
        if (frmmain.cbRMFromNameStart.Checked) then begin
            name := rightstr(name, length(name) - SafeInt(frmmain.txtRMFromNameStart.text));
        end;
        if (frmmain.cbRMFromNameEnd.Checked) then begin
            name := leftstr(name, length(name) - SafeInt(frmmain.txtRMFromNameend.text));
        end;
        if (frmmain.cbRMFromExtensionStart.Checked) then begin
            // make sure the leading period persists
            ext := rightstr(ext, length(ext) - 1);
            ext := rightstr(ext, length(ext) - SafeInt(frmmain.txtRMFromExtensionStart.text));
            if pos('.',ext) <> 1 then ext := '.' + ext;
        end;
        if (frmmain.cbRMFromExtensionEnd.checked) then begin
            ext := leftstr(ext, length(ext) - SafeInt(frmmain.txtRMFromExtensionEnd.text));
            if pos('.',ext) <> 1 then ext := '.' + ext;
            // make sure the leading period persists
        end;

        if (frmmain.cbrmfromname.checked) then begin

            if (not frmmain.cbRMWordList.Checked) then begin
                name := rwc.SearchReplaceWith(name,
                    ReplaceMetaVariables(frmmain.txtrmfromname.Text, rec),
                    '',
                    true);
            end else begin
                s2 := frmmain.txtrmfromname.Text;
                s := TokenString(s2, ' ');
                while s <> '' do begin
                    name := rwc.SearchReplaceWith(name,
                        ReplaceMetaVariables(s, rec),
                        '',
                        true);

                    s := TokenString(s2, ' ');
                end;

            end;
        end;
        if (frmmain.cbRMFromExtension.Checked) then begin
            if (not frmmain.cbRMWordList.Checked) then begin

            ext := rwc.SearchReplaceWith(ext,
                ReplaceMetaVariables(frmmain.txtRMFromExtension.text, rec),
                '',
                true);
            end else begin
                s2 := frmmain.txtRMFromExtension.text;
                s := TokenString(s2, ' ');
                while s <> '' do begin
                    name := rwc.SearchReplaceWith(ext,
                        ReplaceMetaVariables(s, rec),
                        '',
                        true);

                    s := TokenString(s2, ' ');
                end;
            end;
        end;


        //
        // replacements
        //
        if (frmmain.cbRPNameWith.Checked) then begin
            name := ReplaceMetaVariables(frmmain.txtRPNameWith.Text, rec);
        end;

        if (frmmain.cbRPExtensionWith.Checked) then begin
            ext := ReplaceMetaVariables(frmmain.txtRPExtensionWIth.Text, rec);
        end;
        if (frmmain.cbRPThis.Checked) then begin
            name := rwc.SearchReplaceWith(name,
                ReplaceMetaVariables(frmmain.txtRPThis.text, rec),
                ReplaceMetaVariables(frmmain.txtRPWith.Text, rec),
                 true);
            ext := rwc.SearchReplaceWith(ext,
                ReplaceMetaVariables(frmmain.txtRPThis.text, rec),
                ReplaceMetaVariables(frmmain.txtRPWith.Text, rec),
                 true);
        end;

        //
        // replacements 2
        //
        if (frmmain.cbRP2Characters.Checked) then begin
            s :=  Copy(name, frmmain.upRP2X.Position, (frmmain.upRP2Y.Position - frmmain.upRP2X.Position) + 1);
            Delete(name, frmmain.upRP2X.Position, (frmmain.upRP2Y.Position - frmmain.upRP2X.Position) + 1);
            Insert( rwc.SearchReplaceWith(s,
                        ReplaceMetavariables(frmmain.txtRP2This.text, rec),
                        ReplaceMetaVariables(frmmain.txtRP2With.text, rec), true),
                    name,
                    frmmain.upRP2X.Position);
        end;

        if (frmmain.cbRP2CharacterList.checked) then begin
            fillchar(charlist, length(charlist), false);
            for j := 1 to length(name) do begin
                charlist[name[j]] := true;
            end;

            for j := 1 to length(frmmain.txtRP2sc.Text) do begin

                if j <= length(frmmain.txtRP2scWith.text) then begin
                    s := frmmain.txtRP2scWith.Text[j];
                end else begin
                    s := '';
                end;

                c := frmmain.txtRP2sc.Text[j];
                if (s = '') or (charlist[c]) then begin
                    name := StringReplace(
                        name, c, s,
                        [rfreplaceall, rfignorecase]);
                end;
            end;

            fillchar(charlist, length(charlist), false);
            for j := 1 to length(ext) do begin
                charlist[ext[j]] := true;
            end;
            for j := 1 to length(frmmain.txtRP2sc.Text) do begin
                if j <= length(frmmain.txtRP2scWith.text) then begin
                    s := frmmain.txtRP2scWith.Text[j];
                end else begin
                    s := '';
                end;

                c := frmmain.txtRP2sc.Text[j];
                if (s = '') or (charlist[c]) then begin
                    ext := StringReplace(
                        ext, c, s,
                        [rfreplaceall, rfignorecase]);
                end;
            end;
        end;

        if frmmain.cbRPNumbers.checked then begin
            j := 0;
            case frmmain.cbxRPNumbers.ItemIndex of
            0: j := length(name);
            1: j := 1;
            end;

            while (j > 0) and (j<=length(name)) and  (CharInSet(name[j], ['0' .. '9'])) do begin
                case frmmain.cbxRPNumbers.ItemIndex of
                0: dec(j);
                1: inc(j);
                end;
            end;

            case frmmain.cbxRPNumbers.ItemIndex of
            0:
                begin
                    s := rightstr(name,length(name)-j);
                    name := midstr(name,1,j);
                end;
            1:
                begin
                    s := leftstr(name,j);
                    name := midstr(name,1,j);
                end;
            end;
            replacestr := '';
            if (s <> '') then begin
                replaceStr := replacenum.Values[s];
                if (replaceStr = '') then begin
                    replaceStr := IntToStr(replaceSeq);
                    replacenum.Values[s]  := replaceStr;
                    inc(replaceSeq);
                end;
            end;

            while length(s) > length(replacestr) do begin
                replacestr := '0' + replacestr;
            end;

            name := name + replacestr;
        end;

        //
        // additions 2
        //

        if frmmain.cbA2Find.Checked then begin
            if frmmain.txta2Find.text <> '' then begin
                rwc.SetSearchString(name);
                rwc.SetSearchPattern(
                    ReplaceMetaVariables(frmmain.txta2Find.text, rec)
                );
                s := name;
                if rwc.FindFirst(startidx, endidx) then begin
                    case frmmain.cbA2Insert.ItemIndex of
                    0:

                        Insert(
                            ReplaceMetaVariables(frmmain.txtA2Insert.text, rec),
                            name,
                            endidx +1
                        );
                    1:

                        Insert(
                            ReplaceMetaVariables(frmmain.txtA2Insert.text, rec),
                            name,
                            startidx
                        );
                    2..3:
                        begin
                            if (frmmain.txtA2Insert.Text <> '') then begin
                                s2 := copy(name, startidx, endidx-startidx+1);
                                delete(name, startidx, endidx-startidx+1);
                                rwc.SetSearchString(name);
                                rwc.SetSearchPattern(
                                    ReplaceMetaVariables(frmmain.txtA2Insert.Text, rec)
                                );


                                if rwc.FindFirst(startidx, endidx) then begin
                                    if (frmmain.cba2insert.ItemIndex = 3) then begin
                                        startidx := endidx+1;
                                    end;
                                    Insert(s2, name, startidx);
                                end else begin
                                    name := s;
                                end;
                            end;
                        end;
                    end;
                end;
            end;
        end;

        //
        // additions
        //
        if (frmmain.cbAEInsert.Checked) then begin
            // skip the leading period
            Insert(ReplaceMetaVariables(frmmain.txtAEInsert.Text, rec), ext, frmmain.udAEPosition.Position + 1);
        end;
        if (frmmain.cbANInsert.Checked) then begin
            Insert(ReplaceMetaVariables(frmmain.txtANInsert.Text, rec), name, frmmain.udANPosition.Position );
        end;

        if (frmmain.cbAEndOfName.Checked) then begin
            name := name + ReplaceMetaVariables(frmmain.txtAEndOfName.Text, rec);
        end;

        if (frmmain.cbAEndOfExtension.Checked) then begin
            ext :=  ext + ReplaceMetaVariables(frmmain.txtAEndOfExtension.Text, rec);
        end;

    end;
    procedure GetRecordFor(
        var myrec : RenameRecord;
        index,
        currentselected : integer;
        curtd : TTagData;
        fld : TFileListData
    );
    begin
        myrec.shortname := fld.altname;

        myrec.listposition := index + 1;
        myrec.selectedposition := selcount;
        myrec.filedate := UnitColumnManager.GetDate(index);
        if (length(myrec.filedate) > 0) and (myrec.filedate[1] = '[')then begin
            myrec.filedate := copy(myrec.filedate,2,length(myrec.filedate) - 2);
        end;
        myrec.filesize := UnitColumnManager.GetSize(index);
        //curtd := FrmDummyShowPreview.GetTagData(i);
        myrec.Created := curtd.Created;
        myrec.Modified := curtd.Modified;

        myrec.columnname := UnitColumnManager.GetName(index);;
        myrec.foldertotal := FoldersCounter.GetCount(
            ExtractFilePath(
                IncludeTrailingPathDelimiter(frmmain.GetPath) + myrec.columnname));
        //
        // collect file data

        myrec.ext := ExtractFileExt(myrec.columnname);
        if (myrec.ext <> '') then begin
            if (myrec.ext[1] = '.') then myrec.ext := rightstr(myrec.ext, length(myrec.ext) - 1);
        end;
        myrec.name := ExtractFileName(myrec.columnname);
        if curtd.IsFolder then begin
            myrec.ext := '';
            myrec.name := ExtractFileName(myrec.columnname);
        end;

        // (bugfix - no extension = truncates name by 1 char)
        if (myrec.ext <> '') then begin
            myrec.name := leftstr(myrec.name, length(myrec.name) - (length(myrec.ext) + 1));
        end;
        myrec.FullName := IncludeTrailingPathDelimiter(frmmain.GetPath) + myrec.columnname;

    end;
    procedure UpdateSelectCount(var selcount : cardinal; NewPath : boolean);
    begin
        selcount := selcount + 1;
        // clear on new folder?

        if (frmmain.cbClearCount.Checked) then begin
            if (NewPath) then begin
                if (not frmmain.rbUsePositionInList.checked) then begin
                    selcount := 1;
                end;
            end;
        end;
    end;
    procedure TextFileLoad;
    begin
        if not FileExists(FrmMain.FrameTextfield1.txtTName.text) then EXIT;

        TextFileFormat.SkipedLines := SafeInt(FrmMain.FrameTextfield1.txtTSkip.text);
        TextFileFormat.Delim := ',';
        if FrmMain.FrameTextfield1.txtTDelim.text <> '' then begin
            if lowercase(FrmMain.FrameTextfield1.txtTDelim.Text) = 'tab' then begin
                TextFileFormat.Delim := #9;
            end else begin
                TextFileFormat.Delim := FrmMain.FrameTextfield1.txtTDelim.text[1];
            end;
            TextFileFormat.Delim := FrmMain.FrameTextfield1.txtTDelim.text[1];
        end;
        TextFileFormat.LoadFromFile(FrmMain.FrameTextfield1.txtTName.text);
    end;
    procedure TextFileGetVariables(rec : RenameRecord);
    var fields : integer;
        searchterm : string;
    begin
        if not FileExists(FrmMain.FrameTextfield1.txtTName.text) then EXIT;

        UnitMetaVariables.ClearTextFields;
        if FrmMain.FrameTextfield1.rbTMatch.checked  then begin
            if FrmMain.FrameTextfield1.cbTDelim.checked then begin
               for fields := 0 to TextFileFormat.FieldCount - 1 do begin
                    UnitMetaVariables.SetTextFields(
                        fields, TextFileFormat.GetField(selcount-1, fields) );
               end;
            end else begin
                UnitMetaVariables.SetTextFields(0, TextFileFormat.Strings[selcount-1]);
            end;
        end else if FrmMain.FrameTextfield1.rbTFind.Checked then begin
            case frmmain.FrameTextfield1.cbxSearchWith.ItemIndex  of
            0: searchterm := VAR_CUR_FILENAME;
            1: searchterm := VAR_CUR_FILENAME_NOEXT;
            2: searchterm := FrmMain.FrameTextfield1.txtTFind.text;
            end;

            line := TextFileFormat.FindLine(
                UnitMetaVariables.ReplaceMetaVariables(searchterm,rec)
            );
            if line <> -1 then
               if FrmMain.FrameTextfield1.cbTDelim.checked then begin
                    for fields := 0 to TextFileFormat.FieldCount - 1 do begin
                        UnitMetaVariables.SetTextFields(fields, TextFileFormat.GetField(line, fields) );
                    end;
               end else begin
                    UnitMetaVariables.SetTextFields(0, TextFileFormat.Strings[selcount-1]);
               end;

        end;
    end;
    procedure GenerateCustomColumns(fld : TFileListData);
    var k : integer;
        s, s2 : string;
    begin
        fld.CustomData.Clear;
        for k := 0 to UnitColumnManager.CustomColumnCount-1 do begin
            s := UnitColumnManager.GetCustomColumnData(k);

                UnitMetaVariables.SetNoCleanOnce;
                s2 := ReplaceMetaVariables(s,rec);

                UnitMetaVariables.ClearNoCleanOnce;

            if LowerCase(s2) = LowerCase(s) then begin
                fld.CustomData.Add( '' );
            end else begin
                fld.CustomData.Add( s2 );
            end;
        end;
    end;
    procedure AlterCase;
    begin
        //
        // optinally alter case
        //
        if (frmmain.cbOverrideCase.checked) then begin
            s := ext;
            if (frmmain.rbCUppercase.Checked) then begin
                name := ansiuppercase(name);
                ext := ansiuppercase(ext);
            end else if (frmmain.rbCLowercase.Checked) then begin
                name := ansilowercase(name);
                ext := ansilowercase(ext);
            end else if (FrmMain.rbCCapitalizeWords.Checked) then begin
                if frmmain.cbSplit.checked then begin
                    name := UnitToken.CapitalizeWords(name, frmmain.txtSplit.text);
                    ext := UnitToken.CapitalizeWords(ext, frmmain.txtSplit.text);
                end else begin
                    name := UnitToken.CapitalizeWords(name);
                    ext := UnitToken.CapitalizeWords(ext);
                end;
            end else if (FrmMain.rbCCapitalizeSentence.Checked) then begin
                name := lowercase(name);
                if (name <> '') then begin
                    name[1] := ansiuppercase(name)[1];
                end;

                ext := lowercase(ext);
            end else if (FrmMain.rbCTitleCase.Checked) then begin
                name := self.TitleCase(name);
                ext := self.TitleCase(ext)
            end;
            if frmmain.cbCRoman.checked then begin
                name := self.RomanNumberals(name);
            end;


            if frmmain.cbCPreserverExtension.Checked then begin
                ext := s;
            end;
        end else begin
            if frmmain.cbOverrideCase.checked then begin
                if frmmain.rbCUppercase.checked then begin
                    name := ansiuppercase(name);
                    ext := ansiuppercase(ext);
                end else if frmmain.rbCLowercase.checked then begin
                    name := ansilowercase(name);
                    ext := ansilowercase(ext);
                end else begin
                    name := UnitToken.CapitalizeWords(name);
                end;
            end;
        end;
    end;
    procedure HandleCollisionsOld;
    begin
        if (not FrmMain.GetScriptingInterface) then begin
            if FrmCollisionDialog.Exists(IncludeTrailingPathDelimiter(frmmain.GetPath) + newname) then begin
                if (frmmain.cbCollision.Checked) then begin
                    curtd.IsCollision := false;
                    appendCount := 0;
                    {if frmmain.rbStartCountingAt.checked then begin
                        appendcount := SafeInt(frmmain.txtStartCountingAt.text) - 1;
                    end;
                    }
                    //move(rec,temprec,sizeof(rec));
                    temprec.FullName := rec.FullName;
                    temprec.name := rec.name;
                    temprec.shortname := rec.shortname;
                    temprec.ext := rec.ext;
                    temprec.listposition := 0;
                    temprec.listtotal := rec.listtotal;
                    temprec.selectedposition := 0; //
                    temprec.selectedtotal := rec.selectedtotal;
                    temprec.foldertotal := rec.foldertotal;
                    temprec.incrementby := rec.incrementby;
                    temprec.filedate := rec.filedate;
                    temprec.filesize := rec.filesize;
                    temprec.Created := rec.created;
                    temprec.Modified := rec.Modified;

                    repeat
                        inc(appendCount);
                        temprec.selectedposition := appendCount;
                        temprec.listposition := appendCount;
                        s := name + ReplaceMetaVariables(
                            frmmain.txtCollision.text,
                            temprec
                        ) + ext;

                        b := FrmCollisionDialog.Exists( IncludeTrailingPathDelimiter(frmmain.GetPath) + path + s);
                    until (not b)
                        or (not UnitUtil.ContainsTerm('?',frmmain.txtCollision.text))
                        or (appendCount > 50000);

                    newname := path + s;
                end;
            end;
        end;
    end;
var cr : TCursor;
    oldname, oldext : string;
begin
    Application.ProcessMessages;
    if showPreviewRunning then EXIT;

    try
    showPreviewRunning := true;

    replaceNum := TStringList.Create;
    replaceNum.Duplicates := dupError;
    replaceSeq := SafeInt(FrmMain.txtStartCountingAt.text,1);

    //try
    if frmmain.rbRegExp.Checked then begin
        rwc := TRegExpWildcards.Create;
    end else begin
        rwc := TWildcards.Create;
    end;
    DisableCollisions := FrmConfig.GetDisableCollisionDetection;


    CorrectTextNumbers;
    DoCancelRefreshFileList := false;
    selcount := 0;
    ChangedCount := 0;
    CollisionCount := 0;
    IllegalCount := 0;

    rec.listtotal := DataList.Count;
    rec.selectedtotal := 0;
    rec.changedcount := 1; // the startcounting and increment need positions to be 1 based, not zero
    rec.incrementby := SafeInt(frmmain.txtIncrementBy.Text, 1);
    rec.selectedtotal := MyCheckedCount;
    FrmCollisionDialog.Clear(DataList.Count*3);

    lastPath := '';
    path := '';
    lastsequence := '';

    frmmain.ProgressBar1.Visible := true;
    FrmMain.ProgressBar1.Min := 0;
    frmmain.ProgressBar1.Max := 95; // KLLUDGE: bar doesn't update with data often enough
    FrmMain.ProgressBar1.Smooth := true;
    frmmain.ProgressBar1.Refresh;
    Application.ProcessMessages;

    TextFileLoad;


    GetFileList.Items.BeginUpdate;
    k := min(50, trunc(datalist.Count / 10)+1);
    for i := 0 to DataList.Count - 1 do begin
        if DoCancelRefreshFileList then BREAK;

        j := i mod k;
        if (j = 0) then begin
            FrmMain.ProgressBar1.Position := trunc(100 * (i/max(1,(datalist.count-1))));
            frmmain.ProgressBar1.Invalidate;

            if ((i mod REPORT_AT)  = 0) then begin
                frmmain.SetStatusText(
                    'Preview ... ['+inttostr(i) + '/' + IntToStr(DataList.Count) +']',
                    TRUE
                )
            end;
        end;

        curtd := FrmDummyShowPreview.GetTagData(i);
        curtd.IsCollision := false;
        curtd.OriginalCollision := false;

        fld := self.GetFileData(i);

        if not (GetFileList.Items[i].Checked) then begin
            // add current name to collision detection list
            // blank the new name
            // add blank custom data
            ColumnNameData := UnitColumnManager.GetName(i);

            name := ColumnNameData;
            l := FrmCollisionDialog.Add(IncludeTrailingPathDelimiter(frmmain.GetPath) + name,
                IncludeTrailingPathDelimiter(frmmain.GetPath) +  name, i);
            if l <> -1 then begin
                curtd.IsCollision := true;
                td := FrmDummyShowPreview.GetTagData(l);
                td.IsCollision := true;
                inc(self.CollisionCount);
            end;
            UnitColumnManager.SetNewName(i,'');

            fld.CustomData.Clear;
            for k := 0 to UnitColumnManager.CustomColumnCount-1 do begin
                fld := self.GetFileData(i);
                fld.CustomData.Add( '' );
            end;
            CONTINUE;
        end;

        //============================
        // Retreive renaming record
        // - data generatation
        //============================
        GetRecordFor(rec, i, selcount, curtd, fld);

        lastPath := path;
        path := fld.GetBaseFolder;
        UpdateSelectCount(selcount, (lastPath<>path));
        rec.selectedposition := selcount;
        TextFileGetVariables(rec);

        //====================================
        // the meat of the renaming engine
        // NEW NAME generation
        // CUSTOM COLUMN generation
        //====================================
        name := rec.name;
        ext := '.' + rec.ext;

        if FrmMain.GetScriptingInterface then begin
            oldname := name;
            oldext := ext;
            rec.ScriptChanged := false;

            // Count's need to be 1 based
            if (lastPath<>path) and (FrmMain.cbClearCount.Checked) then begin;
                //rec.changedcount := SafeInt(FrmMain.txtStartCountingAt.text,1)
                rec.changedcount := 1;
            end;

            FrmScriptActions.RunScriptOn(name,ext,rec, curTD, collisionIndex);

            if (collisionIndex <> -1) then begin
                miscTD := self.GetTagData(collisionIndex);
                miscTD.IsCollision := false;
                miscTD.OriginalCollision := true;
            end;
            if (oldname<>name) or (oldext<>ext) or rec.ScriptChanged then begin
                Inc(rec.changedcount, 1);
            end;
        end else begin
            ApplyRenamingOptions;
        end;

        GenerateCustomColumns(fld);
        AlterCase;


        newname := path + name + ext;
        if AnsiEndsText('.', newname) then begin
            newname := leftstr(newname,length(newname)-1);
        end;
        // don't show subfolder durring a recursive scan in the new name
        // but allow newly created folder names.

        // update the column sizes for items changed during a preview
        HandleCollisionsOld;

        // format newname with partial folders
        newname := stringreplace(
            IncludeTrailingPathDelimiter(frmmain.GetPath) + newname,
            ExtractFilePath(rec.FullName),
            '', [rfIgnoreCase]);
        UnitColumnManager.SetNewName(i, newname);

        curTD.IsIllegalNewname := IsIllegalFilename(newname);
        if curTD.IsIllegalNewname then inc(IllegalCount);

        ColSize[1] := max(ColSize[1], GetFileList.Canvas.TextWidth(newname)+16);
        fld := self.GetFileData(i);
        for l := 0 to fld.CustomData.count - 1 do begin
            ColSize[6+l] := max(ColSize[6+l], GetFileList.Canvas.TextWidth( fld.CustomData[l] )+16);
        end;

        l := FrmCollisionDialog.Add(
            rec.FullName,
            IncludeTrailingPathDelimiter(frmmain.GetPath) + UnitColumnManager.GetNewName(i),
            i
        );
        if (l<>-1) then begin
            curtd.IsCollision := true;
            td := self.GetTagData(l);
            td.IsCollision := true;
            inc(self.CollisionCount);
        end;

        if (ExtractFilename(rec.FullName)) <> (newname) then begin
            inc(ChangedCount);
            curtd.changed := true; // currently unused
        end else begin
            curtd.changed := false;
        end;
        name := '';
        ext := '';
    end;


    //
    // optionally fix the original collision file name
    if (frmmain.GetScriptingInterface) and not DoCancelRefreshFileList then begin
        selcount := 0;
        lastPath := '';
        path := '';
        for i := 0 to DataList.Count - 1 do begin
        	if ((i mod REPORT_AT)  = 0) then begin
                frmmain.SetStatusText(
                    'Detecting Collisions.... ['+inttostr(i) + '/' + IntToStr(DataList.Count) +']',
                    TRUE
                );
            end;
            if not (GetFileList.Items[i].Checked) then CONTINUE;


            fld := self.GetFileData(i);

            lastPath := path;
            path := fld.GetBaseFolder;
            UpdateSelectCount(selcount, (lastPath<>path));

            td := self.GetTagData(i);
            if td.OriginalCollision then begin
                GetRecordFor(rec, i, selcount, td, fld);
                rec.selectedposition := selcount;

                TextFileGetVariables(rec);
                name := rec.name;
                ext := '.' + rec.ext;
                FrmScriptActions.RunScriptOn(name,ext,rec, td, collisionIndex);
                newname := path + name + ext;
                if AnsiEndsText('.', newname) then begin
                    newname := leftstr(newname,length(newname)-1);
                end;
                AlterCase;
                newname := stringreplace(
                    IncludeTrailingPathDelimiter(frmmain.GetPath) + newname,
                    ExtractFilePath(rec.FullName),
                    '', [rfIgnoreCase]);
                UnitColumnManager.SetNewName(i, newname);
            end;
        end;
    end;

    GetFileList.Items.EndUpdate;
    FrmCollisionDialog.EndUpdate;
    frmmain.ProgressBar1.Visible := false;
    frmmain.SetSubStatusText('', false);
    if rwc is TRegExpWildcards then begin
        i := TRegExpWildcards(rwc).LastError;
        if (i <> 0) and (i <> 1003) then begin
            frmmain.SetSubStatusText(
                ' Error in Regular Expression: ' + IntToStr(i),
                false
            );
        end;
    end;
    rwc.Free;
    self.SelectedInitFromControl;
    replaceNum.Free;

    FrmMain.HighlightCollisions1.Enabled := CollisionCount > 0;
    frmmain.HighlightIllegal1.Enabled := IllegalCount > 0;

    self.SelectedRedraw;
    finally
        showPreviewRunning := false;
    end;
end;





procedure TFrmDummyShowPreview.FormCreate(Sender: TObject);
begin
    self.tagdata := TObjectList.Create();
    self.wc := TWildCards.Create();
    self.DataList := TObjectList.Create();
    self.DataList.OwnsObjects := true;
    self.SelectedList := TObjectList.Create();
    self.SelectedList.OwnsObjects := false;


//    self.RefreshDriveIcons;
end;

function TFrmDummyShowPreview.GetTagData(li : TListItem): TTagData;
var i : integer;
begin
    i := Integer(li.data);
    if i >= tagdata.count then begin
        showmessage('Unexpected tagdata error');
        result := nil;
        EXIT;
    end;

    result :=TTagData(self.tagdata.Items[ i ]);
end;
function TFrmDummyShowPreview.GetTagData(index: integer): TTagData;
var lv : TListView;
begin
    lv := GetFileList;

    result := GetTagData(lv.items[index]);


end;
function TFrmDummyShowPreview.GetTagDataAbsolute(index: integer): TTagData;
begin
    result :=TTagData(self.tagdata.Items[ index ]);
end;
function TFrmDummyShowPreview.IsFolder(li: TListItem): boolean;
var td : TTagData;
begin
    td := self.GetTagData(li);

    result := td.IsFolder;
end;
function TFrmDummyShowPreview.ListContainsJPEGs: boolean;
begin
    result := self.containsJPEGs;
end;
function TFrmDummyShowPreview.ListContainsMP3s: boolean;
begin
    result := self.containsMP3s;
end;
procedure TFrmDummyShowPreview.CancelRefreshFileList;
begin
    self.DoCancelRefreshFileList := true;
    application.ProcessMessages;
end;
function TFrmDummyShowPreview.GetCancelRefresh : boolean;
begin
    result := self.DoCancelRefreshFileList;
end;



function TFrmDummyShowPreview.GetReplacementSequence(numbers: string): string;
begin
   result := '';
    if (numbers <> '') then begin
        result := replacenum.Values[numbers];
        if (result = '') then begin
            result := IntToStr(replaceSeq);
            replacenum.Values[numbers]  := result;
            inc(replaceSeq, SafeInt(frmMain.txtIncrementBy.text,1));
        end;
    end;
end;




procedure TFrmDummyShowPreview.RebuildIcons;
var i : integer;
    li : TListItem;
    fld : TFileListData;
begin
    if self.IgnoreInterface then EXIT;
    self.IgnoreInterface := true;

    UnitFileIcon.ClearCache;

    try
        ILDynamic.Clear;

        ILDynamic.AddImages(ilFolderBrowser);
        if frmmain.thumbnaillist1.Checked then begin
            FrmDummyThumbExtract.Show;
        end;

        GetFileList.Items.Count := 0;
        for i := 0 to DataList.Count - 1 do begin
            li := GetFileList.items[i];
            fld := FrmDummyShowPreview.GetFileData(li);
            fld.ImageIndex :=  ExtractIconCached(ildynamic, IncludeTrailingPathDelimiter(frmmain.GetPath) +  fld.caption);

            if FrmDummyThumbExtract.Visible then begin
                Application.ProcessMessages;
            end;
        end;
        GetFileList.Items.Count := DataList.Count;

        if (not FrmDummyThumbExtract.WasHidePressed) or (not frmmain.thumbnaillist1.Checked) then begin

        end;
        if frmmain.thumbnaillist1.Checked then begin
            FrmDummyThumbExtract.Hide;
        end;
    finally
        self.IgnoreInterface := false;
    end;
end;
procedure TFrmDummyShowPreview.AssignIcon(li: TListItem);
begin
end;



procedure TFrmDummyShowPreview.PopulateListItem(li: TListItem);
begin
    PopulateListItem(li, li.index);
end;
procedure TFrmDummyShowPreview.PopulateListItem(li: TListItem; index: integer);
var fld : TFileListData;
    i : integer;
begin
    if index >= datalist.count then EXIT;

    fld := TFileListData( datalist[index] );
    fld.index := index;
    li.Caption := fld.Caption;
    li.ImageIndex := fld.ImageIndex;
    li.Data := fld.data;

    for i := 0 to fld.SubItems.Count - 1 do begin
        li.SubItems.Add(fld.subitems[i]);
    end;

    for i := 0 to fld.CustomData.count  - 1 do begin
        li.SubItems.Add( fld.CustomData[i] );
    end;

    if fld.Checked then li.StateIndex := 1
    else li.StateIndex := 0;
    li.Checked := li.StateIndex = 1;
end;
procedure TFrmDummyShowPreview.UpdateFileData(li: TListItem);
  var fld : TFileListData;

    index : integer;
begin
    index := li.index;
    if index >= datalist.count then EXIT;

    fld := TFileListData( datalist[index] );
    fld.Caption := li.Caption;
    fld.ImageIndex := li.ImageIndex;
    fld.Data := li.data;

    //li.StateIndex := fld.StateIndex;
    fld.SetNewName(li.SubItems[0]);
    fld.SetSize(li.subitems[1]);
    fld.SetType(li.subitems[2]);
    fld.SetDate(li.subitems[3]);
    fld.SetDateCreated(li.SubItems[4]);
     
    fld.Selected := li.Selected; //li.Selected or (li.index = self.selecteindex);
    //li.SubItems.AddStrings(fld.SubItems);
    if li.Checked then fld.StateIndex := 1
    else fld.StateIndex := 0;
    fld.Checked := fld.StateIndex = 1;
    //if li.checked then li.stateindex := 1;

    GetFileList.UpdateItems(li.Index, li.index);
    li := GetFileList.ItemFocused;
    if li <> nil then
        GetFileList.UpdateItems(li.Index, li.index);
end;


procedure TFrmDummyShowPreview.FindListItem(Find: TItemFind;
  const FindString: String; const FindPosition: TPoint; FindData: Pointer;
  StartIndex: Integer; Direction: TSearchDirection; Wrap: Boolean;
  var Index: Integer);
var
    i : integer;
    found: boolean;
    fld : TFileListData;
begin

    i := StartIndex;
    if (Find = ifExactString) or (Find = ifPartialString) then  begin
    repeat
        if (i = (datalist.Count)) then begin
            if not Wrap then EXIT;
            i := 0
        end;

        fld := TFileListData(DataList[i]);
        found := Pos(UpperCase(FindString), UpperCase( fld.caption )) = 1;
        Inc(i);
    until Found or (I = StartIndex);

    if Found then index := (i-1);
  end;


end;
procedure TFrmDummyShowPreview.StateChangeItems(StartIndex,
  EndIndex: Integer; OldState, NewState: TItemStates);
begin
   if StartIndex >= DataList.Count then EXIT;
   if EndIndex >= DataList.count then EXIT;
  {
    for i := StartIndex to EndIndex do begin
        fld := TFileListData(DataList[i]);
        td := TTagData(self.tagdata.Items[ Integer(fld.data)]);

        fld.Selected := isSelected in OldState;
        td.DirectSelected := true;
    end;
    }
    Self.SelectedInitFromControl;
end;

function TFrmDummyShowPreview.GetFileData(li: TListItem): TFileListData;
begin
    result := nil;
    if li.Index >= datalist.count then EXIT;

    result := GetFileData(li.index);
end;
function TFrmDummyShowPreview.GetFileData(i : integer) : TFileListData;
begin
    result := TFileListData(datalist[i]);
end;
procedure TFrmDummyShowPreview.DeleteListItem(index : integer);
var fld : TFileListData;
begin
    if index >= DataList.Count then EXIT;

    fld := self.GetFileData(index);
    if fld.Checked then begin
        dec(self.MyCheckedCount)
    end;
    DataList.Delete(index);
    GetFileList.Items.count := DataList.count;
end;


{ FileListData }
constructor TFileListData.Create;
begin
    self.caption := '';
    self.ImageIndex := 0;
    self.StateIndex := 0;
    self.Checked := false;

    SubItems := TStringList.Create;
    SubItems.Add('');
    SubItems.Add('');
    SubItems.Add('');
    SubItems.Add('');
    SubItems.Add('');

    self.SetSize('0');
    customdata := TStringList.Create;
end;

function TFileListData.GetDateCreated: string;
begin
    result := subitems[integer(COL_TYPE_DATECREATED) - 1];
end;
function TFileListData.GetAltName: string;
begin
    result := self.altname;
end;
function TFileListData.GetDate: string;
begin
    result := subitems[integer(COL_TYPE_DATE) -1];
end;
function TFileListData.GetName: string;
begin
    result := caption;
end;
function TFileListData.GetNewName: string;
begin
    result := subitems[integer(COL_TYPE_NEWNAME) - 1];
end;
function TFileListData.GetNewNameWithDir : string;
begin
    result := basefolder + GetNewName;
end;
function TFileListData.GetSize: string;
begin
    result := subitems[integer(COL_TYPE_SIZE) - 1];
end;
function TFileListData.GetType: string;
begin
    result := subitems[integer(COL_TYPE_TYPE) - 1];
end;

procedure TFileListData.SetDate(s: string);
begin
    subItems[integer(COL_TYPE_DATE)-1] := s;
end;
procedure TFileListData.SetDateCreated(s: string);
begin
    subItems[Integer(COL_TYPE_DATECREATED)-1] := s;
end;
procedure TFileListData.SetName(s: string);
begin
    Caption := s;
    s := ExtractFilePath(s);
    if s <> '' then IncludeTrailingPathDelimiter( s );
    basefolder := s;
end;
procedure TFileListData.SetNewName(s: string);
begin
    SubItems[integer(col_TYPE_NEWNAME)-1] := s;
end;
procedure TFileListData.SetSize(s: string);
begin
    SubItems[integer(COL_TYPE_SIZE)-1] := s;
end;
procedure TFileListData.SetType(s: string; isDir: boolean);
begin
    if isDir then
        SubItems[integer(COL_TYPE_TYPE)-1] := FOLDER_TYPE_STRING
    else
        SubItems[integer(COL_TYPE_TYPE)-1] := s;
end;
function TFileListData.IsFolder : boolean;
begin
    result := GetType = FOLDER_TYPE_STRING;
end;
function TFileListData.GetBaseFolder : string;
begin
    result := basefolder;
end;



procedure TFrmDummyShowPreview.SwapItems(index1, index2: integer);
begin
    if (index1 > datalist.count-1) or (index2 > datalist.Count-1) then EXIT;

    datalist.Move(index1, index2);
    GetFileList.UpdateItems(index1,index1);
    GetFileList.UpdateItems(index2,index2);
end;

procedure TFrmDummyShowPreview.MoveUp;
var li, li2 : TListItem;
    i : integer;
begin
    if (GetFileList.SelCount = 1) then begin
        li := GetFileList.Selected;

        if (li <> nil) and (li.index >0) then begin
                datalist.Move(li.index-1, li.index);
                //dec(selecteindex);

                GetFileList.UpdateItems(li.Index - 1, li.index);

                li.Selected := false;
                li.Focused := false;
                li2 := GetFileList.Items[li.index-1];


                GetFileList.Selected := li2;
                GetFileList.ItemFocused := li2;
                li2.Selected := true;
                li2.Focused := false;

                GetFileList.UpdateItems(li.Index - 1, li.index);
        end;
    end else begin
        if GetFileList.Items[0].Selected then EXIT;

        for i := 1 to DataList.count - 1 do begin
            li  := GetFileList.Items[i];

            if li.Selected then begin
                // for some reason ANY deviation in order makes
                // this not work - don't change it!
                datalist.Move(li.index-1, li.index);

                GetFileList.UpdateItems(li.Index - 1, li.index);

                li.Selected := false;
                li.Focused := false;

                li2 := GetFileList.Items[li.index-1];
                GetFileList.Selected := li2;
                GetFileList.ItemFocused := li2;
                li2.Selected := true;
                li2.Focused := false;

                GetFileList.UpdateItems(li.Index - 1, li.index);
            end;
        end;
    end;
    self.AfterMove;
    if frmmain.lvFileList.selected <> nil then
        frmmain.lvFileList.selected.MakeVisible(false);
    
end;
procedure TFrmDummyShowPreview.MoveDown;
var li, li2 : TListItem;
    i : integer;
begin
    if (GetFileList.SelCount = 1) then begin
        li := GetFileList.Selected;
        if (li <> nil) and (li.index < DataList.count-1) then begin
            datalist.Move(li.index, li.index+1);
            //dec(selecteindex);

            GetFileList.UpdateItems(li.Index + 1, li.index);

            li.Selected := false;
            li.Focused := false;
            li2 := GetFileList.Items[li.index+1];

            GetFileList.Selected := li2;
            GetFileList.ItemFocused := li2;
            li2.Selected := true;
            li2.Focused := false;

            GetFileList.UpdateItems(li.Index, li.index+1);
        end;
    end else begin
        if GetFileList.Items[Datalist.count-1].Selected then EXIT;

        for i := DataList.count - 2 downto 0 do begin
            li  := GetFileList.Items[i];

            if li.Selected then begin
                // for some reason ANY deviation in order makes
                // this not work - don't change it!
                datalist.Move(li.index+1, li.index);

                GetFileList.UpdateItems(li.Index +1, li.index);

                li.Selected := false;
                li.Focused := false;

                li2 := GetFileList.Items[li.index+1];
                GetFileList.Selected := li2;
                GetFileList.ItemFocused := li2;
                li2.Selected := true;
                li2.Focused := false;

                GetFileList.UpdateItems(li.Index +1, li.index);
            end;
        end;
    end;
    self.AfterMove;
    if frmmain.lvFileList.selected <> nil then
        frmmain.lvFileList.selected.MakeVisible(false);
    
end;
procedure TFrmDummyShowPreview.MoveBottom(BottomIndex :integer = 0);
var li, li2 : TlistItem;
    i, j : integer;
    ol : TObjectList;
    fld : TFileListData;
begin
    if BottomIndex = 0 then
        BottomIndex := DataList.Count-1;

    if GetFileList.SelCount = 1 then begin
        li := GetFileList.Selected;
        if (li <> nil) and (li.index < BottomIndex) then begin
            datalist.Move(li.index, BottomIndex);
            //dec(selecteindex);

            GetFileList.UpdateItems(li.index, BottomIndex);

            li.Selected := false;
            li.Focused := false;
            li2 := GetFileList.Items[BottomIndex];

            GetFileList.Selected := li2;
            GetFileList.ItemFocused := li2;
            li2.Selected := true;
            li2.Focused := false;

            GetFileList.UpdateItems(BottomIndex, BottomIndex);
        end;
    end else begin
        ol := TObjectList.Create;

        j := BottomIndex;
        if GetFileList.Items[j].Selected then EXIT;

        for i := DataList.count - 2 downto 0 do begin
            li  := GetFileList.Items[i];

            if li.Selected then begin
                li.Selected := false;

                fld := TFileListData(DataList.Extract(DataList[i]));
                ol.Add(fld);
            end;
        end;
        j := ol.Count;
        for i := ol.count - 1 downto 0 do begin
            DataList.Insert(BottomIndex-(j-1), ol.Extract(ol[i]) );
        end;
        for i := BottomIndex downto BottomIndex - j do begin
            li := GetFileList.Items[i];
            li.Selected := true;
        end;


        GetFileList.Items[BottomIndex].MakeVisible(false);
        GetFileList.Invalidate;
    end;
    self.AfterMove;
    if frmmain.lvFileList.selected <> nil then
        frmmain.lvFileList.selected.MakeVisible(false);
    
end;
procedure TFrmDummyShowPreview.MoveTop(TopIndex : integer = 0);
var li, li2 : TlistItem;
    i, j : integer;
    ol : TObjectList;
    fld : TFileListData;

begin
    if GetFileList.SelCount = 1 then begin
        li := GetFileList.Selected;
        if (li <> nil) and (li.index >0) then begin
            datalist.Move(li.index,TopIndex);
            //dec(selecteindex);

            GetFileList.UpdateItems(TopIndex,li.index);

            li.Selected := false;
            li.Focused := false;
            li2 := GetFileList.Items[TopIndex];

            GetFileList.Selected := li2;
            GetFileList.ItemFocused := li2;
            li2.Selected := true;
            li2.Focused := false;

            GetFileList.UpdateItems(TopIndex, TopIndex);
        end;
    end else begin
        ol := TObjectList.Create;


        if GetFileList.Items[0].Selected then EXIT;

        for i := DataList.count - 1 downto 1 do begin
            li  := GetFileList.Items[i];

            if li.Selected then begin
                li.Selected := false;

                fld := TFileListData(DataList.Extract(DataList[i]));
                ol.Add(fld);
            end;
        end;
        j := ol.count;
        while ol.Count > 0 do begin
            DataList.Insert(TopIndex, ol.Extract(ol[0]) );
        end;

        for i := topindex + (j - 1) downto topindex do begin
            li := GetFileList.Items[i];
            li.Selected := true;
        end;


        GetFileList.Items[datalist.count-1].MakeVisible(false);
        GetFileList.Invalidate;
    end;

    self.AfterMove;
    if frmmain.lvFileList.selected <> nil then
        frmmain.lvFileList.selected.MakeVisible(false);
    
end;
procedure TFrmDummyShowPreview.AfterMove;
var
    liSelected, liFocused : TListItem;
begin
    liSelected := GetFileList.Selected;
    liFocused := getfilelist.ItemFocused;

    if (liFocused = nil) then begin
        getFileList.ItemFocused := liSelected;
    end;
end;


procedure TFrmDummyShowPreview.SortData;
begin
    DataList.Sort(CompareData);
    GetFileList.Invalidate;
end;





function TFrmDummyShowPreview.RomanNumberals(ins: string): string;
var wc, wc2 : TRegExpWildcards;
    x1,x2, a1,a2 : cardinal;
    s, s2  : string;
    b, roman : boolean;
    i : integer;
begin

    result := ins;
    wc := TRegExpWildcards.Create;
    wc2 := TRegExpWildcards.Create;
    wc.SetSearchString(ins);
    wc.SetSearchPattern('[^A-Za-z]*[A-Za-z]+[^A-Za-z]*');     // letters and following delimiters
    if wc.FindFirst(x1,x2) then begin
        result := '';
        // detect case where file begins with non-letters

        repeat
            s := wc.GetStr(x1,x2);
            wc2.SetSearchString(s);
            wc2.SetSearchPattern('[A-Za-z]+'); // letter parts only
            if wc2.FindFirst(a1,a2) then begin
                s2 := lowercase(wc2.GetStr(a1,a2));
                // detect non roman numerals
                roman := true;
                for i := 1 to length(s2) do begin
                    if not ( charinset(s2[i], ['i','v','x','l','c','d','m'])) then begin
                        roman := false;
                        BREAK;
                    end;
                end;

                if roman then s := UpperCase(s);
                
            end;
            result := result + s;

            b :=  wc.FindNext(x1,x2);
        until not b;
    end;
    wc.Free;
    wc2.free;
end;
function TFrmDummyShowPreview.TitleCase(ins: string): string;
var wc, wc2 : TRegExpWildcards;
    x1,x2, a1,a2 : cardinal;
    s, s2  : string;
    first, b : boolean;
const str_articles = ' a an the ';
      str_preps = ' amid as at atop but by down for from in into like near next of off on onto out over past per plus than to up upon via with ';
      str_conjunctions = ' and both even if nor not now once only or so that till when yet ';
      str_ignored = str_articles + str_preps + str_conjunctions;
begin
    result := ins;

    if frmmain.cbSplit.checked then begin
        ins := UnitToken.CapitalizeWords(ins, frmmain.txtSplit.text);
    end else begin
        ins := UnitToken.CapitalizeWords(ins);
    end;
    wc := TRegExpWildcards.Create;
    wc2 := TRegExpWildcards.Create;
    wc.SetSearchString(ins);
    wc.SetSearchPattern('[^A-Za-z]*[A-Z]+[^A-Z]*'); // non-letters, Capital, anything but capitals
    first := true;
    if wc.FindFirst(x1,x2) then begin
        result := wc.GetStr(x1,x2); // always accept first capital word

        if x1 <> 1 then begin
            result := wc.GetStr(0,x1-1) + result;
        end;

        repeat
            s := wc.GetStr(x1,x2);
            b :=  wc.FindNext(x1,x2);
            if not b then begin
                if not first then begin
                    result := result + s; // always accept last capital word
                end;
            end else if not first then begin
                wc2.SetSearchString(s);
                wc2.SetSearchPattern('[A-Za-z]+');
                if wc2.FindFirst(a1,a2) then begin
                    s2 := lowercase(wc2.GetStr(a1,a2));
                    if ContainsTerm(' ' + s2 + ' ',str_ignored) then begin
                        s[1] := lowercase(s[1])[1];
                    end;
                end;
                result := result + s;
            end;

            first := false;
        until not b;
    end;
    wc.Free;
    wc2.free;
end;

function TFrmDummyShowPreview.GetSelected(index: integer): boolean;
begin
    result := false;
    if index > DataList.count - 1 then EXIT;


    result := SelectedList.IndexOf(DataList[index]) <> -1;
end;
function TFrmDummyShowPreview.GetItemFromSelected(selindex: integer): TListItem;
var fld : TFileListData;
begin
	result := nil;
	if selindex >= selectedlist.count then EXIT;

    fld := TFileListData(SelectedList.Items[selindex]);
	result := TListItem( GetFileList.Items[fld.index]);
end;
procedure TFrmDummyShowPreview.ClearAllSelected(includeHovered : boolean = False);
begin
    SelectedList.clear;
    if includeHovered then begin
        GetFileList.ClearSelection;
    end;
end;
procedure TFrmDummyShowPreview.ClearSelected(index : integer);
begin
    if index > DataList.count - 1 then EXIT;

    SelectedList.Extract(DataList[index]);
    GetFileList.Items[index].Selected := false;
end;
procedure TFrmDummyShowPreview.SetSelected(index: integer);
begin
    if index > DataList.count - 1 then EXIT;
    if index < 0 then EXIT;


    SelectedList.Add(DataList[index]);
    GetFileList.Items[index].Selected := true;
end;
procedure TFrmDummyShowPreview.SetSelected(index : integer; value : boolean);
begin
    if index > DataList.count - 1 then EXIT;

    if value then
        SelectedList.Add(DataList[index])
    else
        SelectedList.Extract(DataList[index]);

    GetFileList.Items[index].Selected := value;
    GetFileList.UpdateItems(index,index);
end;
procedure TFrmDummyShowPreview.ToggleSelected(index: integer);
var sel : boolean;
begin
    if index > DataList.count - 1 then EXIT;

    //sel := GetFileList.items[index].Selected;
    sel := self.GetSelected(index);
    SetSelected(index, not sel);
end;
function TFrmDummyShowPreview.ToggleChecked(index: integer) : integer;
var fld : TFileListData;
begin
    result := -1;
    if index > DataList.count - 1 then EXIT;

    fld := TFileListData(DataList[index]);
    fld.Checked := not fld.checked;
    if fld.checked then begin
        fld.StateIndex := 1;
        result := 1;
    end else begin
        fld.StateIndex := 0;
        result := -1;
    end;

    inc(MyCheckedCount, result);

    GetFileList.UpdateItems(index, index);
end;
function TFrmDummyShowPreview.SetChecked(index: integer): integer;
var fld : TFileListData;
begin
    result := 0;
    if index > DataList.count - 1 then EXIT;

    fld := TFileListData(DataList[index]);
    if fld.checked = false then begin
        fld.Checked := true;
        fld.StateIndex := 1;
        result := 1;
    end else begin
        result := 0
    end;

    inc(MyCheckedCount, result);
    GetFileList.UpdateItems(index, index);
end;
function TFrmDummyShowPreview.GetChecked(index: integer): boolean;
var fld : TFileListData;
begin
    result := false;
    if index > DataList.count - 1 then EXIT;

    fld := TFileListData(DataList[index]);
    result := fld.checked;
end;
function TFrmDummyShowPreview.SetChecked(index: integer; value: boolean): integer;
var fld : TFileListData;
begin
    result := -1;
    if index > DataList.count - 1 then EXIT;

    fld := TFileListData(DataList[index]);
    if fld.Checked = value then EXIT;

    if value then begin
        fld.Checked := true;
        fld.StateIndex := 1;
        result := 1;
    end else begin
        fld.Checked := false;
        fld.StateIndex := 0;
        result := -1
    end;

    inc(MyCheckedCount, result);
    GetFileList.UpdateItems(index, index);
end;

var SelectToggleRunning : boolean;
function TFrmDummyShowPreview.SelectedToggleCheckboxes : integer;
var i : integer;
    li : TListItem;
    fld : TFileListData;
begin
    result := 0;
    if SelectToggleRunning then EXIT;
    try
        SelectToggleRunning := true;

        for i := 0 to SelectedList.Count - 1  do begin
            fld := TFileListData(SelectedList[i]);
            fld.Checked := not fld.checked;
            if fld.checked then begin
                fld.StateIndex := 1;
                inc(result);
            end else begin
                fld.StateIndex := 0;
                dec(result);
            end;
            li := GetFileList.Items[fld.index];
            li.Selected := true;
        end;

        self.SelectedRedraw;
        inc(MyCheckedCount, result);
    finally
        SelectToggleRunning := false;
    end;
end;


procedure TFrmDummyShowPreview.SelectedRedraw;
var i : integer;
//    li : TListItem;
    fld : TFileListData;
begin
    for i := 0 to SelectedList.Count - 1  do begin
        fld := TFileListData(SelectedList[i]);
        GetFileList.Items[fld.index].Selected := true;
        GetFileList.UpdateItems(fld.index,fld.index);
    end;
end;
procedure TFrmDummyShowPreview.SelectedRedrawFull;
var i : integer;
    li : TListItem;
begin
    // redraw all the unselected items instead of just
    // clearing the selection
    for i := 0 to getFileList.Items.count - 1 do begin
        li := getFileList.Items[i];
        if li.Selected then begin
            li.Selected := false;
            GetFileLIst.UpdateItems(i,i);
        end;
    end;
    self.SelectedRedraw;
end;
procedure TFrmDummyShowPreview.SelectedInitFromControl;
var i : integer;
    li : TListItem;
    td : TTagData;

begin
    SelectedList.clear;
    for i := 0 to GetFileList.Items.Count - 1 do begin
        li := GetFileList.Items[i];
        if li.Selected then begin
            td := self.GetTagData(i);
            td.DirectSelected := true;
            SelectedList.Add(DataList[i]);
        end;
    end;
    GetFileList.UpdateItems(0,DataList.Count-1);
end;

function TFrmDummyShowPreview.GetVirtualItem(index: integer): TVirtualListItem;
var li : TListItem;
begin
    if index > datalist.count-1 then EXIT;

    li := GetFileList.Items[index];
    result.index := li.index;
    result.selected := li.Selected;
    result.focused := li.Focused;
end;

function TFrmDummyShowPreview.SelectedMyCount: integer;
begin
    result := SelectedList.Count;
end;
function TFrmDummyShowPreview.GetCheckedCount: integer;
begin
    result := MyCheckedCount;
end;
function TFrmDummyShowPreview.GetCollisionCount: integer;
begin
    result := 0;
    if self.CollisionCount > 0 then
        result := self.CollisionCount + 1;
end;
function TFrmDummyShowPreview.GetIllegalCount: integer;
begin
    result := 0;
    if self.IllegalCount > 0 then
    	result := self.IllegalCount;
end;
function TFrmDummyShowPreview.GetColSize(index: integer): integer;
begin
    result := 0;
    if (index = 0) and (frmmain.thumbnaillist1.Checked) then begin
        inc(result, 64 + 30);
    end;
    inc(result, colsize[index]);
    //inc(result, frmmain.lvFileList.Columns[index].Width);
end;
function TFrmDummyShowPreview.GetDateStringFormatted(td: TTagData): string;
var i : integer;
begin
    i :=  td.modified;
    result := self.GetDateStringFormatted(i);
end;
function TFrmDummyShowPreview.GetDateStringFormatted(i : integer): string;
begin
    if (FrmConfig.UseFullDateTime) then begin
        //result := DateToStr(FileDateToDateTime(i)) + ' '
        //            + FormatDateTime('tt', FileDateToDateTime(i));
        result := FormatDateTime('c', FileDateToDateTime(i));
    end else begin
        result := DateToStr(FileDateToDateTime(i));
    end;

end;

end.
