unit UnitMyRichEdit;

interface

uses
    Windows, Messages, Classes, ComCtrls, RichEdit, SysUtils, VCL.Controls,
    Vcl.ExtCtrls, System.Generics.Collections, Graphics;


//
// -Fix problems with text being cleared when displayed
// -Format variables
// -Handle variable clicks and hovers
//
type TMyRichEdit = class(TRichEdit)
    private
        ignoreChanges : boolean;
        flastText : string;
        variablePositions : TList<TPoint>;
        variableCodes : TDictionary<string, string>;
        lastMoveTime : cardinal;
        lastX, lastY : integer;
        timHover : TTimer;
        myhint : THintWindow;
        originalWidth : integer;
        originalParentWidth : integer;
        disableFormatting : boolean;
        shiftDown : boolean;
        ctrlDown : boolean;
        LastSelStartClick : integer;
        isRegEx : boolean;
        fBackgroundColor : TColor;
    protected
        function ClientXToIndex(x : integer) : integer;
        function VariableAtClientX(x : integer) : string;
        function VariablePointFromClientX(x : integer) : TPoint;
        procedure FormatVariables;
        procedure InsertStream(const Stream : TStream);
        procedure timHoverFire(Sender: TObject);
        function GetPositionList(s : string) : TList<TPoint>;


        // event triggers
        procedure KeyDown(var Key: Word; Shift: TShiftState); override;
        procedure WmKeyUp(var Msg: TWMKeyUp); message WM_KEYUP;
        procedure WmKeyDown(var Msg: TWMKeyDown); message WM_KEYDOWN;
        procedure WMGetText(var msg: TWMGetText); message WM_GETTEXT;
        procedure WMGetTextLength(var Mgs: TWMGetTextLength); message WM_GETTEXTLENGTH;
        procedure CMMouseEnter(var msg: TMessage); message CM_MOUSEENTER;
        procedure CMMouseLeave(var msg: TMessage); message CM_MOUSELEAVE;
        procedure Change;  override;
        procedure MouseMove(shift : TShiftState; x, y : integer); override;
        procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
        procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
        procedure HandleClick(button : TMouseButton; X : Integer);
        procedure DoEnter; override;
        procedure DoExit; override;
        function GetSelText: string; override;


        procedure GrowIfNeeded(forceshrink : boolean = false);
        function SelectIndexIsVariable(index : integer) : boolean; overload;
        function SelectIndexIsVariable(index : integer; var pt : TPoint) : boolean; overload;
        function SelectIsVariableEnd(var pt:tpoint;UseSelectEnd:boolean=false) : boolean;
        function SelectIsVariableStart(var pt:tpoint;UseSelectEnd:boolean=false) : boolean;

        procedure SetSelection(pt : TPoint);
        procedure SetSelectionLength(pt : TPoint);
        procedure SetSelectionStartToX(pt : TPoint);
        procedure SetSelectionStartToY(pt : TPoint);
        function GetSelectionPoint : TPoint;
        function GetDisplayedText(pt : TPoint) : string; overload;
        function PointIsVariable(pt : TPoint) : boolean;

        function GetDisplayedSelText : string;
        function GetDisplayedText : string; overload;

        property lastText : string read flastText write flastText;

        procedure DblClick; override;
        procedure Click; override;

        function GetDelimiter : char;
        function GetTextCode(displaytext : string) : string;
    public
        procedure InsertLink(url, description : string);
        constructor Create(owner: TWinControl);

        function CaretInVariable : boolean;
        procedure HighlightVariableAroundCaret;
        function HighlightIsVariable : boolean;
        procedure DisableVariableFormatting;
        procedure EnableVariableFormatting;
        function getVariableFormatting : boolean;

        procedure DoRegExMode;

        property DisplayedText : string read GetDisplayedText;
        property VarDelimiter : char read GetDelimiter;
end;
// extends the SelAttributes class in TRichEdit
type
	  __TTextAttributes = class
	  private
	    RichEdit: TCustomRichEdit;
	    FType: TAttributeType;
	  end;
	  TTextAttributesH = class helper for TTextAttributes
	  private
	    function GetBackColor: TColor;
	    procedure SetBackColor(const Value: TColor);
	    procedure GetAttributes(var Format: TCharFormat2);
	    procedure SetAttributes(var Format: TCharFormat2);
	  public
	    property BackColor: TColor read GetBackColor write SetBackColor;
	  end;

implementation

uses
    UnitWildcards,  StrUtils, Math, UnitDebug, System.Types,
     Forms, System.Character, UnitMetaVariables, UnitUtil, Clipbrd;

const
    // nobreak space
    META_CHAR_UNI = $00A0;
    META_CHAR = '\x{00A0}';

constructor TMyRichEdit.Create(owner: TWinControl);
var mask : integer;
    i : integer;
begin
    // double buffer causes invisible text - hmmm
    inherited Create(owner);
    self.Parent := owner;
    self.Clear;
    self.PlainText := false;
    self.ParentShowHint := false;
    self.ShowHint := true;
    self.WantTabs := false;
    self.WantReturns := false;
    self.HideSelection := false;
    //
    // workarounds for issue were text is erased when showing the control
    // the first time
    //
    timHover := TTimer.Create(self);
    timHover.Enabled := false;
    timHover.Interval := 300;
    timHover.OnTimer := self.timHoverFire;

    myhint := THintWindow.Create(self);

    variablePositions := TList<TPoint>.Create;
    variableCodes := TDictionary<string,string>.Create;
end;

// Public Interface
//

procedure TMyRichEdit.DoRegExMode;
begin
    //A good idea that just breaks everything right now - need more research
    {
    isRegEx := true;

    ignoreChanges := true;
    self.EnableVariableFormatting;
    self.Text := VAR_REGEXP + self.text;
    self.FormatVariables;

    ignorechanges := false;
    }
end;
procedure TMyRichEdit.DisableVariableFormatting;
begin
    disableFormatting := true;
    self.fBackgroundColor := self.Color;
    self.Color := dimcolor(clWindow, 0.93);
end;
procedure TMyRichEdit.EnableVariableFormatting;
begin
    if disableFormatting then
        self.Color := self.fBackgroundColor;

    disableFormatting := false;
end;
function TMyRichEdit.getVariableFormatting : boolean;
begin
    result := not disableFormatting;
end;

function TMyRichEdit.HighlightIsVariable : boolean;
var
    pt : TPoint;
begin
    pt := self.GetSelectionPoint;
    result := self.PointIsVariable(pt)
end;
function TMyRichEdit.CaretInVariable : boolean;
begin
    result := SelectIndexIsVariable(self.SelStart)
end;
procedure TMyRichEdit.HighlightVariableAroundCaret;
var pt : TPoint;
begin
    if self.SelectIndexIsVariable(self.SelStart,pt) then begin
        self.SetSelection(pt);
    end;
end;
function TMyRichEdit.GetDelimiter : char;
begin
    result := Char(META_CHAR_UNI);
end;

// Translate string positions to selection positions (1based, 0based)
//
procedure TMyRichEdit.SetSelection(pt : TPoint);
begin
    self.SetSelectionStartToX(pt);
    self.SetSelectionLength(pt);
end;
procedure TMyRichEdit.SetSelectionLength(pt : TPoint);
begin
    self.SelLength := pt.Y-pt.X+1;
end;
procedure TMyRichEdit.SetSelectionStartToX(pt : TPoint);
begin
    self.SelStart := pt.X - 1;
end;
procedure TMyRichEdit.SetSelectionStartToY(pt : TPoint);
begin
    self.SelStart := pt.Y;
end;
function TMyRichEdit.GetSelectionPoint : TPoint;
begin
    result.X := self.SelStart+1;
    result.Y := max(0,result.X + self.SelLength-1);
end;

// add background color to the SelAttributes
//
procedure TTextAttributesH.GetAttributes(var Format: TCharFormat2);
var
    RichEdit: TCustomRichEdit;
begin
    RichEdit := __TTextAttributes(Self).RichEdit;
    InitFormat(Format);
    if RichEdit.HandleAllocated then begin
        SendGetStructMessage(
            RichEdit.Handle, EM_GETCHARFORMAT,
            WPARAM(__TTextAttributes(Self).FType = atSelected), Format, True
        );
    end;
end;
function TTextAttributesH.GetBackColor: TColor;
var
    Format: TCharFormat2;
begin
    GetAttributes(Format);
    with Format do begin
        if (dwEffects and CFE_AUTOBACKCOLOR) <> 0 then begin
            result := clWindow
        end else begin
            result := crBackColor;
        end;
    end;
end;
procedure TTextAttributesH.SetAttributes(var Format: TCharFormat2);
var
    Flag: Longint;
    RichEdit: TCustomRichEdit;
begin
    RichEdit := __TTextAttributes(Self).RichEdit;
    if __TTextAttributes(Self).FType = atSelected then begin
        Flag := SCF_SELECTION
    end else begin
        Flag := 0;
    end;
    if RichEdit.HandleAllocated then begin
        SendStructMessage(RichEdit.Handle, EM_SETCHARFORMAT, Flag, Format);
    end;
end;
procedure TTextAttributesH.SetBackColor(const Value: TColor);
var
    Format: TCharFormat2;
begin
    InitFormat(Format);
    with Format do begin
        dwMask := CFM_BACKCOLOR;
        if Value = clWindowText then begin
            dwEffects := CFE_AUTOBACKCOLOR
        end else begin
            crBackColor := ColorToRGB(Value);
        end;
    end;
    SetAttributes(Format);
end;

// auto-expand
// grow (if needed) when gets keyboard focus
// always try to shrink when loses keyboard focus
//
procedure TMyRichEdit.DoEnter;
begin
    inherited;
    perform(WM_HSCROLL,SB_LEFT,0);
    self.GrowIfNeeded;
end;
procedure TMyRichEdit.DoExit;
begin
    self.GrowIfNeeded(true);
end;
procedure TMyRichEdit.GrowIfNeeded(forceshrink:boolean=false);
var i,wd : integer;
begin
    if not self.Focused and not forceshrink then exit;

    wd := self.Width;
    if (originalWidth <> 0) then begin
        wd := originalWidth;
    end;
    i := ClientXToIndex(wd-10);
    if (not forceshrink) and
        (i <> 0) and (originalParentWidth=0) then begin
        originalParentWidth := self.Parent.Width;
        originalWidth := self.Width;

        perform(WM_HSCROLL,SB_LEFT,0);
        self.Parent.Width := self.Parent.Width + 90;
        self.Parent.BringToFront;
    end else if (forceshrink) or (i=0) then begin
        if (originalParentWidth<>0) then begin
            self.Parent.Width := originalParentWidth;
            originalWidth := 0;
            originalParentWidth := 0;
        end;
    end;
end;

// Hover Events
// show a friendly tooltip when over a variable
//
procedure TMyRichEdit.CMMouseEnter(var msg: TMessage);
begin
    inherited;
    timHover.enabled := true;
end;
procedure TMyRichEdit.CMMouseLeave(var msg: TMessage);
begin
    inherited;
    timHover.Enabled := false;
    self.Hint := '';
    lastX := -1;
    lastY := -1;
    myhint.ReleaseHandle;
end;
procedure TMyRichEdit.MouseMove(shift : TShiftState; x, y : integer);
begin
    inherited MouseMove(shift, x, y);

    if (lastX<>-1) and ( (x<>lastX) or (y<>lastY)) then begin
        timHover.enabled := false;
        timHover.enabled := true;
        myhint.ReleaseHandle;
    end;
    lastX := x;
    lastY := y;
end;
procedure TMyRichEdit.timHoverFire(Sender: TObject);
    function TextUnderCursor : string;
    begin
         result := self.VariableAtClientX(lastX);
    end;
var
    s, flags : string;
    re : TRegExpWildcards;
    pt : TPoint;
    x,y : cardinal;
    r : TRect;
    i : integer;
begin
    timHover.Enabled := false;
    s := self.VariableAtClientX(lastX);
    if s = '' then EXIT;

    re := TRegExpWildcards.Create;
    re.SetSearchString(s);
    re.SetSearchPattern('\:[^?]*');
    if re.FindFirst(x, y) then begin
        flags := lowercase(copy(s,X,Y-X+1));
    end;
    s := re.SearchReplaceWith('',true);


    myhint.Caption := s;
    if (flags <> '') and (length(flags)>1)then begin
        case flags[2] of
        'l': flags := 'left pad';
        'r': flags := 'right pad';
        'd': flags := 'decimal';
        'f': flags := 'formatted ' + midstr(flags,3,length(flags)-2) ;
        'c': flags := 'fixed date/time';
        'h': flags := 'hex';
        's': flags := 'substring';
        't': flags := 'as letters';
        end;
    end;
    if flags <> '' then begin
        myhint.Caption := myhint.Caption + ': ' + flags;
    end;
    pt := self.ClientToScreen(point(self.Left+lastX, self.Top+self.Height));
    r.Top := pt.Y;
    r.Left := pt.x;
    r.Bottom := pt.Y+myhint.Canvas.TextHeight('ZZ');
    r.Right := pt.X+myhint.Width;
    myhint.ActivateHint(
        r, myhint.Caption
    );
end;
// (hover helpers)
function TMyRichEdit.ClientXToIndex(x : integer) : integer;
var pt : TPoint;
begin
    pt := Point(x, self.Height div 2);
    result := Perform(Messages.EM_CHARFROMPOS, 0, Integer(@Pt));
    inc(result);
    if result > length(self.GetDisplayedText) then result := 0;
end;
function TMyRichEdit.VariablePointFromClientX(x : integer) : TPoint;
var
    charIndex : integer;
    i : integer;
    pt : TPoint;
begin
    result.x := 0;
    result.y := 0;
    charindex := ClientXToIndex(x);
    if charindex = 0 then EXIT;

    for i := 0 to variablePositions.Count -1  do begin
        pt := variablePositions[i];
        if (charIndex >= pt.X) and (charindex <=pt.Y) then begin
            result := pt;
            BREAK;
        end;
    end;
end;
function TMyRichEdit.VariableAtClientX(x : integer) : string;
var
    charIndex : integer;
    pointlist : TList<TPoint>;
    i : integer;
    pt : TPoint;
    s : string;
begin
    pt := self.VariablePointFromClientX(x);
    if pt.X=0 then EXIT;

    result := '';
    s := self.GetDisplayedText(pt);
    variableCodes.TryGetValue(s, result);
end;

//====================
// Format incoming text, remove special formatting for outgoing text
//====================
// (event triggers)
procedure TMyRichEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
    // handle Copy (C) and Cut (X)
    if (char(key) in ['C','X']) and (shift=[ssCtrl]) then begin
        clipboard.SetTextBuf(PChar(self.GetSelText));
        if char(key)='X' then
            SelText := '';

        key := 0;
    end;
end;
procedure TMyRichEdit.Change;
var s : string;
begin
    if ignoreChanges then begin
        s := '';
        EXIT;
    end;
    try
        ignoreChanges := true;

        self.GrowIfNeeded;
        if not disableFormatting then begin
            s := self.GetDisplayedText;
            if isRegEx then begin

                if not StrUtils.StartsText(VAR_REGEXP, s) then begin
                    self.DisableVariableFormatting;
                    self.isRegEx := false;
                    self.variablePositions.Clear;
                    self.variableCodes.Clear;

                end else begin
                    self.FormatVariables;
                end;
            end else begin
                self.FormatVariables;
            end;

        end;
        inherited;
    finally
        ignoreChanges := false;
    end;
end;
procedure TMyRichEdit.WMGetText(var msg: TWMGetText);
var
    s : string;
begin
    s := self.GetTextCode(self.GetDisplayedText);

    msg.Result := Min(Length(s), msg.TextMax);
    if (msg.Result > 0) then begin
        Move(s[1], msg.Text^, msg.Result * sizeof(s[1]));
    end;
end;
procedure TMyRichEdit.WMGetTextLength(var Mgs: TWMGetTextLength);
begin
    mgs.result := length( self.GetTextCode(self.GetDisplayedText) );
end;

function TMyRichEdit.GetTextCode(displaytext : string) : string;
var pt : TPoint;
    key, val : string;
    i : integer;
begin
    // replace friendly variables with meta-variable codes
    result := displaytext;
    if result = '' then EXIT;

    for key in variableCodes.Keys do begin
        if variableCodes.TryGetValue(key, val) then begin
            result := ReplaceAll(result, key, val);
        end;
    end;
    // non-varaibles
    result := ReplaceAll(result, char(META_CHAR_UNI), '?');
end;
function TMyRichEdit.GetSelText: string;
begin
    result := self.GetTextCode( self.GetDisplayedSelText );
end;
function TMyRichEdit.GetDisplayedSelText : string;
var
    len : Integer;
begin
    SetLength(result, self.GetSelLength + 1);
    len := self.Perform(EM_GETSELTEXT, 0, Longint(PChar(result)));
    SetLength(result, len);
end;
// (helpers for triggers)
function TMyRichEdit.GetPositionList(s : string) : TList<TPoint>;
var
    re : TRegExpWildcards;
    x, y : cardinal;
const
    RE_QUESTION_VAR = '\?{1}[^?'+META_CHAR+']+\?{1}';
    RE_META_VAR =  META_CHAR+'{1}[^?'+META_CHAR+'}]+'+META_CHAR+'{1}';
begin
    re := TRegExpWildcards.Create;
    re.SetSearchString(s);
    re.SetSearchPattern(
        RE_QUESTION_VAR+'|'+RE_META_VAR
    );

    result := self.variablePositions;
    result.Clear;
    if re.FindFirst(x, y) then begin
        repeat
            result.Add(point(x,y));
        until not re.FindNext(x,y);
    end;

    re.Free;
end;
function TMyRichEdit.GetDisplayedText : string;
var
    GetTextLengthEx: TGetTextLengthEx;
    GetTextEx: TGetTextEx;
    Len: Integer;
const
    UNICODE_PAGE = 1200;
begin
    result := '';
    GetTextLengthEx.flags := GTL_DEFAULT;
    GetTextLengthEx.codepage := UNICODE_PAGE;    // UNIC
    len := self.Perform(
        EM_GETTEXTLENGTHEX,
        WPARAM(@GetTextLengthEx), 0
    );

    if len=E_INVALIDARG then begin
        raise Exception.Create('EM_GETTEXTLENGTHEX failed');
    end;
    SetLength(result, len);
    if len=0 then EXIT;

    GetTextEx.cb := (Length(Result)+1)*SizeOf(WideChar);
    GetTextEx.flags := GTL_DEFAULT;
    GetTextEx.codepage := UNICODE_PAGE;
    GetTextEx.lpDefaultChar := nil;
    GetTextEx.lpUsedDefChar := nil;
    self.Perform(
        EM_GETTEXTEX,
        WPARAM(@GetTextEx), LPARAM(PWideChar(result))
    );
end;
function TMyRichEdit.GetDisplayedText(pt : TPoint) : string;
begin
    result := GetDisplayedText;
    result := MidStr(result, pt.x, pt.y-pt.x+1);
end;

procedure TMyRichEdit.FormatVariables;
var
    re : TRegExpWildcards;
    originalstart, updatedstart : integer;
    originallen : integer;
    originaltext, cut : string;
    s, varcode, chunk : string;
    i  : integer;
    pt : TPoint;
    SelectAtNewVariableEnd, SelStartAtEnd : boolean;
    offset : integer;
const
    META_CHAR_SIZE = 9;
    function IsSame : boolean;
    var
        i : integer;
    begin
        // workaround for stupid control trying to
        // change Unicode to ansi
        result := false;
        if lasttext = '' then EXIT;
        if (length(lastText) = length(lines[0])) then begin
            result := true;
            for i := 1 to length(lastText) do begin
                result := result and (
                    (lasttext[i] = lines[0][i]) or
                    ((lasttext[i]= char(META_CHAR_UNI)) and
                     (lines[0][i]=' '))
                );
            end;
        end;
    end;
    procedure DefaultFormat;
    begin
        SelStart := 0;
        SelLength := length(originaltext);
        SelAttributes.Name := 'Sans Serif';
        SelAttributes.Style := [];
        SelAttributes.Size := 10;
        SelAttributes.Color := clWindowText;
        SelAttributes.BackColor := clWindow;
    end;
    function GetFriendlyVariable(displaytext : string) : string;
    var
        s : string;
        id : integer;
    begin
        s := UnitMetaVariables.GetFriendlyName(displaytext);
        result := s;
        id := 2;
        while pos(
            char(META_CHAR_UNI) + result + char(META_CHAR_UNI),
            lines[0]
        ) <> 0 do begin
            result := s + IntToStr(id);
            inc(id);
        end;
        result := char(META_CHAR_UNI) + result + char(META_CHAR_UNI);
    end;
    procedure UnderlineLeadingTrailingSpace;
    var
        x,y : cardinal;
    begin
        re := TRegExpWildcards.Create;
        re.SetSearchPattern('^\s+|\s+$');
        re.SetSearchString(self.GetDisplayedText);
        if re.FindFirst(x, y) then begin
            repeat
                pt.x := x;
                pt.Y := y;
                self.SetSelection(pt);
                SelAttributes.Name := 'Sans Serif';
                SelAttributes.Style := [fsUnderline];
                SelAttributes.Color := clWindowText;
            until not re.FindNext(x,y);
        end;
    end;
    procedure FormatVariable(pt : TPoint);
    begin
        self.SetSelection(pt);
        SelAttributes.Name := 'Tahoma';
        SelAttributes.Size := 8;
        SelAttributes.Color := ColorToRGB(clBlack);
        SelAttributes.BackColor := dimcolor(self.Color, 0.92);
        SelAttributes.Style := [fsUnderline];
        self.SelStart := pt.X;
        self.SelLength := pt.Y-pt.X-1;

        self.SelLength := 0;
    end;
    function GetUpdatedPoint(positionIndex : integer; offset : integer) : TPoint;
    begin
        result := variablePositions.Items[i];
        inc(result.X,offset);
        inc(result.Y,offset);
    end;
begin
    // ignore if unchanged
    if (lines[0] = '') then begin
        lastText := lines[0];
    end;
    if (lines[0] = '') or IsSame then begin
        lines[0] := lastText;
        EXIT;
    end;

    originaltext := self.GetDisplayedText;
    originalstart := SelStart;
    originallen := SelLength;
    updatedstart := originalstart;
    SelStartAtEnd := length(originaltext) = originalstart;
    offset := 0;

    Lines.BeginUpdate;
    DefaultFormat;
    self.GetPositionList(originaltext);
    for i := 0 to (variablePositions.Count-1) do begin
        pt := GetUpdatedPoint(i, offset);
        chunk := self.GetDisplayedText(pt);
        variableCodes.TryGetValue(chunk, varcode);
        SelectAtNewVariableEnd := (originalstart) = pt.Y;

        if (chunk[1] = '?') then begin
            self.SetSelection(pt);
            s := GetFriendlyVariable(chunk);
            SelText := s;

            pt.Y := pt.X + (length(s) - 1);
            updatedstart := pt.Y;

            variableCodes.AddOrSetValue(s, chunk);
            inc(offset, length(s) - length(chunk));
        end;

        FormatVariable(pt);
        variablePositions.Items[i] := pt;
        if SelectAtNewVariableEnd then updatedstart := pt.Y;
    end;
    UnderlineLeadingTrailingSpace;

    SelStart := updatedstart;
    sellength := originallen;
    if (SelStartAtEnd) then begin
        SelStart := length(self.GetDisplayedText);
    end;

    lastText := lines[0];
    Lines.EndUpdate;
end;

// Treat variables as a single character for key navigation
// left/right/backspace/delete + shift events
//
function TMyRichEdit.SelectIsVariableStart(var pt:tpoint;UseSelectEnd:boolean=false) : boolean;
var
    offset :integer;
    pt2 : TPoint;
begin
    offset := 0;
    if UseSelectEnd then begin
        offset := SelLength;
    end;
    result := self.SelectIndexIsVariable(self.SelStart+offset, pt) and
        (self.SelStart+1=pt.X);
end;
function TMyRichEdit.SelectIsVariableEnd(var pt:tpoint;UseSelectEnd:boolean=false) : boolean;
var
    offset : integer;
    pt2 : TPoint;
begin
    offset := 0;
    if UseSelectEnd then begin
        offset := SelLength;
    end;
    result := self.SelectIndexIsVariable(self.SelStart+offset-1, pt) and
        (self.SelStart=pt.Y);
end;
function TMyRichEdit.SelectIndexIsVariable(index : integer; var pt : TPoint) : boolean;
var
    i : integer;
    p : Tpoint;
    function isBetween(i:integer; pt : TPoint) : boolean; inline;
    begin
        result := (i >= pt.X) and (i<= pt.Y);
    end;
begin
    inc(index);
    for p in variablePositions do begin
        pt := p;
        if isbetween(index, p) then begin
            result := true;
            EXIT;
        end;
    end;
    result := false;
end;
function TMyRichEdit.SelectIndexIsVariable(index : integer) : boolean;
var pt : TPoint;
begin
    result := SelectIndexIsVariable(index, pt);
end;
procedure TMyRichEdit.WmKeyUp(var Msg: TWMKeyUp);
var
    pt : TPoint;
    b : boolean;
    procedure ClearResult(var msg : TWMKeyUp); inline
    begin
        msg.Result := 0;
        msg.CharCode := 0;
    end;
begin
    msg.Result := 1;
    case msg.CharCode of
    VK_SHIFT: shiftDown := false;
    VK_MENU: ctrlDown := false;
    VK_LEFT:
        begin
            if self.SelectIsVariableEnd(pt) then begin
                ClearResult(msg);
            end;
        end;
    VK_RIGHT:
        begin
            if self.SelectIsVariableStart(pt,shiftDown) then begin
                ClearResult(msg);
            end;
        end;
    VK_BACK:
        begin
            if SelLength = 0 then begin
                if self.SelectIsVariableEnd(pt) then begin
                    ClearResult(msg);
                end;
            end;
        end;
    VK_DELETE:
        begin
            if SelLength = 0 then begin
                if self.SelectIsVariableStart(pt) then begin
                    ClearResult(msg);
                end;
            end;
        end;
    VK_RETURN:
        begin
            ClearResult(msg);
        end;
    end;

    if msg.Result <> 0 then begin
        inherited;
    end;
end;
procedure TMyRichEdit.WmKeyDown(var Msg: TWMKeyDown);
var
    pt : TPoint;
    b : boolean;
    procedure ClearResult(var msg : TWMKeyUp); inline
    begin
        msg.Result := 0;
        msg.CharCode := 0;
    end;
begin
    msg.Result := 1;

    case msg.CharCode of
    VK_SHIFT: shiftDown := true;
    VK_MENU: ctrlDown := true;
    VK_LEFT:
        begin
            if self.SelectIsVariableEnd(pt) then begin
                self.SetSelectionStartToX(pt);
                if shiftdown then SetSelection(pt);
                ClearResult(msg);
            end;
        end;
    VK_RIGHT:
        begin
            if self.SelectIsVariableStart(pt,ShiftDown) then begin
                self.SetSelectionStartToY(pt);
                if shiftdown then SetSelection(pt);
                ClearResult(msg);
            end;
        end;
    VK_BACK:
        begin
            if SelLength = 0 then begin
                if self.SelectIsVariableEnd(pt) then begin
                    self.SetSelection(pt);
                    ClearResult(msg);
                end;
            end;
        end;
    VK_DELETE:
        begin
            if SelLength = 0 then begin
                if self.SelectIsVariableStart(pt) then begin
                    self.SetSelection(pt);
                    ClearResult(msg);
                end;
            end;
        end;
    VK_RETURN:
        begin
            ClearResult(msg);
        end;
    end;

    if msg.Result <> 0 then begin
        inherited;
    end;
end;

// variable clicks
// highlight a clicked variable,
//
function TMyRichEdit.PointIsVariable(pt : TPoint) : boolean;
var
    p : Tpoint;
begin
    result := false;
    for p in variablePositions do begin
        result := p = pt;
        if result then EXIT;
    end;
end;
procedure TMyRichEdit.HandleClick(button : TMouseButton; X :integer);
var i : integer;
    pt, pt2 : TPoint;
begin
    if not (Button=mbLeft) then EXIT;

    // Make it easy to click around variables
    // -let users click between variables
    // -ignore clicks on the leading/trailing blanks of the variable
    // -handle special case where of 1 character variable
    if self.SelectIsVariableEnd(pt) and self.SelectIsVariableStart(pt2) and
        not self.HighlightIsVariable and (pt.Y=pt2.X) then begin
        self.SelLength := 0;
    end else if Self.SelLength = 0 then begin
        pt := self.VariablePointFromClientX(x);
        if pt.X = 0 then EXIT;
        i := Self.ClientXToIndex(x);
        if (pt.Y-pt.X) = 2 then begin
            if (i=pt.X+1) then begin
                setselection(pt);
            end else if (i=pt.X) then begin
                self.SetSelectionStartToX(pt);
            end else begin
                self.SetSelectionStartToY(pt);
            end;
        end else begin
            if (i > pt.X+1) and (i<pt.Y) then begin
                self.SetSelection(pt);
            end else begin
                if (i=pt.Y) then begin
                    self.SetSelectionStartToY(pt);
                end else begin
                    self.SetSelectionStartToX(pt);
                end;
            end;
        end;
    end;
end;
procedure TMyRichEdit.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
    inherited;
    self.HandleClick(button, x);
    if (button=mbleft) and assigned(onclick) then begin
        OnClick(Self);
    end;
end;
procedure TMyRichEdit.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
    procedure CorrectSelectionInVariables;
    var
        selpt, pt : TPoint;
        offset : integer;
    begin
        selpt := self.GetSelectionPoint;
        if self.SelectIndexIsVariable(selpt.X-1, pt) then begin
            selpt.X := pt.X;
            self.SetSelection(selpt);
        end;

        if self.SelectIndexIsVariable(selpt.Y-1, pt) then begin
            selpt.Y := pt.Y;
            self.SetSelection(selpt);
        end;
    end;
begin
    inherited;
    CorrectSelectionInVariables;
    self.HandleClick(button, x);
end;
procedure TMyRichEdit.Click;
begin
    LastSelStartClick := self.SelStart;

    if not Assigned(OnClick) then begin
        inherited;
    end;
end;
procedure TMyRichEdit.DblClick;
var i : integer;
    pt : TPoint;
    s : string;
    IsWhiteSpace : boolean;
begin
    try
        inherited;

        i := LastSelStartClick + 1;
        self.SelStart := LastSelStartClick;
        if self.CaretInVariable then begin
            self.HighlightVariableAroundCaret;
            EXIT;
        end;

        s := self.GetDisplayedText;
        s := ReplaceAll(s,char(META_CHAR_UNI),'?');
        IsWhiteSpace := true;
        pt.x := i;

        while (pt.X > 1) and ( CharInSet(s[pt.X-1], [' '])) do begin
            dec(pt.X);
        end;
        if (pt.X=i) then begin
            IsWhiteSpace := false;
            while (pt.X > 1) and ( not CharInSet(s[pt.X-1], [' ','?'])) do begin
                dec(pt.X);
            end;
        end;


        pt.Y := i;
        if IsWhiteSpace then begin
            while (pt.y <= length(s)) and ( CharInSet(s[pt.y], [' '])) do begin
                inc(pt.y);
            end;
        end else begin
            while (pt.y <= length(s)) and (not CharInSet(s[pt.y], [' ','?'])) do begin
                inc(pt.y);
            end;
        end;
        dec(pt.Y);

        self.SetSelection(pt);
    except
    end;

end;


//
//  Inserting Streams -
//
procedure TMyRichEdit.InsertLink(url, description : string);
var s : ansistring;
    ss : TMemoryStream;
begin
    ss := TMemoryStream.Create;
    s :=
        '{\rtf1\ansi\deff0{\field{\*\fldinst{HYPERLINK "'
            + url +
        '"}{\fldrslt{\cf1\ul '
            + description +
        '}}}}}'
    ;
    ss.Write(PChar(s)^, length(s)*sizeof(s[1]));
    self.InsertStream(ss);
    ss.Free;
end;
type
  TStreamCallback =
  function(dwCookie: Longint; pbBuff: PByte; cb: Longint; var pcb: Longint): DWORD; stdcall;
TStreamData = record
    dwCookie: Longint;
    dwError: Longint;
    pfnCallback: TStreamCallback;
end;
function GetStreamDataCallback(dwCookie: DWORD; pBuff: Pointer; cb: LongInt; pcb: PLongInt): DWORD; stdcall;
begin
    try
        pcb^ := TStream(dwCookie).Read(pBuff^, cb);
        DebugLog.Append('pbuff=' + PAnsiChar(pBuff));
        result := 0;
    except
        result := cardinal(32);
    end;
end;
procedure TMyRichEdit.InsertStream(const Stream : TStream);
    function SendInsertMessage : TStreamData;
    begin
        result.dwCookie := DWORD(Stream);
        result.dwError := 0;
        result.pfnCallback := @GetStreamDataCallback;
        self.Perform(
            EM_STREAMIN,
            (SFF_SELECTION or SF_RTF),
            LongInt(@result)
        );
    end;
var
    sd : TStreamData;
begin
    self.Lines.BeginUpdate;
    try
        //MaxLength := self.MaxLength + Stream.Size;
        MaxLength := 0;
        stream.Position := 0;
        sd := SendInsertMessage;
        if sd.dwError <> 0 then begin
           // raise Exception.Create('MyRichEdit: Error inserting stream');
        end;
    finally
        self.Lines.EndUpdate;
    end;
end;

// question mark
//    META_CHAR_UNI = $003F;
//    META_CHAR = '\x{003F}';
// decent quote symbols
//    META_CHAR_UNI = $00A8;
//    META_CHAR = '\x{00A8}';
// thin space
//    META_CHAR_UNI = $2009;
//    META_CHAR = '\x{2009}';
// bullet
//    META_CHAR_UNI = $2022;
//    META_CHAR = '\x{2022}';
// small carat
//    META_CHAR_UNI = $2C6;
//    META_CHAR = '\x{2C6}';
// small dot
//    META_CHAR_UNI = $2219;
//    META_CHAR = '\x{2219}';
end.
