﻿unit UnitFrmScriptActions;

{
    Purpose:
        Needed a way to fake creating a custom control to get around
        the limitations of the free version of Delphi.


    NOTES:
        ScriptControlList contains all generic script steps controls located in a grouping control (lastParentControl)
        - .Tag for the controls points to an index in this collection
        - these controls are common to all step controls and expand/enable/move/delete a script step

        ScriptControlOrder contains the order of the ScriptControlList items


        StepAdd/Replace/Remove/OtherList collections reference the step specific controls
        - located under TScriptStepControl.ScriptStep
        - .Tag for these controls are for each of these lists respectively


        DoubleBuffered controls don't refresh correctly
        Toggling visibility is a hack that forces a repaint of the controls        
}
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, ComCtrls, Buttons,Contnrs, UnitMetaVariables, UnitFrmDummyShowPreview,
  Menus, System.Generics.collections, Vcl.Imaging.pngimage, UnitWildcards;



{$TYPEINFO ON}
//{$METHODINFO ON}
{$M+}
type TScriptType = (
    stAddBeginningEnding,
    stAddPosition,
    stAddBeforeAfter,
    stReplaceNameExtension,
    stReplaceCharacters,
    stReplaceNumbers,
    stRemoveWords,
    stRemoveBeginningEnding,
    stRemoveSubstring,
    stRemoveBeforeAfter,

    // new step types below
    stAddCounter,
    stRemoveExpression,
    stConditionalExecute,
    stRemoveWordNumber,
    stSplitAndMove,
    stSelectAndMove,
    stSwap,
    stFindAndEdit
);
{since the scripttype is saved into a file, any new steps must be placed at
the end}

type TRMSpeedButton = class(TSpeedButton)
    procedure Paint(); override;
end;

type TScriptSuperType = (
    sstAdd, sstReplace, sstRemove, sstOther
);
type
TScriptStepControl = class;
TMyToplinePanel = class;
TTransPanel = class;
TMyPanelBottom = class;
TTransFlowPanel = class;
TScriptStep = class(TComponent)
    private
        FScriptType : TScriptType;
        fenabled : boolean;
        fKeepExpanded : boolean;

        procedure AppendSentence(ssc : TScriptStepControl; wc : TControl);
        function GetShortSentence : string; virtual; abstract;
    public
        constructor Create;
        procedure Execute(var sname, ext : string; var rec : RenameRecord); virtual; abstract;
        procedure Reset; virtual; abstract;
        procedure GetSentence(ssc : TScriptStepControl); virtual; abstract;
        function GetWildcards : TWildcards;
    published

    property ScriptType : TScriptType  read FScriptType write FScriptType;
    property Enabled : Boolean read fenabled write fenabled;
    property KeepExpanded : boolean read fKeepExpanded write fKeepExpanded;
end;
TScriptStepControl = class(TObject)
    private
        fDeleted : boolean;
        function GetTag : integer;
        function getisExpanded : boolean;
        function getHeight : integer;
    public
        pnlTop : TMyToplinePanel;
        outerpnl, pnlLeft, pnlRight : TTransPanel;
        pnlBtm : TMyPanelBottom;


    pnlMid : TTransFlowPanel;
    sbmore, sbup, sbdown, sbdelete, sbenable, sbhelp : TSpeedButton;
    imgPinned : TSpeedButton;
    ScriptStep : TScriptStep;

    property Deleted : boolean read fDeleted write fDeleted;
    property Tag : integer read GetTag;
    property IsExpanded : boolean read getIsExpanded;
    property Height : integer read getHeight;

    function GetHint : string;
end;
TMyToplinePanel = Class(TPanel)
        procedure Paint(); override;
        procedure Resize(); override;
    private
        clr : TColor;
    public
        RoundBottom : boolean;
        Hovered : boolean;
         procedure Invalidate; override;
        property shade : TColor read clr write clr;
End;
TMyPanelBottom = Class(TPanel)
    private
        clr : TColor;
    public

        procedure Paint; override;
        procedure Invalidate; override;
        property shade : TColor read clr write clr;
End;

TTransPanel = Class(TPanel)
    private
        clr : TColor;
    public
        procedure Invalidate; override;

    protected
        procedure CreateParams(var Params: TCreateParams); override;
        procedure Paint; override;
        procedure Resize; override;
        property shade : TColor read clr write clr;
end;
TTransFlowPanel = Class(TFlowPanel)
    private
        fExpanded : boolean;
        clr : TColor;
        fEnabled : boolean;
        fSelected : boolean;
        procedure SetShade(value : TColor);
    public
        procedure Invalidate; override;
        property Expanded : boolean read fExpanded write fExpanded;
        property shade : TColor read clr write SetShade;
        property Enabled : boolean read fEnabled write fEnabled;
        property Selected : Boolean read fSelected write fSelected;
    protected
        procedure CreateParams(var Params: TCreateParams); override;
        procedure Paint; override;

End;





type TScriptStepAdd = class(TScriptStep)
    private
    flblAdd, flblToThe : TLabel;
    ftxtAdd : TCustomEdit;  ftxtAddPosition, ftxtAfter : TCustomEdit;
    fcbxAddPart : TComboBox;

    public
        //procedure Execute(var sname, ext : string; rec : RenameRecord); override;
    published
    property lblAdd : TLabel read flblAdd write flblAdd;
    property lblToThe : TLabel read flblToThe write flblToThe;
    property txtAdd : TCustomEdit read ftxtAdd write ftxtAdd;
    property txtAddPosition : TCustomEdit read ftxtAddPosition write ftxtAddPosition;
    property txtAfter : TCustomEdit read ftxtAfter write ftxtAfter;
    property cbxAddPart : TComboBox read fcbxAddPart write fcbxAddPart;

end;
type TScriptStepAddCounter = class(TScriptStep)
{NOTE: This add doesn't use the same leading parts and doesn't need to inherit from TSCriptAdd like the other parts}
    private
        flblAddCounter, flblZeros, flblAOfTheName : TLabel;
        ftxtAZeros : TCustomEdit;
        fAUDZeros : TUpDown;
        fpnlAZero : TPanel;
        fcbxBeginningEnding : TComboBox;
        function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
        property lblAddCounter : TLabel read flblAddCounter write flblAddCounter;
        property lblZeros : TLabel read flblZeros write flblZeros;
        property lblAOfTheName : TLabel read flblAOfTheName write flblAOfTheName;
        property txtAZeros : TCustomEdit read ftxtAZeros write ftxtAZeros;
        property AUDZeros : TUpDown read fAUDZeros write fAUDZeros;
        property pnlAZero : TPanel read fpnlAZero write fpnlAZero;
        property cbxBeginningEnding : TComboBox read fcbxBeginningEnding write fcbxBeginningEnding;
end;
type TScriptStepAddBeginningEnding = Class(TScriptStepAdd)
    private
    flblAtThe : TLabel;
    fcbxAddAtThe : TComboBox;
    fcbOnCollision : TCheckbox;
    function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;

    published
    property lblAtThe : TLabel read flblAtThe write flblAtThe;
    property cbxAddAtThe : TComboBox read fcbxAddAtThe write fcbxAddAtThe;
    property cbOnCollision : TCheckBox read fcbOnCollision write fcbOnCollision;

end;
type TScriptStepAddPosition = Class(TScriptStepAdd)
    private
    flblAtCharacterPosition : TLabel;
    ftxtAddPosition : TCustomEdit;
    ftbPosition : TTrackBar;
    flblAddFromThe : TLabel;
    fcbxFromTheBeginningEnding : TComboBox;
    function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
    property lblAtCharacterPosition : TLabel read flblAtCharacterPosition write flblAtCharacterPosition;
    property txtAddPosition : TCustomEdit read ftxtAddPosition write ftxtAddPosition;
    property tbPosition : TTrackBar read ftbPosition write ftbPosition;
    property lblAddFromThe : TLabel read flblAddFromThe write flblAddFromThe;
    property cbxFromTheBeginningEnding : TComboBox read fcbxFromTheBeginningEnding write fcbxFromTheBeginningEnding;
end;
type TScriptStepAddBeforeAfter = Class(TScriptStepAdd)
    private
    flblPlaceIt : TLabel;
    ftxtAfter : TCustomEdit;
    fcbxAddPlace : TComboBox;
    function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
    property lblPlaceIt : TLabel read flblPlaceIt write flblPlaceIt;
    property txtAfter : TCustomEdit read ftxtAfter write ftxtAfter;
    property cbxAddPlace : TComboBox read fcbxAddPlace write fcbxAddPlace;

end;


type TScriptStepReplace = class(TScriptStep)
end;

type TScriptStepReplaceNameExtension = class(TScriptStepReplace)
    private
    flblReplaceThe, flblWith, flblThru, flblRMInThe : TLabel;
    fcbxReplacePart, fcbxReplaceTarget : TComboBox;
    ftxtReplacePhrase, ftxtReplaceWith,
    ftxtRP2X, ftxtRP2Y : TCustomEdit;
    fudRP2X, fudRP2Y : TUpDown;
    fpnlReplaceOnly : TPanel;
    fcbReplaceOnly : TCheckbox;
    function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
    property lblReplaceThe : TLabel read flblReplaceThe write flblReplaceThe;
    property lblWith : TLabel read flblWith write flblWith;
    property lblThru : TLabel read flblThru write flblThru;
    property cbxReplacePart : TComboBox read fcbxReplacePart write fcbxReplacePart;
    property txtReplacePhrase : TCustomEdit read ftxtReplacePhrase write ftxtReplacePhrase;
    property txtReplaceWith : TCustomEdit read ftxtReplaceWith write ftxtReplaceWith;
    property txtRP2X : TCustomEdit read ftxtRP2X write ftxtRP2X;
    property txtRP2Y : TCustomEdit read ftxtRP2Y write ftxtRP2Y;
    property udRP2X : TUpDown read fudRP2X write fudRP2X;
    property udRP2Y : TUpDown read fudRP2Y write fudRP2Y;
    property pnlReplaceOnly : TPanel read fpnlReplaceOnly write fpnlReplaceOnly;
    property cbReplaceOnly : TCheckBox read fcbReplaceOnly write fcbReplaceOnly;
    property lblRMInThe : TLabel read flblRmInThe write flblRmInThe;
    property cbxReplaceTarget : TComboBox read fcbxReplaceTarget write fcbxReplaceTarget;
end;
type TScriptStepReplaceCharacters = class(TScriptStepReplace)
    private
    flblReplaceEach,
    flblWithChar  : TLabel;
    ftxtReplaceChar1, ftxtReplaceChar2 : TCustomEdit;
    flblInThe : TLabel;
    fcbxReplaceCharPart : TComboBox;
    function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
    property lblReplaceEach : TLabel read flblReplaceEach write flblReplaceEach;
    property lblWithChar : TLabel read flblWithChar write flblWithChar;
    property txtReplaceChar1 : TCustomEdit read ftxtReplaceChar1 write ftxtReplaceChar1;
    property txtReplaceChar2 : TCustomEdit read ftxtReplaceChar2 write ftxtReplaceChar2;
    property lblInThe : TLabel read flblInThe write flblInThe;
    property cbxReplaceCharPart : TComboBox read fcbxReplaceCharPart write fcbxReplaceCharPart;
end;
type TScriptStepReplaceNumbers = class(TScriptStepReplace)
    private
    flblReplaceNumbers : TLabel;
    fcbxNumbers : TComboBox;
    flblOfTheName : TLabel;
    fcbxRNSequence : TComboBox;
    ftxtRNWith : TCustomEdit;
    function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
    property lblReplaceNumbers : TLabel read flblReplaceNumbers write flblReplaceNumbers;
    property cbxNumbers : TComboBox read fcbxNumbers write fcbxNumbers;
    property lblOfTheName : TLabel read flblOfTheName write flblOfTheName;
    property cbxRNSequence : TComboBox read fcbxRNSequence write fcbxRNSequence;
    property txtRNWith : TCustomEdit read ftxtRNWith write ftxtRNWith;
end;
type TScriptStepSelectAndMove = class(TScriptStepReplace)
    private
        fpnlMove: TPanel;
        flblSMMoveTo: TLabel;
        frbSMBeginningEnding2: TRadioButton;
        fcbxSMBeginningEnding2: TComboBox;
        frbSMBeforeAfter2: TRadioButton;
        fcbxSMBeforeAfter2: TComboBox;
        ftxtSMPhrase2: TCustomEdit;
        frbSMPosition2: TRadioButton;
        fpnlSMX2: TPanel;
        ftxtSMX2: TCustomEdit;
        fudSMX2: TUpDown;
        fpnlSelect: TPanel;
        flblSMSelect: TLabel;
        flblSMCharacters: TLabel;
        flblSMThru: TLabel;
        frbSMBeforeAfter: TRadioButton;
        fcbxSMBeforeAfter: TComboBox;
        ftxtSMPhrase: TCustomEdit;
        frbSMBeginningEnding: TRadioButton;
        fcbxSMBeginningEnding: TComboBox;
        fpnlSMZ1: TPanel;
        ftxtSMZ1: TCustomEdit;
        fudSMZ1: TUpDown;
        frbSMPosition: TRadioButton;
        fpnlSMX1: TPanel;
        ftxtSMX1: TCustomEdit;
        fupSMX1: TUpDown;
        fpnlSMY1: TPanel;
        ftxtSMY1: TCustomEdit;
        fudSMY1: TUpDown;
        function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
    property pnlMove: TPanel read fpnlMove write fpnlMove;
    property lblSMMoveTo: TLabel read flblSMMoveTo write flblSMMoveTo;
    property rbSMBeginningEnding2: TRadioButton read frbSMBeginningEnding2 write frbSMBeginningEnding2;
    property cbxSMBeginningEnding2: TComboBox read fcbxSMBeginningEnding2 write fcbxSMBeginningEnding2;
    property rbSMBeforeAfter2: TRadioButton read frbSMBeforeAfter2 write frbSMBeforeAfter2;
    property cbxSMBeforeAfter2: TComboBox read fcbxSMBeforeAfter2 write fcbxSMBeforeAfter2;
    property txtSMPhrase2: TCustomEdit read ftxtSMPhrase2 write ftxtSMPhrase2;
    property rbSMPosition2: TRadioButton read frbSMPosition2 write frbSMPosition2;
    property pnlSMX2: TPanel read fpnlSMX2 write fpnlSMX2;
    property txtSMX2: TCustomEdit read ftxtSMX2 write ftxtSMX2;
    property udSMX2: TUpDown read fudSMX2 write fudSMX2;
    property pnlSelect: TPanel read fpnlSelect write fpnlSelect;
    property lblSMSelect: TLabel read flblSMSelect write flblSMSelect;
    property lblSMCharacters: TLabel read flblSMCharacters write flblSMCharacters;
    property lblSMThru: TLabel read flblSMThru write flblSMThru;
    property rbSMBeforeAfter: TRadioButton read frbSMBeforeAfter write frbSMBeforeAfter;
    property cbxSMBeforeAfter: TComboBox read fcbxSMBeforeAfter write fcbxSMBeforeAfter;
    property txtSMPhrase: TCustomEdit read ftxtSMPhrase write ftxtSMPhrase;
    property rbSMBeginningEnding: TRadioButton read frbSMBeginningEnding write frbSMBeginningEnding;
    property cbxSMBeginningEnding: TComboBox read fcbxSMBeginningEnding write fcbxSMBeginningEnding;
    property pnlSMZ1: TPanel read fpnlSMZ1 write fpnlSMZ1;
    property txtSMZ1: TCustomEdit read ftxtSMZ1 write ftxtSMZ1;
    property udSMZ1: TUpDown read fudSMZ1 write fudSMZ1;
    property rbSMPosition: TRadioButton read frbSMPosition write frbSMPosition;
    property pnlSMX1: TPanel read fpnlSMX1 write fpnlSMX1;
    property txtSMX1: TCustomEdit read ftxtSMX1 write ftxtSMX1;
    property udSMX1: TUpDown read fupSMX1 write fupSMX1;
    property pnlSMY1: TPanel read fpnlSMY1 write fpnlSMY1;
    property txtSMY1: TCustomEdit read ftxtSMY1 write ftxtSMY1;
    property udSMY1: TUpDown read fudSMY1 write fudSMY1;


end;
type TScriptStepSwap = class(TScriptStepReplace)
        flblSSwap : TLabel;
        ftxtSPhrase : TCustomEdit;
        fcbSTrim : TCheckBox;
       function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
        property lblSSwap : TLabel read flblSSwap write flblSSwap;
        property txtSPhrase : TCustomEdit read ftxtSPhrase write ftxtSPhrase;
        property cbSTrim : TCheckbox read fcbSTrim write fcbSTrim;

end;
type TScriptStepFindAndEdit = class(TScriptStepReplace)
        flblFEFindINTheName : TLabel;
        ftxtFEPhrase1 : TCustomEdit;

        fcbFEFromThe : TCheckBox;
        fcbxFEBeginningEnding : TComboBox;
        flblFEReplacePosition: TLabel;
        fpnlFEX1 : TPanel;
        ftxtFEX1 : TCustomEdit;
        fudFEX1 : TUpDown;
        flblFELength : TLabel;
        fpnlFEY1 : TPanel;
        ftxtFEY1 : TCustomEdit;
        fudFEY1 : TUpDown;

        flblFEWith : TLabel;
        ftxtFEPhrase2 : TCustomEdit;

        fcbFEInsert : TCheckBox;
        ftxtFEPhrase3 : TCustomEdit;

        fcbxFEBeforeAFter : TComboBox;
        ftxtFEPhrase4 : TCustomEdit;
        function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
        property lblFEFindINTheName : TLabel read flblFEFindINTheName write flblFEFindINTheName;
        property txtFEPhrase1 : TCustomEdit read ftxtFEPhrase1 write ftxtFEPhrase1;

        property cbFEFromThe : TCheckBox read fcbFEFromThe write fcbFEFromThe;
        property cbxFEBeginningEnding : TComboBox read fcbxFEBeginningEnding write fcbxFEBeginningEnding;
        property lblFEReplacePosition : TLabel read flblFEReplacePosition write flblFEReplacePosition;
        property pnlFEX1 : TPanel read fpnlFEX1 write fpnlFEX1;
        property txtFEX1 : TCustomEdit read ftxtFEX1 write ftxtFEX1;
        property udFEX1 : TUpDown read fudFEX1 write fudFEX1;
        property lblFELength : TLabel read flblFELength write flblFELength;
        property pnlFEY1 : TPanel read fpnlFEY1 write fpnlFEY1;
        property txtFEY1 : TCustomEdit read ftxtFEY1 write ftxtFEY1;
        property udFEY1 : TUpDown read fudFEY1 write fudFEY1;

        property lblFEWith : TLabel read flblFEWith write flblFEWith;
        property txtFEPhrase2 : TCustomEdit read ftxtFEPhrase2 write ftxtFEPhrase2;

        property cbFEInsert : TCheckBox read fcbFEInsert write fcbFEInsert;
        property txtFEPhrase3 : TCustomEdit read ftxtFEPhrase3 write ftxtFEPhrase3;

        property cbxFEBeforeAFter : TComboBox read fcbxFEBeforeAFter write fcbxFEBeforeAFter;
        property txtFEPhrase4 : TCustomEdit read ftxtFEPhrase4 write ftxtFEPhrase4;

end;

type TScriptStepOther = class(TScriptStep)
end;
type TScriptStepConditionExecute = class(TScriptStepOther)
    private
        fcbxCESkipRemove : TComboBox;
        flblCESKipTheNExt : TLabel;
        ftxtCE1 : TCustomEdit;
        fudCE1 : TUpDown;
        fpnlCE1 : TPanel;
        flblCESteps : TLabel;
        flblCEIf : TLabel;
        fcbxCECondition : TComboBox;
        ftxtCEText : TCustomEdit;
        function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
    property cbxCESkipRemove: TCombobox read fcbxCESkipRemove write fcbxCESkipRemove ;
    property lblCESkipTheNext : TLabel read flblCESkipTheNext write flblCESkipTheNext;
    property txtCE1 : TCustomEdit read ftxtCE1 write ftxtCE1;
    property udCE1 : TUpDown read fudCE1 write fudCE1;
    property pnlCE1 : TPanel read fpnlCE1 write fpnlCE1;
    property lblCESteps : TLabel read flblCESteps write flblCESteps;
    property lblCEIf : TLabel read flblCEIf write flblCEIf;
    property cbxCECondition : TComboBox read fcbxCECondition write fcbxCECondition;
    property txtCEText : TCustomEdit read ftxtCEText write ftxtCEText;
end;

type TScriptStepRemove = class(TScriptStep)
end;

type TScriptStepRemoveExpression = class(TScriptStepRemove)
    private
        fcbxRERemoveReplace : TComboBox;
        flblreRemoveTheExpression : TLabel;
        fcbreAtThe : TCheckBox;
        fcbxreBegginingEnding : TComboBox;
        flblreOfTheName : TLabel;
        fcbreThatSTartsWIth : TCheckBox;
        fcbxrePhraseCharacters : TComboBox;
        ftxtrePhraseCharacters : TCustomEdit;

        fcbxreHasInTheMiddle : TCheckBox;
        fcbxrePhraseCharacters2 : TComboBox;
        ftxtrePhraseCharacters2 : TCustomEdit;

        fcbreAndEndsWith : TCheckBox;
        fcbxrePhraseCharacters3 : TComboBox;
        ftxtrePhraseCharacters3 : TCustomEdit;
        fcbreAndStore : TCheckBox;

        flblReplaceExpressionWith : TLabel;
        ftxtREWith : TCustomEdit;

        flblRERemoveThePart : TLabel;
        fcbxREParts : TComboBox;

        fcbreRemoveNumber : TCheckBox;
        fpnlRE1 : TPanel;
        ftxtRE1 : TCustomEdit;
        fudRE1 : TUpDown;
        function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
        property cbxRERemoveReplace : TComboBox read fcbxRERemoveReplace write fcbxRERemoveReplace;
        property lblreRemoveTheExpression : TLabel read flblreRemoveTheExpression write flblreRemoveTheExpression;
        property cbreAtThe : TCheckBox read fcbreAtThe write fcbreAtThe;
        property cbxreBegginingEnding : TComboBox read fcbxreBegginingEnding write fcbxreBegginingEnding;
        property lblreOfTheName : TLabel read flblreOfTheName write flblreOfTheName;
        property cbreThatSTartsWIth : TCheckBox read fcbreThatSTartsWIth write fcbreThatSTartsWIth;
        property cbxrePhraseCharacters : TComboBox read fcbxrePhraseCharacters write fcbxrePhraseCharacters;
        property txtrePhraseCharacters : TCustomEdit read ftxtrePhraseCharacters write ftxtrePhraseCharacters;

        property cbreHasInTheMiddle : TCheckBox read fcbxreHasInTheMiddle write fcbxreHasInTheMiddle;
        property cbxrePhraseCharacters2 : TComboBox read fcbxrePhraseCharacters2 write fcbxrePhraseCharacters2;
        property txtrePhraseCharacters2 : TCustomEdit read ftxtrePhraseCharacters2 write ftxtrePhraseCharacters2;

        property cbreAndEndsWith : TCheckBox read fcbreAndEndsWith write fcbreAndEndsWith;
        property cbxrePhraseCharacters3 : TComboBox read fcbxrePhraseCharacters3 write fcbxrePhraseCharacters3;
        property txtrePhraseCharacters3 : TCustomEdit read ftxtrePhraseCharacters3 write ftxtrePhraseCharacters3;

        property lblReplaceExpressionWith : TLabel read flblReplaceExpressionWith write flblReplaceExpressionWith;
        property txtREWith : TCustomEdit read ftxtREWith write ftxtREWith;

        property lblRERemoveTheParts  : TLabel read flblRERemoveThePart write flblRERemoveThePart;
        property cbxREParts : TComboBox read fcbxREParts write fcbxREParts;

        property cbreAndStore : TCheckBox read fcbreAndStore write fcbreAndStore;

        property cbreRemoveNumber : TCheckBox read fcbreRemoveNumber write fcbreRemoveNumber;
        property pnlRE1 : TPanel read fpnlRE1 write fpnlRE1;
        property txtRE1 : TCustomEdit read ftxtRE1 write ftxtRE1;
        property udRE1 : TUpDown read fudRE1 write fudRE1;
end;
type TScriptStepRemoveWords = class(TScriptStepRemove)
    private
    flblRemoveWords : TLabel;
    fpnlRW1 : TPanel;
    ftxtRW1 : TCustomEdit;
    fudWR1 : TupDown;
    fcbxRemovePhrase : TComboBox;
    ftxtRemovePhrase : TCustomEdit;
    flblFromThe : TLabel;
    fcbxRemovePhrasePart : TComboBox;
    function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
    property lblRemoveWords : TLabel read flblRemoveWords write flblRemoveWords;
    property pnlRW1 : TPanel read fpnlRW1 write fpnlRW1;
    property txtRW1 : TCustomEdit read ftxtRW1 write ftxtRW1;
    property upWR1 : TUpDown read fudWR1 write fudWR1;
    property cbxRemovePhrase : TComboBox read fcbxRemovePhrase write fcbxRemovePhrase;
    property txtRemovePhrase : TCustomEdit read ftxtRemovePhrase write ftxtRemovePhrase;
    property lblFromThe : TLabel read flblFromThe write flblFromThe;
    property cbxRemovePhrasePart : TComboBox read fcbxRemovePhrasePart write fcbxRemovePhrasePart;

end;
type TScriptStepRemoveCharacters = class(TScriptStepRemove)
    private
    flblRemoveChars : TLabel;
    ftxtR2X: TCustomEdit;
    fpnlR2X: TPanel;
    fudR2X: TUpDown;
    flblCharsFromThe : TLabel;
    fcbxRemoveCharPosition : TComboBox;
    flblOfThe : TLabel;
    fcbxRemoveCharactersPart : TComboBox;
    fcbAndSTore : TCheckbox;
    fcbRCReplace : TCheckBox;
    ftxtRCReplace : TCustomEdit;

    function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
    property lblRemoveChars : TLabel read flblRemoveChars write flblRemoveChars;
    property txtR2X : TCustomEdit read ftxtR2X write ftxtR2X;
    property pnlR2X : TPanel read fpnlR2X write fpnlR2X;
    property udR2X : TUpDown read fudR2X write fudR2X;
    property lblCharsFromThe : TLabel read flblCharsFromThe write flblCharsFromThe;
    property cbxRemoveCharPosition : TComboBox read fcbxRemoveCharPosition write fcbxRemoveCharPosition;
    property lblOfThe : TLabel read flblOfThe write flblOfThe;
    property cbxRemoveCharactersPart : TComboBox read fcbxRemoveCharactersPart write fcbxRemoveCharactersPart;
    property cbRCReplace : TCheckBox read fcbRCReplace write fcbRCReplace;
    property txtRCReplace : TCustomEdit read ftxtRCReplace write ftxtRCReplace;
    property cbAndStore : TCheckbox read fcbandstore write fcbandstore;

end;
type TScriptStepRemoveSubstring = class(TScriptStepRemove)
    private
    flblRemoveSubstring : TLabel;
    fcbxRemoveType : TComboBox;
    flblChars, flblCharThru : TLabel;
     ftxtR3X, ftxtR3Y, ftxtR3Z : TCustomEdit;
     fpnlR3X, fpnlR3Y, fpnlR3Z : TPanel;
     fudR3X, fudR3Y, fudR3Z : TUpDown;

    fcbAndMove : TCheckBox;
    fcbAndStoreRM : TCheckBox;
    function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
    property lblRemoveSubstring : TLabel read flblRemoveSubstring write flblRemoveSubstring;
    property cbxRemoveType : TComboBox read fcbxRemoveType write fcbxRemoveType;
    property lblChars : TLabel read flblChars write flblChars;
    property lblCharThru : TLabel read flblCharThru write flblCharThru;
    property txtR3X : TCustomEdit read ftxtR3X write ftxtR3X;
    property txtR3Y : TCustomEdit read ftxtR3Y write ftxtR3Y;
    property txtR3Z : TCustomEdit read ftxtR3Z write ftxtR3Z;
    property pnlR3X : TPanel read fpnlR3X write fpnlR3X;
    property pnlR3Y : TPanel read fpnlR3Y write fpnlR3Y;
    property pnlR3Z : TPanel read fpnlR3Z write fpnlR3Z;
    property udR3X : TUpDown read fudR3X write fudR3X;
    property udR3Y : TUpDown read fudR3Y write fudR3Y;
    property udR3Z : TUpDown read fudR3Z write fudR3Z;
    property cbAndMove : TCheckBox read fcbAndMove write fcbAndMove;
    property cbAndStoreRM : TCheckBox read fcbAndStoreRM write fcbAndStoreRM;

end;
type TScriptStepRemoveBeforeAfter = class(TScriptStepRemove)
    private
        flblRemoveEverything : TLabel;
        fcbxBeforeAfter : TComboBox;
        flblThePhrase : TLabel;
        ftxtRemove2Phrase : TCustomEdit;
        flblAndThePhrase : TLabel;
        ftxtRemove2Phrase2 : TCustomEdit;
        fcbAndStoreRM2 : TCheckBox;
        fcbRBReplace : TCheckBox;
        ftxtRBReplace : TCustomEdit;
        function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
        property lblRemoveEverything : TLabel read flblRemoveEverything write flblRemoveEverything;
        property cbxBeforeAfter : TComboBox read fcbxBeforeAfter write fcbxBeforeAfter;
        property lblThePhrase : TLabel read flblThePhrase write flblThePhrase;
        property txtRemove2Phrase : TCustomEdit read ftxtRemove2Phrase write ftxtRemove2Phrase;

        property lblAndThePhrase : TLabel read flblAndThePhrase write flblAndThePhrase;
        property txtRemove2Phrase2 : TCustomEdit read ftxtRemove2Phrase2 write ftxtRemove2Phrase2;

        property cbAndStoreRM2 : TCheckBox read fcbAndStoreRM2 write fcbAndStoreRM2;
        property cbRBReplace : TCheckBox read fcbRBReplace write fcbRBReplace;
        property txtRBReplace : TCustomEdit read ftxtRBReplace write ftxtRBReplace;

end;
type TScriptStepRemoveWordNumber = class(TScriptStepRemove)
    private
      flblrwRemoveWordNumber : TLabel;
      ftxtRW1 : TCustomEdit;
      fudRW1 : TupDown;
      fpnlRW1 : TPanel;
      flblrwAWordContains : TLabel;
      fcbxrwCharactersType : TComboBox;
      ftxtrwCharacters : TCustomEdit;
      fcbrwStoreTemp : TCheckBox;
      function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
        property lblrwRemoveWordNumber : TLabel read flblrwRemoveWordNumber write flblrwRemoveWordNumber;
        property txtRW1 : TCustomEdit read ftxtRW1 write ftxtRW1;
        property udRW1 : TupDown read fudRW1 write fudRW1;
        property pnlRW1 : TPanel read fpnlRW1 write fpnlRW1;
        property lblrwAWordContains : TLabel read flblrwAWordContains write flblrwAWordContains;
        property cbxrwCharactersType : TComboBox read fcbxrwCharactersType write fcbxrwCharactersType;
        property txtrwCharacters : TCustomEdit read ftxtrwCharacters write ftxtrwCharacters;
        property cbrwStoreTemp : TCheckBox read fcbrwStoreTemp write fcbrwStoreTemp;
end;


type TScriptStepSplitAndMove = class(TScriptStepOther)
    private
        flblSplitTheName : TLabel;
        ftxtFilemask : TCustomEdit;
        fbtnAutoSplit : TButton;
        flblMoveTheParts : TLabel;
        ftxtMove : TCustomEdit;
        function GetShortSentence : string; override;
    public
        procedure Execute(var sname, ext : string; var rec : RenameRecord); override;
        procedure Reset; override;
        procedure GetSentence(ssc : TScriptStepControl); override;
    published
        property lblSplitTheName : TLabel read flblSplitTheName write flblSplitTheName;
        property txtFilemask : TCustomEdit read ftxtFilemask write ftxtFilemask;
        property lblMoveTheParts : TLabel read flblMoveTheParts write flblMoveTheParts;
        property txtMove : TCustomEdit read ftxtMove write ftxtMove;
        property btnAutoSplit : TButton read fbtnAutoSplit write fbtnAutoSplit;
end;


type TScriptEventHandler = interface
    procedure NavigationEvent(controlRect : TRect);
    procedure NewSelectionEvent(controlRect : TRect);
end;

type
  TFrmScriptActions = class(TForm)
    {$REGION 'default'}
    imgExpand: TImage;
    imgCollapse: TImage;
    imgDelete: TImage;
    imgUp: TImage;
    imgDown: TImage;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    TabSheet3: TTabSheet;
    TabSheet4: TTabSheet;
    TabSheet5: TTabSheet;
    pnlAdd: TPanel;
    lblAdd: TLabel;
    lblToThe: TLabel;
    txtAdd: TEdit;
    cbxAddPart: TComboBox;
    pnlPlaceItAfter: TPanel;
    lblPlaceIt: TLabel;
    cbxAddPlace: TComboBox;
    txtAfter: TEdit;
    pnlBeginning: TPanel;
    cbxAddAtThe: TComboBox;
    cbOnCollision: TCheckBox;
    pnlAtCharPosition: TPanel;
    lblAtCharacterPosition: TLabel;
    txtAddPosition: TEdit;
    tbPosition: TTrackBar;
    pnlReplace: TPanel;
    lblWith: TLabel;
    lblReplaceThe: TLabel;
    cbxReplacePart: TComboBox;
    txtReplacePhrase: TEdit;
    txtReplaceWith: TEdit;
    pnlReplaceOnly: TPanel;
    lblThru: TLabel;
    cbReplaceOnly: TCheckBox;
    txtRP2X: TEdit;
    udRP2X: TUpDown;
    txtRP2Y: TEdit;
    udRP2Y: TUpDown;
    pnlReplaceChar: TPanel;
    lblWithChar: TLabel;
    lblReplaceEach: TLabel;
    lblInThe: TLabel;
    txtReplaceChar1: TEdit;
    txtReplaceChar2: TEdit;
    cbxReplaceCharPart: TComboBox;
    pnlReplaceNumbers: TPanel;
    lblReplaceNumbers: TLabel;
    lblOfTheName: TLabel;
    cbxNumbers: TComboBox;
    pnlRemove: TPanel;
    lblFromThe: TLabel;
    lblRemoveWords: TLabel;
    cbxRemovePhrase: TComboBox;
    txtRemovePhrase: TEdit;
    cbxRemovePhrasePart: TComboBox;
    TabSheet6: TTabSheet;
    pnlRemoveChars: TPanel;
    lblCharsFromThe: TLabel;
    lblOfThe: TLabel;
    lblRemoveChars: TLabel;
    pnlR2X: TPanel;
    txtR2X: TEdit;
    udR2X: TUpDown;
    cbxRemoveCharPosition: TComboBox;
    cbxRemoveCharactersPart: TComboBox;
    TabSheet7: TTabSheet;
    pnlRemoveSubstring: TPanel;
    lblChars: TLabel;
    lblCharThru: TLabel;
    lblRemoveSubstring: TLabel;
    cbxRemoveType: TComboBox;
    pnlR3X: TPanel;
    txtR3X: TEdit;
    udR3X: TUpDown;
    pnlR3Y: TPanel;
    txtR3Y: TEdit;
    udR3Y: TUpDown;
    cbAndMove: TCheckBox;
    pnlR3Z: TPanel;
    txtR3Z: TEdit;
    udR3Z: TUpDown;
    cbAndStore: TCheckBox;
    cbAndStoreRM: TCheckBox;
    imgCheck: TImage;
    imgBlank: TImage;
    TabSheet8: TTabSheet;
    pnlRemoveBeforeAfter: TPanel;
    lblRemoveEverything: TLabel;
    cbxBeforeAfter: TComboBox;
    cbAndStoreRM2: TCheckBox;
    txtRemove2Phrase: TEdit;
    lblThePhrase: TLabel;
    poRightClick: TPopupMenu;
    Cut1: TMenuItem;
    Copy1: TMenuItem;
    Paste1: TMenuItem;
    Delete2: TMenuItem;
    MenuItem13: TMenuItem;
    FormatMetaVariable1: TMenuItem;
    TabSheet9: TTabSheet;
    pnlAddCounter: TPanel;
    lblAddCounter: TLabel;
    pnlAZero: TPanel;
    txtAZeros: TEdit;
    AUDZeros: TUpDown;
    lblZeros: TLabel;
    cbxBeginningEnding: TComboBox;
    lblAOfTheName: TLabel;
    TabSheet10: TTabSheet;
    pnlRemoveExpression: TPanel;
    lblreRemoveTheExpression: TLabel;
    lblreOfTheName: TLabel;
    cbxreBegginingEnding: TComboBox;
    cbreAndStore: TCheckBox;
    txtrePhraseCharacters: TEdit;
    cbreAtThe: TCheckBox;
    cbreThatSTartsWIth: TCheckBox;
    cbxrePhraseCharacters: TComboBox;
    cbreHasInTheMiddle: TCheckBox;
    cbxrePhraseCharacters2: TComboBox;
    txtrePhraseCharacters2: TEdit;
    cbreAndEndsWith: TCheckBox;
    cbxrePhraseCharacters3: TComboBox;
    txtrePhraseCharacters3: TEdit;
    TabSheet11: TTabSheet;
    pnlConditionalExecute: TPanel;
    lblCESkipTheNext: TLabel;
    lblCESteps: TLabel;
    cbxCESkipRemove: TComboBox;
    lblCEIf: TLabel;
    pnlCE1: TPanel;
    txtCE1: TEdit;
    udCE1: TUpDown;
    lblRMInThe: TLabel;
    cbxReplaceTarget: TComboBox;
    TabSheet12: TTabSheet;
    pnlRemoveWord: TPanel;
    lblrwRemoveWordNumber: TLabel;
    pnlRW1: TPanel;
    txtRW1: TEdit;
    udrw1: TUpDown;
    lblrwAWordContains: TLabel;
    cbrwStoreTemp: TCheckBox;
    cbxrwCharactersType: TComboBox;
    txtrwCharacters: TEdit;
    pnlRE1: TPanel;
    txtRE1: TEdit;
    udRE1: TUpDown;
    cbreRemoveNumber: TCheckBox;
    imgReplace: TImage;
    imgAdd: TImage;
    ImgRemove: TImage;
    imgRemoveSide: TImage;
    imgAddSide: TImage;
    imgReplaceSide: TImage;
    pmFormat: TPopupMenu;
    FormatVariable1: TMenuItem;
    More1: TMenuItem;
    lblAddFromThe: TLabel;
    cbxFromTheBeginningEnding: TComboBox;
    imgInsert: TImage;
    TabSheet13: TTabSheet;
    pnlSplit: TPanel;
    txtFilemask: TEdit;
    lblSplitTheName: TLabel;
    btnAutoSplit: TButton;
    lblMoveTheParts: TLabel;
    txtMove: TEdit;
    timScrollUp: TTimer;
    timScrollDown: TTimer;
    cbxCECondition: TComboBox;
    txtCEText: TEdit;
    imgHelp: TImage;
    cbxRNSequence: TComboBox;
    txtRNWith: TEdit;
    lblReplaceExpressionWith: TLabel;
    txtREWith: TEdit;
    cbxRERemoveReplace: TComboBox;
    cbRCReplace: TCheckBox;
    txtRCReplace: TEdit;
    cbRBReplace: TCheckBox;
    txtRBReplace: TEdit;
    TabSheet15: TTabSheet;
    pnlSelectAndMove: TPanel;
    pnlMove: TPanel;
    lblSMMoveTo: TLabel;
    rbSMBeginningEnding2: TRadioButton;
    cbxSMBeginningEnding2: TComboBox;
    rbSMBeforeAfter2: TRadioButton;
    cbxSMBeforeAfter2: TComboBox;
    txtSMPhrase2: TEdit;
    rbSMPosition2: TRadioButton;
    pnlSelect: TPanel;
    lblSMSelect: TLabel;
    lblSMCharacters: TLabel;
    lblSMThru: TLabel;
    rbSMBeforeAfter: TRadioButton;
    cbxSMBeforeAfter: TComboBox;
    txtSMPhrase: TEdit;
    rbSMBeginningEnding: TRadioButton;
    cbxSMBeginningEnding: TComboBox;
    pnlSMZ1: TPanel;
    txtSMZ1: TEdit;
    udSMZ1: TUpDown;
    rbSMPosition: TRadioButton;
    pnlSMX1: TPanel;
    txtSMX1: TEdit;
    udSMX1: TUpDown;
    pnlSMY1: TPanel;
    txtSMY1: TEdit;
    udSMY1: TUpDown;
    pnlSMX2: TPanel;
    txtSMX2: TEdit;
    udSMX2: TUpDown;
    pmStepRightClick1: TPopupMenu;
    MIToggleEnable1: TMenuItem;
    MIDelete1: TMenuItem;
    MIReset1: TMenuItem;
    MIMoveUp1: TMenuItem;
    MIMoveDown1: TMenuItem;
    N1: TMenuItem;
    N2: TMenuItem;
    timIgnoreChanges: TTimer;
    tsSwap: TTabSheet;
    pnlSwap: TPanel;
    lblSSwap: TLabel;
    txtSPhrase: TEdit;
    cbSTrim: TCheckBox;
    TabSheet14: TTabSheet;
    pnlFindAndEdit: TPanel;
    lblFEFindInTheName: TLabel;
    txtFEPhrase1: TEdit;
    cbFEFromThe: TCheckBox;
    cbxFEBeginningEnding: TComboBox;
    lblFEReplacePosition: TLabel;
    pnlFEX1: TPanel;
    txtFEX1: TEdit;
    udFEX1: TUpDown;
    pnlFEY1: TPanel;
    txtFEY1: TEdit;
    udFEY1: TUpDown;
    lblFELength: TLabel;
    lblFEWith: TLabel;
    txtFEPhrase2: TEdit;
    cbFEInsert: TCheckBox;
    txtFEPhrase3: TEdit;
    cbxFEBeforeAfter: TComboBox;
    txtFEPhrase4: TEdit;
    imgInsertWildcards: TImage;
    AddPadding1: TMenuItem;
    AddDecimal1: TMenuItem;
    FormatCorrectDate1: TMenuItem;
    Hex1: TMenuItem;
    Substring1: TMenuItem;
    Letters1: TMenuItem;
    N3: TMenuItem;
    MIKeepOpen: TMenuItem;
    imgMore: TImage;
    MIScriptStepHelp1: TMenuItem;
    imgPin: TImage;
    MIDuplicate1: TMenuItem;
    N4: TMenuItem;
    lblAndThePhrase: TLabel;
    txtRemove2Phrase2: TEdit;
    lblRERemoveTheParts: TLabel;
    cbxREParts: TComboBox;
    Edit1: TEdit;
    Edit2: TEdit;
    Edit3: TEdit;
    Edit4: TEdit;
    Edit5: TEdit;
    Edit6: TEdit;
    Label1: TLabel;
    pmTextType: TPopupMenu;
    miEnableRegExp1: TMenuItem;
    miEnabledWildCardsnoformatting1: TMenuItem;
    N5: TMenuItem;
    miInsertzeroormorecharacters1: TMenuItem;
    miInsertexactly1character1: TMenuItem;
    miInsertoneormorecharacters1: TMenuItem;
    imgFormatInsert: TImage;
    procedure tbPositionChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure cbxReplacePartChange(Sender: TObject);
    procedure cbxRemoveTypeChange(Sender: TObject);
    procedure cbxRemovePhraseChange(Sender: TObject);
    procedure txtAddChange(Sender: TObject);
    procedure cbxAddPartChange(Sender: TObject);
    procedure txtReplacePhraseChange(Sender: TObject);
    //procedure txtReplaceWithChange(Sender: TObject);
    procedure txtReplaceChar1Change(Sender: TObject);
    procedure txtReplaceChar2Change(Sender: TObject);
    procedure Cut1Click(Sender: TObject);
    procedure Copy1Click(Sender: TObject);
    procedure Paste1Click(Sender: TObject);
    procedure Delete2Click(Sender: TObject);
    procedure FormatMetaVariable1Click(Sender: TObject);
    procedure txtCE1Change(Sender: TObject);
    procedure FormatVariable1Click(Sender: TObject);
    procedure More1Click(Sender: TObject);
    procedure AUDZerosChanging(Sender: TObject; var AllowChange: Boolean);
    procedure btnAutoSplitClick(Sender: TObject);
    procedure timScrollDownTimer(Sender: TObject);
    procedure timScrollUpTimer(Sender: TObject);

    procedure GenericClickChangeEvent(Sender : TObject);
    procedure GenericKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure GenericOnEnter(Sender: TObject);

    procedure cbxCEConditionChange(Sender: TObject);
    procedure txtCETextChange(Sender: TObject);
    procedure cbxRNSequenceChange(Sender: TObject);
    procedure cbxRERemoveReplaceChange(Sender: TObject);
    procedure MIToggleEnable1Click(Sender: TObject);
    procedure MIDelete1Click(Sender: TObject);
    procedure MIMoveUp1Click(Sender: TObject);
    procedure MIMoveDown1Click(Sender: TObject);
    procedure MIReset1Click(Sender: TObject);
    procedure timIgnoreChangesTimer(Sender: TObject);
    procedure AddPadding1Click(Sender: TObject);
    procedure AddDecimal1Click(Sender: TObject);
    procedure FormatCorrectDate1Click(Sender: TObject);
    procedure Hex1Click(Sender: TObject);
    procedure Substring1Click(Sender: TObject);
    procedure Letters1Click(Sender: TObject);
    procedure MIKeepOpenClick(Sender: TObject);
    procedure pmStepRightClick1Popup(Sender: TObject);
    procedure sbMoreClick(Sender: TObject);
    procedure sbChangeTextFormalClick(Sender: TObject);
    procedure MIScriptStepHelp1Click(Sender: TObject);
    procedure MIDuplicate1Click(Sender: TObject);
    procedure cbxBeforeAfterChange(Sender: TObject);
    procedure GenerickKeypress(Sender: TObject; var Key: Char);
    procedure miEnableRegExp1Click(Sender: TObject);
    procedure miEnabledWildCardsnoformatting1Click(Sender: TObject);
    procedure pmTextTypePopup(Sender: TObject);
    procedure miInsertzeroormorecharacters1Click(Sender: TObject);
    procedure miInsertexactly1character1Click(Sender: TObject);
    procedure miInsertoneormorecharacters1Click(Sender: TObject);


    {$ENDREGION}
  private
    { Private declarations }
        ScriptControlList : TObjectList;
        ScriptControlOrder : TObjectList;
        StepAddList : TObjectList;
        StepReplaceList : TObjectList;
        StepRemoveList : TObjectList;
        StepOtherList : TObjectList;
        SkipAutoCollapse : boolean;
        DontExpand : boolean;
        IgnoreMouse : boolean;
        LastText : TCustomEdit;
        Loading : boolean;
        Expanding : boolean;

        LastParentControl : TWinControl;

        LastSelStart, lastSelLength : integer;
        LastSelText : string;
        //fFilterChanged, fPathChanged : boolean;

        fStepClicked : TPanel;
        fIsDragging : boolean;
        fStartingPoint : TPoint;
        fSavedCursor : TCursor;
        fRightClick : boolean;

        wildcardcontrols : TList<TCustomEdit>;

        fSavePathChecked : boolean;
//        fSaveFilterChecked : boolean;

        DefaultStepWidth : integer;

        fLastSelectedControl : TScriptStepControl;
        fLastSelectedEdit : TCustomEdit;
        fLastSelectedEditControlIndex : integer;

        NoClickExpand : boolean;
        fListeners : TList<TScriptEventHandler>;

        fDragHovered : TScriptStepControl;

        //procedure FormatVariableText(t : TCustomEdit);
        procedure AddInsert(t : TCustomEdit);
        procedure AddWildCards(t: TCustomEdit);
        function SpeedButtonToEdit(sb : TSpeedbutton) : TWinControl;
        function EditToChangeTextButton(wc : TWinControl) : TSpeedButton;

        procedure HandleKeypress(sender : Tobject);
        procedure TextMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
        procedure TextMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
        procedure TextContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);

        procedure btnHelpClick(Sender: TObject);

        procedure GetRemoveWord(p: TPanel; sac: TScriptStepControl);

        procedure RefreshOrder;
        function UpdateSentence(index : integer; conditional : boolean = false) : string; overload;
        function UpdateSentence(sac : TScriptStepControl; conditional : boolean = false) : string; overload;
        procedure RefreshPanel;



        procedure DragDropHook(c :TControl);
        procedure StepMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
        procedure StepMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
        procedure StepMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
        procedure StepMouseLeave(Sender: TObject);
        procedure MoveStep(p : TPanel; newindex : integer);


        procedure ShowButtons(sac : TScriptStepControl; visible : boolean = true);
        procedure OuterPanelClick(Sender: TObject);
        procedure LabelClickEvent(Sender : TObject);
        procedure ToggleScriptDetailsClick(Sender: TObject);
        procedure ToggleSingleStep(sac : TScriptStepControl);
        procedure MoveScriptUpClick(Sender: TObject);
        procedure MoveScriptDownClick(Sender: TObject);
        procedure DeleteScriptClick(Sender: TObject);
        procedure EnableClick(sender : TObject); overload;
        procedure EnableClick(tag : integer); overload;
        procedure MyToggle(sac : TScriptStepControl);
        procedure DeleteScript(tag : integer);

        function CreateNewScriptControl(scrlScriptbox : twincontrol) : TPanel;
        function GetScriptStepAdd(i : integer) : TScriptStepAdd;
        function GetScriptControl(i : integer) : TScriptStepControl;

        procedure GetAddBeginningEndingStep(p : TPanel; sac : TScriptStepControl);
        procedure GetAddCounter(p : TPanel; sac : TScriptStepControl);
        procedure GetAddPositionStep(p : TPanel; sac : TScriptStepControl);
        procedure GetAddBeforeAfterStep(p : TPanel; sac : TScriptStepControl);
        procedure GetReplaceNameExtensionStep(p : TPanel; sac : TScriptStepControl );
        procedure GetReplaceCharactersStep(p : TPanel; sac : TScriptStepControl );
        procedure GetReplaceNumbersStep(p : TPanel; sac : TScriptStepControl );
        procedure GetRemoveWordStep(p : TPanel; sac : TScriptStepControl );
        procedure GetRemoveCharactersStep(p : TPanel; sac : TScriptStepControl );
        procedure GetRemoveSubstringStep(p : TPanel; sac : TScriptStepControl );
        procedure GetRemoveBeforeAfter(p : TPanel; sac : TScriptStepControl );
        procedure GetRemoveExpression(p : TPanel; sac : TScriptStepControl );
        procedure GetRemoveWordNumber(p : TPanel; sac : TScriptStepControl );
        procedure GetConditionExecute(p : TPanel; sac : TScriptStepControl );
        procedure GetSplitAndMove(p : TPanel; sac : TScriptStepControl );
        procedure GetSelectAndMove(p : TPanel; sac : TScriptStepControl );
        procedure GetSwap(p : TPanel; sac : TScriptStepControl );
        procedure GetFindAndEdit(p : TPanel; sac : TScriptStepControl );

        procedure setLastSelectedControl(sac : TScriptStepControl);
        procedure BtnPinnedClick(sender : Tobject);

        function GetWildCards : TWildcards;
        procedure FocusFirstTabOrderControl(sac : TScriptStepControl);
        procedure NavigationNotify(sac : TScriptStepControl);
        procedure SelectionNotify(sac : TScriptStepControl);

        procedure ClearHovered;
  public
    { Public declarations }
        procedure CollapseAll(forceToggle : Boolean=False);
        function GetText(t : TCustomEdit) : string;
        procedure TextClick(Sender: TObject);
        procedure TextKeyPress(Sender: TObject; var key : char);
        procedure TextKeyDown(Sender: TObject; var key : word; shift : TShiftState);
        procedure NewScriptControl(scrlScriptbox : twincontrol; ScriptType : TScriptType; Collapse : boolean = false; Disabled : boolean = false);

        procedure AlterStep(ScriptType : TScriptType; ControlName, Value : string);
        procedure Clear(wc : TWinControl);
        procedure ClearDisabled;
        procedure LoadOldSettingAsScript(wc : TWinControl);
        procedure SaveToFile(fullnameNoEXT : string);
        procedure FixKeepExpanded;
        procedure LoadFromFile(wc: TWinControl; fullNameNoEXT : string);
        procedure RunScriptOn(var sname, ext : string; var rec : RenameRecord; curTD : TTagData; var collisionindex : integer);

        procedure DisableAll;
        procedure ResetAll;
        function AnyEnabled : boolean;
        function GetScriptExtension : string;
        function FilterChanged : boolean;
        function PathChanged : boolean;

        procedure RefreshAllButtons;
        procedure RefreshColors(redraw : boolean=True);
        procedure RedoSentences;

        procedure BeginUpdate;
        procedure EndUpdate;
        procedure ForceAlign;

        function IsWildcardEdit(T : TCustomEdit) : boolean;

        function SavePathChecked : boolean;
        function SaveFilterChecked : boolean;
        function SaveFilesFoldersChecked : boolean;
        procedure ResizeEvent;

        procedure NavigateNext;
        procedure NagivatePrev;

        procedure RegisterEventHandler(seh : TScriptEventHandler);
  end;

var
  FrmScriptActions: TFrmScriptActions;


implementation

{$R *.dfm}

{ TForm1 }
    uses UnitFrmMain, Math, UnitUtil, UnitFrmConfig,  StrUtils,
  UnitToken, UnitFrmDummyPreview,
  UnitFrmCollisionDialog, IniFiles, TypInfo, UnitRMScript, Themes,
  UnitFrmMetaVariables, UnitAutoSplit, UnitFrmScriptStepHelp, uxtheme, GraphUtil
, UnitMyRichEdit, Clipbrd, UnitFrmMetaVariables2, UnitFormScaling, UnitDebug;

//
// ComboBox indexes used in the script steps
//
{$REGION 'ComboBOx Indexes'}
const ADDPART_NAME = 0;
    ADDPART_EXT = 1;
    ADDATTHE_BEGINNING = 0;
    ADDATTHE_ENDING = 1;
    ADDZEROS_BEGINNING = 0;
    ADDZERO_ENDING = 1;
    ADDPLACE_BEFORE = 1;
    ADDPLACE_AFTER = 0;
    REPLACEPART_NAME = 0;
    REPLACEPART_EXT = 1;
    REPLACEPART_PHRASE = 2;
    REPLACETARGET_NAME = 1;
    REPLACETARGET_EXT = 2;
    REPLACETARGET_NAME_OR_EXT = 0;
    REPLACECHARPART_NAME =0;
    REPLACECHARPART_EXT = 1;
    REPLACENUMBER_RIGHT = 0;
    REPLACENUMBER_LEFT = 1;
    REPLACENUMBER_SEQUENCE = 0;
    REPLACENUMBER_PHRASE = 1;



    REMOVEPHRASE_PHRASE = 0;
    REMOVEPHRASE_WORDS = 1;
    REMOVEPHRASE_NUMBERS = 2;
    REMOVEPHRASEPART_NAME = 0;
    REMOVEPHRASEPART_EXT = 1;
    REMOVECHARPOS_BEGINNING = 0;
    REMOVECHARPOS_ENDING = 1;
    REMOVECHARPART_NAME = 0;
    REMOVECHARPART_EXT = 1;
    REMOVETYPE_EVERYTHINGBUT = 0;
    REMOVETYPE_ONLY = 1;
    REMOVEBEFOREAFTER_BEFORE = 0;
    REMOVEBEFOREAFTER_AFTER = 1;
    REMOVEBEFOREAFTER_BEFORE_INC = 2;
    REMOVEBEFOREAFTER_AFTER_INC = 3;
    REMOVEBEFOREAFTER_BETWEEN = 4;

    REMOVEEXPRESSION_BEGINNNING = 2;
    REMOVEEXPRESSION_ENDING = 1;
	REMOVEEXPRESSION_ANYWHERE = 0;

    REMOVEEXPRESSION_PHRASE = 0;
    REMOVEEXPRESSION_CHARACTERS = 1;
    REMOVEEXPRESSION_CHARACTERS_EXCEPT = 2;
    REMOVEEXPRESSION_NOT_PHRASE = 3;

    REMOVEEXPRESSION_REMOVE = 0;
    REMOVEEXPRESSION_REPLACE = 1;

    REMOVEEXPRESSION_PARTS_ALL = 0;
    REMOVEEXPRESSION_PARTS_STARTING = 1;
    REMOVEEXPRESSION_PARTS_MIDDLE = 2;
    REMOVEEXPRESSION_PARTS_ENDING = 3;
    REMOVEEXPRESSION_PARTS_BUT_STARTING = 4;
    REMOVEEXPRESSION_PARTS_BUT_MIDDLE = 5;
    REMOVEEXPRESSION_PARTS_BUT_ENDING = 6;


    CONTITIONALEXECUTE_EXECUTE = 0;
    CONDITIONALEXECUTE_SKIP = 1;
    REMOVEWORDNUMBER_CHARACTERS = 0;
    REMOVEWORDNUMBER_CHARACTERS_EXCEPT = 1;

    //SKIPEXECUTE_DOESNT_CHANGE = 0;
    SKIPEXECUTE_CHANGES = 0;
    SKIPEXECUTE_NAME_MATCH = 1;
    //SKIPEXECUTE_NAME_DOESNT_MATCH = 3;
    SKIPEXECUTE_NAME_CONTAINS = 2;
    //SKIPEXECUTE_NAME_DOESNT_CONTAIN = 5;

    FINDEDIT_BEFORE = 0;
    FINDEDIT_AFTER = 1;
    FINDEDIT_REPLACING = 2;
    FINDEDIT_BEGINNING = 0;
    FINDEDIT_ENDING = 1;
{$ENDREGION}

    SCRIPTSECTIONNAME = 'SCRIPT';
    SCRIPTEXTENSION = '.rmscr';


function DimColor(c : TCOlor; saturation : double) : TColor;
    var r,g,b : byte;
    begin
        c := ColorToRGB(c);
        r := GetRValue(c);
        g := GetGValue(c);
        b := GetBValue(c);

        result := RGB(
            min(trunc(r * saturation),255),
            min(trunc(g * saturation),255),
            min(trunc((b ) * saturation),255)
        );

    end;
const
    DEFAULT_DIM = 1;
    GREY_DIM = 0.58;
    SHADOW_HEIGHT_PC = 0.5;
    SHADOW_DELTA_PC = 0.06;
    SHADOW_PC = DEFAULT_DIM - SHADOW_DELTA_PC;
    GRADIENT_PC = 0.88;
    OUTLINE_COLOR = clGray;
    OUTLINE_EXPANDED_COLOR = clGrayText;
    CLIENTWIDTH_DELTA = 9;

var
    RemoveColor : TColor;
    ReplaceColor : TColor;
    AddColor : TColor;
    OtherColor : TColor;



//
// Textbox Events
//
const
// question mark
//    META_CHAR_UNI = $003F;
    META_CHAR = '\x{003F}';//



procedure TFrmScriptActions.FormCreate(Sender: TObject);
begin
    ScriptControlList := TObjectList.Create;
    StepAddList := TObjectList.Create;
    StepReplaceList := TObjectList.Create;
    StepRemoveList := TObjectList.Create;
    StepOtherList := TObjectList.create;
    ScriptControlOrder := TObjectList.Create;
    ScriptControlOrder.OwnsObjects := false;

    wildcardcontrols := TList<TCustomEdit>.Create;
    TFormScaling.RegisterSelf(self);
end;
procedure TFrmScriptActions.ResizeEvent;
var i, diff, limit : integer;
    sac2 : TScriptStepControl;
begin
    if LastParentControl = nil then EXIT;

    TPanel(self.LastParentControl).AutoSize := false;
    if self.LastParentControl.Width <> self.LastParentControl.Parent.ClientWidth then begin
        limit := self.LastParentControl.Constraints.MaxWidth;
        if self.LastParentControl.width < limit then begin
            self.LastParentControl.Width := self.LastParentControl.Parent.ClientWidth;
        end;
    end;

    for i := 0 to ScriptControlOrder.Count-1 do begin
        sac2 := TScriptStepControl( ScriptControlOrder.Items[i]);
        sac2.outerpnl.AutoSize := false;

        diff := Abs(sac2.outerpnl.Width - sac2.outerpnl.Parent.ClientWidth);
        if diff > CLIENTWIDTH_DELTA then begin
            sac2.outerpnl.Width := sac2.outerpnl.Parent.ClientWidth-CLIENTWIDTH_DELTA;
        end;

        if sac2.pnlTop.Width <> sac2.outerpnl.Width then begin
            sac2.pnlTop.Width := sac2.outerpnl.Width;
            sac2.pnlBtm.Width := sac2.outerpnl.Width;
        end;
        sac2.outerpnl.AutoSize := true;
        sac2.outerpnl.AutoSize := false;
    end;

    TPanel(self.LastParentControl).AutoSize := true;
end;
procedure TFrmScriptActions.Clear(wc : TWinControl);
var i, j : integer;
    sc : TScriptStepControl;
    c : TControl;
    b : boolean;
begin
    self.LastParentControl := wc;
    self.BeginUpdate;

    for I := (ScriptControlList.count - 1) downto 0 do begin
        sc := TScriptStepControl(ScriptControlList.Items[i]);
        if not sc.Deleted  then
            self.DeleteScriptClick(sc.sbdelete);
    end;

    ScriptControlList.Clear;
    ScriptControlOrder.Clear;
    StepAddList.Clear;
    StepReplaceList.Clear;
    StepRemoveList.Clear;
    StepOtherList.clear;

    fLastSelectedEdit := nil;
    fLastSelectedEditControlIndex := -1;

    if self.LastParentControl <> nil then
        if self.LastParentControl.ControlCount <> 0 then begin
            for i := 0 to self.LastParentControl.ControlCount - 1 do begin
                j := LastParentControl.ControlCount - 1;
                c := self.LastParentControl.Controls[j];
                if c.Name <> 'pnlScriptTop' then begin
                    c.Free;
                end;
            end;
        end;

    self.EndUpdate;

    wildcardcontrols.Clear;
end;
procedure TFrmScriptActions.ClearDisabled;
var i : integer;
    sc : TScriptStepControl;
begin
    FrmScriptActions.BeginUpdate;

    for I := (ScriptControlOrder.count - 1) downto 0 do begin
        sc := TScriptStepControl(ScriptControlOrder.Items[i]);
        if (not sc.ScriptStep.Enabled) and (not sc.Deleted) then
            self.DeleteScriptClick(sc.sbdelete);
    end;

    for I := (ScriptControlOrder.count - 1) downto 0 do begin
        self.UpdateSentence(i);
    end;

    FrmScriptActions.EndUpdate;
end;


//=====================
{$region '// Navigation Routines //'}
procedure TFrmScriptActions.RegisterEventHandler(seh : TScriptEventHandler);
begin
    if not assigned(fListeners) then begin
        fListeners := TList<TScriptEventHandler>.Create;
    end;

    fListeners.Add(seh);
end;
procedure TFrmScriptActions.NavigateNext;
var i, j, k : integer;
    ssc, prev : TScriptStepControl;
    sb : TScrollBox;
var
    Info: TScrollInfo;
    b : boolean;
begin
    if self.fLastSelectedControl = nil then begin
        if self.ScriptControlList.Count > 0 then begin
            ssc := TScriptStepControl(self.ScriptControlList.Items[0]);
            self.fLastSelectedControl := ssc;
            self.ToggleScriptDetailsClick(ssc.pnlMid);
            self.FocusFirstTabOrderControl(ssc);
        end;
        EXIT;
    end;

    prev := self.fLastSelectedControl;
    i := prev.GetTag;
    inc(i);
    if not (i < self.ScriptControlList.Count) then EXIT;

    self.Expanding := true;
    ssc := TScriptStepControl(self.ScriptControlList.Items[i]);

    //self.LastParentControl.Invalidate;

    if prev.pnlMid.Expanded then begin
        self.ToggleSingleStep(prev);
    end;
    if not ssc.pnlMid.Expanded then begin
        self.ToggleSingleStep(ssc);
    end;




    self.FocusFirstTabOrderControl(ssc);
    self.ForceAlign;
    self.Expanding := false;

    self.setLastSelectedControl(ssc);
    self.NavigationNotify(ssc);
end;
procedure TFrmScriptActions.NagivatePrev;
var i : integer;
    ssc, prev : TScriptStepControl;
    b : boolean;
begin
    if self.fLastSelectedControl = nil then EXIT;

    prev := self.fLastSelectedControl;
    i := prev.GetTag;
    dec(i);
    if not (i >= 0) then EXIT;


    self.Expanding := true;
    ssc := TScriptStepControl(self.ScriptControlList.Items[i]);
    if prev.pnlMid.Expanded then begin
        self.ToggleSingleStep(prev);
    end;
    if not ssc.pnlMid.Expanded then begin
        self.ToggleSingleStep(ssc);
    end;
    self.FocusFirstTabOrderControl(ssc);
    self.ForceAlign;
    self.Expanding := false;

    self.setLastSelectedControl(ssc);
    self.NavigationNotify(ssc);
end;
procedure TFrmScriptActions.NavigationNotify(sac : TScriptStepControl);
var
    seh : TScriptEventHandler;
begin
    if assigned(fListeners) then
        for seh in fListeners do begin
            seh.NavigationEvent(sac.outerpnl.BoundsRect);
        end;
end;
procedure TFrmScriptActions.FocusFirstTabOrderControl(sac : TScriptStepControl);
begin
    sac.pnlBtm.SelectFirst;
    //sac.pnlBtm.Invalidate;
end;
procedure TFrmScriptActions.setLastSelectedControl(sac : TScriptStepControl);
var NewSelection : boolean;
    prev : TScriptStepControl;
    seh : TScriptEventHandler;
    r : TRect;
begin
    sac.pnlMid.Selected := true;
    NewSelection := sac <> self.fLastSelectedControl;
    if self.fLastSelectedControl = nil then begin
        Self.fLastSelectedControl := sac;
        self.fLastSelectedControl.pnlMid.Invalidate;
        ShowButtons(self.fLastSelectedControl, true);
        Exit;
    end else if not NewSelection then begin
        EXIT;
    end else begin
        Self.fLastSelectedControl.pnlMid.Selected := false;
        self.fLastSelectedControl.pnlMid.Invalidate;
    end;


    prev := self.fLastSelectedControl;
    Self.fLastSelectedControl := sac;
    sac.pnlMid.Selected := true;
    sac.pnlMid.Invalidate;
    if NewSelection then begin
        ShowButtons(sac, true);
        ShowButtons(prev, false);
    end;
end;
{$endregion}
//=====================

function TFrmScriptActions.GetWildCards : TWildcards;
begin
    // use global setting to generate
    if frmmain.rbRegExp.Checked then begin
        result := TRegExpWildcards.Create;
    end else begin
        result := TWildcards.Create;
    end;
end;

function TFrmScriptActions.AnyEnabled: boolean;
var i : integer;
    sc : TscriptStepControl;
begin
    result := false;
    for i := (ScriptControlList.count - 1) downto 0 do begin
        sc := TScriptStepControl(ScriptControlList.Items[i]);
        if sc.ScriptStep.Enabled then begin
            result := true;
            EXIT;
        end;
    end;
end;
function TFrmScriptActions.IsWildcardEdit(T : TCustomEdit) : boolean;
begin
    result := false;
    if wildcardcontrols = nil then EXIT;
    result := wildcardcontrols.Contains(t);
end;

function TFrmScriptActions.GetScriptExtension: string;
begin
    result := SCRIPTEXTENSION;
end;
function TFrmScriptActions.PathChanged: boolean;
begin
    result := RMScript.GetUsePath;
end;

procedure TFrmScriptActions.pmTextTypePopup(Sender: TObject);
begin
    miEnableRegExp1.Checked := ContainsText(self.fLastSelectedEdit.Text, '?reg?');
    miEnableRegExp1.Enabled := wildcardcontrols.Contains(self.fLastSelectedEdit);
    miEnabledWildCardsnoformatting1.Checked := not TMyRichEdit(self.fLastSelectedEdit).getVariableFormatting;
    miEnabledWildCardsnoformatting1.enabled := miEnableRegExp1.Enabled;


    miInsertzeroormorecharacters1.Enabled := miEnabledWildCardsnoformatting1.Checked;
    miInsertexactly1character1.Enabled := miEnabledWildCardsnoformatting1.Checked;
    miInsertoneormorecharacters1.Enabled := miEnabledWildCardsnoformatting1.Checked;
end;

function TFrmScriptActions.FilterChanged: boolean;
begin
    result := RMScript.GetUseFilter;
end;

procedure TFrmScriptActions.DisableAll;
var i : integer;
    sc : TscriptStepControl;
begin
    self.BeginUpdate;
    for i := (ScriptControlList.count - 1) downto 0 do begin

        sc := TScriptStepControl(ScriptControlList.Items[i]);
        if not Sc.Deleted then begin
            if sc.ScriptStep.Enabled then begin
                EnableClick(sc.sbenable);
            end;
            sc.pnlMid.Invalidate;
        end;
    end;

    self.RedoSentences;
    self.EndUpdate;
end;
procedure TFrmScriptActions.ResetAll;
var i : integer;
    sc : TscriptStepControl;
begin
    self.BeginUpdate;
    for i := (ScriptControlList.count - 1) downto 0 do begin
        sc := TScriptStepControl(ScriptControlList.Items[i]);
        sc.ScriptStep.Reset;
        if sc.ScriptStep.Enabled then begin
        	EnableClick(sc.sbenable);
        end;
        sc.pnlMid.Invalidate;
    end;

    self.RedoSentences;
    self.EndUpdate;
end;
procedure TFrmScriptActions.AlterStep(ScriptType: TScriptType; ControlName, Value: string);
var i, j : integer;
    ssc : TScriptStepControl;

    pl : TPropList;
    ppl : PPropList;
    pti : PTypeInfo;
    cnt : integer;
    cmp : TComponent;
begin
    for i := 0 to ScriptControlOrder.Count - 1 do begin

        ssc := TScriptStepControl(ScriptControlOrder.items[ i ]);
        if ssc.Deleted then CONTINUE;
        if not ssc.ScriptStep.Enabled  then CONTINUE;



        ppl := @pl;
        case ssc.ScriptStep.ScriptType  of
            stAddBeginningEnding:
                pti := PTypeInfo(TScriptStepAddBeginningEnding(ssc.ScriptStep).ClassInfo);
            stAddCounter:
                pti := PTypeInfo(TScriptStepAddCounter(ssc.ScriptStep).ClassInfo);
            stAddPosition:
                pti := PTypeInfo(TScriptStepAddPosition(ssc.ScriptStep).ClassInfo);
            stAddBeforeAfter:
                pti := PTypeInfo(TScriptStepAddBeforeAfter(ssc.ScriptStep).ClassInfo);
            stReplaceNameExtension:
                pti := PTypeInfo(TScriptStepReplaceNameExtension(ssc.ScriptStep).ClassInfo);
            stReplaceCharacters:
                pti := PTypeInfo(TScriptStepReplaceCharacters(ssc.ScriptStep).ClassInfo);
            stReplaceNumbers:
                pti := PTypeInfo(TScriptStepReplaceNumbers(ssc.ScriptStep).ClassInfo);
            stRemoveWords:
                pti := PTypeInfo(TScriptStepRemoveWords(ssc.ScriptStep).ClassInfo);
            stRemoveBeginningEnding:
                pti := PTypeInfo(TScriptStepRemoveCharacters(ssc.ScriptStep).ClassInfo);
            stRemoveSubstring:
                pti := PTypeInfo(TScriptStepRemoveSubstring(ssc.ScriptStep).ClassInfo);
            stRemoveBeforeAfter:
                pti := PTypeInfo(TScriptStepRemoveBeforeAfter(ssc.ScriptStep).ClassInfo);
            stRemoveExpression:
                pti := PTypeInfo(TScriptStepRemoveExpression(ssc.ScriptStep).ClassInfo);
            stConditionalExecute:
                pti := PTypeInfo(TScriptStepOther(ssc.scriptstep).ClassInfo)
        end;

        if ScriptTYpe = ssc.ScriptStep.ScriptType  then begin
            cnt := GetPropList(pti,[tkClass], ppl);
            if cnt <> 0 then begin
                for j := 0 to cnt - 1 do begin
                    if (pl[j]^.name = 'Name') or (pl[j]^.name ='Tag') then CONTINUE;
                    
                    cmp := TComponent(TypInfo.GetObjectProp(ssc.scriptstep, pl[j]^.name));
                    if cmp <> nil then begin
                        if cmp is TCustomEdit then begin

                            if pl[j]^.name = ControlName then begin
                                TCustomEdit(cmp).text := value;

                                if assigned(TRichEdit(cmp).OnChange) then begin
                                    TRichEdit(cmp).OnChange(tedit(cmp));
                                end;

                                EXIT;
                            end;
                        end else if cmp is TComboBox then begin
                            if pl[j]^.name = ControlName then begin
                                tComboBox(cmp).ItemIndex := StrToInt(Value);
                                if assigned(tComboBox(cmp).OnChange) then begin
                                    tComboBox(cmp).OnChange(tedit(cmp));
                                end;

                                EXIT;
                            end;

                        end else if cmp is TCheckbox then begin

                        end;
                    end;
                end;
            end;
        end;
    end;
end;
procedure TFrmScriptActions.RefreshColors(redraw : boolean=True);
var i : integer;
    ssc : TScriptStepControl;
begin
    RemoveColor := frmconfig.lblRemoveStep.color;
    ReplaceColor := frmConfig.lblReplaceStep.Color;
    AddColor := frmConfig.lblAddStep.color;
    OtherColor := RGB(133,133,133);


    FrmMain.scrlScriptbox.Color := frmConfig.lblStepContainer.Color;
    if self.LastParentControl is TPanel then begin
        TPanel(self.LastParentControl).Color := frmConfig.lblStepContainer.Color;
    end;
    //FrmMain.FlowPanel1.Color := frmConfig.lblStepContainer.Color;

    if redraw then
        for i := 0 to ScriptControlOrder.Count-1 do begin
            ssc := TScriptStepControl(ScriptControlOrder.Items[i]);

            if (ssc.ScriptStep is TScriptStepAdd) or (ssc.ScriptStep is TScriptStepAddCounter) then begin
                ssc.pnlTop.shade := AddColor;
                ssc.pnlMid.shade := AddColor;
            end else if (ssc.ScriptStep is TScriptStepReplace) or (ssc.ScriptStep is TScriptStepSplitAndMove) then begin
                ssc.pnlTop.shade := ReplaceColor;
                ssc.pnlMid.shade := ReplaceColor;
            end else if ssc.ScriptStep is TScriptStepRemove then begin
                ssc.pnlTop.shade := RemoveColor;
                ssc.pnlMid.shade := RemoveColor;
            end else begin
                ssc.pnlTop.shade := OtherColor;
                ssc.pnlMid.shade := OtherColor;
            end;

            ssc.pnlMid.Invalidate;
            ssc.pnlTop.Invalidate;
        end;
end;
procedure TFrmScriptActions.BeginUpdate;
begin
    self.LastParentControl.DisableAlign;
end;
procedure TFrmScriptActions.EndUpdate;
begin
    self.LastParentControl.EnableAlign;
end;
procedure TFrmScriptActions.ForceAlign;
var i : integer;
    sc : TscriptStepControl;
    top : integer;
begin
    top := 0;
    for i := 0 to ScriptControlOrder.Count -1 do begin
        sc := TscriptStepControl(ScriptControlOrder.Items[i]);
        sc.outerpnl.Top := top;

        inc(top, sc.Height);
        if not sc.IsExpanded then begin
            sc.outerpnl.AutoSize := true;  // corrects the outer panel size
            sc.outerpnl.AutoSize := false;
        end;
    end;
end;
procedure TFrmScriptActions.CollapseAll(forceToggle : Boolean=False);
var i : integer;
    sac : TScriptStepControl;
    prevSac : TScriptStepControl;
begin
    if FrmMain.IsBatchMode then EXIT;

    Self.SkipAutoCollapse := true;
    for i := ScriptControlList.count - 1 downto 0 do begin
        sac := self.GetScriptControl( i );
        if (not sac.deleted) then begin
            if sac.IsExpanded or forceToggle then begin
                self.MyToggle(sac);
                if sac.IsExpanded then
                    self.MyToggle(sac);
            end;
        end;
    end;
    ForceAlign;
    Self.SkipAutoCollapse := false;
end;


procedure TFrmScriptActions.LoadFromFile(wc : TWinControl; fullNameNoEXT: string);
var ini : TMyINIFile;
    i,j, cnt : integer;
    section, s : string;
    sc : TScriptStepControl;
    ne : TNotifyEvent;
    wasVisible : boolean;

    pl : TPropList;
    ppl : PPropList;
    pti : PTypeInfo;
    cmp : TComponent;
    stype : TScriptType;
    p : TPanel;

    // for workaround purposes
    tally : integer;
    allenabled, autocollapse : boolean;
    pname : string;
    tick1, top, bottom : cardinal;
    b : boolean;

    cbs : TComboBoxStyle;
    mainwc : TWinControl;

    function SameMethod(AMethod1, AMethod2: TNotifyEvent): boolean;
    begin
        result := (TMethod(AMethod1).Code = TMethod(AMethod2).Code) and
                (TMethod(AMethod1).Data = TMethod(AMethod2).Data);
    end;
begin

    try
        DebugLog.Append('ScriptActions LoadFromFile');
        DebugLog.StartTimer;
        self.LastParentControl := wc;
        //SendMessage(self.LastParentControl.Handle, WM_SETREDRAW, 0, 0);
        ne := nil;
        if self.LastParentControl is TPanel then begin
            ne := TPanel(self.LastParentControl).onresize;
            TPanel(self.LastParentControl).onresize := nil;
        end;
        wasVisible := self.LastParentControl.Visible;
        self.LastParentControl.Visible := false;


        self.DefaultStepWidth := FrmMain.scrlScriptbox.ClientWidth-CLIENTWIDTH_DELTA;
        self.DefaultStepWidth := min(self.DefaultStepWidth, self.LastParentControl.Constraints.MaxWidth);

        // self.clear would take care of the "last selected", so this is technically redundant
        self.fLastSelectedControl := nil;
        self.fLastSelectedEdit := nil;
        self.fLastSelectedEditControlIndex := -1;

        Self.BeginUpdate;
        Loading := true;
        if ExtractFileExt(fullNameNoEXT) <> ''  then begin
            fullnameNoEXT := leftstr(fullNameNoEXT, length(fullNameNoEXT) - length(ExtractFileExt(fullNameNoEXT)));
        end;

        fullnameNoEXT := IncludeTrailingPathDelimiter(ExtractFilePath(fullnameNoEXT))  + ExtractFileName(fullnameNoEXT) + SCRIPTEXTENSION;
        if not FileExists(fullnameNoEXT) then begin
            EXIT;
        end;



        rmscript.SetFilename(fullNameNoEXT);
        rmscript.LoadSettings(false);


        ini := TMyINIFile.Create(fullnameNoEXT);
        self.clear(self.LastParentControl);


        i := 0;
        tally := 0;
        allenabled := true;
        // NewScriptControl will autocollapse on every call if set
        // override this temporarily
        autocollapse := frmconfig.cbAutoCollapse.Checked;
        frmconfig.cbAutoCollapse.Checked := false;
        repeat


            section := SCRIPTSECTIONNAME + IntToStr(i);
            cnt := ini.ReadInteger(section,'TYPE', -1);
            if cnt = -1 then begin
                BREAK;
            end;
            stype := TScriptType(cnt);
            inc(tally);

            // given a type, create a step control and create the details view
            self.NewScriptControl(self.LastParentControl, stype, true, true);
            sc := Self.GetScriptControl(ScriptControlOrder.Count-1);

            p := sc.outerpnl;

            sc.ScriptStep.Enabled := ini.ReadBool(section,'ENABLED',true);
            if not sc.ScriptStep.Enabled  then begin
                allenabled := false;
            end;
            self.ShowButtons(sc, false); // set the correct glyphs, default to invisible

            //DebugLog.Append('StepType: ' + IntToStr(Integer(sc.ScriptStep.ScriptType)));
            ppl := @pl;
            case sc.ScriptStep.ScriptType  of
                stAddBeginningEnding:
                    pti := PTypeInfo(TScriptStepAddBeginningEnding(sc.ScriptStep).ClassInfo);
                stAddPosition:
                    pti := PTypeInfo(TScriptStepAddPosition(sc.ScriptStep).ClassInfo);
                stAddBeforeAfter:
                    pti := PTypeInfo(TScriptStepAddBeforeAfter(sc.ScriptStep).ClassInfo);
                stAddCounter:
                    pti := PTypeInfo(TScriptStepAddCounter(sc.ScriptStep).ClassInfo);
                stReplaceNameExtension:
                    pti := PTypeInfo(TScriptStepReplaceNameExtension(sc.ScriptStep).ClassInfo);
                stReplaceCharacters:
                    pti := PTypeInfo(TScriptStepReplaceCharacters(sc.ScriptStep).ClassInfo);
                stReplaceNumbers:
                    pti := PTypeInfo(TScriptStepReplaceNumbers(sc.ScriptStep).ClassInfo);
                stRemoveWords:
                    pti := PTypeInfo(TScriptStepRemoveWords(sc.ScriptStep).ClassInfo);
                stRemoveBeginningEnding:
                    pti := PTypeInfo(TScriptStepRemoveCharacters(sc.ScriptStep).ClassInfo);
                stRemoveSubstring:
                    pti := PTypeInfo(TScriptStepRemoveSubstring(sc.ScriptStep).ClassInfo);
                stRemoveBeforeAfter:
                    pti := PTypeInfo(TScriptStepRemoveBeforeAfter(sc.ScriptStep).ClassInfo);
                stRemoveExpression:
                    pti := PTypeInfo(TScriptStepRemoveExpression(sc.ScriptStep).ClassInfo);
                stConditionalExecute:
                    pti := PTypeInfo(TScriptStepConditionExecute(sc.ScriptStep).ClassInfo);
                stSplitAndMove:
                    pti := PTypeInfo(TScriptStepSplitAndMove(sc.ScriptStep).ClassInfo);
                stSelectAndMove:
                    pti := PTypeInfo(TScriptStepSelectAndMove(sc.ScriptStep).ClassInfo);
                stSwap:
                    pti := PTypeInfo(TScriptStepSwap(sc.ScriptStep).ClassInfo);
                stFindAndEdit:
                    pti := PTypeInfo(TScriptStepFindAndEdit(sc.ScriptStep).ClassInfo);


            end;
            sc.ScriptStep.KeepExpanded := ini.ReadBool(section,'KEEPEXPANDED', false);

            // iterate all published properties of each StepScript
            // and load only those that have user input
            // each is loaded using a unique integer
            cnt := GetPropList(pti,[tkClass], ppl);
            if cnt <> 0 then begin
                for j := 0 to cnt - 1 do begin
                    pname := pl[j]^.name;
                    if (pname = 'Name') or (pname ='Tag') then CONTINUE;

                    try
                        cmp := TComponent(TypInfo.GetObjectProp(sc.scriptstep, pname));
                        if cmp <> nil then begin
                            if cmp is TCustomEdit then begin
                                s := ini.ReadString(section,pname, '');

                                TCustomEdit(cmp).Clear;
                                TCustomEdit(cmp).Text := s;
                                // KLUDGE - RichEdit ingores text unless Clear is called
                                // - this happens before the control is drawn on screen

                                ne := TRichEdit(cmp).OnChange;
                                if assigned(ne) and (not SameMethod(ne,self.GenericClickChangeEvent)) then begin
                                    TRichEdit(cmp).OnChange(cmp);
                                end;
                            end else if cmp is TComboBox then begin
                                cbs := TComboBox(cmp).Style;
                                TComboBox(cmp).Style := csSimple;
                                tComboBox(cmp).Items.BeginUpdate;
                                tComboBox(cmp).ItemIndex := ini.ReadInteger(section,pname, 0);
                                tComboBox(cmp).Items.EndUpdate;

                                ne := tcombobox(cmp).OnChange;
                                if assigned(ne) and (not SameMethod(ne,self.GenericClickChangeEvent)) then begin
                                    tcombobox(cmp).OnChange(cmp);
                                end;

                                TComboBox(cmp).Style := cbs;
                            end else if cmp is TCheckbox then begin
                                TCheckBox(cmp).Checked := ini.ReadBool(section,pname, false);
                            end else if cmp is TRadioButton then begin
                                TRadioButton(cmp).Checked := ini.ReadBool(section,pname, False);
                            end;
                        end;
                    except
                        // a bug in a Public Beta incorrectly used non-existant controls
                        // for the new Add Counter script and caused a crash [even thought
                        // I did ALL the correct error checking]

                    end;
                end;
            end;

            //
            // FIX: The up/down is overriding the value read by the script
            //
            case sc.ScriptStep.ScriptType   of
            stAddCounter:    with TScriptStepAddCounter(sc.ScriptStep) do begin
                    AUDZeros.position := SafeInt(txtAZeros.text,2);
                end;
            stRemoveBeginningEnding: with TScriptStepRemoveCharacters(sc.ScriptStep) do begin
                    udr2x.Position := SafeInt(txtR2X.text,1);
                end;

            stRemoveSubstring: with TScriptStepRemoveSubstring(sc.scriptstep) do begin
                    udR3X.Position := SafeInt(txtR3X.text,1);
                    udR3Y.Position := SafeInt(txtR3Y.text,10);
                    udR3Z.Position := SafeInt(txtR3Z.text,1);
                end;
            stRemoveExpression: with TScriptStepRemoveExpression(sc.scriptstep) do begin
                    udRE1.Position := SafeInt(txtRE1.text,1);
                end;
            stReplaceNameExtension: with TScriptStepReplaceNameExtension(sc.scriptstep) do begin
                    udRP2X.Position := SafeInt(txtRP2X.text,1);
                    udRP2Y.position := SafeInt(txtRP2Y.text,10);
                end;
            stSelectAndMove: with TScriptStepSelectAndMove(sc.ScriptStep) do begin
                    udSMZ1.position := SafeInt(txtSMZ1.Text,1);
                    udSMX1.position := SafeInt(txtSMX1.Text,1);
                    udSMY1.position := SafeInt(txtSMY1.Text,10);
                    udSMX2.position := SafeInt(txtSMX2.Text,1);
                end;
            stFindAndEdit: with TScriptStepFindAndEdit(sc.ScriptStep) do begin
                udFEX1.Position := SafeInt(txtFEX1.Text,1);
                udFEY1.Position := SafeInt(txtFEY1.Text,0);
            end;
            end;

            sc.ScriptStep.Enabled := ini.ReadBool(section,'ENABLED',true);
            sc.pnlMid.Enabled := sc.ScriptStep.Enabled;

            inc(i);
        until false;


        self.RedoSentences;
        // WORKAROUND - detect an old DEFAULT with all items enabled
        //
        if (ini.ReadString('Version','Current','')='' ) and
             (allenabled and (tally >= 10)) then begin
            for i := 0 to ScriptControlOrder.count  - 1 do begin
                sc := GetScriptControl(i);
                self.EnableClick(sc.Tag);
            end;
        end;
        self.FixKeepExpanded;


        // TODO - investigate when or if this is needed elsewhere
        //if frmmain.scrlScriptbox.CanFocus and frmmain.Visible then frmmain.scrlScriptbox.SetFocus;

        // no write, update not needed
        ini.free;
        frmmain.SetScriptName(ExtractFilename(fullNameNoEXT));


    finally
        DebugLog.CurrentTimer('End of Load');


        loading := false;
        self.LastParentControl.Visible := wasVisible;

        self.ForceAlign;
        self.EndUpdate;
        self.LastParentControl.Invalidate;


        frmconfig.cbAutoCollapse.Checked := autocollapse;

        if self.LastParentControl is TPanel then begin
            TPanel(self.LastParentControl).onresize := ne;
        end;

        DebugLog.EndTimer('End of Finally');
        // TODO make sure this isn't neede
       // FrmDummyAutoPreview.InformOfKeypress;
    end;
end;
procedure TFrmScriptActions.SaveToFile(fullnameNoEXT: string);
var ini : TMyINIFile;
    i,j, cnt : integer;
    section : string;
    sc : TScriptStepControl;


    pl : TPropList;
    ppl : PPropList;
    pti : PTypeInfo;

    cmp : TComponent;
begin

    try
        if ExtractFileExt(fullNameNoEXT) <> ''  then begin
            fullnameNoEXT := leftstr(fullNameNoEXT, length(fullNameNoEXT) - length(ExtractFileExt(fullNameNoEXT)));
        end;
        fullnameNoEXT := IncludeTrailingPathDelimiter(ExtractFilePath(fullnameNoEXT))  + ExtractFileName(fullnameNoEXT) + SCRIPTEXTENSION;
        if FileExists(fullnameNoEXT) then begin
            DeleteFile(fullnameNoEXT);
        end;
        ini := TMyINIFile.Create(fullnameNoEXT);

        for i := 0 to ScriptControlOrder.count  - 1 do begin
            // load the full properties for each ScriptStep class
            sc := TScriptStepControl(ScriptControlOrder.items[i]);
            ppl := @pl;
            case sc.ScriptStep.ScriptType  of
                stAddBeginningEnding:
                    pti := PTypeInfo(TScriptStepAddBeginningEnding(sc.ScriptStep).ClassInfo);
                stAddCounter:
                    pti := PTypeInfo(TScriptStepAddCounter(sc.ScriptStep).ClassInfo);
                stAddPosition:
                    pti := PTypeInfo(TScriptStepAddPosition(sc.ScriptStep).ClassInfo);
                stAddBeforeAfter:
                    pti := PTypeInfo(TScriptStepAddBeforeAfter(sc.ScriptStep).ClassInfo);
                stReplaceNameExtension:
                    pti := PTypeInfo(TScriptStepReplaceNameExtension(sc.ScriptStep).ClassInfo);
                stReplaceCharacters:
                    pti := PTypeInfo(TScriptStepReplaceCharacters(sc.ScriptStep).ClassInfo);
                stReplaceNumbers:
                    pti := PTypeInfo(TScriptStepReplaceNumbers(sc.ScriptStep).ClassInfo);
                stRemoveWords:
                    pti := PTypeInfo(TScriptStepRemoveWords(sc.ScriptStep).ClassInfo);
                stRemoveBeginningEnding:
                    pti := PTypeInfo(TScriptStepRemoveCharacters(sc.ScriptStep).ClassInfo);
                stRemoveSubstring:
                    pti := PTypeInfo(TScriptStepRemoveSubstring(sc.ScriptStep).ClassInfo);
                stRemoveBeforeAfter:
                    pti := PTypeInfo(TScriptStepRemoveBeforeAfter(sc.ScriptStep).ClassInfo);
                stRemoveExpression:
                    pti := PTypeInfo(TScriptStepRemoveExpression(sc.ScriptStep).ClassInfo);
                stConditionalExecute:
                    pti := PTypeInfo(TScriptStepConditionExecute(sc.ScriptStep).ClassInfo);
                stSplitAndMove:
                    pti := PTypeInfo(TScriptStepSplitAndMove(sc.ScriptStep).ClassInfo);

                stSelectAndMove:
                    pti := PTypeInfo(TScriptStepSelectAndMove(sc.ScriptStep).ClassInfo);
                stSwap:
                    pti := PTypeInfo(TScriptStepSwap(sc.ScriptStep).ClassInfo);
                stFindAndEdit:
                    pti := PTypeInfo(TScriptStepFindAndEdit(sc.ScriptStep).ClassInfo);
            end;

            // iterate all published properties of each StepScript
            // and save only the user input using the name of the control
            // - each saved using a unique integer
            section := SCRIPTSECTIONNAME + IntToStr(i);
            ini.WriteInteger(section,'TYPE', Integer(sc.ScriptStep.ScriptType));
            ini.WriteBool(section,'ENABLED', sc.ScriptStep.Enabled);
            ini.WriteBool(section,'KEEPEXPANDED', sc.ScriptStep.KeepExpanded);
            cnt := GetPropList(pti,[tkClass], ppl);
            if cnt <> 0 then begin
                for j := 0 to cnt - 1 do begin
                    if (pl[j]^.name = 'Name') or (pl[j]^.name ='Tag') then CONTINUE;

                    cmp := TComponent(TypInfo.GetObjectProp(sc.scriptstep, pl[j]^.name));
                    if cmp <> nil then begin
                        if cmp is TMyRichEdit then begin
                            ini.WriteString(section,pl[j]^.name, TMyRichEdit(cmp).Text);
                        end else if cmp is TCustomEdit then begin
                            ini.WriteString(section,pl[j]^.name, TCustomEdit(cmp).Text);
                        end else if cmp is TComboBox then begin
                            ini.WriteInteger(section,pl[j]^.name, tComboBox(cmp).ItemIndex);
                        end else if cmp is TCheckbox then begin
                            ini.WriteBool(section,pl[j]^.name, TCheckBox(cmp).Checked);
                        end else if cmp is TRadioButton then begin
                            ini.WriteBool(section,pl[j]^.name, TRadioButton(cmp).Checked);
                        end;
                    end;
                end;
            end;
        end;

        ini.UpdateFile;
        ini.free;

        RMScript.SetFilename(fullnameNoEXT);
        RMScript.SaveSettings(false);
    finally

    end;
end;
procedure TFrmScriptActions.FixKeepExpanded;
var
    i : integer;
    ssc : TScriptStepControl;
begin
    for i := 0 to ScriptControlOrder.Count - 1 do begin
        ssc := TScriptStepControl(ScriptControlOrder.items[ i ]);
        if ssc.ScriptStep.KeepExpanded then begin
            NoClickExpand := true;
            FrmScriptActions.ToggleScriptDetailsClick(ssc.pnlTop);
            self.ShowButtons(ssc, true);
        end;
    end;

end;


function TFrmScriptActions.SavePathChecked : boolean;
begin
    result := RMScript.GetUsePath;
end;
function TFrmScriptActions.SaveFilterChecked : boolean;
begin
    result := RMScript.GetUseFilter;
end;
function TFrmScriptActions.SaveFilesFoldersChecked : boolean;
begin
    result := RMScript.GetUseFilesFolders;
end;



//====================
{$region '// Create Step Controls and Steps //'}
procedure TFrmScriptActions.NewScriptControl(scrlScriptbox: twincontrol;
  ScriptType: TScriptType; Collapse : boolean = false; Disabled : boolean = false);
var p : TPanel;
    ssc : TScriptStepControl;
    wasvisible : boolean;

    procedure SetDefaultGlyphs;
    begin
        ssc.sbmore.Glyph := Self.imgMore.Picture.Bitmap;
        ssc.sbmore.Glyph.Transparent := true;
        ssc.sbmore.Glyph.TransparentMode := tmauto;

        ssc.imgPinned.Glyph := self.imgPin.Picture.Bitmap;
        ssc.imgPinned.Visible := True;
        ssc.imgPinned.Glyph.Transparent := true;
        ssc.imgPinned.Glyph.TransparentMode := tmAuto
    end;
    procedure AddGlyph;
    begin
        SetDefaultGlyphs;
        ssc.pnlTop.shade := AddColor;
        ssc.pnlMid.shade := AddColor;
    end;
    procedure RemoveGlyph;
    begin
        SetDefaultGlyphs;
        ssc.pnlTop.shade := RemoveColor;
        ssc.pnlMid.shade := RemoveColor;
    end;
    procedure ReplaceGlyph;
    begin
        SetDefaultGlyphs;
        ssc.pnlTop.shade := ReplaceColor;
        ssc.pnlMid.shade := ReplaceColor;
    end;
    procedure OtherInit;
    begin
        SetDefaultGlyphs;
        ssc.pnlTop.shade := otherColor;
        ssc.pnlMid.shade := ssc.pnlTop.shade;
    end;

begin
    //
    // For some reason, the container panel must be visible while controls are added
    // Otherwise, alignment gets screwed up for each control
    //

    if frmConfig.cbAutoCollapse.Checked then begin
        self.CollapseAll;
    end;

    self.RefreshColors(false);
    wasvisible := self.LastParentControl.Visible;


    p := self.CreateNewScriptControl(scrlScriptbox);




    ssc := self.GetScriptControl(p.tag);
    ssc.outerpnl.DisableAlign;
   // SendMessage(ssc.outerpnl.Handle, WM_SETREDRAW, 0, 0);

    case ScriptType of
        stAddBeginningEnding: self.GetAddBeginningEndingStep( ssc.pnlbtm, ssc );
        stAddCounter: self.GetAddCounter(ssc.pnlbtm, ssc );
        stAddPosition:  self.GetAddPositionStep( ssc.pnlbtm, ssc ) ;
        stAddBeforeAfter: self.GetAddBeforeAfterStep( ssc.pnlbtm, ssc ) ;
        stReplaceNameExtension: self.GetReplaceNameExtensionStep( ssc.pnlbtm, ssc ) ;
        stReplaceCharacters: self.GetReplaceCharactersStep( ssc.pnlbtm, ssc );
        stReplaceNumbers: self.GetReplaceNumbersStep( ssc.pnlbtm, ssc ) ;
        stRemoveWords: self.GetRemoveWordStep( ssc.pnlbtm, ssc ) ;
        stRemoveBeginningEnding: self.GetRemoveCharactersStep( ssc.pnlbtm, ssc ) ;
        stRemoveSubstring: self.GetRemoveSubstringStep( ssc.pnlbtm, ssc ) ;
        stRemoveBeforeAfter: self.GetRemoveBeforeAfter( ssc.pnlbtm, ssc);
        stRemoveExpression: self.GetRemoveExpression(ssc.pnlBtm, ssc );
        stConditionalExecute: self.GetConditionExecute( ssc.pnlbtm, ssc );
        stRemoveWordNumber: self.GetRemoveWordNumber( ssc.pnlBtm, ssc );
        stSplitAndMove: self.GetSplitAndMove(ssc.pnlbtm, ssc);
        stSelectAndMove: self.GetSelectAndMove(ssc.pnlBtm, ssc);
        stSwap: self.GetSwap(ssc.pnlbtm, ssc);
        stFindAndEdit: self.GetFindAndEdit(ssc.pnlBtm, ssc);
    end;


    case ScriptType of
        stAddBeginningEnding: AddGlyph;
        stAddCounter: AddGlyph;
        stAddPosition: AddGlyph;
        stAddBeforeAfter: AddGlyph;
        stReplaceNameExtension: ReplaceGlyph;
        stReplaceCharacters: ReplaceGlyph;
        stReplaceNumbers: ReplaceGlyph;
        stRemoveWords: RemoveGlyph;
        stRemoveBeginningEnding: RemoveGlyph;
        stRemoveSubstring: RemoveGlyph;
        stRemoveBeforeAfter: RemoveGlyph;
        stRemoveExpression: RemoveGlyph;
        stConditionalExecute: OtherInit;
        stRemoveWordNumber: RemoveGlyph;
        stSplitAndMove: ReplaceGlyph;
        stSelectAndMove: ReplaceGlyph;
        stSwap: ReplaceGlyph;
        stFindAndEdit: ReplaceGlyph;
    end;

    if Disabled and (ssc.ScriptStep.Enabled = true) then
        self.EnableClick(ssc.tag);

    if Collapse then begin

        //self.MyToggle(ssc);
        //self.MyToggle(ssc);
    end else begin
        if wasvisible then begin
            self.MyToggle(ssc);
        end;
    end;

    //SendMessage(ssc.outerpnl.Handle, WM_SETREDRAW, -1, 0);
    ssc.outerpnl.EnableAlign;

end;
function TFrmScriptActions.CreateNewScriptControl(scrlScriptbox : twincontrol): TPanel;
var
    sac, sac2 : TScriptStepControl;
    i : integer;
const BUTTON_TOP_MARGIN = 3;
    PANEL_PADDING = 3;

    procedure OuterPanel;
    begin
        with sac do begin
        if scrlScriptbox = nil then begin
            outerpnl := TTransPanel.Create(self);
        end else begin
            outerpnl := TTransPanel.Create(scrlScriptbox.parent);
        end;

        //outerpnl.Parent := scrlScriptbox;
        outerpnl.FullRepaint := false;
        outerpnl.Constraints.MinWidth := 450;
        outerpnl.Constraints.MaxWidth := max(0,scrlScriptbox.Constraints.MaxWidth);
        if self.LastParentControl.Visible then begin
            outerpnl.Width := scrlScriptbox.ClientWidth;
        end else begin
            outerpnl.Width := Self.DefaultStepWidth;
        end;


        outerpnl.height := 226;
        outerpnl.Align := alNone;
        outerpnl.Top := 0;
        if ScriptControlOrder.count > 1 then begin
            sac2 := TScriptStepControl(ScriptControlOrder.Items[ScriptControlOrder.count-2]);
            outerpnl.Top := sac2.outerpnl.Top+sac2.outerpnl.Height;
        end;
        outerpnl.Tag := i;
        outerpnl.Padding.Top := 0;
        outerpnl.padding.bottom := 0;
        outerpnl.Padding.Left := 0;
        outerpnl.Padding.Right := 0;
        outerpnl.Margins.Top := 0;
        outerpnl.Margins.Bottom := 0;
        outerpnl.BevelEdges := [];
        outerpnl.BevelInner := bvNone;
        outerpnl.BevelKind := bkNone;
        outerpnl.BevelOuter := bvNone;
        outerpnl.BorderStyle := bsNone;
        outerpnl.BevelWidth := 1;
        outerpnl.BorderWidth := 0;
        // testing
        //outerpnl.ParentBackground := true;
        //outerpnl.ParentColor := true;

        outerpnl.OnClick := self.OuterPanelClick;
        outerpnl.OnMouseMove := self.StepMouseMove;
        outerpnl.OnMouseLeave := self.StepMouseLeave;
        outerpnl.AlignWithMargins := False;
        outerpnl.DoubleBuffered := true;
        outerpnl.DisableAlign;

        //outerpnl.TabStop := false;
        end;
    end;
    procedure TopLine;
    begin
        with sac do begin
            // kludge to place new panel on top, but last in the
            // list of ontop panels
            pnlTop := TMyToplinePanel.Create(outerpnl);
            pnltop.FullRepaint := false;

            pnlTop.Parent := outerpnl;
            pnlTop.Width := outerpnl.Width;
            pnlTop.Align := alTop;
            pnlTop.Height := 27;     // was 28
            pnlTop.BevelKind := bkNone;
            pnlTop.BevelInner := bvNone;
            pnlTop.BevelOuter := bvNone;
            pnlTop.BorderStyle := bsNone;
            pnlTop.BorderWidth := 0;
            pnlTop.Margins.left := 1; //was 4
            pnlTop.Padding.Top := 0; // was 1
            pnlTop.Padding.Bottom := 0;
            pnlTop.Margins.Top := 0;
            pnlTop.Margins.Bottom := 0;
            pnlTop.BevelEdges := [];
            pnlTop.BevelWidth  := 1;
            pnlTop.Ctl3D := false;
            pnlTop.ParentBackground := true;
            pnlTop.ParentColor := true;
            pnlTop.Tag := i;
            pnlTop.OnClick := self.ToggleScriptDetailsClick;
            pnlTop.AlignWithMargins := false;
            //pnlTop.TabStop := false;

            self.DragDropHook(pnlTop);
        end;
    end;
    procedure BottomPanel;
    begin
        with sac do begin
            pnlbtm := TMyPanelBottom.Create(outerpnl);
            pnlbtm.FullRepaint := false;
            pnlbtm.Parent := outerpnl;
            pnlbtm.Height := 16;
            pnlbtm.Align:= alBottom;
            pnlbtm.Autosize := false;
            pnlbtm.Visible := false;
            pnlbtm.BevelKind := bkNone;
            pnlbtm.BevelOuter := bvNone;
            pnlbtm.BevelInner := bvNone ;
            pnlbtm.BevelEdges := [];
            pnlBtm.BorderStyle  := bsNone;
            pnlBtm.Padding.Bottom := 0;
            pnlBtm.Padding.Top := 0;
            pnlBtm.Margins.Top := 0;
            pnlBtm.Margins.Bottom := 0;
            pnlBtm.Tag := i;
            pnlBtm.AlignWithMargins := false;
            //pnlBtm.TabStop := false;
        end;
    end;
    procedure Buttons;
    begin

    end;
    procedure TopMiddle;
    begin
        with sac do begin
            pnlMid := TTransFlowPanel.Create(pnlTop);
            pnlMid.FlowStyle := fsLeftRightTopBottom;
            pnlMid.FullRepaint := false;

            pnlmid.AutoWrap := false;
            pnlMid.parent := pnlTop;
            pnlMid.Height := 22; // was 28 // was 21
            pnlMid.BevelOuter := bvNone;
            pnlMid.BevelInner := bvNone;
            pnlMid.BevelKind := bkNone;
            pnlMid.BorderStyle := bsNone;
            pnlMid.OnClick := self.ToggleScriptDetailsClick;
            pnlMid.ShowHint := false;




            pnlMid.Width := pnlTop.Width - 50; // was 45
            pnlMid.Top := 1; // was 2 // was 0
            // at zero, the top line is drawn over

            pnlMid.Height := pnlTop.Height - (pnlMid.Top + 2);
            pnlMid.Anchors := [akLeft, akRight, akTop];
            pnlMid.tag := i;
            pnlMid.Padding.top := 3;
            pnlMid.Padding.Right := 5;
            pnlMid.Padding.Left := 6;
            pnlMid.Padding.Bottom := 0;

            pnlMid.Margins.Top := 2;
            pnlMid.Margins.right := 0;
            pnlMid.Margins.Left := 0;
            pnlMid.Margins.Bottom := 0;
            pnlMid.AlignWithMargins := false; //was true
            pnlMid.Ctl3D := false;
            pnlMid.BevelEdges := [];

            self.DragDropHook(pnlMid);
            //pnlMid.TabStop := false;
        end;
    end;
begin
    self.RefreshColors(false);
    if scrlScriptbox <> nil then
        self.LastParentControl := scrlScriptbox;
    TPanel(self.LastParentControl).OnMouseMove := self.StepMouseMove;

    // This is a static list so any TAG will always point towards a single object
    // The order is changed to represent the visible order of items in the script
    sac := TScriptStepControl.Create;
    i := ScriptControlList.Add(sac);
    ScriptControlOrder.add(sac);
    with sac do begin
        OuterPanel;
        TopLine;
        BottomPanel;
        Buttons;
        TopMiddle;

        outerpnl.AutoSize := true;
        outerpnl.AutoSize := false;
        outerpnl.EnableAlign;
        result := outerpnl;
        self.ShowButtons(sac,false);
    end;

    sac.outerpnl.parent := scrlScriptbox;
    if (i=0) then begin
        self.ResizeEvent;
    end;
end;
procedure TFrmScriptActions.ShowButtons(sac : TScriptStepControl; visible : boolean = true);
var i : integer;

    procedure SetCommon(sb : TSpeedButton);
    begin
        sb.parent := sac.pnlTop;
        sb.Flat := true;
        sb.Tag := i;
        sb.top := 4;
        sb.width := 20;
        sb.Height := 18+1;
        sb.anchors := [akTop, akRight];
        sb.Margins.Right := 2;
        sb.Margins.left := 2;
        sb.margins.top := 8;
        sb.margins.Bottom := 8;
        sb.Transparent := true;
    end;
var
    left : integer;
const
    SPACER = 2;
    ENABLE_SPACER = 2;
    procedure CreateButtons;
    begin
        with sac do
        if not assigned(sbdelete) {and (not loading)} then begin
            sbdelete := TSpeedButton.Create(pnlTop);
            SetCommon(sbdelete);
            sbdelete.glyph := imgDelete.picture.Bitmap;
            sbdelete.OnClick := self.DeleteScriptClick;
            sbdelete.Hint := 'Delete   (Shift+Click = Reset)';
            sbdelete.Left := pnlTop.Width - (sbdelete.Width + SPACER*2);
            sbdelete.Anchors := [akTop, akRight];
            left := sbdelete.left;


            sbdown := TSpeedButton.Create(pnlTop);
            SetCommon(sbdown);
            sbdown.glyph := imgDown.Picture.bitmap;
            sbDown.OnClick := self.MoveScriptDownClick;
            sbdown.Hint := 'Move down (shift=bottom)';
            sbdown.Visible := frmconfig.cbShowUpDown.checked;

            sbup := TSpeedButton.Create(pnlTop);
            SetCommon(sbup);
            sbup.Glyph := imgUp.picture.bitmap;
            sbup.OnClick := self.MoveScriptUpClick;
            sbup.Hint := 'Move up (shift=top)';
            sbup.Visible := frmconfig.cbShowUpDown.checked;

            sbenable := TSpeedButton.Create(pnlTop);
            setcommon(sbenable);
            sbenable.anchors := [akTop, akleft];
            imgCheck.Picture.Bitmap.Transparent := true;
            imgCheck.Picture.Bitmap.TransparentColor  := clWhite;
            imgCheck.Picture.Bitmap.TransparentMode := tmAuto;
            sbenable.Glyph := imgBlank.Picture.Bitmap;
            sbenable.Hint := 'Enable/Disable';
            left := SPACER;

            sbenable.left := ENABLE_SPACER;
            sbenable.OnClick := self.EnableClick;
            sbenable.Spacing := 0;

            sbmore := TRMSpeedButton.Create(pnltop);
            setcommon(sbmore);
            sbmore.Anchors := [akTop, akRight];
            sbmore.Glyph := imgmore.Picture.Bitmap;
            sbmore.Glyph.TransparentMode := tmauto;
            sbmore.Glyph.TransparentColor := clGrayText;
            sbmore.Hint := 'More';
            sbmore.Left := sbenable.Left + sbenable.Width + SPACER;
            sbmore.OnClick := Self.sbMoreClick;
            sbmore.Spacing := 0;


            imgPinned := TSpeedButton.Create(pnlTop);
            SetCommon(imgPinned);
            imgPinned.Anchors := [akTop, akLeft];
            imgPinned.Transparent := true;

            imgPinned.Left := sbenable.Left + sbenable.Width + SPACER;
            imgPinned.Hint := 'Unpin';
            imgPinned.OnClick := BtnPinnedClick;

            imgPinned.ShowHint := True;
        end;
    end;
begin
    if frmconfig.cbAlwaysShowStepButtons.checked then visible := true;
    i := sac.pnlTop.Tag;


    Self.BeginUpdate;
    if Assigned(sac.sbdelete) then begin
        left := sac.sbdelete.left;
    end;

    CreateButtons;
    with sac do begin
        sbmore.Visible := fLastSelectedControl = sac;
        sbdelete.Visible := fLastSelectedControl = sac;
        if frmconfig.cbShowUpDown.checked then begin
            if not sbdelete.Visible then
                sbdelete.Visible := visible;
            left := sbdelete.Left;
            left := left - (sbdelete.Width + SPACER);
            sbup.Visible := visible;
            sbup.Left := left;

            left := left - (sbdelete.Width + SPACER);
            sbdown.Visible := visible;
            sbdown.Left := left;

            left := left - (sbdelete.Width + SPACER);

            sbenable.left := ENABLE_SPACER;


        end else begin
            if not sbdelete.Visible then
                sbdelete.Visible := visible;
            left := sbdelete.Left;
            left := left - (sbdelete.Width + SPACER);
            sbup.Visible := false;
            sbdown.Visible := false;

            sbenable.left := ENABLE_SPACER;
        end;
        sbmore.Left := left - SPACER;


        if (sac.ScriptStep <> nil) and sac.ScriptStep.fKeepExpanded then begin
            imgPinned.Visible := True;
            imgPinned.Left := sbenable.Left + sbenable.Width + SPACER;
            pnlMid.Left := imgPinned.Left + imgPinned.Width + SPACER;
            pnlMid.Width := pnlMid.Width  - (imgPinned.Width + SPACER);
        end else begin
            imgPinned.Visible := false;
            pnlMid.Left := sbenable.Left + sbenable.Width + SPACER;
            sac.pnlMid.Width := left - sac.pnlMid.left - 1;
        end;
        pnlMid.Invalidate;

        if (sac.ScriptStep <> nil) then begin
            if sac.ScriptStep.Enabled then begin
                sac.sbenable.Glyph := self.imgCheck.Picture.Bitmap;
            end;
        end;
    end;
    self.EndUpdate;

end;


procedure DupePosition(p : TWinControl; src, target : TControl; tag : integer = 0);
var controlname : string;
begin
    target.Visible := false;
    target.Top := src.Top;
    target.Left := src.left;
    target.Width := src.Width;
    target.Height := src.Height;


    target.tag := tag;


    target.Parent := p;

    if (src is TLabel) then begin
        controlname := 'Label';
        TLabel(target).caption := TLabel(src).Caption;
        TLabel(target).Transparent := TLabel(src).Transparent;

    end else if (src is TRadioButton) then begin
        controlname := 'RadioButton';
        TRadioButton(target).Caption := TRadioButton(src).Caption;
        TRadioButton(target).Checked  := TRadioButton(src).Checked;
        TRadioButton(target).ParentColor := true;

        TRadioButton(target).OnClick := TRadioButton(src).OnClick;
        TRadioButton(target).OnEnter := TRadioButton(src).OnEnter;
        TRadioButton(target).TabOrder := TRadioButton(src).TabOrder;
        TradioButton(target).TabStop := TradioButton(src).TabStop;

        TRadioButton(target).OnEnter :=  FrmScriptActions.GenericOnEnter;
    end else if (src is TEdit) then begin
        if target is TRichEdit then begin
            controlname := 'RichEdit';
            TRichEdit(target).text := TEdit(src).text;
            TRichEdit(target).font := TEdit(src).font;
            TRichEdit(target).TabOrder := TEdit(src).TabOrder;

            TRichEdit(target).OnChange := TEdit(src).OnChange;
            TRichEdit(target).OnExit := TEdit(src).OnExit;
            TRichEdit(target).OnEnter := TEdit(src).OnEnter;
            if not assigned(TRichEdit(target).OnEnter) then
                TRichEdit(target).OnEnter :=  FrmScriptActions.GenericOnEnter;
        end else begin
            controlname := 'Edit';
            TEdit(target).text := TEdit(src).text;
            TEdit(target).font := TEdit(src).font;
            TEdit(target).TabOrder := TEdit(src).TabOrder;

            TEdit(target).OnChange := TEdit(src).OnChange;
            TEdit(target).OnExit := TEdit(src).OnExit;
            TEdit(target).OnEnter := TEdit(src).OnEnter;
            TEdit(target).OnKeyPress := TEdit(src).OnKeyPress;
            if not assigned(TEdit(target).OnEnter ) then
                TEdit(target).OnEnter :=  FrmScriptActions.GenericOnEnter;
        end;

    end else if (src is TComboBox) then begin
        //DebugLog.StartTimer();
        controlname := 'ComboBox';

        // speeds up by a significant ammount
        TComboBox(target).Style := csSimple; // about 30 ms

        //TComboBox(target).Style := csDropDown; // about 100 ms
        //TComboBox(target).Style := csDropDownList; // about 100 ms
        //TComboBox(target).Style := csOwnerDrawFixed; // about 120 ms


        TComboBox(target).OnChange := nil;
        TComboBox(target).OnEnter := nil;

        TComboBox(target).Items.BeginUpdate;
        TComboBox(target).Items.AddStrings(TComboBox(src).Items);
        TComboBox(target).ItemIndex := TComboBox(src).ItemIndex;
        TComboBox(target).Items.EndUpdate;


        TComboBox(target).OnChange := TComboBox(src).OnChange;
        TComboBox(target).OnEnter := TComboBox(src).OnEnter;
        TComboBox(target).TabOrder := TComboBox(src).TabOrder;
        TComboBox(target).TabStop := TComboBox(src).TabStop;
        if not assigned(TComboBox(target).OnEnter) then
            TComboBox(target).OnEnter :=  FrmScriptActions.GenericOnEnter;
        TComboBox(target).Style := csDropDownList;

        //DebugLog.EndTimer('ComboBox:');
    end else if (src is TUpDown) then begin
        controlname := 'UpDown';
        TUpDown(target).position := TUpDown(src).position;
        TUpDown(target).Min := TUpDown(src).Min;
        TUpDown(target).Max := TUpDown(src).Max;
        TUpDown(target).TabOrder := TUpDown(src).TabOrder;

        if not assigned(TUpDown(target).OnEnter) then
            TUpDown(target).OnEnter :=  FrmScriptActions.GenericOnEnter;
    end else if (src is TCheckBox) then begin
        controlname := 'CheckBox';
        TCheckbox(target).caption := TCheckbox(src).caption;
        TCheckBox(Target).Checked := TCheckBox(src).Checked;

        TCheckBox(Target).OnClick := TCheckBox(src).OnClick;
        TCheckBox(Target).OnEnter := TCheckBox(src).OnEnter;
        TCheckBox(Target).tabOrder := TCheckBox(src).tabOrder;

        //TCheckBox(Target).Color := ColorToRGB(clWhite);
        if not assigned(TCheckBox(target).OnEnter) then
            TCheckBox(target).OnEnter :=  FrmScriptActions.GenericOnEnter;
        if p is TPanel then begin
            TPanel(p).ParentBackground := false;
        end;
    end else if (src is TPanel) then begin
        controlname := 'Panell';
        TPanel(target).BevelEdges := TPanel(src).BevelEdges;
        TPanel(target).BevelKind := TPanel(src).BevelKind;
        TPanel(target).BevelOuter := TPanel(src).BevelOuter;
        TPanel(target).Padding := TPanel(src).Padding;
        TPanel(target).Margins := TPanel(src).Margins;
        TPanel(target).TabOrder := TPanel(src).TabOrder;

        //TPanel(target).Color := TPanel(src).Color;
        TPanel(target).Color := clWhite;
        TPanel(target).ParentColor := false;
        TPanel(target).ParentBackground := false;
    end else if (src is TTrackBar) then begin
        controlname := 'TrackBar';
        TTrackBar(target).TickMarks := TTrackBar(src).TickMarks;
        TTrackBar(target).Position := TTrackBar(src).Position;
        TTrackBar(target).Min := TTrackBar(src).Min;
        TTrackBar(target).Max := TTrackBar(src).Max;
        TTrackBar(target).taborder := TTrackBar(src).taborder;

        TTrackBar(target).OnChange := TTrackBar(src).OnChange;
        if not assigned(TTrackBar(target).OnEnter ) then
            TTrackBar(target).OnEnter :=  FrmScriptActions.GenericOnEnter;
    end else if (src is TButton) then begin
        controlname := 'Button';
        TButton(target).caption := TButton(src).caption;
        TButton(target).taborder := TButton(src).taborder;

        TButton(target).OnClick := TButton(src).OnClick;
        TButton(target).OnEnter := TButton(src).OnEnter;
        if not assigned(TButton(target).OnEnter ) then
            TButton(target).OnEnter :=  FrmScriptActions.GenericOnEnter;
    end;

    target.hint := src.hint;
    target.ShowHint := src.ShowHint;
    target.Visible := src.visible;

    target.Align := src.Align;
end;
procedure TFrmScriptActions.AddWildCards(t: TCustomEdit);
var wc : TWinControl;
    i : integer;
begin
    wildcardcontrols.Add(t);
    if (t is TEdit) then
        TEdit(t).Color := $FFF8F3;
    t.Hint := 'Supports Wildcards';

    wc := TControl(t).parent;
    for I := 0 to wc.ControlCount  - 1 do begin
        if not (wc.controls[i] is TSpeedButton) then CONTINUE;
        if not (wc.controls[i].Tag = 0) then CONTINUE;

        TSpeedButton(wc.controls[i]).Glyph := imgInsertWildcards.Picture.Bitmap;
    end;

    // this was the default, I think wildcards aren't used often enough
    // to default to formatted variables

//    if t is TMyRichEdit then begin
//        TMyRichEdit(t).DisableVariableFormatting;
//    end;
end;
procedure TFrmScriptActions.AddInsert(t: TCustomEdit);
var p : TPanel;
    sb, sb2 : TSpeedButton;
begin
    p := TPanel.Create(t.Parent);
    sb := TSpeedButton.Create(p);
    sb2 := TSpeedButton.Create(p);

    p.ParentBackground := false;
    p.Parent := t.parent;
    p.Width := t.Width{ +10};
    p.Height := t.Height + 3;
    p.top := t.top-1;
    p.left := t.left;
    p.Visible := t.Visible;
    p.BevelEdges := [beRight, beBottom];

    p.BevelKind := bkFlat;
    p.BevelOuter := bvNone;
    p.Margins.Top := 0;
    p.Padding.Left := 1;
    p.Padding.Bottom := 1;
    p.padding.top := 1;
    p.ShowHint := true;

    p.TabOrder := t.TabOrder;

    sb.Parent := p;
    sb.Align := alRight;
    sb.Glyph :=imgInsert.Picture.Bitmap;
    sb.Flat := true;
    sb.OnClick := frmmain.InsertButtonClick;
    sb.Hint := 'Insert File/Date variables';
    sb.ShowHint := true;
    sb.Tag := 0;

    sb2.Parent := p;
    sb2.Align := alRight;
    sb2.Glyph :=imgFormatInsert.Picture.Bitmap;
    sb2.Flat := true;
    sb2.OnClick := self.sbChangeTextFormalClick;
    sb2.Hint := 'Formatting / Wildcards';
    sb2.ShowHint := true;
    sb2.Visible := false;
    sb2.Tag := 1;


    if t is TRichEdit then
    with t as TRichEdit do begin
        trichedit(t).Update;
        t.Parent := p;
        t.Align := alClient;
        t.ShowHint := true;
        t.Hint := '';
        OnClick :=   self.TextClick;
        OnMouseDown := self.TextMouseDown;
        PopupMenu := self.poRightClick;
        OnContextPopup := self.TextContextPopup;
        HideSelection := false;
        OnKeyPress := self.TextKeyPress;
        TRichEdit(t).WordWrap := false;
        OnKeyDown := TextKeyDown;
        OnKeyUp := GenericKeyUp;

        t.Tag := integer(@sb);
    end;

    if t is TEdit then
    with t as TEdit do begin
        Parent := p;
        Align := alClient;
        ShowHint := true;
        Hint := '';
        OnClick :=   self.TextClick;
        OnMouseDown := self.TextMouseDown;
        PopupMenu := self.poRightClick;
        OnContextPopup := self.TextContextPopup;
        HideSelection := false;
        OnKeyPress := self.TextKeyPress;

        OnKeyDown := TextKeyDown;

        Tag := integer(@sb);
    end;

end;

function TFrmScriptActions.SpeedButtonToEdit(sb : TSpeedbutton) : TWinControl;
var parent : TWinControl;
    i : integer;
begin
    result := nil;
    parent := sb.Parent;
    for i := 0 to parent.ControlCount - 1 do begin
        if not (parent.Controls[i] is TEdit) and
            not (parent.Controls[i] is TRichEdit) then CONTINUE;

        result := TWinControl(parent.Controls[i]);
    end;
end;
function TFrmScriptActions.EditToChangeTextButton(wc : TWinControl) : TSpeedButton;
var parent : TWinControl;
    i : integer;
begin
    result := nil;
    parent := wc.Parent;
    for i := 0 to parent.ControlCount - 1 do begin
        if not (parent.Controls[i] is TSpeedButton) then CONTINUE;
        if not (TSpeedButton(parent.Controls[i]).tag = 1) then CONTINUE;

        result := TSpeedButton(parent.Controls[i]);
    end;
end;


procedure TFrmScriptActions.GetAddBeforeAfterStep(p: TPanel; sac: TScriptStepControl);
var saa : TScriptStepAddBeforeAfter;
    i : integer;
begin
    p.Height := self.pnlPlaceItAfter.Height;

    saa := TScriptStepAddBeforeAfter.Create;
    saa.ScriptType := stAddBeforeAfter;
    sac.ScriptStep := saa;


    i := StepAddList.add(saa);
    with saa do begin
        lblAdd := TLabel.Create(p);
        DupePosition(p,self.lblAdd, lblAdd, i);

        txtAdd := TMyRichEdit.Create(p);
        DupePosition(p,self.txtAdd, txtAdd, i);
        AddInsert(txtAdd);


        lblToThe := TLabel.Create(p);
        DupePosition(p,self.lblToThe, lblToThe, i);

        cbxAddPart := TComboBox.Create(p);
        DupePosition(p,self.cbxAddPart, cbxAddPart, i);

        lblPlaceIt := TLabel.Create(p);
        DupePosition(p,self.lblPlaceIt, lblPlaceIt, i);

        cbxAddPlace := TComboBox.Create(p);
        DupePosition(p,self.cbxAddPlace, cbxAddPlace, i);

        txtAfter  := TMyRichedit.Create(p);
        DupePosition(p,self.txtAfter, txtAfter, i);
        AddInsert(txtAfter);
        AddWildCards(txtAfter);
    end;
end;
procedure TFrmScriptActions.GetAddBeginningEndingStep(p: TPanel; sac: TScriptStepControl);
var saa : TScriptStepAddBeginningEnding;
    i : integer;
begin
    p.Height := self.pnlBeginning.Height;

    saa := TScriptStepAddBeginningEnding.Create;
    saa.ScriptType := stAddBeginningEnding;
    sac.ScriptStep := saa;

    i := StepAddList.add(saa);
    with saa do begin
        lblAdd := TLabel.Create(p);
        DupePosition(p,self.lblAdd, lblAdd, i);


        txtAdd := TMyRichEdit.Create(p);
        txtAdd.Clear;
        txtAdd.Text := 'hello';
        DupePosition(p,self.txtAdd, txtAdd, i);
        AddInsert(txtAdd);

        lblToThe := TLabel.Create(p);
        DupePosition(p,self.lblToThe, lblToThe, i);

        cbxAddPart := TComboBox.Create(p);
        DupePosition(p,self.cbxAddPart, cbxAddPart, i);
        cbxAddPart.Items[0] := 'Name''s';
        cbxAddPart.Items[1] := 'Extensions''s';
        cbxAddPart.ItemIndex := 0;

//        lblAtThe := TLabel.Create(p);
//        DupePosition(p,self.lblAtThe, lblAtThe, i);

        cbxAddAtThe := TComboBox.Create(p);
        DupePosition(p,self.cbxAddAtThe, cbxAddAtThe, i);

        cbOnCollision := TCheckBox.Create(p);
        DupePosition(p,self.cbOnCollision, cbOnCollision, i);

    end;
end;
procedure TFrmScriptActions.GetAddCounter(p: TPanel; sac: TScriptStepControl);
var saa : TScriptStepAddCounter;
    i : integer;
begin
    p.Height := self.pnlAddCounter.Height;

    saa := TScriptStepAddCounter.Create;
    saa.ScriptType := stAddCounter;
    sac.ScriptStep := saa;

    i := StepAddList.add(saa);
    with saa do begin
        lblAddCounter := TLabel.Create(p);
        DupePosition(p,self.lblAddCounter, lblAddCounter, i);

        pnlAZero := TPanel.Create(p);
        DupePosition(p,self.pnlAZero, pnlAZero);


            txtAZeros := TEdit.Create(pnlAZero);
            DupePosition(pnlAZero,self.txtAZeros, txtAZeros, i);

            AUDZeros := TUpDown.Create(pnlAZero);
            DupePosition(pnlAZero,self.AUDZeros, AUDZeros, i);
            AUDZeros.Associate := txtAZeros;

        lblZeros := TLabel.Create(p);
        DupePosition(p,self.lblZeros, lblZeros, i);

        cbxBeginningEnding := TComboBox.Create(p);
        DupePosition(p,self.cbxBeginningEnding, cbxBeginningEnding, i);

        lblAOfTheName  := TLabel.Create(p);
        DupePosition(p,self.lblAOfTheName, lblAOfTheName, i);
    end;
end;
procedure TFrmScriptActions.GetAddPositionStep(p: TPanel; sac: TScriptStepControl);
var saa : TScriptStepAddPosition;
    i : integer;
begin
    p.Height := self.pnlAtCharPosition.Height;

    saa := TScriptStepAddPosition.Create;
    saa.ScriptType := stAddPosition;
    sac.ScriptStep := saa;

    i := StepAddList.add(saa);
    with saa do begin
        lblAdd := TLabel.Create(p);
        DupePosition(p,self.lblAdd, lblAdd, i);


        txtAdd := TMyRichedit.Create(p);
        DupePosition(p,self.txtAdd, txtAdd, i);
        AddInsert(txtAdd);

        lblToThe := TLabel.Create(p);
        DupePosition(p,self.lblToThe, lblToThe, i);

        cbxAddPart := TComboBox.Create(p);
        DupePosition(p,self.cbxAddPart, cbxAddPart, i);

        lblAtCharacterPosition := TLabel.Create(p);
        DupePosition(p,self.lblAtCharacterPosition, lblAtCharacterPosition, i);

        txtAddPosition := TEdit.Create(p);
        DupePosition(p,self.txtAddPosition, txtAddPosition, i);

        tbPosition := TTrackBar.Create(p);
        DupePosition(p,self.tbPosition, tbPosition, i);
        tbPosition.OnChange := self.tbPosition.OnChange;
        tbPosition.TickMarks := self.tbPosition.TickMarks;
        tbPosition.Position := self.tbPosition.Position;
        tbPosition.Min := self.tbPosition.Min;
        tbPosition.Max := self.tbPosition.Max;
        tbPosition.PageSize := Self.tbPosition.PageSize;

        lblAddFromThe := TLabel.Create(p);
        DupePosition(p,self.lblAddFromThe, lblAddFromThe, i);

        cbxFromTheBeginningEnding:= TComboBox.Create(p);
        DupePosition(p,self.cbxFromTheBeginningEnding, cbxFromTheBeginningEnding, i);
    end;
end;

procedure TFrmScriptActions.GetConditionExecute(p: TPanel; sac: TScriptStepControl);
var sce : TScriptStepConditionExecute;
    i : integer;
begin
    p.Height := self.pnlConditionalExecute.Height;

    sce := TScriptStepConditionExecute.Create;
    sce.ScriptType := stConditionalExecute;
    sac.ScriptStep := sce;

    i := StepOtherList.Add(sce); // ?? What todo here ??
    with sce do begin
        cbxCESkipRemove := TComboBox.create(p);
        DupePosition(p, self.cbxCESkipRemove, cbxCESkipRemove, i);

        lblCESkipTheNext := Tlabel.Create(p);
        DupePosition(p, self.lblCESkipTheNext, lblCESkipTheNext, i);

        pnlCE1 := TPanel.Create(p);
        DupePosition(p,self.pnlCE1, pnlCE1, i);

            txtCE1 := TEdit.Create(pnlCE1);
            DupePosition(pnlCE1,self.txtCE1, txtCE1, i);
            udce1 := TUpDown.Create(pnlCE1);
            DupePosition(pnlCE1,self.udce1, udce1, i);
            udce1.Associate := txtce1;
        
        lblCESteps := Tlabel.Create(p);
        DupePosition(p, self.lblCESteps, lblCESteps, i);

        lblCEIf := Tlabel.Create(p);
        DupePosition(p, self.lblCEIf, lblCEIf, i);

        cbxCECondition := TCombobox.Create(p);
        DupePosition(p, self.cbxCECondition, cbxCECondition, i);

        txtCEText := TMyRichedit.Create(p);
        DupePosition(p, self.txtCEText, txtCEText, i);
        AddInsert(txtCEText);
        AddWildCards(txtCEText);
    end;
end;

procedure TFrmScriptActions.GetRemoveBeforeAfter(p: TPanel; sac: TScriptStepControl);
var sar : TScriptStepRemoveBeforeAfter;
    i : integer;
begin
    p.height := self.pnlRemoveBeforeAfter.Height;

    sar := TScriptStepRemoveBeforeAfter.Create;
    sar.ScriptType := stRemoveBeforeAfter;
    sac.ScriptStep := sar;

    i := StepRemoveList.add(sar);
    p.tag := i;
    with sar do begin
        lblRemoveEverything := TLabel.Create(p);
        DupePosition(p, self.lblRemoveEverything, lblRemoveEverything);

        sar.cbxBeforeAfter := TComboBox.create(p);
        DupePosition(p, self.cbxBeforeAfter, sar.cbxBeforeAfter, i);

        sar.lblThePhrase := TLabel.Create(p);
        DupePosition(p, self.lblThePhrase, sar.lblThePhrase);

        sar.txtRemove2Phrase := TMyRichedit.Create(p);
        DupePosition(p, self.txtRemove2Phrase, sar.txtRemove2Phrase);
        AddInsert(sar.txtRemove2Phrase);
        AddWildCards(sar.txtRemove2Phrase);

        sar.lblAndThePhrase  := TLabel.Create(p);
        DupePosition(p, self.lblAndThePhrase, sar.lblAndThePhrase);

        sar.txtRemove2Phrase2 := TMyRichedit.Create(p);
        DupePosition(p, self.txtRemove2Phrase2, sar.txtRemove2Phrase2);
        AddInsert(sar.txtRemove2Phrase2);
        AddWildCards(sar.txtRemove2Phrase2);


        cbRBReplace := TCheckBox.Create(p);
        DupePosition(p, self.cbRBReplace, cbRBReplace);

        txtRBReplace := TMyRichedit.Create(p);
        DupePosition(p, self.txtRBReplace, txtRBReplace);
        AddInsert(txtRBReplace);

        cbAndStoreRM2 := TCheckBox.Create(p);
        DupePosition(p, self.cbAndStoreRM2, cbAndStoreRM2);
    end;
end;
procedure TFrmScriptActions.GetRemoveCharactersStep(p: TPanel; sac: TScriptStepControl);
var sar : TScriptStepRemoveCharacters ;
    i : integer;
begin
    p.Height := self.pnlRemoveChars.Height;

    sar := TScriptStepRemoveCharacters.Create;
    sar.ScriptType := stRemoveBeginningEnding;
    sac.ScriptStep := sar;

    i := StepRemoveList.add(sar);
    p.tag := i;
    with sar do begin
        lblRemoveChars := TLabel.Create(p);
        DupePosition(p,self.lblRemoveChars, lblRemoveChars);

            pnlR2X := TPanel.Create(p);
            DupePosition(p,self.pnlR2X, pnlR2X);
                txtR2X := TEdit.Create(pnlR2X);
                DupePosition(pnlR2X,self.txtR2X, txtR2X);
                udR2X := TUpDown.Create(pnlR2X);
                DupePosition(pnlR2X,self.udR2X, udR2X);
                udR2X.Associate := txtR2X;

        lblCharsFromThe  := TLabel.Create(p);
        DupePosition(p,self.lblCharsFromThe, lblCharsFromThe);

        cbxRemoveCharPosition  := TComboBox.Create(p);
        DupePosition(p,self.cbxRemoveCharPosition, cbxRemoveCharPosition);

        lblOfThe  := TLabel.Create(p);
        DupePosition(p,self.lblOfThe, lblOfThe);

        cbxRemoveCharactersPart  := TComboBox.Create(p);
        DupePosition(p,self.cbxRemoveCharactersPart, cbxRemoveCharactersPart);

        cbRCReplace := TCheckBox.Create(p);
        DupePosition(p, self.cbRCReplace, cbRCReplace);

        txtRCReplace := TMyRichedit.Create(p);
        DupePosition(p, Self.txtRCReplace, txtRCReplace);
        AddInsert(txtRCReplace);

        cbAndStore := TCheckBox.Create(p);
        DupePosition(p, self.cbAndStore, cbAndStore);

    end;
end;
procedure TFrmScriptActions.GetRemoveExpression(p: TPanel; sac: TScriptStepControl);
var sar : TScriptStepRemoveExpression;
    i : integer;
begin
    p.Height := self.pnlRemoveExpression.Height;
    p.ParentColor := false;
    p.Color := ColorToRGB(clWHite);


    sar := TScriptStepRemoveExpression.create;
    sar.ScriptType := stRemoveExpression;
    sac.ScriptStep := sar;

    i := StepRemoveList.add(sar);
    p.tag := i;

    with sar do begin
        cbxRERemoveReplace := TComboBox.Create(p);
        DupePosition(p, self.cbxRERemoveReplace, cbxRERemoveReplace, i);

        lblreRemoveTheExpression := TLabel.Create(p);
        DupePosition(p,self.lblreRemoveTheExpression, lblreRemoveTheExpression);

        cbreAtThe := TCheckBox.Create(p);
        DupePosition(p,self.cbreAtThe, cbreAtThe);
        cbxreBegginingEnding := TComboBox.Create(p);
        DupePosition(p,self.cbxreBegginingEnding, cbxreBegginingEnding);
        lblreOfTheName := TLabel.Create(p);
        DupePosition(p,self.lblreOfTheName, lblreOfTheName);



        cbreThatSTartsWIth := TCheckBox.Create(p);
        DupePosition(p,self.cbreThatSTartsWIth, cbreThatSTartsWIth);
        cbxrePhraseCharacters := TComboBox.Create(p);
        DupePosition(p,self.cbxrePhraseCharacters, cbxrePhraseCharacters);
        txtrePhraseCharacters := TMyRichedit.Create(p);
        DupePosition(p,self.txtrePhraseCharacters, txtrePhraseCharacters);
        AddInsert(txtrePhraseCharacters);

        cbreHasInTheMiddle := TCheckBox.Create(p);
        DupePosition(p,self.cbreHasInTheMiddle, cbreHasInTheMiddle);
        cbxrePhraseCharacters2 := TComboBox.Create(p);
        DupePosition(p,self.cbxrePhraseCharacters2, cbxrePhraseCharacters2);
        txtrePhraseCharacters2 := TMyRichedit.Create(p);
        DupePosition(p,self.txtrePhraseCharacters2, txtrePhraseCharacters2);
        AddInsert(txtrePhraseCharacters2);



        cbreAndEndsWith := TCheckBox.Create(p);
        DupePosition(p,self.cbreAndEndsWith, cbreAndEndsWith);
        cbxrePhraseCharacters3 := TComboBox.Create(p);
        DupePosition(p,self.cbxrePhraseCharacters3, cbxrePhraseCharacters3);
        txtrePhraseCharacters3 := TMyRichedit.Create(p);
        DupePosition(p,self.txtrePhraseCharacters3, txtrePhraseCharacters3);
        AddInsert(txtrePhraseCharacters3);

        lblReplaceExpressionWith := TLabel.Create(p);
        DupePosition(p, self.lblReplaceExpressionWith, lblReplaceExpressionWith);

        txtREWith := TMyRichedit.Create(p);
        DupePosition(p, self.txtREWith, txtREWith);
        AddInsert(txtREWith);



        sar.lblRERemoveTheParts := TLabel.Create(p);
        DupePosition(p, self.lblRERemoveTheParts, sar.lblRERemoveTheParts);
        self.lblRERemoveTheParts.Visible := true;

        sar.cbxREParts := TComboBox.Create(p);
        DupePosition(p, self.cbxREParts, sar.cbxREParts, i);

        cbreRemoveNumber := TCheckBox.Create(p);
        DupePosition(p, self.cbreRemoveNumber, cbreRemoveNumber);
        pnlRE1 := TPanel.Create(p);
        DupePosition(p, self.pnlRE1, pnlRE1);
        txtRE1 := TEdit.Create(pnlRE1);
        DupePosition(pnlRE1, self.txtRE1, txtRE1);
        udRE1 := TUpDown.create(pnlRE1);
        DupePosition(pnlRE1, self.udRE1, udRE1);
        udRE1.Associate := txtRE1;

        cbreAndStore := TCheckBox.Create(p);
        DupePosition(p,self.cbreAndStore, cbreAndStore);

    end;
end;
procedure TFrmScriptActions.GetRemoveSubstringStep(p: TPanel; sac: TScriptStepControl);
var sar : TScriptStepRemoveSubstring;
    i : integer;
begin
    p.Height := self.pnlRemoveSubstring.Height;

    sar := TScriptStepRemoveSubstring.Create;
    sar.ScriptType := stRemoveSubstring;
    sac.ScriptStep := sar;

    i := StepRemoveList.add(sar);
    p.tag := i;
    with sar do begin
        lblRemoveSubstring   := TLabel.Create(p);
        DupePosition(p,self.lblRemoveSubstring, lblRemoveSubstring);

            cbxRemoveType := TComboBox.Create(p);
            DupePosition(p, self.cbxRemoveType, cbxRemoveType, i);


            lblChars := TLabel.Create(p);
            DupePosition(p, self.lblChars, lblChars);

            pnlR3X := TPanel.Create(p);
            DupePosition(p,self.pnlR3X, pnlR3X);
                txtR3X := TEdit.Create(pnlR3X);
                DupePosition(pnlR3X,self.txtR3X, txtR3X);
                udR3X := TUpDown.Create(pnlR3X);
                DupePosition(pnlR3X,self.udR3X, udR3X);
                udR3X.Associate := txtR3X;

            lblCharThru := TLabel.Create(p);
            DupePosition(p, self.lblCharThru, lblCharThru);

            pnlR3Y := TPanel.Create(p);
            DupePosition(p,self.pnlR3Y, pnlR3Y);
                txtR3Y := TEdit.Create(pnlR3Y);
                DupePosition(pnlR3Y,self.txtR3Y, txtR3Y);
                udR3Y := TUpDown.Create(pnlR3Y);
                DupePosition(pnlR3Y,self.udR3Y, udR3Y);
                udR3Y.Associate := txtR3Y;


            cbAndMove := TCheckBox.Create(p);
            DupePosition(p, self.cbAndMove, cbAndMove);
            pnlR3Z := TPanel.Create(p);
            DupePosition(p,self.pnlR3Z, pnlR3Z);
                txtR3Z := TEdit.Create(pnlR3Z);
                DupePosition(pnlR3z,self.txtR3Z, txtR3Z);
                udR3Z := TUpDown.Create(pnlR3Z);
                DupePosition(pnlR3z,self.udR3Z, udR3Z);
                udR3Z.Associate := txtR3Z;
        cbAndStoreRM := TCheckBox.Create(p);
        DupePosition(p, self.cbAndSToreRM, cbAndStoreRM);
    end;
end;

procedure TFrmScriptActions.GetRemoveWord(p: TPanel; sac: TScriptStepControl);
begin

end;
procedure TFrmScriptActions.GetRemoveWordNumber(p: TPanel; sac: TScriptStepControl);
var sar : TScriptStepRemoveWordNumber;
    i : integer;
begin
    p.height := self.pnlRemoveWord.Height;

    sar := TScriptStepRemoveWordNumber.create;
    sar.ScriptType := stRemoveWordNumber;
    sac.ScriptStep := sar;

    i := StepRemoveList.add(sar);
    p.Tag := i;

    with sar do begin
        lblrwRemoveWordNumber := Tlabel.Create(p);
        DupePosition(p, self.lblrwRemoveWordNumber, lblrwRemoveWordNumber);

        pnlRW1 := TPanel.Create(p);
        DupePosition(p, self.pnlRW1, pnlRW1);
        txtRW1 := TEdit.create(pnlRW1);
        DupePosition(pnlRW1, self.txtRW1, txtRW1);
        udRW1 := TUpDown.Create(pnlRW1);
        DupePosition(pnlRW1, self.udRW1, udRW1);
        udRW1.Associate := txtRW1;

        lblrwAWordContains := TLabel.Create(p);
        DupePosition(p, self.lblrwAWordContains, lblrwAWordContains);

        cbxrwCharactersType := TComboBox.create(p);
        DupePosition(p, self.cbxrwCharactersType, cbxrwCharactersType, i);

        txtrwCharacters := TEdit.create(p);
        DupePosition(p, self.txtrwCharacters, txtrwCharacters);

        cbrwStoreTemp := TCheckBox.create(p);
        DupePosition(p, self.cbrwStoreTemp, cbrwStoreTemp);
    end;
end;
procedure TFrmScriptActions.GetRemoveWordStep(p: TPanel; sac: TScriptStepControl);
var sar : TScriptStepRemoveWords;
    i : integer;
begin
    p.Height := self.pnlRemove.Height;

    sar := TScriptStepRemoveWords.Create;
    sar.ScriptType := stRemoveWords;
    sac.ScriptStep := sar;

    i := StepRemoveList.add(sar);
    p.tag := i;
    with sar do begin
        lblRemoveWords  := TLabel.Create(p);
        DupePosition(p, self.lblRemoveWords, lblRemoveWords);

        cbxRemovePhrase := TComboBox.Create(p);
        DupePosition(p, self.cbxRemovePhrase, cbxRemovePhrase, i);

        txtRemovePhrase := TMyRichedit.Create(p);
        DupePosition(p, self.txtRemovePhrase, txtRemovePhrase);
        AddInsert(txtRemovePhrase);
        AddWildCards(txtRemovePhrase);

        lblFromThe  := TLabel.Create(p);
        DupePosition(p, self.lblFromThe, lblFromThe);


        cbxRemovePhrasePart := TComboBox.Create(p);
        DupePosition(p, self.cbxRemovePhrasePart, cbxRemovePhrasePart);
    end;
end;

procedure TFrmScriptActions.GetReplaceNameExtensionStep(p: TPanel; sac: TScriptStepControl);
var sar : TScriptStepReplaceNameExtension;
    i : integer;
begin
    p.Height := self.pnlReplace.Height;

    sar := TScriptStepReplaceNameExtension.Create;
    sar.ScriptType := stReplaceNameExtension;
    sac.ScriptStep := sar;

    i := StepReplaceList.add(sar);
    with sar do begin
        lblReplaceThe := TLabel.Create(p);
        DupePosition(p,self.lblReplaceThe, lblReplaceThe);

        cbxReplacePart := TComboBox.Create(p);
        DupePosition(p,self.cbxReplacePart,cbxReplacePart);
        cbxReplacePart.Tag := i;


        txtReplacePhrase := TMyRichedit.Create(p);
        DupePosition(p,self.txtReplacePhrase, txtReplacePhrase);
        AddInsert(txtReplacePhrase);
        AddWildCards(txtReplacePhrase);

        lblWith := TLabel.Create(p);
        DupePosition(p,self.lblWith, lblWith);

        txtReplaceWith  := TMyRichedit.Create(p);
        DupePosition(p,self.txtReplaceWith, txtReplaceWith);
        AddInsert(txtReplaceWith);

        pnlReplaceOnly := TPanel.Create(p);
        pnlReplaceOnly.BevelOuter := bvNone;
        DupePosition(p,self.pnlReplaceOnly, pnlReplaceOnly);

            cbReplaceOnly  := TCheckBox.Create(pnlReplaceOnly);
            DupePosition(pnlReplaceOnly,self.cbReplaceOnly, cbReplaceOnly);

            txtRP2X  := TEdit.Create(pnlReplaceOnly);
            DupePosition(pnlReplaceOnly,self.txtRP2X, txtRP2X);
            udRP2X   := TUpDown.Create(pnlReplaceOnly);
            DupePosition(pnlReplaceOnly,self.udRP2X, udRP2X);
            udRP2X.Associate := txtRP2X;

            lblThru  := TLabel.Create(pnlReplaceOnly);
            DupePosition(pnlReplaceOnly,self.lblThru, lblThru);

            txtRP2Y  := TEdit.Create(pnlReplaceOnly);
            DupePosition(pnlReplaceOnly,self.txtRP2Y, txtRP2Y);
            udRP2Y   := TUpDown.Create(pnlReplaceOnly);
            DupePosition(pnlReplaceOnly,self.udRP2Y, udRP2Y);
            udRP2Y.Associate := txtRP2Y;

            lblRMInThe := TLabel.Create(pnlReplaceOnly);
            DupePosition(pnlReplaceOnly,self.lblRMInThe, lblRMInThe);
            cbxReplaceTarget := TComboBox.Create(pnlReplaceOnly);
            DupePosition(pnlReplaceOnly,self.cbxReplaceTarget, cbxReplaceTarget);

        cbxReplacePart.ItemIndex :=  REPLACEPART_PHRASE; // default to Replace Phrase
        cbxReplacePartChange(cbxReplacePart);
        cbxReplaceTarget.ItemIndex := REPLACETARGET_NAME_OR_EXT;
        // default to Name or Extension, since this was the previous only option
        // this won't break existing scripts
    end;
end;
procedure TFrmScriptActions.GetReplaceCharactersStep(p: TPanel; sac: TScriptStepControl);
var sar : TScriptStepReplaceCharacters ;
    i : integer;
begin
    p.Height := self.pnlReplaceChar.Height;

    sar := TScriptStepReplaceCharacters.Create;
    sar.ScriptType := stReplaceCharacters;
    sac.ScriptStep := sar;

    i := StepReplaceList.add(sar);
    p.tag := i;
    with sar do begin
        lblReplaceEach := TLabel.Create(p);
        DupePosition(p,self.lblReplaceEach, lblReplaceEach);

        txtReplaceChar1 := TEdit.Create(p);
        DupePosition(p,self.txtReplaceChar1, txtReplaceChar1);

        lblWithChar := TLabel.Create(p);
        DupePosition(p,self.lblWithChar, lblWithChar);

        txtReplaceChar2 := TEdit.Create(p);
        DupePosition(p,self.txtReplaceChar2, txtReplaceChar2);

        lblInThe  := TLabel.Create(p);
        DupePosition(p,self.lblInThe, lblInThe);

        cbxReplaceCharPart :=  TComboBox.Create(p);
        DupePosition(p,self.cbxReplaceCharPart, cbxReplaceCharPart);

    end;
end;
procedure TFrmScriptActions.GetReplaceNumbersStep(p: TPanel; sac: TScriptStepControl);
var sar : TScriptStepReplaceNumbers;
    i : integer;
begin
    p.Height := self.pnlReplaceNumbers.Height;

    sar := TScriptStepReplaceNumbers.Create;
    sar.ScriptType := stReplaceNumbers;
    sac.ScriptStep := sar;

    i := StepReplaceList.add(sar);
    //p.tag := i;
    with sar do begin
        lblReplaceNumbers := TLabel.Create(p);
        DupePosition(p,self.lblReplaceNumbers, lblReplaceNumbers);

        cbxNumbers := TComboBox.Create(p);
        DupePosition(p,self.cbxNumbers, cbxNumbers);

        lblOfTheName := TLabel.Create(p);
        DupePosition(p,self.lblOfTheName, lblOfTheName);

        cbxRNSequence := TComboBox.Create(p);
        DupePosition(p,self.cbxRNSequence, cbxRNSequence, i);

        txtRNWith := TMyRichedit.Create(p);
        DupePosition(p,self.txtRNWith, txtRNWith, i);
        self.AddInsert(txtRNWith);

        Reset;
    end;
end;
procedure TFrmScriptActions.GetSplitAndMove(p: TPanel; sac: TScriptStepControl);
var ssm : TScriptStepSplitAndMove;
    i : integer;
begin
    p.Height := self.pnlSplit.Height;

    ssm := TScriptStepSplitAndMove.Create;
    ssm.ScriptType := stSplitAndMove;
    sac.ScriptStep := ssm;

    i := StepOtherList.Add(ssm); // ?? What todo here ??
    with ssm do begin
        lblSplitTheName := Tlabel.Create(p);
        DupePosition(p, self.lblSplitTheName, lblSplitTheName, i);

        txtFilemask := TEdit.Create(p);
        DupePosition(p, self.txtFilemask, txtFilemask, i);

        btnAutoSplit := TButton.Create(p);
        DupePosition(p, self.btnAutoSplit, btnAutoSplit, i);

        lblMoveTheParts := Tlabel.Create(p);
        DupePosition(p, self.lblMoveTheParts, lblMoveTheParts, i);

        txtMove := TMyRichedit.Create(p);
        DupePosition(p, self.txtMove, txtMove, i);
        AddInsert(txtMove);

        Reset;
    end;

end;
procedure TFrmScriptActions.GetSelectAndMove(p : TPanel; sac : TScriptStepControl );
var ssm : TScriptStepSelectAndMove;
    i : integer;
begin
    p.Height := self.pnlSelectAndMove.Height;

    ssm := TScriptStepSelectAndMove.Create;
    ssm.ScriptType := stSelectAndMove;
    sac.ScriptStep := ssm;

    i := StepRemoveList.Add(ssm); // ?? What todo here ??
    with ssm do begin
    pnlSelect := TPanel.Create(p);
    DupePosition(p,self.pnlSelect, pnlSelect,i);

        rbSMBeforeAfter:= TRadioButton.Create(pnlSelect);
        DupePosition(pnlSelect,self.rbSMBeforeAfter,rbSMBeforeAfter,i);

        rbSMBeginningEnding:= TRadioButton.Create(pnlSelect);
        DupePosition(pnlSelect,self.rbSMBeginningEnding,rbSMBeginningEnding,i);

        rbSMPosition:= TRadioButton.Create(pnlSelect);
        DupePosition(pnlSelect,self.rbSMPosition,rbSMPosition,i);


        cbxSMBeforeAfter:= TComboBox.Create(pnlSelect);
        DupePosition(pnlSelect,self.cbxSMBeforeAfter,cbxSMBeforeAfter,i);

        txtSMPhrase:= TMyRichedit.Create(pnlSelect);
        DupePosition(pnlSelect,self.txtSMPhrase,txtSMPhrase,i);
        AddInsert(txtSMPhrase);
        AddWildCards(txtSMPhrase);


        cbxSMBeginningEnding:= TComboBox.Create(pnlSelect);
        DupePosition(pnlSelect,self.cbxSMBeginningEnding,cbxSMBeginningEnding,i);

        pnlSMZ1:= TPanel.Create(pnlSelect);
        DupePosition(pnlSelect,self.pnlSMZ1,pnlSMZ1,i);
            txtSMZ1:= TEdit.Create(pnlSMZ1);
            DupePosition(pnlSMZ1,self.txtSMZ1,txtSMZ1,i);
            udSMZ1:= TUpDown.Create(pnlSMZ1);
            DupePosition(pnlSMZ1,self.udSMZ1,udSMZ1,i);
            udSMZ1.Associate := txtSMZ1;


        pnlSMX1:= TPanel.Create(pnlSelect);
        DupePosition(pnlSelect,self.pnlSMX1,pnlSMX1,i);
            txtSMX1:= TEdit.Create(pnlSMX1);
            DupePosition(pnlSMX1,self.txtSMX1,txtSMX1,i);
            udSMX1:= TUpDown.Create(pnlSMX1);
            DupePosition(pnlSMX1,self.udSMX1,udSMX1,i);
            udSMX1.Associate := txtSMX1;

        pnlSMY1:= TPanel.Create(pnlSelect);
        DupePosition(pnlSelect,self.pnlSMY1,pnlSMY1,i);
            txtSMY1:= TEdit.Create(pnlSMY1);
            DupePosition(pnlSMY1,self.txtSMY1,txtSMY1,i);
            udSMY1:= TUpDown.Create(pnlSMY1);
            DupePosition(pnlSMY1,self.udSMY1,udSMY1,i);
            udSMY1.Associate := txtSMY1;

        lblSMSelect:= TLabel.Create(pnlSelect);
        DupePosition(pnlSelect,self.lblSMSelect,lblSMSelect,i);

        lblSMCharacters:= TLabel.Create(pnlSelect);
        DupePosition(pnlSelect,self.lblSMCharacters,lblSMCharacters,i);

        lblSMThru:= TLabel.Create(pnlSelect);
        DupePosition(pnlSelect,self.lblSMThru,lblSMThru,i);


    pnlMove := TPanel.Create(p);
    DupePosition(p, self.pnlMove, pnlMove, i);
        lblSMMoveTo:= TLabel.Create(pnlMove);
        DupePosition(pnlMove, Self.lblSMMoveTo, lblSMMoveTo,i);

        rbSMBeforeAfter2:= TRadioButton.Create(pnlMove);
        DupePosition(pnlMove, self.rbSMBeforeAfter2, rbSMBeforeAfter2,i);
        rbSMBeginningEnding2:= TRadioButton.Create(pnlMove);
        DupePosition(pnlMove,Self.rbSMBeginningEnding2 ,rbSMBeginningEnding2 ,i);
        rbSMPosition2:= TRadioButton.Create(pnlMove);
        DupePosition(pnlMove, self.rbSMPosition2, rbSMPosition2,i);


        cbxSMBeforeAfter2:= TComboBox.Create(pnlMove);
        DupePosition(pnlMove,self.cbxSMBeforeAfter2 ,cbxSMBeforeAfter2 ,i);
        txtSMPhrase2:= TMyRichedit.Create(pnlMove);
        DupePosition(pnlMove, self.txtSMPhrase2, txtSMPhrase2,i);
        AddInsert(txtSMPhrase2);
        AddWildCards(txtSMPhrase2);

        cbxSMBeginningEnding2:= TComboBox.Create(pnlMove);
        DupePosition(pnlMove,self.cbxSMBeginningEnding2 , cbxSMBeginningEnding2,i);


        pnlSMX2:= TPanel.Create(pnlMove);
        DupePosition(pnlMove,self.pnlSMX2,pnlSMX2,i);
            txtSMX2:= TEdit.Create(pnlSMX2);
            DupePosition(pnlSMX2,self.txtSMX2,txtSMX2,i);
            udSMX2:= TUpDown.Create(pnlSMX2);
            DupePosition(pnlSMX2,self.udSMX2,udSMX2,i);
            udSMX2.Associate := txtSMX2;
    end;
end;
procedure TFrmScriptActions.GetSwap(p : TPanel; sac : TScriptStepControl );
var ssm : TScriptStepSwap;
    i : integer;
begin
    p.Height := self.pnlSwap.Height;

    ssm := TScriptStepSwap.Create;
    ssm.ScriptType := stSwap;
    sac.ScriptStep := ssm;


    i := StepReplaceList.Add(ssm); // ?? What todo here ??
    with ssm do begin

        lblSSwap:= TLabel.Create(p);
        DupePosition(p,self.lblSSwap,lblSSwap,i);


        txtSPhrase:= TMyRichedit.Create(p);
        DupePosition(p,self.txtSPhrase,txtSPhrase,i);
        AddInsert(txtSPhrase);

        cbSTrim:= TCheckBox.Create(p);
        DupePosition(p,self.cbSTrim,cbSTrim,i);
    end;
end;
procedure TFrmScriptActions.GetFindAndEdit(p : TPanel; sac : TScriptStepControl );
var ssm : TScriptStepFindAndEdit;
    i : integer;
begin
    p.Height := self.pnlFindAndEdit.Height;

    ssm := TScriptStepFindAndEdit.Create;
    ssm.ScriptType := stFindAndEdit;
    sac.ScriptStep := ssm;


    i := StepReplaceList.Add(ssm);
    with ssm do begin
        flblFEFindINTheName := TLabel.Create(p);
        DupePosition(p, self.lblFEFindINTheName, lblFEFindINTheName, i);

        ftxtFEPhrase1 := TMyRichedit.Create(p);
        DupePosition(p, self.txtFEPhrase1, txtFEPhrase1, i);
        AddInsert(ftxtFEPhrase1);
        AddWildCards(ftxtFEPhrase1);

        fcbFEFromThe := TCheckBox.Create(p);
        DupePosition(p, self.cbFEFromThe, cbFEFromThe, i);

        fcbxFEBeginningEnding := TComboBox.Create(p);
        DupePosition(p, self.cbxFEBeginningEnding, cbxFEBeginningEnding, i);

        flblFEReplacePosition := TLabel.Create(p);
        DupePosition(p, self.lblFEReplacePosition, lblFEReplacePosition, i);

        fpnlFEX1 := TPanel.Create(p);
        DupePosition(p, self.pnlFEX1, pnlFEX1, i);

            ftxtFEX1 := TEdit.Create(fpnlFEX1);
            DupePosition(fpnlFEX1, self.txtFEX1, txtFEX1, i);
            fudFEX1 := TUpDown.Create(fpnlFEX1);
            DupePosition(fpnlFEX1, self.udFEX1, udFEX1, i);
            fudFEX1.Associate := txtFEX1;

        flblFELength := TLabel.Create(p);
        DupePosition(p, self.lblFELength, lblFELength, i);

        fpnlFEY1 := TPanel.Create(p);
        DupePosition(p, self.pnlFEY1, pnlFEY1, i);

            ftxtFEY1 := TEdit.Create(fpnlFEY1);
            DupePosition(fpnlFEY1, self.txtFEY1, txtFEY1, i);
            fudFEY1 := TUpDown.Create(fpnlFEY1);
            DupePosition(fpnlFEY1, self.udFEY1, udFEY1, i);
            fudFEY1.Associate := txtFEY1;

        flblFEWith := TLabel.Create(p);
        DupePosition(p, self.lblFEWith, lblFEWith, i);

        ftxtFEPhrase2 := TMyRichedit.Create(p);
        DupePosition(p, self.txtFEPhrase2, txtFEPhrase2, i);
        AddInsert(ftxtFEPhrase2);

        fcbFEInsert := TCheckBox.Create(p);
        DupePosition(p, self.cbFEInsert, cbFEInsert, i);

        ftxtFEPhrase3 := TMyRichedit.Create(p);
        DupePosition(p, self.txtFEPhrase3, txtFEPhrase3, i);
        AddInsert(ftxtFEPhrase3);

        fcbxFEBeforeAFter := TComboBox.Create(p);
        DupePosition(p, self.cbxFEBeforeAFter, cbxFEBeforeAFter, i);


        fcbxFEBeforeAFter := TComboBox.Create(p);
        DupePosition(p, self.cbxFEBeforeAFter, cbxFEBeforeAFter, i);

        ftxtFEPhrase4 := TMyRichedit.Create(p);
        DupePosition(p, self.txtFEPhrase4, txtFEPhrase4, i);
        AddInsert(ftxtFEPhrase4);
        AddWildCards(ftxtFEPhrase4);
    end;

end;


function TFrmScriptActions.GetScriptStepAdd(i: integer): TScriptStepAdd;
begin
    result := TScriptStepAdd(StepAddList.items[i]);
end;
function TFrmScriptActions.GetScriptControl(i: integer): TScriptStepControl;
begin
    result := nil;
    if (i >= ScriptControlList.Count) or (i<0) then EXIT;
    result := TScriptStepControl(ScriptControlList.Items[i]);
end;
{$endregion}
//====================


//===============
{$region '// Text Context Menu stuff - ADDINSERT//'}
procedure TFrmScriptActions.FormatVariable1Click(Sender: TObject);
begin
    frmMetaVariables.ShowFormFormat(LastText);
end;
procedure TFrmScriptActions.AddPadding1Click(Sender: TObject);
begin
    frmMetaVariables.ShowFormFormat(LastText, 1);
end;
procedure TFrmScriptActions.AddDecimal1Click(Sender: TObject);
begin
frmMetaVariables.ShowFormFormat(LastText, 2);
end;
procedure TFrmScriptActions.FormatCorrectDate1Click(Sender: TObject);
begin
frmMetaVariables.ShowFormFormat(LastText, 3);
end;
procedure TFrmScriptActions.Hex1Click(Sender: TObject);
begin
frmMetaVariables.ShowFormFormat(LastText, 4);
end;
procedure TFrmScriptActions.Substring1Click(Sender: TObject);
begin
frmMetaVariables.ShowFormFormat(LastText, 5);
end;
procedure TFrmScriptActions.Letters1Click(Sender: TObject);
begin
frmMetaVariables.ShowFormFormat(LastText, 6);
end;

procedure TFrmScriptActions.More1Click(Sender: TObject);
var p : TPoint;
begin
    p := LastText.ClientToScreen(LastText.BoundsRect.TopLeft);
    poRightClick.Popup(p.X, p.y+ LastText.Height)
end;
procedure TFrmScriptActions.Copy1Click(Sender: TObject);
begin
    if LastText is TMyRichEdit then begin
        clipboard.SetTextBuf(pwidechar(TMyRichEdit(LastText).SelText));
    end else begin
        LastText.CopyToClipboard;
    end;
end;
procedure TFrmScriptActions.Cut1Click(Sender: TObject);
begin
    Copy1Click(Sender);
    Delete2Click(Sender);
end;
procedure TFrmScriptActions.Delete2Click(Sender: TObject);
begin
    LastText.ClearSelection;
end;
procedure TFrmScriptActions.Paste1Click(Sender: TObject);
begin
    LastText.PasteFromClipboard;
end;
procedure TFrmScriptActions.FormatMetaVariable1Click(Sender: TObject);
begin
    frmMetaVariables.ShowFormFormat(LastText);
end;
{$endregion}
//===============


//======================
{$region '// Generic Event Handlers for all controls //'}
var ignorechange : boolean;
procedure TFrmScriptActions.SelectionNotify(sac : TScriptStepControl);
var
    seh : TScriptEventHandler;
begin
    if assigned(fListeners) then
        for seh in fListeners do begin
            seh.NewSelectionEvent(sac.outerpnl.BoundsRect);
        end;
end;


procedure TFrmScriptActions.GenericClickChangeEvent(Sender: TObject);
begin
    if not LastParentControl.Visible then EXIT;

    self.HandleKeypress(sender);

    {
    if ignorechange then EXIT;
    if sender is TMyRichEdit then begin
        ignorechange := true;
        FormatVariableText(TRichEdit(Sender));
        ignorechange := false;
    end;
    }
end;
procedure TFrmScriptActions.GenericKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
    {
    // ignore all tab navigations
    if (key = VK_TAB) then EXIT;
    // ignore single modified keys
    if (shift = []) then begin
        if (key in [VK_SHIFT, VK_CONTROL, VK_MENU]) then EXIT;
    end;
    // ignore out Step navigation keystroke
    if (key in [VK_PRIOR, VK_NEXT]) and (shift = [ssCtrl]) then EXIT;
    }

    // React to cursor navigation keystrokes to delay the auto-preview
    if (key in [VK_LEFT, VK_RIGHT, VK_HOME, VK_END]) then
        self.HandleKeypress(sender);
end;
procedure TFrmScriptActions.GenerickKeypress(Sender: TObject; var Key: Char);
begin
//
end;
procedure TFrmScriptActions.GenericOnEnter(Sender: TObject);

    function GetMainStepControlIndex(sender: TObject) : integer;
    var basewc, wc : TWinControl;
    begin
        result := -1;
        basewc := TWinControl(sender);
        wc := basewc;
        repeat
            wc := wc.parent;
        until (wc = nil) or (wc is TTransPanel);
        if wc = nil then begin
            EXIT;
        end;
        result := wc.Tag;
    end;
    function GetChangeTextBoxButton(editControl : TWinControl) : TSpeedButton;
    var parent : TWinControl;
        i : integer;
    begin
        result := nil;
        parent := editcontrol.Parent;
        for i := 0 to parent.ControlCount - 1 do begin
            if not (parent.Controls[i] is TSpeedButton) then CONTINUE;
            if not (TSpeedButton(parent.Controls[i]).tag = 1) then CONTINUE;

            result := TSpeedButton(parent.Controls[i]);
        end;
    end;

    procedure SetButtonVisibile(sb : TSpeedButton; value : boolean);
    begin
        sb.Visible := value;
    end;
var i : integer;
    sc : TScriptStepControl;
    newSelection : boolean;
    sb : TSpeedButton;

begin
    // use the control to figure out which Step Control is selected
    i := GetMainStepControlIndex(sender);
    if (i <> -1) then begin
        sc := self.GetScriptControl(i);
        newSelection := sc <> self.fLastSelectedControl;
        self.setLastSelectedControl(sc);
        if newSelection then begin
            self.SelectionNotify(sc);
        end;
    end;




    // will reenable this when I figure out the details when changing
    // type is needed/possible
    if (sender is TCustomEdit) then
        if (fLastSelectedEdit <> TControl(sender)) then begin
            if (fLastSelectedEdit <> nil)  then begin
                sb := self.EditToChangeTextButton(fLastSelectedEdit);
                if (sb <> nil) then begin
                    SetButtonVisibile(sb, false);
                end;
            end;

            fLastSelectedEdit := TCustomEdit(sender);
            if (wildcardcontrols.Contains(fLastSelectedEdit)) then begin
                sb := self.EditToChangeTextButton(fLastSelectedEdit);
                if (sb <> nil) then begin
                    SetButtonVisibile(sb, true);
                end;
                fLastSelectedEditControlIndex := i;
            end;
        end;

end;
procedure TFrmScriptActions.HandleKeypress(sender: Tobject);
var
    sc : TScriptStepControl;
    wc : TWinControl;
begin
    if not frmmain.Visible then EXIT;
    if loading then EXIT;
    if expanding then
        EXIT;
    if timIgnoreChanges.Enabled then EXIT;

    wc := TWinControl(sender);
    repeat
        wc := wc.parent;

    until (wc.parent = self.LastParentControl) or (wc = nil);
    if (wc = nil) then begin
        // TODO - log the error
        FrmDummyAutoPreview.InformOfKeypress;
        EXIT;
    end;

    sc := TScriptStepControl(ScriptControlList.Items[wc.Tag]);
    if not sc.ScriptStep.Enabled  then
        self.EnableClick(sc.pnlTop.Tag );

    FrmDummyAutoPreview.InformOfKeypress;

    if frmConfig.cbAlwaysShowStepDescriptions.Checked then begin
        self.UpdateSentence(sc);
    end;

    {
    if Sender is TCustomEdit then begin
        with Sender as TCustomEdit do begin
            if AnsiStartsText(' ', Text) or AnsiEndsText(' ', Text) then begin
                Font.Style := Font.Style + [fsUnderline]
            end else begin
                Font.style := Font.Style - [fsUnderline];
            end;
        end;
    end;
    }
end;
{$endregion}
//======================


// Format / Wildcard popup stuff
procedure TFrmScriptActions.miEnableRegExp1Click(Sender: TObject);
var ce : TCustomEdit;
begin
    miEnableRegExp1.Checked := not miEnableRegExp1.Checked;
    if miEnableRegExp1.Checked then begin
        TMyRichEdit(self.fLastSelectedEdit).EnableVariableFormatting;
    end;



    ce := TCustomEdit(self.fLastSelectedEdit);
    if not ContainsText(ce.text, '?reg?') then begin
        // add because doesn't exist
        if miEnableRegExp1.Checked then begin
            ce.Text := '?reg?' + ce.Text;
        end;
    end else begin
        // remove because does exists
        if not miEnableRegExp1.Checked then begin
            ce.Text := StringReplace(ce.text,'?reg?','', [rfReplaceAll]);
        end;
    end;
end;
procedure TFrmScriptActions.miInsertexactly1character1Click(Sender: TObject);
begin
    fLastSelectedEdit.SelText := '?';

end;
procedure TFrmScriptActions.miInsertoneormorecharacters1Click(Sender: TObject);
begin
    fLastSelectedEdit.SelText := '?*';
end;
procedure TFrmScriptActions.miInsertzeroormorecharacters1Click(Sender: TObject);
begin
    fLastSelectedEdit.SelText := '*';
end;

procedure TFrmScriptActions.miEnabledWildCardsnoformatting1Click(Sender: TObject);
var ce : TCustomEdit;
    s : string;
begin
    miEnabledWildCardsnoformatting1.Checked := not miEnabledWildCardsnoformatting1.Checked;
    //miEnableRegExp1.Checked := not miEnabledWildCardsnoformatting1.Checked;

    if miEnabledWildCardsnoformatting1.Checked then begin
        TMyRichEdit(self.fLastSelectedEdit).DisableVariableFormatting;
        self.fLastSelectedEdit.text := StringReplace(self.fLastSelectedEdit.text,'?reg?','', [rfReplaceAll]);
    end else begin
        TMyRichEdit(self.fLastSelectedEdit).EnableVariableFormatting;
    end;

    s := self.flastSelectedEdit.text;
    self.flastSelectedEdit.text := '';
    self.flastSelectedEdit.text :=s;
end;
procedure TFrmScriptActions.pmStepRightClick1Popup(Sender: TObject);
var sc : TScriptStepControl;
begin
    if fStepClicked <> nil then begin
        sc := self.GetScriptControl(self.fStepClicked.tag);
        MIKeepOpen.Checked := sc.ScriptStep.KeepExpanded;
    end;
end;



const
    METAVARIABLE_REGEXP = '\?[a-zA-Z0-9: _.]+\?';
    METAVARIABLE_REGEXP_BETTER = META_CHAR+'[a-zA-Z0-9_.]+(:[^'+META_CHAR+']*)?'+META_CHAR;


function TFrmScriptActions.GetText(t : TCustomEdit) : string;
begin
    if t is TMyRichEdit then begin
        result := TMyRichEdit(t).Text;
    end else begin
        result := t.text;
    end;
end;
procedure TFrmScriptActions.TextKeyPress(Sender: TObject; var key : char);
begin
    if key = #13 then begin
        key := #0;
    end;

    self.HandleKeypress(sender);
end;
procedure TFrmScriptActions.TextKeyDown(Sender: TObject; var key : word; shift : TShiftState);
begin
    if key = VK_RETURN then key := 0;
end;
procedure TFrmScriptActions.TextClick(Sender: TObject);
var
    i : integer;
    wc : TWildCards;
    b : boolean;
    xid,yid : cardinal;
    p : TPoint;
    s : string;
    re : TMyRichEdit;
begin
    if not (Sender is TMyRichEdit) then EXIT;

    re := TMyRichEdit(Sender);
    if re.HighlightIsVariable then begin
        if re.CaretInVariable then begin
            re.HighlightVariableAroundCaret;
            s := re.SelText;
            if Pos(':',s) <> 0 then begin
                FormatVariable1.Visible := true;
                AddPadding1.Visible := false;
                AddDecimal1.Visible := false;
                FormatCorrectDate1.Visible := false;
                Hex1.Visible := false;
                Substring1.Visible := false;
                Letters1.Visible := false;
            end else begin
                FormatVariable1.Visible := false;
                AddPadding1.Visible := true;
                AddDecimal1.Visible := true;
                FormatCorrectDate1.Visible := true;
                Hex1.Visible := true;
                Substring1.Visible := true;
                Letters1.Visible := true;

            end;
            p := re.ClientToScreen(re.BoundsRect.TopLeft);
            lasttext := re;

            pmFormat.Popup(p.X, p.y + re.Height);
        end;
    end;


    LastSelStart := re.SelStart;
    lastSelLength := re.SelLength;
    LastSelText := re.SelText;

    {
    t := TEdit(Sender);
    if t.seltext <> '' then begin
        wc := TRegExpWildcards.create;
        wc.SetSearchPattern(METAVARIABLE_REGEXP_BETTER);
        wc.SetSearchString(t.SelText);
        if wc.findstring = t.seltext then begin
            // current selection is a metavariable, so let the user
            // change the cursor position
            wc.free;
            EXIT;
        end;
        wc.free;
    end;

    i := t.SelStart + 1; // zero based index, strings are 1 based
    if (true) then begin
        wc := TRegExpWildcards.Create;
        // find the starting and ending ?? marks

        wc.SetSearchPattern(METAVARIABLE_REGEXP_BETTER);
        wc.SetSearchString(t.text);

        b := wc.FindFirst(xid,yid);
        while b do begin
            if (i > xid)  and (i <= yid) then begin
                BREAK;
            end;
            b := wc.FindNext(xid,yid);
        end;
        // hilight the metavariable clicked on
        t.Hint := '';
        if b then begin
            s := StrUtils.MidStr(t.text,xid,yid-xid+1);
            if (s = LastSelText) and
                (LastSelStart = (xid-1)) and
                (lastSelLength = (yid-xid+1)) then begin
                LastSelStart := -1;
                lastSelLength := -1;
                LastSelText := '';
            end else begin
                t.SelStart := xid-1;
                t.SelLength := yid-xid+1;

                t.hint := 'Right Click to format';
                formatvariable1.Caption := 'Format Variable' ;
                p := t.ClientToScreen(t.BoundsRect.TopLeft);
                lasttext := t;

                if frmConfig.cbAutoFormatPopup.checked then begin
                    pmFormat.Popup(p.X, p.y+ t.Height);
                end;

                LastSelStart := t.SelStart;
                lastSelLength := t.SelLength;
                LastSelText := t.SelText;
            end;
        end;
    end else begin
    end;
    }
end;
procedure TFrmScriptActions.TextContextPopup(Sender:TObject; MousePos: TPoint; var Handled: Boolean);
var wc : TWildCards;
    t : TMyRichEdit;
    xid, yid : cardinal;
    b : boolean;
begin
    t := TMyRichEdit(sender);
    self.LastText := t;

    FormatMetaVariable1.Enabled := false;
    if t.SelText <> '' then begin
        wc := TRegExpWildcards.Create;
        wc.SetSearchPattern(METAVARIABLE_REGEXP_BETTER);
        wc.SetSearchString(t.SelText);

        FormatMetaVariable1.Enabled := wc.FindFirst(xid, yid) and
            ((yid-xid)+1 = t.SelLength);
    end;

    handled := true;

    // these only work if there's a selection
    b := t.SelLength<>0;
    cut1.Enabled := b;
    copy1.Enabled := b;
    delete2.Enabled := b;


    poRightClick.Popup(mouse.CursorPos.X -5, mouse.CursorPos.Y + 11);
end;
procedure TFrmScriptActions.TextMouseDown(Sender:TObject; Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
    //

end;
procedure TFrmScriptActions.TextMouseUp(Sender:TObject; Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
end;



//===============
{$region '// Command Control Buttons - expand/collapse //'}
procedure TFrmScriptActions.ToggleScriptDetailsClick(Sender: TObject);
var
    ht : integer;
    sac, sac2, previousLastSelected : TScriptStepControl;
var i : integer;

    enabled, manualClickCollapse, wasExpanded : boolean;
    wc : TWinControl;
    autopreview : Boolean;

    pt : TPoint;
begin
    sac := self.GetScriptControl( TWinControl(sender).tag );
    Expanding := true;
    wc := self.LastParentControl;
    autopreview := FrmDummyAutoPreview.GetEnabled;

    if assigned(sac.pnlTop) then begin
        pt := sac.pnltop.ScreenToClient(mouse.CursorPos);
        if PtInRect(sac.pnlTop.ClientRect, pt ) then begin
            if pt.X > sac.pnlTop.Width - (((10+3)*4)+10) then begin
                self.ShowButtons(sac);
                expanding := false;
                EXIT;
            end;
        end;
    end;


    timIgnoreChanges.Enabled := true;
    wc.DisableAlign;
    wasExpanded := sac.IsExpanded;





    if (NoClickExpand) then begin
        NoClickExpand := false;
    end else begin
        previousLastSelected := self.fLastSelectedControl;
        setLastSelectedControl(sac);

    end;
    self.ToggleSingleStep(sac);
    if frmconfig.cbAutoCollapse.Checked {and not wasExpanded} then begin
        for i := 0 to ScriptControlOrder.Count-1 do begin
            sac2 := TScriptStepControl( ScriptControlOrder.Items[i]);
            if sac2 <> sac then begin
                if sac2.IsExpanded then
                    self.ToggleSingleStep(sac2);
            end;
        end;
    end;


    if (self.fLastSelectedControl <> nil) then begin
        for i := 0 to ScriptControlOrder.Count-1 do begin
            sac2 := TScriptStepControl( ScriptControlOrder.Items[i]);
            if sac2 <> self.fLastSelectedControl then begin
                if sac2.sbmore.Visible then begin
                    sac2.sbmore.Visible := false;
                    //sac2.pnlMid.Left := sac2.sbenable.Left + sac2.sbenable.Width;
                end;
            end else begin
                sac2.sbdelete.Visible := true;
            end;
        end;
    end;

    // splitting the two operations here makes it smoother, for some reason

    Self.ForceAlign;
    //if not wasExpanded then self.FocusFirstTabOrderControl(sac);
    expanding := false;

    wc.EnableAlign;
    FrmMain.scrlScriptbox.ScrollInView(sac.outerpnl);
end;
procedure TFrmScriptActions.ToggleSingleStep(sac : TScriptStepControl);
var
    ht : integer;
    pnl : TPanel;
CONST EXPANDED_SPACING = 4;
    procedure UpdateMoreButton;
    begin
//        sac.pnlMid.Left := sac.sbenable.Left + sac.sbenable.Width;
    end;
begin
    if (sac.IsExpanded and sac.ScriptStep.KeepExpanded) then begin
        sac.sbmore.Visible := true;
        sac.pnlMid.Invalidate;
        UpdateMoreButton;
        EXIT;
    end;

    ht :=  0;
    pnl := sac.outerpnl;
    sac.pnlBtm.Align := alnone;
    sac.pnlTop.Align := alnone;

    sac.pnlBtm.Visible := not sac.pnlBtm.Visible;
    sac.pnlTop.RoundBottom := not sac.pnlBtm.Visible;
    sac.pnlMid.Expanded := sac.pnlBtm.Visible;

    sac.pnlTop.Top := ht;
    inc(ht, sac.pnlTop.height);



    if sac.IsExpanded then begin
        sac.pnlBtm.Top := ht;
        inc(ht, sac.pnlbtm.Height);
        if sac.Tag <> 0 then begin
            sac.outerpnl.Padding.Top := EXPANDED_SPACING;
        end;
        sac.outerpnl.Padding.bottom := EXPANDED_SPACING;

        if sac.pnlMid.Hint <> '' then
            sac.pnlMid.Caption := sac.pnlMid.Hint;
        sac.pnlTop.color :=  SetBlueColor(clBtnFace, 1.14);
        sac.pnlTop.BevelKind := bkTile;

        self.ShowButtons(sac);
    end else begin
        sac.outerpnl.Padding.Top := 0;
        sac.outerpnl.Padding.bottom := 0;

        {*
        if sac.ScriptStep.ScriptType in [stAddBeginningEnding,stAddPosition,stAddBeforeAfter,stAddCounter] then begin
            sac.sbtoggle.Glyph := self.imgAdd.Picture.Bitmap;
            sac.imgReplace.Picture := self.imgAddSide.Picture;
        end else if sac.ScriptStep.ScriptType in [stRemoveWords,stRemoveBeginningEnding,stRemoveSubstring,stRemoveBeforeAfter,stRemoveExpression,stRemoveWordNumber] then begin
            sac.sbtoggle.Glyph := self.ImgRemove.Picture.Bitmap;
            sac.imgReplace.Picture := self.imgRemoveSide.Picture;
        end else if sac.ScriptStep.ScriptType in [stReplaceNameExtension,stReplaceCharacters,stReplaceNumbers, stSplitAndMove] then begin
            sac.sbtoggle.Glyph := self.imgReplace.Picture.Bitmap;
            sac.imgReplace.Picture := self.imgReplaceSide.Picture;
        end else begin
            sac.sbtoggle.Glyph := imgExpand.Picture.Bitmap;
            sac.imgReplace.Picture := nil;
        end;
        *}

        sac.pnlMid.Hint := sac.pnlMid.Caption;
        sac.pnlMid.Caption := '';

        sac.pnlTop.Height := max(sac.pnlMid.Height,sac.pnlTop.Height);
        sac.pnlTop.color := DimColor(clBtnFace, 0.93);
        sac.pnlTop.RoundBottom := true;
        sac.pnlMid.Expanded := false;
        self.ShowButtons(sac, false);
    end;


    UpdateMoreButton;

    sac.pnlMid.AutoWrap := true;


    sac.pnlTop.hint := self.UpdateSentence(sac);

    sac.pnlTop.Width := sac.outerpnl.Width;
    sac.pnlbtm.Width := sac.outerpnl.Width;

    sac.outerpnl.Height := ht;
    sac.outerpnl.AutoSize := true; // this triggers the padding
    sac.outerpnl.AutoSize := false;
end;
procedure TFrmScriptActions.timIgnoreChangesTimer(Sender: TObject);
begin
    timIgnoreChanges.Enabled := False;
end;

procedure TFrmScriptActions.MoveScriptUpClick(Sender: TObject);
var sac : TScriptStepControl;
    i, j : integer;
    sc : TScriptStepControl;

    shift : boolean;
label exitcode;
begin
    shift := GetKeyState(VK_SHIFT)<0;

    sac := self.GetScriptControl( TSpeedButton(sender).tag );

    // and the target control, move it up in the order list
    // refresh the display with the new order
    for i := 0 to ScriptControlOrder.count  - 1 do begin
        sc := TScriptStepControl(ScriptControlOrder.Items[i]);
        if sc = sac then begin
            if (i = 0) then goto exitcode;

            if shift then begin
                j := 0;
            end else begin
                j := i-1;
            end;
            ScriptControlOrder.Move(i, j);

            if (self.LastParentControl is TFlowPanel) then begin
                with TFlowPanel(self.LastParentControl) do begin
                    SetControlIndex(sc.outerpnl, j);
                end;
                sc.outerpnl.Changed;
            end;

            BREAK;
        end;
    end;

    FrmScriptActions.BeginUpdate;

    self.RedoSentences;
    self.ForceAlign;

exitcode:
    frmmain.scrlScriptbox.ScrollInView(sac.outerpnl);
    application.ProcessMessages;
    FrmDummyAutoPreview.InformOfKeypress;

   FrmScriptActions.EndUpdate;
end;
procedure TFrmScriptActions.MoveScriptDownClick(Sender: TObject);
var sac, sc : TScriptStepControl;
    i, j : integer;

    shift : boolean;
label exitcode;
begin
    sc := nil;
    shift := GetKeyState(VK_SHIFT)<0;
    sac := self.GetScriptControl( TSpeedButton(sender).tag );

    // and the target control, move it down in the order list
    // refresh the display with the new order
    for i := 0 to ScriptControlOrder.count  - 1 do begin
        sc := TScriptStepControl(ScriptControlOrder.Items[i]);
        if sc = sac then begin
            if (i = ScriptControlOrder.Count-1) then goto exitcode;

            if shift then begin
                j := ScriptControlOrder.count  - 1;
            end else begin
                j := i+1;
            end;
            ScriptControlOrder.Move(i, j);
            if (self.LastParentControl is TFlowPanel) then begin
                with TFlowPanel(self.LastParentControl) do begin
                    SetControlIndex(sc.outerpnl, j);
                end;
            end;

            BREAK;
        end;
    end;

    FrmScriptActions.BeginUpdate;

    sc.outerpnl.Changed;

    self.RedoSentences;
    self.ForceAlign;

exitcode:
    frmmain.scrlScriptbox.ScrollInView(sac.outerpnl);
    application.ProcessMessages;
    FrmDummyAutoPreview.InformOfKeypress;

    FrmScriptActions.EndUpdate;
end;
procedure TFrmScriptActions.LabelClickEvent(Sender: TObject);
var p : TPanel;
begin
    p := TPanel(TLabel(sender).parent);

    while p.parent <> LastParentControl do begin
        p := TPanel(p.parent);
    end;

    self.ToggleScriptDetailsClick(p);
end;

procedure TFrmScriptActions.DeleteScriptClick(Sender: TObject);
var sac : TScriptStepControl;
    b : boolean;
    tag : integer;
begin
    if sender = NIL then EXIT;

    tag := TSpeedButton(sender).tag;
    sac := GetScriptControl( tag );
    if sac = NIL then EXIT;
    if sac.Deleted  then EXIT;
    if GetKeyState(VK_SHIFT)<0 then begin
    	sac.ScriptStep.Reset;
        self.UpdateSentence(sac);
    	EXIT;
    end;;


    self.BeginUpdate;
    if assigned(sac.outerpnl) then  begin
        sac.outerpnl.Parent := nil;
        sac.outerpnl.Free;

        // For some reason, directly freeing the control causes an exception
        // when the parent is still set

    end;

    self.DeleteScript(tag);

    self.RefreshOrder;
    self.ForceAlign;



    self.EndUpdate;
    FrmDummyAutoPreview.InformOfKeypress;

    // NOTE: the associated TScriptStepControl and TScriptStep objects
    // are not removed to preserve indexes for all the other objects in the
    // object lists
end;
procedure TFrmScriptActions.DeleteScript(tag : integer);
var sac : TScriptStepControl;
begin
    // NOTES: This must be the ONLY place a script item is deleted

    sac := self.GetScriptControl( tag );
    if sac = NIL then EXIT;
    if sac.Deleted  then EXIT;
    sac.Deleted := true;
    ScriptControlOrder.Remove(sac);

    // if this is a "last selected" item, clear the references
    if (sac = self.fLastSelectedControl) then begin
        self.fLastSelectedControl := nil;
    end;
    if (tag = self.fLastSelectedEditControlIndex) then begin
        self.fLastSelectedEdit := nil;
        self.fLastSelectedEditControlIndex := -1;
    end;
end;



procedure TFrmScriptActions.EnableClick(tag: integer);
var sac : TScriptStepControl;
begin
    sac := self.GetScriptControl( tag);
    self.BeginUpdate;
    sac.ScriptStep.Enabled := not sac.ScriptStep.Enabled;
    sac.pnlMid.Enabled := sac.ScriptStep.Enabled;

    if (assigned(sac.sbenable))  then begin
        if  (sac.ScriptStep.Enabled) then begin
            sac.sbenable.Glyph := self.imgCheck.Picture.Bitmap;
        end else begin
            sac.sbenable.Glyph := self.imgBlank.Picture.Bitmap;
        end;
    end;

//    for i := (sac.pnlMid.ControlCount - 1) downto 0 do begin
//        if sac.ScriptStep.Enabled then begin
//            TLabel(sac.pnlMid.Controls[i]).Font.Color := clWindowText;
//        end else begin
//            TLabel(sac.pnlMid.Controls[i]).Font.Color := clGrayText;
//        end;
//    end;

    self.UpdateSentence(sac);
    self.EndUpdate;
    FrmDummyAutoPreview.InformOfKeypress;
end;
procedure TFrmScriptActions.EnableClick(sender: TObject);
begin
    self.EnableClick(TSpeedButton(sender).tag);
    FrmDummyAutoPreview.InformOfKeypress;
end;


procedure TFrmScriptActions.OuterPanelClick(Sender: TObject);
begin
    if frmmain.Visible and frmmain.scrlScriptbox.CanFocus then Frmmain.scrlScriptbox.SetFocus;
end;

procedure TFrmScriptActions.sbMoreClick(Sender: TObject);
var p : TPanel;
begin
    p := TPanel(sender);
    repeat
        p := TPanel(p.Parent);
    until p.Parent = self.LastParentControl;

    self.fStepClicked := p;

    pmStepRightClick1.Popup(mouse.CursorPos.X, Mouse.CursorPos.y);
end;
procedure TFrmScriptActions.sbChangeTextFormalClick(Sender: TObject);
var p : TPanel;
begin
    p := TPanel(sender);
    repeat
        p := TPanel(p.Parent);
    until p.Parent = self.LastParentControl;

    //self.fStepClicked := p;

    pmTextType.Popup(mouse.CursorPos.X, Mouse.CursorPos.y);
end;


procedure TFrmScriptActions.BtnPinnedClick(sender : Tobject);
var p : TPanel;
    sc : TScriptStepControl;
begin
    p := TPanel(sender);
    repeat
        p := TPanel(p.Parent);
    until p.Parent = self.LastParentControl;

    self.fStepClicked := p;
    sc := self.GetScriptControl(self.fStepClicked.tag);

    sc.ScriptStep.KeepExpanded := false;
    ShowButtons(sc);

    FixKeepExpanded;
end;
{$endregion}
//===============

//
// refresh step contents
//
procedure TFrmScriptActions.MyToggle(sac : TScriptStepControl);
begin
    self.IgnoreMouse := true;
    self.ToggleSingleStep(sac);
end;

function TFrmScriptActions.UpdateSentence(sac: TScriptStepControl; conditional : boolean = false): string;
begin
    result := self.UpdateSentence(ScriptControlOrder.IndexOf(sac), conditional);
end;
function TFrmScriptActions.UpdateSentence(index: integer; conditional : boolean = false) : string;
const
    FONT_SIZE = 7;
    FONT_HEIGHT = 15;
    function GetSentence(sac : TScriptStepControl ) : string;
        function GetText(wc : TControl) : string;
        begin
            if wc is TLabel then
                result := TLabel(wc).caption
            else if wc is TCheckbox then
                result := TCheckbox(wc).Caption
            else if wc is TCustomEdit then
                result := '''' + TCustomEdit(wc).text + ''''
            else if wc is TComboBox  then
                result := TComboBox(wc).text
            else if wc is TRadioButton then begin
                if TRadioButton(wc).checked then begin
                    result := TRadioButton(wc).Caption;
                end;
            end;

        end;
        procedure Append(wc : TControl);   overload;
        var lbl : TLabel;
            t : string;
        begin
            lbl := TLabel.Create(sac.pnlmid);
            frmConfig.GetScriptFont(lbl.font);
            //lbl.Font.Height := FONT_HEIGHT;
            //lbl.Font.Name := 'Times New Roman';

            lbl.Parent := sac.pnlmid;
            lbl.AlignWithMargins := true;
            lbl.Margins.Left := 0;
            lbl.Margins.Right := 3;
            if sac.ScriptStep.Enabled then begin
                lbl.Font.Color := clWindowText;
            end else begin
                lbl.Font.Color := DimColor(clGrayText, GREY_DIM);
            end;
                        lbl.Transparent := true;

            t := GetText(wc);

            lbl.Caption := t;
            if wc is TCustomEdit then begin
                lbl.Font.Style := lbl.font.style + [fsBold]
            end else if wc is TComboBox then begin
                lbl.Font.Style := lbl.font.style + [fsUnderline];
            end;
            lbl.Tag := index;
            lbl.OnClick := self.LabelClickEvent;

            self.DragDropHook(lbl);
            result := result + t + ' ';
        end;
        procedure Append(s : string);   overload;
        var
            l : TLabel;
        begin
            l := TLabel.Create(sac.pnlMid);
            l.caption := s;
            l.Font.Height := FONT_HEIGHT;

            Append(l);
            l.free;
        end;
        var i : integer;
            num : string;
            s : string;
    begin
        result := '';

        i := self.ScriptControlOrder.IndexOf(sac);
        num := Format('[%0.2d]', [i+1]) + ' ';
        if conditional then begin
            num := num + ' >> ';
        end;
        Append(num);

        if (sac.ScriptStep.Enabled and not sac.IsExpanded)
            or frmConfig.cbAlwaysShowStepDescriptions.checked then begin
            sac.ScriptStep.GetSentence(sac);
            s := sac.GetHint;
        end else begin
            s := sac.ScriptStep.GetShortSentence;
            Append(s);
        end;

        result := s;
    end;
var sc : TScriptStepControl;
    i : integer;
begin
    sc := TScriptStepControl(ScriptControlOrder.Items[index]);

    for i := (sc.pnlMid.ControlCount - 1) downto 0 do begin
        if assigned(sc.pnlmid.Controls[i]) then begin
            sc.pnlMid.RemoveControl(sc.pnlMid.Controls[i]);
        end;
    end;
    sc.pnlTop.Hint := GetSentence(sc);
    sc.pnlTop.Invalidate;
    result := sc.pnlTop.Hint;
end;
procedure TFrmScriptActions.RedoSentences;
var i, cnt : integer;
    sc : TScriptStepControl;
    sce : TScriptStepConditionExecute;
begin

    self.BeginUpdate;
    cnt := 0;
    for i := 0 to ScriptControlOrder.count - 1 do begin
        sc := TScriptStepControl(ScriptControlOrder.Items[i]);
        sc.outerpnl.DisableAlign;
        if sc.ScriptStep.ScriptType = stConditionalExecute  then begin
            sce := TScriptStepConditionExecute(sc.scriptstep);
            cnt := 0;
            if sce.Enabled then begin
                cnt := SafeInt(sce.txtCE1.text, 0) + 1;
            end;
            self.UpdateSentence(i, FALSE);
        end else begin
            self.UpdateSentence(i, (cnt<>0));
        end;
        if frmconfig.cbShowUpDown.Checked then begin
            sc.pnlMid.Width := sc.pnlTop.Width - 100;
        end else begin
            sc.pnlMid.Width := sc.pnlTop.Width - 70;
        end;
        if sc.sbmore.Visible then begin
            sc.pnlMid.Width := (sc.pnlMid.Width - sc.sbmore.Width) - 2;
        end;

        //self.UpdateSentence(i);
        if (cnt > 0) then dec(cnt);
        sc.outerpnl.EnableAlign;
    end;

    Self.EndUpdate;
end;
procedure TFrmScriptActions.RefreshAllButtons;
var i : integer;
    sc : TscriptStepControl;
begin
    self.BeginUpdate;
    for i := (ScriptControlOrder.count - 1) downto 0 do begin
        sc := TScriptStepControl(ScriptControlOrder.Items[i]);
        showbuttons(sc, frmconfig.cbAlwaysShowStepButtons.checked);
    end;
    self.EndUpdate;
end;
procedure TFrmScriptActions.RefreshOrder;
begin
    self.ForceAlign;
    self.RedoSentences;
end;
procedure TFrmScriptActions.RefreshPanel;
begin
    
end;


//
// Drag and Drop Routines  & right-click
//

//===============
{$region '// Drag and Drop Stuff - Right-Click //'}
procedure TFrmScriptActions.ClearHovered;
begin
    if self.fDragHovered = nil then EXIT;

    self.fDragHovered.pnlTop.Hovered := false;
    self.fDragHovered.outerpnl.Invalidate;
    self.fDragHovered := nil;
end;

procedure TFrmScriptActions.DragDropHook(c: TControl);
begin
    if c is TLabel then begin
        with TLabel(c) do begin
            OnMouseDown := self.StepMouseDown;
            OnMouseUp := self.StepMouseUp;
            OnMouseMove := self.StepMouseMove;
            OnMouseLeave := self.StepMouseLeave;
        end;
    end else if (c is TFlowPanel) then begin
        with TFlowPanel(c) do begin
            OnMouseDown := self.StepMouseDown;
            OnMouseUp := self.StepMouseUp;
            OnMouseMove := self.StepMouseMove;
            OnMouseLeave := self.StepMouseLeave;
        end;
    end;
end;
procedure TFrmScriptActions.StepMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var p : TPanel;
begin
    p := TPanel(sender);
    repeat
        p := TPanel(p.Parent);
    until p.Parent = self.LastParentControl;

    self.fStepClicked := p;
    self.fStartingPoint := Mouse.CursorPos;
    self.fIsDragging := false;
    self.fSavedCursor := screen.Cursor;
    self.fRightClick := Button = mbRight;
    screen.Cursor := crDrag;

end;
procedure TFrmScriptActions.StepMouseLeave(Sender: TObject);
begin
    Screen.Cursor := crDefault;
    ClearHovered;
end;
procedure TFrmScriptActions.StepMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var wc : TWinControl;
    pt, pt2 : TPoint;

    procedure DetectIfDragging;
    begin
        self.fIsDragging := (
        (abs(mouse.CursorPos.X - self.fStartingPoint.x) > mouse.DragThreshold)
        or
        (abs(mouse.CursorPos.Y - self.fStartingPoint.Y) > mouse.DragThreshold)
        );

        self.fIsDragging := self.fIsDragging and (fStepClicked <> nil);


        if self.fIsDragging then begin
            screen.Cursor := crNoDrop
        end;
    end;
    procedure HandleMouseNotInScrollbox;
    begin
        wc := TWinControl(Sender);
        pt := Point(x,y);
        pt := wc.ClientToScreen(pt);
        pt2 := frmmain.scrlScriptbox.ScreenToClient(pt);

        if not PtInRect(frmmain.scrlScriptbox.ClientRect, pt2)   then begin
            with frmmain.scrlScriptbox.VertScrollBar do
            if (pt2.Y <= frmmain.scrlScriptbox.ClientRect.Top) then begin
                timScrollUp.Enabled := true;
            end else if (pt2.Y >= frmmain.scrlScriptbox.ClientRect.Bottom) then begin
                timScrollDown.Enabled := true;
            end;
        end;
    end;
    procedure HandleHoveredControl(wc : TWinControl);
    var ssc : TScriptStepControl;
        lastDragged : TScriptStepControl;
    begin
        ssc := GetScriptControl(wc.tag);
        if ssc = nil then EXIT;
        if ssc = self.fDragHovered then EXIT;


        lastDragged := nil;
        if self.fDragHovered <> nil then begin
            lastDragged := self.fDragHovered;
            lastDragged.pnlTop.Hovered := false;
        end;

        self.fDragHovered := ssc;
        self.fDragHovered.pnlTop.Hovered := true;


        if lastDragged <> nil then begin
            lastDragged.outerpnl.Invalidate;
        end;
        self.fDragHovered.outerpnl.Invalidate;
    end;
    procedure HandleMouseInScrollbox;
    begin
        pt := self.LastParentControl.ScreenToClient(pt);

        wc := TWinControl(self.LastParentControl.ControlAtPos(pt, false,true,false));
        if wc <> nil then begin
            // get topmost controll
            while (wc.parent <> self.LastParentControl) do begin
                wc := wc.parent;
            end;

            if wc <> self.fStepClicked then begin
                screen.Cursor := crDrag;
                HandleHoveredControl(wc);
            end else if wc = self.fStepClicked then begin
                screen.cursor := crNoDrop;
            end;
        end else begin
            screen.Cursor := crNoDrop;
        end;
    end;
begin
    if (ssleft in shift) then begin
        // detect movement large enough to be a drag
        // detect the correct cursor to use
        if not self.fIsDragging then begin
            DetectIfDragging;
        end else begin
            HandleMouseNotInScrollbox;
            HandleMouseInScrollbox;
        end;
    end else begin
        ClearHovered;
        if (screen.cursor = crNoDrop) or (screen.Cursor =crDrag) then begin
            screen.Cursor := self.fSavedCursor;
        end else begin
            if UnitUtil.MouseInControl(self.LastParentControl) then begin
                //screen.Cursor := crHandPoint;
            end else begin
                screen.Cursor := crDefault;
            end;
        end;
    end;
end;
procedure TFrmScriptActions.StepMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var wc : TWinControl;
    sc : TScriptStepControl;
    pt : TPoint;
begin
    if self.fRightClick then begin
        pmStepRightClick1.Popup(mouse.CursorPos.X, Mouse.CursorPos.y);
    end else if self.fIsDragging then begin
        wc := TWinControl(Sender);
        pt := Point(x,y);
        pt := wc.ClientToScreen(pt);
        pt := self.LastParentControl.ScreenToClient(pt);
        wc := TWinControl(self.LastParentControl.ControlAtPos(pt, false,true,false));
        if wc <> nil then begin
            while (wc.parent <> self.LastParentControl) do begin
                wc := wc.parent;
            end;

            sc := self.GetScriptControl(wc.tag);
            self.MoveStep(fStepClicked, ScriptControlOrder.IndexOf(sc));
        end;
        self.fStepClicked := nil;

        self.ForceAlign;
    end;
    self.fIsDragging := false;
    screen.Cursor := crDefault;
    ClearHovered;

end;


procedure TFrmScriptActions.MoveStep(p: TPanel; newindex: integer);
var sac : TScriptStepControl;
    i : integer;
    sc : TScriptStepControl;
label exitcode;
begin
    sac := self.GetScriptControl( p.tag );

    // and the target control, move it up in the order list
    // refresh the display with the new order
    for i := 0 to ScriptControlOrder.count  - 1 do begin
        sc := TScriptStepControl(ScriptControlOrder.Items[i]);
        if sc = sac then begin
            ScriptControlOrder.Move(i, newindex);

            if (self.LastParentControl is TFlowPanel) then begin
                with TFlowPanel(self.LastParentControl) do begin
                    SetControlIndex(sc.outerpnl, newindex);
                end;
                sc.outerpnl.Changed;
            end;

            BREAK;
        end;
    end;

    FrmScriptActions.BeginUpdate;

    self.RedoSentences;

exitcode:
    frmmain.scrlScriptbox.ScrollInView(sac.outerpnl);
    application.ProcessMessages;
    FrmDummyAutoPreview.InformOfKeypress;

    FrmScriptActions.EndUpdate;
end;


procedure TFrmScriptActions.timScrollDownTimer(Sender: TObject);
begin
    timScrollDown.Enabled := false;
    FrmMain.sbScript.Position := FrmMain.sbScript.Position + 1;
end;
procedure TFrmScriptActions.timScrollUpTimer(Sender: TObject);
begin
    timScrollUp.Enabled := false;
    FrmMain.sbScript.Position := FrmMain.sbScript.Position - 1;
end;


procedure TFrmScriptActions.MIDelete1Click(Sender: TObject);
var sc : TScriptStepControl;
begin
     sc := self.GetScriptControl(self.fStepClicked.tag);
     sc.sbdelete.Click;
end;
procedure TFrmScriptActions.MIDuplicate1Click(Sender: TObject);
var sc, sc2 : TScriptStepControl;
var j, cnt : integer;

    pl, pl2 : TPropList;
    ppl, ppl2 : PPropList;
    pti, pti2 : PTypeInfo;

    cmp, cmp2 : TComponent;

    cl : TList<TcomboBox>;
    cbx : TComboBox;
    s : string;
    function GetTypeInfo(sc : TScriptSTepControl) : PTypeInfo;
    begin
        case sc.ScriptStep.ScriptType  of
            stAddBeginningEnding:
                result := PTypeInfo(TScriptStepAddBeginningEnding(sc.ScriptStep).ClassInfo);
            stAddCounter:
                result := PTypeInfo(TScriptStepAddCounter(sc.ScriptStep).ClassInfo);
            stAddPosition:
                result := PTypeInfo(TScriptStepAddPosition(sc.ScriptStep).ClassInfo);
            stAddBeforeAfter:
                result := PTypeInfo(TScriptStepAddBeforeAfter(sc.ScriptStep).ClassInfo);
            stReplaceNameExtension:
                result := PTypeInfo(TScriptStepReplaceNameExtension(sc.ScriptStep).ClassInfo);
            stReplaceCharacters:
                result := PTypeInfo(TScriptStepReplaceCharacters(sc.ScriptStep).ClassInfo);
            stReplaceNumbers:
                result := PTypeInfo(TScriptStepReplaceNumbers(sc.ScriptStep).ClassInfo);
            stRemoveWords:
                result := PTypeInfo(TScriptStepRemoveWords(sc.ScriptStep).ClassInfo);
            stRemoveBeginningEnding:
                result := PTypeInfo(TScriptStepRemoveCharacters(sc.ScriptStep).ClassInfo);
            stRemoveSubstring:
                result := PTypeInfo(TScriptStepRemoveSubstring(sc.ScriptStep).ClassInfo);
            stRemoveBeforeAfter:
                result := PTypeInfo(TScriptStepRemoveBeforeAfter(sc.ScriptStep).ClassInfo);
            stRemoveExpression:
                result := PTypeInfo(TScriptStepRemoveExpression(sc.ScriptStep).ClassInfo);
            stConditionalExecute:
                result := PTypeInfo(TScriptStepConditionExecute(sc.ScriptStep).ClassInfo);
            stSplitAndMove:
                result := PTypeInfo(TScriptStepSplitAndMove(sc.ScriptStep).ClassInfo);

            stSelectAndMove:
                result := PTypeInfo(TScriptStepSelectAndMove(sc.ScriptStep).ClassInfo);
            stSwap:
                result := PTypeInfo(TScriptStepSwap(sc.ScriptStep).ClassInfo);
            stFindAndEdit:
                result := PTypeInfo(TScriptStepFindAndEdit(sc.ScriptStep).ClassInfo);
        end;

    end;
begin
    sc := self.GetScriptControl(self.fStepClicked.tag);
    self.CollapseAll;
    self.NewScriptControl(self.LastParentControl, sc.ScriptStep.ScriptType, false, false);
    sc2 := self.GetScriptControl(self.ScriptControlList.Count-1);

    try
        cl := TList<TComboBox>.create;

        // load the full properties for each ScriptStep class
        ppl := @pl;
        pti := GetTypeInfo(sc);

        ppl2 := @pl2;
        pti2 := getTypeInfo(sc2);

        // iterate all published properties of each StepScript
        cnt := GetPropList(pti,[tkClass], ppl);
        GetPropList(pti2, [tkClass], ppl2);
        if cnt <> 0 then begin
            for j := 0 to cnt - 1 do begin
                if (pl[j]^.name = 'Name') or (pl[j]^.name ='Tag') then CONTINUE;

                cmp := TComponent(TypInfo.GetObjectProp(sc.scriptstep, pl[j]^.name));
                cmp2 := TComponent(TypInfo.GetObjectProp(sc2.scriptstep, pl2[j]^.name));
                if cmp <> nil then begin
                    if cmp is TMyRichEdit then begin
                        // some weird timing/race condition
                        // the first assign is ignored here
                        s := TMyRichEdit(cmp).Text;
                        TMyRichEdit(cmp2).Text := 'blank';
                        TMyRichEdit(cmp2).Text := s;
                    end else if cmp is TCustomEdit then begin
                        TCustomEdit(cmp2).Text := TCustomEdit(cmp).Text;
                    end else if cmp is TComboBox then begin
                        tComboBox(cmp2).ItemIndex := tComboBox(cmp).ItemIndex;
                        cl.Add(TComboBox(cmp2));
                    end else if cmp is TCheckbox then begin
                        TCheckBox(cmp2).Checked := TCheckBox(cmp).Checked;
                    end else if cmp is TRadioButton then begin
                        TRadioButton(cmp2).Checked := TRadioButton(cmp).Checked;
                    end;
                end;
            end;
        end;

        // changing the index doesn't trigger the "onchange" event
        // manually execute
        for cbx in cl do begin
            If assigned(cbx.OnChange) then begin
                cbx.OnChange(cbx);
            end;
        end;
        FreeAndNil(cl);
    finally

    end;

    self.MoveStep(sc2.outerpnl, ScriptControlOrder.IndexOf(sc)+1);
    ToggleSingleStep(sc2);
    self.FixKeepExpanded;
    LastParentControl.Invalidate;
end;


procedure TFrmScriptActions.MIKeepOpenClick(Sender: TObject);
var sc : TScriptStepControl;
begin
    sc := self.GetScriptControl(self.fStepClicked.tag);
    MIKeepOpen.Checked := not MIKeepOpen.Checked;
    sc.ScriptStep.KeepExpanded := MIKeepOpen.Checked;
    ShowButtons(sc);

    FixKeepExpanded;
end;

procedure TFrmScriptActions.MIMoveDown1Click(Sender: TObject);
var sc : TScriptStepControl;
begin
     sc := self.GetScriptControl(self.fStepClicked.tag);
     sc.sbdown.Click;

     self.SelectionNotify(sc); //scroll to if needed
end;
procedure TFrmScriptActions.MIMoveUp1Click(Sender: TObject);
var sc : TScriptStepControl;
begin
     sc := self.GetScriptControl(self.fStepClicked.tag);
     sc.sbup.Click;

     self.SelectionNotify(sc); //scroll to if needed
end;
procedure TFrmScriptActions.MIReset1Click(Sender: TObject);
var sc : TScriptStepControl;
begin
    sc := self.GetScriptControl(self.fStepClicked.tag);
    sc.ScriptStep.Reset;
    UpdateSentence(sc);
end;

procedure TFrmScriptActions.MIScriptStepHelp1Click(Sender: TObject);
begin
    FrmScriptStepHelp.ShowHelp(self.fLastSelectedControl.ScriptStep.ScriptType);
end;

procedure TFrmScriptActions.MIToggleEnable1Click(Sender: TObject);
var sc : TScriptStepControl;
begin
     sc := self.GetScriptControl(self.fStepClicked.tag);
     sc.sbenable.Click;
     self.UpdateSentence(sc);
end;

procedure TFrmScriptActions.btnHelpClick(Sender: TObject);
var
	sb : TSpeedButton;
	sac : TScriptStepControl;
begin
	sb := TSpeedButton(sender);
    sac := self.GetScriptControl( sb.tag );

	FrmScriptStepHelp.ShowHelp(sac.ScriptStep.ScriptType);
end;
{$endregion}
//===============

//
// AutoPreview helpers
//



procedure TFrmScriptActions.cbxAddPartChange(Sender: TObject);
begin
    self.HandleKeypress(sender);
end;
procedure TFrmScriptActions.cbxBeforeAfterChange(Sender: TObject);
var sar : TScriptStepRemoveBeforeAfter;
begin
    sar := TScriptStepRemoveBeforeAfter(StepRemoveList.Items[TWinControl(sender).Tag]);

    if sar.cbxBeforeAfter.ItemIndex = 4 then begin
        sar.txtRemove2Phrase2.parent.Visible := true;
        sar.lblAndThePhrase.Visible := true;
    end else begin
        sar.txtRemove2Phrase2.parent.Visible := false;
        sar.lblAndThePhrase.Visible := false;
    end;

    self.HandleKeypress(sender);
end;

procedure TFrmScriptActions.cbxCEConditionChange(Sender: TObject);
var ce : TScriptStepConditionExecute;
begin
    ce := TScriptStepConditionExecute(StepOtherList.items[TComboBox(sender).tag]);

    ce.txtCEText.parent.Visible := ce.cbxCECondition.ItemIndex <> SKIPEXECUTE_CHANGES;

    self.HandleKeypress(sender);
end;

procedure TFrmScriptActions.cbxRemovePhraseChange(Sender: TObject);
var sar : TScriptStepRemoveWords;
begin
    sar := TScriptStepRemoveWords(StepRemoveList.Items[TWinControl(sender).tag]);

    if sar.cbxRemovePhrase.ItemIndex = 2  then begin
        sar.txtRemovePhrase.Parent.Visible := false;
    end else begin
        sar.txtRemovePhrase.Parent.Visible := true;

    end;

       self.HandleKeypress(sender);
end;
procedure TFrmScriptActions.cbxRemoveTypeChange(Sender: TObject);
var sar : TScriptStepRemoveSubstring;
begin
    sar := TScriptStepRemoveSubstring(StepRemoveList.Items[TWinControl(sender).tag]);

    sar.cbAndMove.Visible := false;
    sar.pnlR3z.Visible := false;
    sar.txtr3z.Visible := false;
    sar.udr3z.Visible := false;
    sar.cbAndStoreRM.Visible := false;

    if sar.cbxRemoveType.ItemIndex = 1 then begin
        sar.cbAndMove.Visible := true;
        sar.pnlR3z.visible := true;
        sar.txtr3z.Visible := true;
        sar.udr3z.Visible := true;
        sar.cbAndStoreRM.Visible := true;
    end;

    //sar.pnlR3X.Parent.Invalidate;
    self.HandleKeypress(sender);
end;
procedure TFrmScriptActions.cbxReplacePartChange(Sender: TObject);
var sar : TScriptStepReplaceNameExtension;
    procedure SetReplaceOnlyVisible(v : boolean);
    var i : integer;
    begin
        for i := 0 to sar.pnlReplaceOnly.ControlCount - 1 do begin
            sar.pnlReplaceOnly.Controls[i].Visible := v;

        end;
    end;

begin
    sar := TScriptStepReplaceNameExtension(StepReplaceList.Items[TWinControl(sender).Tag]);

    sar.txtReplacePhrase.Visible := false;
    sar.txtReplacePhrase.Parent.Visible := false;

    SetReplaceOnlyVisible(false);


    case sar.cbxReplacePart.ItemIndex of
    0,1:
        begin
        end;
    2:
        begin
            sar.txtReplacePhrase.Visible := true;
            sar.txtReplacePhrase.Parent.Visible := true;

            SetReplaceOnlyVisible(true);
        end;

    end;

    self.HandleKeypress(sender);
end;
procedure TFrmScriptActions.cbxRERemoveReplaceChange(Sender: TObject);
var sar : TScriptStepRemoveExpression;
    b : boolean;
begin
    sar := TScriptStepRemoveExpression(StepRemoveList.Items[TWinControl(sender).tag]);

    b := sar.cbxRERemoveReplace.ItemIndex = REMOVEEXPRESSION_REPLACE;
    if Assigned(sar.txtREWith) then begin
        sar.lblReplaceExpressionWith.Visible := b;
        sar.txtREWith.Visible := b;
        sar.txtREWith.Parent.Visible := b;

        sar.lblRERemoveTheParts.Visible := not b;
        sar.cbxREParts.Visible := not b;
        sar.cbreAndStore.Visible := not b;
    end;

    self.HandleKeypress(sender);
end;

procedure TFrmScriptActions.cbxRNSequenceChange(Sender: TObject);
var sar : TScriptStepReplaceNumbers;
begin
    sar := TScriptStepReplaceNumbers(StepReplaceList.Items[TWinControl(sender).Tag]);

    sar.txtRNWith.Visible := false;
	sar.txtRNWith.Parent.Visible := false;

    case sar.cbxRNSequence.ItemIndex of
    0:
        begin
        end;
    1:
        begin
        	sar.txtRNWith.Visible := true;
            sar.txtRNWith.Parent.Visible := true;
        end;
    end;

    self.HandleKeypress(sender);
end;

procedure TFrmScriptActions.txtAddChange(Sender: TObject);
begin
    self.HandleKeypress(sender);
end;
procedure TFrmScriptActions.txtCE1Change(Sender: TObject);
begin
    self.HandleKeypress(sender);
    self.RedoSentences;
end;
procedure TFrmScriptActions.txtCETextChange(Sender: TObject);
begin
    self.HandleKeypress(sender);
end;

procedure TFrmScriptActions.txtReplaceChar1Change(Sender: TObject);
begin
    self.HandleKeypress(sender);
end;
procedure TFrmScriptActions.txtReplaceChar2Change(Sender: TObject);
begin
    self.HandleKeypress(sender);
end;
procedure TFrmScriptActions.txtReplacePhraseChange(Sender: TObject);
begin
    self.HandleKeypress(sender);
end;


procedure TFrmScriptActions.AUDZerosChanging(Sender: TObject; var AllowChange: Boolean);
begin
    AllowChange := true;
end;
procedure TFrmScriptActions.tbPositionChange(Sender: TObject);
var saa : TScriptStepAddPosition;
begin
    saa := TScriptStepAddPosition(StepAddList.Items[TTrackBar(sender).tag]);
    saa.txtAddPosition.text := IntToStr( saa.tbPosition.Position);
end;

procedure TFrmScriptActions.btnAutoSplitClick(Sender: TObject);
var s : string;
    st : TScriptStepSplitAndMove;
    i : integer;
begin
    self.HandleKeypress(sender);

    st := TScriptStepSplitAndMove(StepOtherList.Items[TWinControl(sender).tag]);

    s := AutoSplit.GetWildcardExpression;
    st.txtFilemask.text := s;

    s := StringReplace(s,'[0-9]','*',[rfReplaceAll]);
    s := StringReplace(s,'[a-z]','*',[rfReplaceAll]);

    i := 1;
    while ContainsTerm('*', s) do begin
        s := StringReplace(s, '*', '$'+IntToStr(i), []);
        inc(i);
    end;

    st.txtMove.Text := s;
end;







procedure TFrmScriptActions.LoadOldSettingAsScript(wc : TWinControl);
var P : TPanel;
    sac : TScriptStepControl;

    procedure NewControl;
    begin
        p := FrmScriptActions.CreateNewScriptControl(wc);
        sac := FrmScriptActions.GetScriptControl(p.tag);
    end;

    procedure Tickel(cbx : TComboBox);
    begin
        cbx.OnChange(cbx);
    end;
begin
    wc.visible := false;
    self.LastParentControl := wc;
    FrmScriptActions.BeginUpdate;
    self.Clear(wc);
    //self.DontExpand := true;

    //wc.Visible := false;

    //=============
    // All Removes
    //=============/
    if frmmain.cbRM2Characters.checked then begin
        NewControl;
        self.GetRemoveSubstringStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepRemoveSubstring do begin
            cbxRemoveType.ItemIndex := REMOVETYPE_ONLY;
            cbxRemoveType.OnChange(cbxRemoveType);
            udR3X.Position := frmmain.udR2CharX.Position;
            udR3Y.Position := frmmain.udR2CharY.Position;
            cbAndMove.Checked := frmmain.cbRM2AndMove.checked;
            udR3Z.Position := frmmain.udR2AndMove.position

        end;
        MyToggle(sac);
    end;
    if frmmain.cbEverything.checked then begin
        NewControl;
        self.GetRemoveSubstringStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepRemoveSubstring do begin
            cbxRemoveType.ItemIndex := REMOVETYPE_EVERYTHINGBUT;
            cbxRemoveType.OnChange(cbxRemoveType);
            udR3X.position := FrmMain.udRM2X.Position;
            udR3Y.Position := frmmain.udRM2Y.Position;
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbRM2NumFromName.Checked) then begin
        NewControl;
        self.GetRemoveWordStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepRemoveWords do begin
            cbxRemovePhrase.itemindex := REMOVEPHRASE_NUMBERS;
            cbxRemovePhrase.OnChange(cbxRemovePhrase);
            cbxRemovePhrasePart.ItemIndex := REMOVEPHRASEPART_NAME;
        end;
        MyToggle(sac);

    end;
    if (frmmain.cbRM2NumFromExt.Checked) then begin
        NewControl;
        self.GetRemoveWordStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepRemoveWords do begin
            cbxRemovePhrase.itemindex := REMOVEPHRASE_NUMBERS;
            cbxRemovePhrase.OnChange(cbxRemovePhrase);
            cbxRemovePhrasePart.ItemIndex := REMOVEPHRASEPART_EXT;
        end;
        MyToggle(sac);

    end;
    if (frmmain.cbRMFromNameStart.Checked) then begin
        NewControl;
        self.GetRemoveCharactersStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepRemoveCharacters do begin
            cbxRemoveCharPosition.ItemIndex := REMOVECHARPOS_BEGINNING;
            cbxRemoveCharactersPart.ItemIndex := REMOVECHARPART_NAME;
            cbxRemoveCharPosition.OnChange(cbxRemoveCharPosition);
            udR2X.position := SafeInt(frmmain.txtRMFromNameStart.text);
            txtR2X.text := IntToStr(SafeInt(frmmain.txtRMFromNameStart.text));

        end;
        MyToggle(sac);
    end;
    if (frmmain.cbRMFromNameEnd.Checked) then begin
        NewControl;
        self.GetRemoveCharactersStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepRemoveCharacters do begin
            cbxRemoveCharPosition.ItemIndex := REMOVECHARPOS_ENDING;
            cbxRemoveCharactersPart.ItemIndex := REMOVECHARPART_NAME;
            cbxRemoveCharPosition.OnChange(cbxRemoveCharPosition);
            udR2X.position := SafeInt(frmmain.txtRMFromNameEnd.text);
            txtR2X.text := IntToStr(SafeInt(frmmain.txtRMFromNameEnd.text));
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbRMFromExtensionStart.Checked) then begin
        NewControl;
        self.GetRemoveCharactersStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepRemoveCharacters do begin
            cbxRemoveCharPosition.ItemIndex := REMOVECHARPOS_BEGINNING;
            cbxRemoveCharactersPart.ItemIndex := REMOVECHARPART_EXT;
            cbxRemoveCharPosition.OnChange(cbxRemoveCharPosition);
            udR2X.position := SafeInt(frmmain.txtRMFromExtensionStart.text);
            txtR2X.text := IntToStr(SafeInt(frmmain.txtRMFromExtensionStart.text));
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbRMFromExtensionEnd.Checked) then begin
        NewControl;
        self.GetRemoveCharactersStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepRemoveCharacters do begin
            cbxRemoveCharPosition.ItemIndex := REMOVECHARPOS_ENDING;
            cbxRemoveCharactersPart.ItemIndex := REMOVECHARPART_EXT;
            cbxRemoveCharPosition.OnChange(cbxRemoveCharPosition);
            udR2X.position := SafeInt(frmmain.txtRMFromExtensionEnd.text);
            txtR2X.text := IntToStr(SafeInt(frmmain.txtRMFromExtensionEnd.text));
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbrmfromname.checked) then begin
        NewControl;
        Self.GetRemoveWordStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepRemoveWords do begin
            cbxRemovePhrasePart.ItemIndex := REMOVEPHRASEPART_NAME;
            Tickel(cbxRemovePhrasePart);
            if (not frmmain.cbRMWordList.Checked) then begin
                cbxRemovePhrase.ItemIndex := REMOVEPHRASE_PHRASE;
                Tickel(cbxRemovePhrase);
                txtRemovePhrase.text := frmmain.txtrmfromname.Text
            end else begin
                cbxRemovePhrase.ItemIndex := REMOVEPHRASE_WORDS;
                Tickel(cbxRemovePhrase);
                txtRemovePhrase.text := frmmain.txtrmfromname.Text
            end;
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbRMFromExtension.Checked) then begin
        NewControl;
        Self.GetRemoveWordStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepRemoveWords do begin
            cbxRemovePhrasePart.ItemIndex := REMOVEPHRASEPART_EXT;
            Tickel(cbxRemovePhrasePart);
            if (not frmmain.cbRMWordList.Checked) then begin
                cbxRemovePhrase.ItemIndex := REMOVEPHRASE_PHRASE;
                Tickel(cbxRemovePhrase);
                txtRemovePhrase.text := frmmain.txtRMFromExtension.Text
            end else begin
                cbxRemovePhrase.ItemIndex := REMOVEPHRASE_WORDS;
                Tickel(cbxRemovePhrase);
                txtRemovePhrase.text := frmmain.txtRMFromExtension.Text
            end;
        end;
        MyToggle(sac);
    end;

    //
    // All Replaces
    //
    if (frmmain.cbRPNameWith.Checked) then begin
        NewControl;
        Self.GetReplaceNameExtensionStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepReplaceNameExtension do begin
            cbxReplacePart.ItemIndex := REPLACEPART_NAME;
            tickel(cbxReplacePart);
            txtReplaceWith.text := frmmain.txtRPNameWith.Text;
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbRPExtensionWith.Checked) then begin
        NewControl;
        Self.GetReplaceNameExtensionStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepReplaceNameExtension do begin
            cbxReplacePart.ItemIndex := REPLACEPART_EXT;
            tickel(cbxReplacePart);
            txtReplaceWith.text := frmmain.txtRPExtensionWIth.Text;
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbRPThis.Checked) then begin
        NewControl;
        Self.GetReplaceNameExtensionStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepReplaceNameExtension do begin
            cbxReplacePart.ItemIndex := REPLACEPART_PHRASE;
            tickel(cbxReplacePart);
            txtReplacePhrase.text := frmmain.txtRPThis.text;
            txtReplaceWith.text := frmmain.txtRPWith.Text;
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbRP2Characters.Checked) then begin
        NewControl;
        Self.GetReplaceNameExtensionStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepReplaceNameExtension do begin
            cbxReplacePart.ItemIndex := REPLACEPART_PHRASE;
            tickel(cbxReplacePart);
            txtReplacePhrase.text := frmmain.txtRP2This.text;
            txtReplaceWith.text := frmmain.txtRP2With.Text;
            cbReplaceOnly.checked := true;
            udRP2X.position := frmmain.upRP2X.Position;
            udRP2Y.Position := frmmain.upRP2Y.Position;
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbRP2CharacterList.checked) then begin
        NewControl;
        Self.GetReplaceCharactersStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepReplaceCharacters do begin
            txtReplaceChar1.text := frmmain.txtRP2sc.Text;
            txtReplaceChar2.text := frmmain.txtRP2scWith.text;
            cbxReplaceCharPart.ItemIndex := REPLACECHARPART_NAME;
        end;
        MyToggle(sac);
    end;
    if frmmain.cbRPNumbers.checked then begin
        NewControl;
        Self.GetReplaceNumbersStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepReplaceNumbers do begin
            cbxNumbers.ItemIndex := frmmain.cbxRPNumbers.ItemIndex;
        end;
        MyToggle(sac);
    end;
    //
    // Additions
    //
    if frmmain.cbA2Find.Checked then begin
        if frmmain.cba2insert.ItemIndex in [0,1] then begin
            NewControl;
            Self.GetAddBeforeAfterStep(sac.pnlbtm,sac);
            with sac.ScriptStep as TScriptStepAddBeforeAfter do begin
                cbxAddPlace.ItemIndex := frmmain.cba2insert.ItemIndex;
                txtAfter.text := frmmain.txta2Find.text;
                txtAdd.Text := frmmain.txtA2Insert.text;
            end;
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbAEInsert.Checked) then begin
        NewControl;
        Self.GetAddPositionStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepAddPosition do begin
            cbxAddPlace.ItemIndex := ADDPART_EXT;
            txtAdd.Text := frmmain.txtAEInsert.Text;
            if frmmain.udAEPosition.Position > tbPosition.max then
                tbPosition.Max := frmmain.udAEPosition.Position;
            tbPosition.Position := frmmain.udAEPosition.Position;
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbANInsert.Checked) then begin
        NewControl;
        Self.GetAddPositionStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepAddPosition do begin
            cbxAddPlace.ItemIndex := ADDPART_NAME;
            txtAdd.Text := frmmain.txtANInsert.Text;
            if frmmain.udAEPosition.Position > tbPosition.max then
                tbPosition.Max := frmmain.udANPosition.Position;
            tbPosition.Position := frmmain.udANPosition.Position;
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbAEndOfName.Checked) then begin
        NewControl;
        Self.GetAddBeginningEndingStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepAddBeginningEnding  do begin
            cbxAddPart.ItemIndex := ADDPART_NAME;
            txtAdd.Text := frmmain.txtAEndOfName.Text;
            cbxAddAtThe.ItemIndex := ADDATTHE_ENDING;
        end;
        MyToggle(sac);
    end;
    if (frmmain.cbAEndOfExtension.Checked) then begin

        NewControl;
        Self.GetAddBeginningEndingStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepAddBeginningEnding  do begin

            cbxAddPart.ItemIndex := ADDPART_EXT;
            txtAdd.Text := frmmain.txtAEndOfExtension.Text;
            cbxAddAtThe.ItemIndex := ADDATTHE_ENDING;
        end;
        MyToggle(sac);
    end;
    if (frmMain.cbCollision.checked) then begin
        NewControl;
        Self.GetAddBeginningEndingStep(sac.pnlbtm,sac);
        with sac.ScriptStep as TScriptStepAddBeginningEnding  do begin

            cbxAddPart.ItemIndex := ADDPART_NAME;
            txtAdd.Text := frmmain.txtCollision.Text;
            cbxAddAtThe.ItemIndex := ADDATTHE_ENDING;
            cbOnCollision.Checked := true;
        end;
        MyToggle(sac);

    end;


    FrmScriptActions.EndUpdate;
    wc.Visible := true;
    self.DontExpand := false;
    if sac <> nil then begin
        MyToggle(sac);
        //MyToggle(sac);
    end;

end;


{ TScriptStep }
constructor TScriptStep.Create;
begin
    FrmDummyAutoPreview.InformOfKeypress;
    fenabled := true;
end;
function TScriptStep.GetWildcards : TWildcards;
begin
    result := FrmScriptActions.GetWildCards;
end;

{ MyPanel }

const
    DIM_BRIGHT_PC = 0.78;
    DIM_DARK_PC = 0.72;

procedure TMyToplinePanel.Invalidate;
begin
    inherited;
end;
procedure TMyToplinePanel.Paint;
var
    r, rHead : TRect;
    ssc : TScriptStepControl;
    c : TColor;
    wd : integer;
    savedColor : TColor;
    useEffects : boolean;
begin
    inherited;

    self.ParentColor := false;
    self.Color := DimColor(clBtnFace, DEFAULT_DIM);
    self.Canvas.Brush.Color := frmmain.scrlScriptbox.Color;
    r := self.ClientRect;
    inc(r.Bottom,3);
    r.top := 0;
    self.Canvas.Pen.Color := self.Canvas.Brush.Color;
    self.Canvas.Rectangle(r);



    // outter gray rounded corners
    if self.RoundBottom  then begin
        self.Canvas.Pen.Color := OUTLINE_COLOR;
    end else begin
        self.Canvas.Pen.Color := OUTLINE_EXPANDED_COLOR;
    end;
    self.Canvas.Brush.Color := DimColor(self.color,   0.97);
    self.Canvas.RoundRect(r,11,7);



    // side lines under rounded header
{
    if self.RoundBottom  then begin
        self.Canvas.Rectangle(
            r.Left,
            r.top + trunc(self.clientrect.Bottom * 0.5),
            r.Right,
            r.Bottom-7
        );
    end else begin
        self.Canvas.Rectangle(
            r.Left,
            r.top + trunc(self.clientrect.Bottom * 0.5),
            r.Right,
            r.Bottom
        );
    end;
}

    useEffects := frmConfig.cbUseEffects.Checked;
    savedColor := clr;
    if Hovered then begin
        clr := DimColor(clr, 1.45);
    end;


    if RoundBottom then begin
        // Closed Step
        // main color
        c := DimColor(self.clr, DEFAULT_DIM);
        self.Canvas.Pen.Color := c;
        self.canvas.Brush.color := c;

        rhead.Left := r.left;
        rhead.right := r.Right;
        rhead.top := 0;
        rhead.Bottom := r.Bottom;

        self.Canvas.Rectangle(rhead);
        if (UseEffects) then begin
            GraphUtil.GradientFillCanvas(
                self.Canvas, c,DimColor(c,GRADIENT_PC),rhead,gdVertical
            );
        end;

//        rhead.Top := self.Parent.Top;
        rhead.Top := 0;
        Inc(rhead.Bottom, rhead.Top+1);
        rhead.right := TMyToplinePanel(self.Parent.parent).ClientRect.right;
        Inc(rhead.left,1);
        if (UseEffects) then begin

            GraphUtil.GradientFillCanvas(
                TMyToplinePanel(self.Parent.parent).Canvas, c,DimColor(self.clr,SHADOW_PC),rhead,gdVertical
            );


            // bottom darker color
            rhead.Left := r.left;
            rhead.right := r.Right;
            rhead.Top := r.top + trunc(r.Bottom * SHADOW_HEIGHT_PC);
            rhead.Bottom := r.Bottom;

            c := DimColor(self.clr,SHADOW_PC);
            self.canvas.Brush.color := c;
            self.Canvas.Pen.Color := c;
//            self.Canvas.RoundRect(rHead,11,4);
        end else begin
//            self.Canvas.Rectangle(r);
//            GraphUtil.GradientFillCanvas(
//                TMyToplinePanel(self.Parent.parent).Canvas, c,c,rhead,gdVertical
//            );
        end;

        if (UseEffects) and FALSE then begin
            //
            // right button shadowbox
            ssc := TScriptStepControl(FrmScriptActions.ScriptControlList.items[self.tag]);
            c := DimColor(self.clr,SHADOW_PC);
            self.Canvas.Pen.Color := c;
            self.canvas.Brush.color := c;
            if ((not assigned(ssc.sbenable)) or (ssc.sbenable.Visible = false)) then begin
                if frmconfig.cbShowUpDown.checked then begin
                    wd := ssc.pnlTop.Width-(((20+3)*4)+8)  +2;
                end else begin
                    wd := ssc.pnlTop.Width-(((20+3)*2)+8) +2;
                end;
               { self.canvas.RoundRect(
                    wd,
                    ssc.sbdelete.Top-2,
                    ssc.pnltop.Width,
                    ssc.pnlTop.Top + ssc.sbdelete.Height+6,
                    11,11
                );}
            end;
        end;

    end else begin
        self.Canvas.Pen.Color := clBlack;
        self.Canvas.brush.Color := clBlack;


        // main color
        c := DimColor(self.clr, DEFAULT_DIM);
        self.Canvas.Pen.Color := OUTLINE_EXPANDED_COLOR;
        self.canvas.Brush.color := c;

        rhead.Left := r.left+1;
        rhead.right := r.Right-1;
        rhead.top := 0;
//        rhead.Bottom := self.ClientRect.Bottom;
        rhead.Bottom := r.Bottom;

        self.Canvas.RoundRect(rhead,11,4);
        if (UseEffects) then begin
            rHead.Top := 3;
            GraphUtil.GradientFillCanvas(
                self.Canvas, c,DimColor(c,GRADIENT_PC),rhead,gdVertical
            );
            // bottom darker color
            rhead.Left := self.ClientRect.left+1;
            rhead.right := self.ClientRect.Right-1;
            rhead.Top := self.ClientRect.top + trunc(self.clientrect.Bottom * SHADOW_HEIGHT_PC);
//            rhead.Bottom := self.ClientRect.Bottom;
            rhead.Bottom := self.BoundsRect.Bottom;

            c := DimColor(self.clr,SHADOW_PC);
            self.canvas.Brush.color := c;
            self.Canvas.Pen.Color := c;
//            self.Canvas.FillRect(rHead);

        end;
    end;

    clr := savedColor;
end;
procedure TMyToplinePanel.Resize;
begin
    canvas.Brush.Style := bsClear;
    inherited;
end;

{ TyPanelBottom }
procedure TMyPanelBottom.Invalidate;
begin
    if not self.Visible then Exit;
    //if not SmoothAni then
    //self.DoubleBuffered := true;
    canvas.Brush.Style := bsClear;
    self.ParentColor := false;
    self.Color := clWhite;
    if FrmMain.timResize.Enabled then EXIT;
    inherited;
    self.Color := clWhite;



end;
procedure TMyPanelBottom.Paint;
var
    ssc : TScriptStepControl;
    r : TRect;
begin
    if not self.Visible then
        EXIT;

    if Self.Height > 30 then begin    // TODO: explain this
    //    inherited;
    end;

    self.Color := clWhite; //clWindow make unrendered sections for some reason

    r := self.ClientRect;

    r.Top := 0;
    self.Canvas.Brush.Style := bsSolid;
    self.Canvas.Brush.Color := frmmain.scrlScriptbox.Color;
    self.Canvas.pen.Color := frmmain.scrlScriptbox.Color;
    self.Canvas.FillRect(r);

    // darker rounded bottom with outline
    ssc := TScriptStepControl(FrmScriptActions.ScriptControlList.items[self.tag]);
    self.Canvas.Pen.Color := OUTLINE_EXPANDED_COLOR;
    self.Canvas.Brush.Color := DimColor(SetBlueColor(clBtnFace, 1.14),   0.95);


    self.Canvas.RoundRect(
        r.Left,
        r.bottom - 12, // was -14
        r.Right,
        r.Bottom,
        11,7
    );

    //left/right lines
    self.Canvas.Rectangle(
        rect(
        r.Left,
        r.top,
        r.Right,
        r.Bottom - 8
        )
    );

    // white fill
    self.Canvas.Brush.Color := clWindow;
    self.Canvas.FillRect(
        rect(
        r.Left+1,
        r.top,
        r.Right-1,
        r.Bottom - 7
        )
    );


end;

//
// Workarounds to get transparent panels
//


procedure TTransFlowPanel.CreateParams(var Params: TCreateParams);
begin
     inherited CreateParams(Params);

    { params.ExStyle := Params.ExStyle + WS_EX_Transparent;
     ControlStyle := ControlStyle - [csOpaque] + [csAcceptsControls];
     canvas.Brush.Style := bsClear;}
     //self.DoubleBuffered := true;

    ParentBackground := true;
    ParentColor := false;

    self.color := DimColor(clBtnFace, 0.95);
    //self.clr := clWhite;
end;
procedure TTransFlowPanel.Invalidate;
begin
    //if not SmoothAni then
    //canvas.Brush.Style := bsClear;
    //inherited;

    {if Assigned(Parent) and Parent.HandleAllocated and HandleAllocated then
    RedrawWindow(Parent.Handle, nil, 0, RDW_ERASE or RDW_FRAME or RDW_INTERNALPAINT or RDW_INVALIDATE
      or RDW_ERASENOW or RDW_UPDATENOW or RDW_ALLCHILDREN);
    }
    inherited invalidate;
end;
procedure TTransFlowPanel.SetShade(value : TColor);
begin
    self.clr := value;
end;
procedure TTransFlowPanel.Paint;
var
    rhead : TRect;
    c : TColor;
    i,x,y,ht : integer;
    ctrl : TControl;
    bld : boolean;

    procedure FindLeftmostAndRightmostChildren(var x : integer; var y : integer; var ht : integer);
    var
        i : integer;
        ctrl : TControl;
    begin
        x := 0;
        y := 0;
        ht := 0;
        for i := 0 to self.ControlCount-1 do begin
            ctrl := self.Controls[i];
            x := Min(ctrl.left,x);
            y := Max(ctrl.left+ctrl.Width,y);
            ht := max(ctrl.Height, ht);
        end;
    end;
begin
    canvas.Brush.Style := bsClear;
    self.BorderStyle := bsNone;

    inherited;

    if parent.Visible = False or
        Self.Visible = false then
        EXIT;


    canvas.Brush.Style := bsClear;

    x := 0;
    y := 0;
    for i :=0 to self.ControlCount-1 do begin
        ctrl := self.Controls[i];
        x := Min(ctrl.left,x);
        y := Max(ctrl.left+ctrl.Width,y);
        ht := ctrl.Height;
    end;

    // NOTE: No Idea why this in procedure form causes colors to fail
    //FindLeftmostAndRightmostChildren(x,y,ht);



    self.Color := DimColor(clBtnFace, DEFAULT_DIM);
    rhead.Top := 0;
    rhead.Bottom := self.BoundsRect.Bottom+1;

    c := DimColor(self.clr, DEFAULT_DIM);

    self.canvas.Brush.color := c;
    self.Canvas.FillRect(rhead);


    if (frmConfig.cbUseEffects.Checked) then begin
        GraphUtil.GradientFillCanvas(
            self.Canvas, c,DimColor(c,GRADIENT_PC),rhead,gdVertical
        );
        // dark bottom
        rhead.Left := self.ClientRect.left;
        rhead.right := self.ClientRect.Right;
        rhead.Top := self.ClientRect.top + trunc(self.clientrect.Bottom * SHADOW_HEIGHT_PC)+2;
        rhead.Bottom := self.BoundsRect.Bottom ;

        c := DimColor(self.clr, SHADOW_PC);
        self.canvas.Brush.color := c;
        self.Canvas.FillRect(rhead);
    end;

    rhead.Left := x;
    rhead.Right := y+7;
    rhead.Top := Self.ClientRect.Top+4;
    rhead.Bottom := rhead.Top + ht+5;


    if self.fSelected then begin
        c := ShadeColor(clWhite, clHighlight, 0.18);
    end else begin
        if fenabled then begin
            c := DimColor(clWhite,0.96);
        end else begin

            c := ShadeColor(clWhite, self.clr, 0.60);
        end;
    end;
    self.canvas.Brush.color := c;
    if self.fSelected then begin
        self.Canvas.Pen.Color := ShadeColor(clBlack, clHighlight, 0.30);
    end else begin

        if fenabled then begin
            self.Canvas.Pen.Color := dimcolor(self.clr, 0.90);
        end else begin
            self.Canvas.Pen.Color := self.clr;
        end;
    end;

    self.Canvas.RoundRect(
        rhead,7,7
    );

end;

{ TTransPanel }
procedure TTransPanel.CreateParams(var Params: TCreateParams);
begin
    inherited CreateParams(Params);
   { params.ExStyle := Params.ExStyle + WS_EX_Transparent;
    ControlStyle := ControlStyle - [csOpaque] + [csAcceptsControls];
    canvas.Brush.Style := bsClear;
    }
    //self.DoubleBuffered := true;

     self.borderstyle := bsSingle;
end;
procedure TTransPanel.Invalidate;
begin
    canvas.Brush.Style := bsClear;
    inherited;
end;
procedure TTransPanel.Paint;
begin
    canvas.Brush.Style := bsClear;
    inherited;
end;
procedure TTransPanel.Resize;
var i : integer;
begin
    canvas.Brush.Style := bsClear;


    for i := 0 to self.ControlCount - 1 do begin
        self.Controls[i].Invalidate;
    end;

    inherited Resize;
end;



procedure TRMSpeedButton.Paint;
var r : TRect;
begin
    //ParentColor := False;
    Canvas.Brush.Color := clBtnFace;
    Canvas.Pen.Color := clWhite;
    //Color := clWhite;

    if self.MouseInControl then begin
        r := self.ClientRect;
        r.Left := r.Left + 1;
        Canvas.Rectangle(r);
    end;

    inherited;


end;

{ TScriptStepControl }

function TScriptStepControl.GetTag: integer;
begin
    result := self.pnlTop.Tag;
end;

{ TScriptStepAdd }




function ToRegHEX(s : string) : string;
var i : integer;
begin
    result := '';
    for i := 1 to length(s)  do begin
        if s[i] in ['[', '\', '^', '$', '.', '|', '?', '*', '+', '(', ')', ']'] then begin
            result := result + '\x' + IntToHex(Ord(s[i]),2);
        end else if s[i] = ' ' then begin
            result := result + '\x20';
        end else begin
            result := result + s[i];
        end;
    end;
end;

procedure TFrmScriptActions.RunScriptOn(var sname, ext: string;
    var rec : RenameRecord;
    curTD : TTagData;
    var collisionindex : integer);
var i : integer;
    ssc : TScriptStepControl;
    tempname, s2: string;
    rwc : TWildcards;
    //RE : TWildcards;
    // collision stuff
    CollisionChecked : boolean;
    AppendCount : integer;
    temprec : RenameRecord;
    b : boolean;
    path : string;
    //used for conditional execute
    lastName : string;
    lastExt : string;
    skipcount : integer;
    ReplaceNumberCounter : integer;


    procedure CollisionCheckIt;
    begin
        if not CollisionChecked then EXIT;
        if ssc.Deleted then EXIT;

        with ssc.ScriptStep as TScriptStepAddBeginningEnding do begin
            path := IncludeTrailingPathDelimiter(ExtractFilePath(rec.FullName));
            if (frmCollisionDialog.Exists(path + sname + ext, collisionindex) ) then begin
                if (cbOnCollision.Checked) then begin
                    curtd.IsCollision := false;
                    if curtd.OriginalCollision then begin
                        appendCount := 0;
                    end else begin
                        appendCount := 1; // reserve position 1 for the original collision name
                    end;

                    temprec.FullName := rec.FullName;
                    temprec.name := rec.name;
                    temprec.ext := rec.ext;
                    temprec.listposition := 0;
                    temprec.listtotal := rec.listtotal;
                    temprec.selectedposition := 0; //
                    temprec.selectedtotal := rec.selectedtotal;
                    temprec.foldertotal := rec.foldertotal;
                    temprec.incrementby := rec.incrementby;
                    temprec.filedate := rec.filedate;
                    temprec.filesize := rec.filesize;
                    temprec.Created := rec.created;
                    temprec.Modified := rec.Modified;

                    repeat
                        inc(appendCount);
                        // fake position data for counters
                        temprec.selectedposition := appendCount;
                        temprec.listposition := appendCount;

                        case cbxAddPart.ItemIndex  of
                        ADDPART_NAME: s2 := sname;
                        ADDPART_EXT:  s2 := ext;
                        end;

                        case cbxAddAtThe.ItemIndex of
                        ADDATTHE_BEGINNING:
                            tempname := ReplaceMetaVariables(
                                GetText(txtAdd),
                                temprec
                            ) + s2;
                        ADDATTHE_ENDING:
                            tempname := s2 + ReplaceMetaVariables(
                                GetText(txtAdd),
                                temprec
                            );
                        end;
                        b := FrmCollisionDialog.Exists(path + tempname + ext);
                    until (not b)
                        or (not UnitUtil.ContainsTerm('?',GetText(txtAdd)))
                        or (not UnitUtil.ContainsTerm('?n', GetText(txtAdd)))
                        or (appendCount > 50000)
                        ; // skip when no counter exists

                        // need a better way
                        // to detect an infinite loop condition

                    curTD.IsCollision := b;
                    case cbxAddPart.ItemIndex  of
                    ADDPART_NAME: sname :=  tempname;
                    ADDPART_EXT: ext := tempname;
                    end;
                end;
            end;
        end;
    end;
    procedure ConditionalExecute;
    var tlastname : string;
        tname : string;

        b : boolean;
    begin
        tname := sname + ext;
        tlastname := lastname + lastext;
        b := false;

        with ssc.ScriptStep as TScriptStepConditionExecute do begin
            skipcount := SafeInt(txtCE1.text,1);
            case cbxCECondition.ItemIndex  of
            SKIPEXECUTE_CHANGES: begin
                b := tname <> tlastname;
            end;
            SKIPEXECUTE_NAME_MATCH: begin
                // search name or name+ext
                rwc.SetSearchPattern(ReplaceMetaVariables(GetText(txtCEText), rec));
                rwc.SetSearchString(tname);

                b := rwc.Matches;
                rwc.SetSearchString(sname);
                if not b then b := rwc.matches;
            end;
            SKIPEXECUTE_NAME_CONTAINS: begin
                rwc.SetSearchPattern(ReplaceMetaVariables(GetText(txtCEText), rec));
                rwc.SetSearchString(tname);
                b := rwc.FindString <> '';
            end;
            end;

            case cbxCESkipRemove.ItemIndex of
            CONDITIONALEXECUTE_SKIP: begin
                if not b then begin
                    skipcount := 0; // Execute
                end else begin
                    // skip
                end;
            end;
            CONTITIONALEXECUTE_EXECUTE: begin
                if not b then begin
                    // skip
                end else begin
                    skipcount := 0; // execute
                end;
            end;
            end;
        end;
    end;
begin
    rwc := self.GetWildCards;


    CollisionChecked := false;
    collisionindex := -1;
    UnitMetaVariables.settemp('');
    ReplaceNumberCounter := 0;

    lastName := sname;
    lastExt := ext;
    skipcount := 0;
    for i := 0 to ScriptControlOrder.Count - 1 do begin
        ssc := TScriptStepControl(ScriptControlOrder.items[ i ]);
        if ssc.Deleted then CONTINUE;
        if not ssc.ScriptStep.Enabled  then CONTINUE;

        if skipcount > 0 then begin
            dec(skipcount);
            lastName := sname;
            lastExt := ext;
        end else begin
            if ssc.ScriptStep.ScriptType <> stConditionalExecute then begin
                lastName := sname;
                lastExt := ext;
            end;
            case ssc.ScriptStep.ScriptType of
                stAddBeginningEnding: begin
                    with ssc.ScriptStep as TScriptStepAddBeginningEnding do
                        CollisionChecked := CollisionChecked or cbOnCollision.Checked;
                    ssc.ScriptStep.Execute(sname, ext, rec);
                end;
                stAddCounter,
                stAddBeforeAfter,
                stAddPosition,
                stReplaceNameExtension,
                stRemoveExpression,
                stRemoveBeforeAfter,
                stRemoveSubstring,
                stRemoveBeginningEnding,
                stRemoveWords,
                stReplaceNumbers,
                stReplaceCharacters,
                stSplitAndMove,
                stSelectAndMove,
                stSwap,
                stFindAndEdit: ssc.ScriptStep.Execute(sname, ext, rec);
                //stRemoveWords: RemoveWords;
                stConditionalExecute: ConditionalExecute;
            end;
        end;

        case ssc.ScriptStep.ScriptType of
             stAddBeginningEnding: CollisionCheckIt;
        end;
    end;
end;


//
{$region '// ScriptStep Execute/Reset/Sentences //'}
procedure TScriptStepAddCounter.Execute(var sname, ext: string; var rec: RenameRecord);
var s2 : string;
begin
    rec.ScriptChanged := true;
    s2 := '?n' + IntTOStr(AUDZeros.position) + '?';
    if length(s2) = 4 then begin
        insert('0',s2,3);
    end;
    case cbxBeginningEnding.itemindex of
        ADDZEROS_BEGINNING:
            begin
                sname := ReplaceMetaVariables(s2, rec) + sname;
            end;
        ADDZERO_ENDING:
            begin
                sname := sname + ReplaceMetaVariables(s2, rec);
            end;
    end;
end;
procedure TScriptStepAddBeginningEnding.Execute(var sname, ext: string; var rec: RenameRecord);
begin
    if not cbOnCollision.Checked then begin
        case cbxAddPart.ItemIndex of
        ADDPART_NAME:
            begin
                case cbxAddAtThe.ItemIndex  of
                ADDATTHE_BEGINNING:
                    Insert(ReplaceMetaVariables(FrmScriptActions.GetText(txtAdd), rec), sname, 1 );
                ADDATTHE_ENDING:
                    sname := sname + ReplaceMetaVariables(FrmScriptActions.GetText(txtAdd) , rec);
                end;
            end;
        ADDPART_EXT:
            begin
                case cbxAddAtThe.ItemIndex  of
                ADDATTHE_BEGINNING:
                    // skip the Dot position
                    Insert(ReplaceMetaVariables(FrmScriptActions.GetText(txtAdd) , rec), ext, 2);
                ADDATTHE_ENDING:
                    ext :=  ext + ReplaceMetaVariables(FrmScriptActions.GetText(txtAdd), rec);
                end;
            end;
        end;
    end;
end;
procedure TScriptStepAddPosition.Execute(var sname, ext: string; var rec: RenameRecord);
var pos : integer;
begin
    pos := 1;
    case cbxFromTheBeginningEnding.ItemIndex  of
        ADDATTHE_BEGINNING: begin
            pos := SafeInt(txtAddPosition.text,1);
        end;
        ADDATTHE_ENDING:begin
            case cbxAddPart.ItemIndex of
            ADDPART_NAME:
                pos := length(sname) - (SafeInt(txtAddPosition.text,1)-2);
            ADDPART_EXT:
                pos := length(ext) - (SafeInt(txtAddPosition.text,1)-2);
            end;
        end;
    end;


    case cbxAddPart.ItemIndex of
    ADDPART_NAME:
        begin
            Insert(ReplaceMetaVariables(FrmScriptActions.GetText(txtAdd), rec), sname, pos );
        end;
    ADDPART_EXT:
        begin
            Insert(ReplaceMetaVariables(FrmScriptActions.GetText(txtAdd), rec), ext, pos+1 );
        end;
    end;
end;
procedure TScriptStepAddBeforeAfter.Execute(var sname, ext: string; var rec: RenameRecord);
var tempname : string;
    rwc : TWildcards;
    startidx, endidx : cardinal;
begin
    rwc := self.GetWildCards;

    if txtAfter.Text <> '' then begin

        rwc.SetSearchPattern(
            ReplaceMetaVariables(FrmScriptActions.GetText(txtAfter), rec)
        );
        rwc.SetSearchString(sname);
        tempname := sname;
        rec.selectedposition := rec.changedcount;
        if rwc.FindFirst(startidx, endidx) then begin
            rec.ScriptChanged := true;
            case cbxAddPlace.ItemIndex of
            ADDPLACE_AFTER:
                Insert(
                    ReplaceMetaVariables(FrmScriptActions.GetText(txtAdd), rec),
                    sname,
                    endidx +1
                );
            ADDPLACE_BEFORE:
                Insert(
                    ReplaceMetaVariables(FrmScriptActions.GetText(txtAdd), rec),
                    sname,
                    startidx
                );
            end;
        end;
    end;

    rwc.free;
end;
procedure TScriptStepReplaceNameExtension.Execute(var sname, ext: string; var rec: RenameRecord);
var tempname : string;
    rwc : TWildcards;
    startidx, endidx : cardinal;
begin
    rwc := self.GetWildCards;

    case cbxReplacePart.ItemIndex of
    REPLACEPART_NAME:
        sname := ReplaceMetaVariables(FrmScriptActions.GetText(txtReplaceWith), rec);

    REPLACEPART_EXT:
        ext := ReplaceMetaVariables(FrmScriptActions.GetText(txtReplaceWith), rec);

    REPLACEPART_PHRASE: begin
        rec.selectedposition := rec.changedcount;
        if cbReplaceOnly.checked  then begin
            startIDX := SafeInt(txtRP2X.text,1);
            endIDX := (SafeInt(txtRP2Y.text,1) - SafeInt(txtRP2X.text,1)) + 1;
            tempname :=  Copy(sname, startIDX, endIDX );
            Delete(sname, startIDX, endIDX);
            tempname := rwc.SearchReplaceWith(
                    tempname,
                    ReplaceMetavariables(FrmScriptActions.GetText(txtReplacePhrase), rec),
                    ReplaceMetaVariables(FrmScriptActions.GetText(txtReplaceWith), rec), true
            );
            rec.ScriptChanged := rwc.FirstFound;

            Insert(
                tempname,
                    sname,
                startIDX);
        end else begin
            case cbxReplaceTarget.ItemIndex  of
            REPLACETARGET_NAME: begin
                sname := rwc.SearchReplaceWith(sname,
                    ReplaceMetaVariables(FrmScriptActions.GetText(txtReplacePhrase), rec),
                    ReplaceMetaVariables(FrmScriptActions.GetText(txtReplaceWith), rec),
                     true);
                rec.ScriptChanged := rwc.FirstFound;
            end;
            REPLACETARGET_EXT: begin
                ext := rwc.SearchReplaceWith(ext,
                    ReplaceMetaVariables(FrmScriptActions.GetText(txtReplacePhrase), rec),
                    ReplaceMetaVariables(FrmScriptActions.GetText(txtReplaceWith), rec),
                     true);
                rec.ScriptChanged := rwc.FirstFound;
            end;
            REPLACETARGET_NAME_OR_EXT: begin
                sname := rwc.SearchReplaceWith(sname,
                    ReplaceMetaVariables(FrmScriptActions.GetText(txtReplacePhrase), rec),
                    ReplaceMetaVariables(FrmScriptActions.GetText(txtReplaceWith), rec),
                     true);
                rec.ScriptChanged := rwc.FirstFound;
                ext := rwc.SearchReplaceWith(ext,
                    ReplaceMetaVariables(FrmScriptActions.GetText(txtReplacePhrase), rec),
                    ReplaceMetaVariables(FrmScriptActions.GetText(txtReplaceWith), rec),
                     true);
                rec.ScriptChanged := rec.ScriptChanged or rwc.FirstFound;
            end;

            end;
        end;
    end;
    end;

    rwc.free;
end;
procedure TScriptStepReplaceCharacters.Execute(var sname, ext: string; var rec: RenameRecord);
var j : integer;
    tempname : string;
    c : char;
    charlist : array[#0 .. #255] of boolean;
begin
    EXIT;

    case cbxReplaceCharPart.ItemIndex  of
    REPLACECHARPART_NAME: begin
        fillchar(charlist, length(charlist), false);
        for j := 1 to length(sname) do begin
            charlist[sname[j]] := true;
        end;

        for j := 1 to length(FrmScriptActions.GetText(txtReplaceChar1)) do begin
            if j <= length(FrmScriptActions.GetText(txtReplaceChar2)) then begin
                tempname := FrmScriptActions.GetText(txtReplaceChar2)[j];
            end else begin
                tempname := '';
            end;

            c := FrmScriptActions.GetText(txtReplaceChar1)[j];
            if (tempname = '') or (charlist[c]) then begin
                sname := StringReplace(
                    sname, c, tempname,
                    [rfreplaceall, rfignorecase]);
            end;
        end;
    end;

    REPLACECHARPART_EXT: begin
        fillchar(charlist, length(charlist), false);
        for j := 1 to length(ext) do begin
            charlist[ext[j]] := true;
        end;
        for j := 1 to length(FrmScriptActions.GetText(txtReplaceChar1)) do begin
            if j <= length(FrmScriptActions.GetText(txtReplaceChar2)) then begin
                tempname := FrmScriptActions.GetText(txtReplaceChar2)[j];
            end else begin
                tempname := '';
            end;

            c := FrmScriptActions.GetText(txtReplaceChar1)[j];
            if (tempname = '') or (charlist[c]) then begin
                ext := StringReplace(
                    ext, c, tempname,
                    [rfreplaceall, rfignorecase]);
            end;
        end;
    end;
    end;
end;
procedure TScriptStepReplaceNumbers.Execute(var sname, ext: string; var rec: RenameRecord);
var tempname, replacestr : string;
    j : integer;
begin

    rec.selectedposition := rec.changedcount;

    // count the left/right numbers
    // and remove them
    case cbxNumbers.ItemIndex of
    REPLACENUMBER_RIGHT: j := length(sname);
    REPLACENUMBER_LEFT: j := 1;
    end;
    while (j > 0) and (j<=length(sname)) and
        (CharInSet(sname[j], ['0' .. '9'])) do begin
        case cbxNumbers.ItemIndex of
        REPLACENUMBER_RIGHT: dec(j);
        REPLACENUMBER_LEFT: inc(j);
        end;
    end;

    case cbxNumbers.ItemIndex of
    REPLACENUMBER_RIGHT: if (j = length(sname)) then EXIT;
    REPLACENUMBER_LEFT: if (j = 1) then EXIT;
    end;

    rec.ScriptChanged := true;
    tempname := '';
    case cbxNumbers.ItemIndex of
    REPLACENUMBER_RIGHT:
        begin
            tempname := rightstr(sname,length(sname)-j);
            sname := midstr(sname,1,j);
        end;
    REPLACENUMBER_LEFT:
        begin
            if j>1 then begin
                tempname := leftstr(sname,j-1);
                sname := midstr(sname,j, (length(sname)-j)+1);
            end;
        end;
    end;

    case cbxRNSequence.ItemIndex   of
    REPLACENUMBER_SEQUENCE: begin
        replacestr := FrmDummyShowPreview.GetReplacementSequence(tempname);
        while length(tempname) > length(replacestr) do begin
            replacestr := '0' + replacestr;
        end;
    end;
    REPLACENUMBER_PHRASE: begin
    	if tempname <> sname then begin
			replacestr := ReplaceMetaVariables(FrmScriptActions.GetText(txtRNWith), rec);
        end;
    end;
    end;

    case cbxNumbers.ItemIndex of
    REPLACENUMBER_LEFT:
        sname := replacestr + sname;
    REPLACENUMBER_RIGHT:
        sname := sname + replacestr;
    end;
end;
procedure TScriptStepConditionExecute.Execute(var sname, ext: string; var rec: RenameRecord);
begin

end;
procedure TScriptStepRemoveExpression.Execute(var sname, ext: string; var rec: RenameRecord);
    function NotEach(s : string) : string;
        var i : integer;
        begin
            result := '';

            for i := 1 to length(s) do begin
                result := result + '^' + s[i];
            end;
        end;

    function SpaceEnc(s : string) : string;
        var i : integer;
        begin
            result := '';
            for i := 1 to length(s)  do begin
                if s[i] = ' ' then begin
                    result := result + '\x20';
                end else begin
                    result := result + s[i];
                end;
            end;
        end;
var i : integer;
    startidx, endidx : cardinal;
    s2, tempname, swith,
    sStarting, sMiddle, sEnding, subExp : string;
    re : TRegExpWildcards;
    b : boolean;
begin
    re := TRegExpWildcards.create;

    s2 := '(?i)';  // Case Insensitive
    RE.SetSearchString(sname);

    case cbxreBegginingEnding.ItemIndex  of
        REMOVEEXPRESSION_BEGINNNING:
            s2 := s2 + '\A';
    end;

    if cbreThatSTartsWIth.checked then begin
        case cbxrePhraseCharacters.ItemIndex of
            REMOVEEXPRESSION_PHRASE:
                sStarting := ToRegHEX(ReplaceMetaVariables(FrmScriptActions.GetText(txtrePhraseCharacters), rec));
            REMOVEEXPRESSION_CHARACTERS:
                sStarting := '[' + ToRegHEX(ReplaceMetaVariables(FrmScriptActions.GetText(txtrePhraseCharacters), rec)){)} + ']+';
            REMOVEEXPRESSION_CHARACTERS_EXCEPT:
                sStarting := '[^'+ToRegHEX(ReplaceMetaVariables(FrmScriptActions.GetText(txtrePhraseCharacters), rec)) + ']+';
            REMOVEEXPRESSION_NOT_PHRASE:
                sStarting := '.+?'+ToRegHex(ReplaceMetaVariables(FrmScriptActions.GetText(txtrePhraseCharacters), rec));
        end;
        s2 := s2 + sStarting;
    end;
    if cbreHasInTheMiddle.checked then begin
        case cbxrePhraseCharacters2.ItemIndex of
            REMOVEEXPRESSION_PHRASE:
                sMiddle :=   ToRegHEX(ReplaceMetaVariables(FrmScriptActions.GetText(txtrePhraseCharacters2), rec));
            REMOVEEXPRESSION_CHARACTERS:
                sMiddle := '[' + ToRegHEX(ReplaceMetaVariables(FrmScriptActions.GetText(txtrePhraseCharacters2), rec)){)} + ']+';
            REMOVEEXPRESSION_CHARACTERS_EXCEPT:
                sMiddle := '[^'+ToRegHEX(ReplaceMetaVariables(FrmScriptActions.GetText(txtrePhraseCharacters2), rec)) + ']+';
            REMOVEEXPRESSION_NOT_PHRASE:
                sMiddle := '.+?'+ToRegHex(ReplaceMetaVariables(FrmScriptActions.GetText(txtrePhraseCharacters2), rec));
        end;
        s2 := s2 + sMiddle;
    end;
    if cbreAndEndsWith.checked then begin
        case cbxrePhraseCharacters3.ItemIndex of
            REMOVEEXPRESSION_PHRASE:
                sEnding := ToRegHEX(ReplaceMetaVariables(FrmScriptActions.GetText(txtrePhraseCharacters3), rec));
            REMOVEEXPRESSION_CHARACTERS:
                sEnding := '[' + ToRegHEX(ReplaceMetaVariables(FrmScriptActions.GetText(txtrePhraseCharacters3), rec)){)} + ']+';
            REMOVEEXPRESSION_CHARACTERS_EXCEPT:
                sEnding := '[^'+ToRegHEX(ReplaceMetaVariables(FrmScriptActions.GetText(txtrePhraseCharacters3), rec)) + ']+';
            REMOVEEXPRESSION_NOT_PHRASE:
                sEnding := '.+?'+ToRegHex(ReplaceMetaVariables(FrmScriptActions.GetText(txtrePhraseCharacters3), rec));
        end;
        s2 := s2 + sEnding;
    end;

    case cbxreBegginingEnding.ItemIndex  of
        REMOVEEXPRESSION_ENDING:
            s2 := s2 + '\Z';
    end;


    if (s2<>'(?i)') and (s2<>'(?i)\Z') and (s2<>'(?i)\A') then begin
        RE.SetSearchPattern(s2);
        if RE.FindFirst(startidx, endidx) then begin
            rec.ScriptChanged := true;
            if cbreRemoveNumber.checked then begin
                // skip ahead to the specified instance,
                // ignore if that number is never found
                // - we're already at instance 1
                i := 1;
                b := true;
                while (i<safeint(txtRE1.Text)) and (b) do begin
                    b := RE.FindNext(startidx, endidx);
                    inc(i);
                end;
                if b then begin
                    tempname := Copy(sname, startidx, (endidx-startidx)+1);
                    Delete(sname, startidx, (endidx-startidx)+1);
                    if cbreAndStore.Checked then begin
                         UnitMetaVariables.SetTemp(tempname);
                    end;

                    if cbxRERemoveReplace.ItemIndex = REMOVEEXPRESSION_REPLACE then begin
                        Insert( ReplaceMetaVariables(FrmScriptActions.GetText(txtREWith),rec), sname, startidx );
                    end;
                end;
            end else begin
                // replace all instances
                // - copy the first instance for the ?temp? variable
                tempname := Copy(sname, startidx, (endidx-startidx)+1);

                case cbxRERemoveReplace.ItemIndex of
                REMOVEEXPRESSION_REMOVE: begin
                    swith := '';
                    RE.SetSearchString(tempname);
                    subExp := '(' + sStarting + ')' +
                        '(' + sMiddle + ')' +
                        '(' + sEnding + ')';
                    case cbxreBegginingEnding.ItemIndex  of
                        REMOVEEXPRESSION_BEGINNNING:
                            subExp := '\A' + subExp;
                        REMOVEEXPRESSION_ENDING:
                            subExp := subExp + '\Z';
                    end;
                    subExp := '(?i)' + subExp;

                    RE.SetSearchPattern( subExp );
                    case cbxREParts.ItemIndex of
                    REMOVEEXPRESSION_PARTS_ALL: ;
                    REMOVEEXPRESSION_PARTS_STARTING:
                        swith := re.SearchReplaceWith('\2\3',false);
                    REMOVEEXPRESSION_PARTS_MIDDLE:
                        swith := re.SearchReplaceWith('\1\3',false);
                    REMOVEEXPRESSION_PARTS_ENDING:
                        swith := re.SearchReplaceWith('\1\2',false);
                    REMOVEEXPRESSION_PARTS_BUT_STARTING:
                        swith := re.SearchReplaceWith('\1',false);
                    REMOVEEXPRESSION_PARTS_BUT_MIDDLE:
                        swith := re.SearchReplaceWith('\2',false);
                    REMOVEEXPRESSION_PARTS_BUT_ENDING:
                        swith := re.SearchReplaceWith('\3',false);
                    end;

                    if cbreAndStore.Checked then begin
                        RE.SetSearchString(tempname);
                        RE.SetSearchPattern(
                            '(' + sStarting + ')' +
                            '(' + sMiddle + ')' +
                            '(' + sEnding + ')'
                        );
                        case cbxREParts.ItemIndex of
                        REMOVEEXPRESSION_PARTS_ALL: ;
                        REMOVEEXPRESSION_PARTS_STARTING:
                            tempname := re.SearchReplaceWith('\1',false);
                        REMOVEEXPRESSION_PARTS_MIDDLE:
                            tempname := re.SearchReplaceWith('\2',false);
                        REMOVEEXPRESSION_PARTS_ENDING:
                            tempname := re.SearchReplaceWith('\3',false);
                        REMOVEEXPRESSION_PARTS_BUT_STARTING:
                            tempname := re.SearchReplaceWith('\2\3',false);
                        REMOVEEXPRESSION_PARTS_BUT_MIDDLE:
                            tempname := re.SearchReplaceWith('\1\3',false);
                        REMOVEEXPRESSION_PARTS_BUT_ENDING:
                            tempname := re.SearchReplaceWith('\1\2',false);
                        end;
                    end;
                end;
                REMOVEEXPRESSION_REPLACE: begin
                    rec.selectedposition := rec.changedcount;
                    swith :=  ReplaceMetaVariables(FrmScriptActions.GetText(txtREWith),rec);
                end;
                end;

                sname := RE.SearchReplaceWith(sname, s2, swith, true);
                if (cbxRERemoveReplace.ItemIndex = REMOVEEXPRESSION_REMOVE) and cbreAndStore.Checked then begin
                     UnitMetaVariables.SetTemp(tempname);
                end;

             end;
        end;
    end;

    re.free;
end;
procedure TScriptStepRemoveWords.Execute(var sname, ext: string; var rec: RenameRecord);
var j : integer;
    s2, tempname : string;
    rwc : TWildcards;
begin
    rwc := self.GetWildCards;

    case cbxRemovePhrase.ItemIndex of
    REMOVEPHRASE_PHRASE: begin
        case cbxRemovePhrasePart.ItemIndex  of
        REMOVEPHRASEPART_NAME: begin
            sname := rwc.SearchReplaceWith(sname,
                ReplaceMetaVariables(FrmScriptActions.GetText(txtRemovePhrase), rec),
                '',
                true
            );
            rec.ScriptChanged := rwc.FirstFound;
        end;
        REMOVEPHRASEPART_EXT: begin
            ext := rwc.SearchReplaceWith(ext,
                ReplaceMetaVariables(FrmScriptActions.GetText(txtRemovePhrase), rec),
                '',
                true
            );
            rec.ScriptChanged := rwc.FirstFound;
        end;
        end;
    end;
    REMOVEPHRASE_WORDS: begin
        s2 := FrmScriptActions.GetText(txtRemovePhrase);
        tempname := TokenString(s2, ' ');
        while tempname <> '' do begin
            case cbxRemovePhrasePart.ItemIndex  of
            REMOVEPHRASEPART_NAME: begin
                sname := rwc.SearchReplaceWith(sname,
                    ReplaceMetaVariables(tempname, rec),
                    '',
                    true
                );
                rec.ScriptChanged := rwc.FirstFound;
            end;
            REMOVEPHRASEPART_EXT: begin
                ext := rwc.SearchReplaceWith(ext,
                    ReplaceMetaVariables(tempname, rec),
                    '',
                    true
                );
                rec.ScriptChanged := rwc.FirstFound;
            end;
            end;

            tempname := TokenString(s2, ' ');
        end;

    end;
    REMOVEPHRASE_NUMBERS: begin
        case cbxRemovePhrasePart.ItemIndex  of
        REMOVEPHRASEPART_NAME:
            tempname := sname;
        REMOVEPHRASEPART_EXT:
            tempname := ext;
        end;

        s2 := '';
        for j := 1 to length(tempname) do begin
            if not (tempname[j] in ['0'..'9']) then begin
                s2 := s2 + tempname[j];
            end;
        end;
        case cbxRemovePhrasePart.ItemIndex  of
        REMOVEPHRASEPART_NAME:
            sname := s2;
        REMOVEPHRASEPART_EXT:
            ext := s2;
        end;
    end;
    end;
    rwc.Free;
end;
procedure TScriptStepRemoveCharacters.Execute(var sname, ext: string;var rec: RenameRecord);
var tempname : string;
begin
    case cbxRemoveCharactersPart.ItemIndex  of
    REMOVECHARPART_NAME:
        tempname := sname;
    REMOVECHARPART_EXT: begin
        // get rid of period
        tempname := ext;
        tempname := rightstr(tempname, length(tempname) - 1);
    end;
    end;
    case cbxRemoveCharPosition.ItemIndex  of
    REMOVECHARPOS_BEGINNING:
        begin
            if cbAndStore.checked  then
                UnitMetaVariables.SetTemp(leftstr(tempname,SafeInt(txtR2x.text)));
            tempname := rightstr(tempname, length(tempname) - SafeInt(txtR2X.text));
            if cbRCReplace.Checked then begin
                rec.selectedposition := rec.changedcount;
                rec.ScriptChanged := true;
                tempname := ReplaceMetaVariables(FrmScriptActions.GetText(txtRCReplace), rec) + tempname;
            end;
        end;
    REMOVECHARPOS_ENDING:
        begin
            if cbAndStore.checked  then
                UnitMetaVariables.SetTemp(rightstr(tempname,SafeInt(txtR2x.text)));
            tempname := leftstr(tempname, length(tempname) - SafeInt(txtR2X.text));
            if cbRCReplace.Checked then begin
                rec.selectedposition := rec.changedcount;
                rec.ScriptChanged := true;
                tempname := tempname + ReplaceMetaVariables(FrmScriptActions.GetText(txtRCReplace), rec);
            end;
        end;
    end;
    case cbxRemoveCharactersPart.ItemIndex  of
    REMOVECHARPART_NAME:
        sname := tempname;
    REMOVECHARPART_EXT: begin
        ext := tempname;
        if pos('.',ext) <> 1 then ext := '.' + ext;
    end;
    end;
end;
procedure TScriptStepRemoveSubstring.Execute(var sname, ext: string; var rec: RenameRecord);
var tempname : string;
begin
    case cbxRemoveType.ItemIndex of
    REMOVETYPE_EVERYTHINGBUT:begin
        sname := copy(
            sname,
            SafeInt(txtR3X.text,1),
            SafeInt(txtR3Y.text,1) - SafeInt(txtR3X.text,1) + 1
        );
    end;
    REMOVETYPE_ONLY: begin
       tempname := copy(
            sname,
            SafeInt(txtR3X.text,1),
            SafeInt(txtR3Y.text,1) - SafeInt(txtR3X.text,1) + 1
       );
       delete(
            sname,
            SafeInt(txtR3X.text,1),
            SafeInt(txtR3Y.text,1) - SafeInt(txtR3X.text,1) + 1
        );
        if cbAndMove.checked then begin
            insert(tempname,sname, SafeInt(txtR3Z.text));
        end;
        if cbAndStoreRM.checked then begin
            UnitMetaVariables.SetTemp(tempname);
        end;
    end;
    end;
end;
procedure TScriptStepRemoveBeforeAfter.Execute(var sname, ext: string;  var rec: RenameRecord);
var tempname, pattern, pattern2 : string;
    startidx, endidx, startidx2, endidx2, len : cardinal;
    rwc, rwc2 : TWildcards;
begin
    rwc := self.GetWildCards;
    rwc2 := self.GetWildcards;

    tempname := sname;
    pattern := ReplaceMetaVariables(FrmScriptActions.GetText(txtRemove2Phrase), rec);
    rwc.SetSearchPattern(pattern);
    if cbxBeforeAfter.ItemIndex = REMOVEBEFOREAFTER_BETWEEN then begin
        pattern2 := ReplaceMetaVariables(FrmScriptActions.GetText(txtRemove2Phrase2), rec);
        rwc2.SetSearchPattern(pattern2);
    end;
    rwc.SetSearchString(sname);
    rwc2.SetSearchString(sname);

    if rwc.FindFirst(startidx, endidx) then begin
        case cbxBeforeAfter.ItemIndex of
            REMOVEBEFOREAFTER_BEFORE, REMOVEBEFOREAFTER_BEFORE_INC: begin
                if cbxBeforeAfter.ItemIndex = REMOVEBEFOREAFTER_BEFORE then begin
                    len := startidx-1;
                end else begin
                    len := endidx;
                end;
                tempname := Copy(sname, 1, len);
                Delete(sname, 1, len);
                if cbAndStoreRM2.checked then begin
                    UnitMetaVariables.SetTemp(tempname);
                end;
                if cbRBReplace.Checked then begin
                    rec.selectedposition := rec.changedcount;
                    rec.ScriptChanged := true;
                    sname := ReplaceMetaVariables(FrmScriptActions.GetText(txtRBReplace), rec) + sname;
                end;
            end;
            REMOVEBEFOREAFTER_AFTER, REMOVEBEFOREAFTER_AFTER_INC: begin
                if  cbxBeforeAfter.ItemIndex = REMOVEBEFOREAFTER_AFTER then begin
                    tempname := copy(sname, endidx+1, length(sname)-endidx);
                    Delete(sname, endidx+1, length(tempname));
                end else begin
                    tempname := copy(sname, startidx, length(sname)-startidx+1);
                    Delete(sname, startidx, length(tempname));
                end;
                if cbAndStoreRM2.checked then begin
                    UnitMetaVariables.SetTemp(tempname);
                end;
                if cbRBReplace.Checked then begin
                    rec.selectedposition := rec.changedcount;
                    rec.ScriptChanged := true;
                    sname := sname + ReplaceMetaVariables(FrmScriptActions.GetText(txtRBReplace), rec);
                end;
            end;
            REMOVEBEFOREAFTER_BETWEEN: begin
                tempname := midstr(sname,endidx+1, length(sname)-endidx);
                rwc2.SetSearchString( tempname );
                tempname := '';
                if rwc2.FindFirst(startidx2, endidx2) then begin
                    len := startidx2-1;
                    tempname := copy(sname, endidx+1, len);
                    Delete(sname, endidx+1, len);

                    if cbAndStoreRM2.checked then begin
                        UnitMetaVariables.SetTemp(tempname);
                    end;
                    if cbRBReplace.Checked then begin
                        Insert(
                            ReplaceMetaVariables(FrmScriptActions.GetText(txtRBReplace), rec),
                            sname,
                            endidx+1
                        );
                    end;
                end;
            end;
        end;
        if cbAndStoreRM2.checked then begin
            UnitMetaVariables.SetTemp(tempname);
        end;
    end;
    rwc.Free;
    rwc2.Free;
end;
procedure TScriptStepRemoveWordNumber.Execute(var sname, ext: string; var rec: RenameRecord);
begin

end;
procedure TScriptStepSplitAndMove.Execute(var sname, ext: string;  var rec: RenameRecord);
var mywc : TRegExpWildcards;
        spattern : string;
        sform : string;
    tempname : string;
    s : string;
begin
    mywc := TRegExpWildcards.create;

    if Enabled then begin
        if (txtFilemask.text <> '') and (txtMove.Text <> '') then begin
            spattern := ToRegHEX(FrmScriptActions.GetText(txtFilemask));

            s := ToRegHex('.*');
            if Pos(s, spattern) = length(spattern)-length(s)+1 then begin
                spattern := MidStr(spattern,1,length(spattern)-length(s)) + ToRegHex('.')+'([^.]*?)';
            end;
            spattern := StringReplace( spattern, ToRegHex('*'), '(.+?)', [rfreplaceall]);
            spattern := StringReplace( spattern, ToRegHex('[0-9]'), '([0-9]+)', [rfreplaceall]);
            spattern := StringReplace( spattern, ToRegHex('[a-z]'), '([a-zA-Z]+)', [rfreplaceall]);
            spattern := '^' + spattern + '$';
            sform := ReplaceMetaVariables(FrmScriptActions.GetText(txtMove),rec);

            tempname := mywc.SearchReplaceWith(sname+ext,spattern,sform,false);
            sname := leftstr(tempname, length(tempname) - length(ExtractFileExt(tempname)));
            ext := ExtractFileExt(tempname);
        end;
    end;

    mywc.free;
end;
procedure TScriptStepSelectAndMove.Execute(var sname, ext: string;  var rec: RenameRecord);
var s, s2 : string;
    tempname : string;
    startidx, endidx : cardinal;
    rwc, rwc2 : TWildcards;
    isRegExPattern : boolean;
    RegExPattern : string;
    RegExResult : string;
begin
    rec.selectedposition := rec.changedcount;
    tempname := sname;

    isRegExPattern := false;
    // remove
    if Self.rbSMBeginningEnding.Checked then begin
        case self.cbxSMBeginningEnding.ItemIndex of
        REMOVECHARPOS_BEGINNING:begin
            s := leftstr(tempname,SafeInt(self.txtSMZ1.text));
            tempname := rightstr(tempname, length(tempname) - SafeInt(txtSMZ1.text));
        end;
        REMOVECHARPOS_ENDING:begin
            s := rightstr(tempname,SafeInt(txtSMZ1.text));
            tempname := leftstr(tempname, length(tempname) - SafeInt(txtSMZ1.text));
        end;
        end;
    end else if self.rbSMBeforeAfter.Checked then begin
        rwc := self.GetWildCards;
        RegExPattern := ReplaceMetaVariables(FrmScriptActions.GetText(txtSMPhrase), rec);
        rwc.SetSearchPattern(RegExPattern);
        rwc.SetSearchString(sname);
        isRegExPattern := rwc.isRegExpPattern;
        if rwc.FindFirst(startidx, endidx) then begin
            RegExResult := rwc.GetStr(startidx, endidx);
            rwc.Free;
            tempname := sname;
            case cbxSMBeforeAfter.ItemIndex of
                REMOVEBEFOREAFTER_BEFORE: begin
                    s := Copy(tempname, 1, startidx-1);
                    Delete(tempname, 1, startidx-1);
                end;
                REMOVEBEFOREAFTER_AFTER: begin
                    s := copy(tempname, endidx+1, length(tempname)-endidx);
                    Delete(tempname, endidx+1, length(s));
                end;
                2:begin //FIND
                    rwc2 := self.GetWildCards;
                    rwc2.SetSearchPattern(ReplaceMetaVariables(FrmScriptActions.GetText(txtSMPhrase), rec));
                    rwc2.SetSearchString(sname);
                    if rwc2.FindFirst(startidx, endidx) then begin
                        s := Copy(tempname, startidx, endidx-startidx+1);
                        Delete(tempname, startidx, endidx-startidx+1);
                    end;
                    rwc2.Free;
                end;

            end;
        end;



        if tempname = sname then EXIT; // DO NOTHING

    end else if self.rbSMPosition.Checked then begin
       s := copy(
            tempname,
            SafeInt(txtSMX1.text,1),
            SafeInt(txtSMY1.text,1) - SafeInt(txtSMX1.text,1) + 1
       );
       delete(
            tempname,
            SafeInt(txtSMX1.text,1),
            SafeInt(txtSMY1.text,1) - SafeInt(txtSMX1.text,1) + 1
        );
    end;

    // paste
    if Self.rbSMBeginningEnding2.Checked then begin
        rec.ScriptChanged := true;
        case cbxSMBeginningEnding2.ItemIndex of
        ADDATTHE_BEGINNING:begin
            sname := s + tempname;
        end;
        ADDATTHE_ENDING:begin
            sname := tempname + s;
        end;
        end;
    end else if self.rbSMBeforeAfter2.Checked then begin

        rwc2 := self.GetWildCards;
        if IsRegExPattern then begin
            s2 := rwc2.SearchReplaceWith(
                RegExResult,
                RegExPattern,
                ReplaceMetaVariables(FrmScriptActions.GetText(txtSMPhrase2), rec),
                false
            );
        end else begin
            s2 := ReplaceMetaVariables(FrmScriptActions.GetText(txtSMPhrase2), rec);
        end;


        rwc2.SetSearchPattern(s2);
        rwc2.SetSearchString(tempname);
        if rwc2.FindFirst(startidx, endidx) then begin
            rec.ScriptChanged := true;
            case cbxSMBeforeAfter2.ItemIndex of
                REMOVEBEFOREAFTER_BEFORE: begin
                    Insert(s,tempname, startidx-1);
                end;
                REMOVEBEFOREAFTER_AFTER: begin
                    Insert(s,tempname, endidx+1);
                end;
            end;
        end else begin
            rwc2.free;
            EXIT; // DO NOTHING
        end;
        rwc2.Free;
        sname := tempname;
    end else if self.rbSMPosition2.Checked then begin
        Insert(s,tempname,SafeInt(self.txtSMX2.text,1));
        sname := tempname;
    end;


end;
procedure TScriptStepSwap.Execute(var sname, ext: string;  var rec: RenameRecord);
var
    tempname, front, back : string;
    rwc : TWildcards;
begin
    rec.selectedposition := rec.changedcount;
    tempname := sname;

    if self.ftxtSPhrase.Text = '' then EXIT;

    front := UnitToken.TokenString(tempname, FrmScriptActions.GetText(ftxtSPhrase), false);
    back := tempname;

    if back = '' then EXIT;

    if self.fcbSTrim.Checked then begin
        front := trim(front);
        back := trim(back);
    end;

    sname := back + self.ftxtSPhrase.Text + front;
end;
type TFoundData = class(TObject)
    public
        x, y : Cardinal;
        substitute : string;
end;
procedure TScriptStepFindAndEdit.Execute(var sname, ext: string;  var rec: RenameRecord);
var tempname, s, s2 : string;
    rwc, rwc2 : TWildcards;
    x,y,x2,y2,
    index, len : cardinal;

    l : TStack<TFoundData>;
    fd : TFoundData;
begin
    rwc := self.GetWildCards;
    rwc2 := self.GetWildCards;

    rec.selectedposition := rec.changedcount;

    rwc.SetSearchPattern( ReplaceMetaVariables(FrmScriptActions.GetText(ftxtFEPhrase1), rec) );
    rwc.SetSearchString( sname );
    if rwc.FindFirst(x,y) then begin
        l := TStack<TFoundData>.Create;
        repeat
            s := rwc.GetStr(x,y);

            // Replace checked - TODO update name to reflect this
            if self.cbFEFromThe.Checked then begin
                index := SafeInt(self.ftxtFEX1.text,1);
                len := SafeInt(self.ftxtFEY1.Text,1);
                if rwc.isRegExpPattern then begin
                   s2 := rwc2.SearchReplaceWith(
                        s,
                        ReplaceMetaVariables(FrmScriptActions.GetText(ftxtFEPhrase1), rec),
                        ReplaceMetaVariables(FrmScriptActions.GetText(txtFEPhrase2),rec),
                        true
                    );
                end else begin
                    s2 := ReplaceMetaVariables(FrmScriptActions.GetText(txtFEPhrase2),rec);
                end;

                if self.cbxFEBeginningEnding.ItemIndex = FINDEDIT_ENDING then begin
                    index := Length(s)+1 - (index-1);

                    Delete(s,index-len,len);
                end else begin
                    Delete(s,index,len);
                end;
                Insert(s2,s,index);
            end;

            // Insert Checked
            if self.cbFEInsert.Checked then begin
                if rwc.isRegExpPattern then begin
                    s2 := rwc2.SearchReplaceWith(
                        s,
                        ReplaceMetaVariables(FrmScriptActions.GetText(ftxtFEPhrase1), rec),
                        ReplaceMetaVariables(FrmScriptActions.GetText(txtFEPhrase4),rec),
                        true
                    );
                    rwc2.SetSearchPattern(s2);
                end else begin
                    rwc2.SetSearchPattern(ReplaceMetaVariables(FrmScriptActions.GetText(txtFEPhrase4),rec));
                end;

                rwc2.SetSearchString(s);
                if (rwc2.FindFirst(x2,y2)) then begin
                    s2 := rwc2.GetStr(x2,y2);
                    index := x2;
                    case self.cbxFEBeforeAFter.ItemIndex of
                    FINDEDIT_BEFORE:
                        begin

                        end;
                    FINDEDIT_AFTER:
                        begin
                            Inc(index, length(s2));
                        end;
                    FINDEDIT_REPLACING:
                        begin
                            Delete(s,index,length(s2));
                        end;
                    end;
                    case self.cbxFEBeforeAFter.ItemIndex of
                    FINDEDIT_BEFORE,
                    FINDEDIT_AFTER,
                    FINDEDIT_REPLACING:
                        begin
                            Insert(ReplaceMetaVariables(FrmScriptActions.GetText(txtFEPhrase3),rec),s,index);
                        end;
                    end;

                end;
            end;
            fd := TFoundData.Create;
            fd.x := x;
            fd.y := y;
            fd.substitute := s;
            l.Push(fd);
        until not rwc.FindNext(x,y);

        tempname := sname;
        while (l.Count <> 0) do begin
            fd := l.Pop;
            Delete(tempname,fd.x, fd.y-fd.x+1);
            Insert(fd.substitute,tempname,fd.x);
        end;
        sname := tempname;

        l.free;
    end;
    rwc.Free;
    rwc2.Free;

end;



procedure TScriptStepSplitAndMove.Reset;
begin
    self.ftxtFilemask.Text := FrmScriptActions.txtFilemask.text;
    self.ftxtMove.Text := FrmScriptActions.txtMove.text;
end;
procedure TScriptStepAddCounter.Reset;
begin
    self.fAUDZeros.Position := FrmScriptActions.AUDZeros.Position;
    self.ftxtAZeros.Text := FrmScriptActions.txtAZeros.text;
    self.fcbxBeginningEnding.ItemIndex := FrmScriptActions.cbxBeginningEnding.ItemIndex;
    self.fcbxBeginningEnding.OnChange(self.fcbxBeginningEnding);
end;
procedure TScriptStepAddBeginningEnding.Reset;
begin
    self.ftxtAdd.Text := '';
    self.fcbxAddAtThe.ItemIndex := FrmScriptActions.cbxAddAtThe.ItemIndex;
    self.fcbOnCollision.Checked := FrmScriptActions.cbOnCollision.Checked;
end;
procedure TScriptStepAddPosition.Reset;
begin
	self.ftxtAddPosition.Text := FrmScriptActions.txtAddPosition.text;
    self.ftbPosition.Position := FrmScriptActions.tbPosition.Position;
    self.fcbxFromTheBeginningEnding.ItemIndex := FrmScriptActions.cbxFromTheBeginningEnding.ItemIndex;
end;
procedure TScriptStepAddBeforeAfter.Reset;
begin
	self.ftxtAfter.text := FrmScriptActions.txtAfter.text;
    self.fcbxAddPlace.ItemIndex := FrmScriptActions.cbxAddPlace.ItemIndex;
end;
procedure TScriptStepReplaceNameExtension.Reset;
begin
	self.fcbxReplacePart.ItemIndex := FrmScriptActions.cbxReplacePart.ItemIndex;
    self.fcbxReplaceTarget.ItemIndex := FrmScriptActions.cbxReplaceTarget.ItemIndex;
    self.ftxtReplacePhrase.Text := FrmScriptActions.txtReplacePhrase.Text;
    self.ftxtReplaceWith.text := FrmScriptActions.txtReplaceWith.text;
    self.ftxtRP2X.text := FrmScriptActions.txtRP2X.text;
    self.ftxtRP2Y.text := FrmScriptActions.txtRP2Y.text;
    self.fudRP2X.Position := FrmScriptActions.udRP2X.Position;
    self.fudRP2Y.Position := FrmScriptActions.udRP2Y.Position;
    self.fcbReplaceOnly.Checked := FrmScriptActions.cbReplaceOnly.Checked;

    self.fcbxReplacePart.OnChange(self.fcbxReplacePart);
end;
procedure TScriptStepReplaceCharacters.Reset;
begin
	self.ftxtReplaceChar1.text := FrmScriptActions.txtReplaceChar1.text;
    self.ftxtReplaceChar2.text := FrmScriptActions.txtReplaceChar2.text;
    self.fcbxReplaceCharPart.ItemIndex := FrmScriptActions.cbxReplaceCharPart.ItemIndex;
end;
procedure TScriptStepReplaceNumbers.Reset;
begin
	self.fcbxNumbers.ItemIndex := FrmScriptActions.cbxNumbers.ItemIndex;
    self.fcbxRNSequence.ItemIndex := FrmScriptActions.cbxRNSequence.ItemIndex;
    self.fcbxRNSequence.OnChange(self.fcbxRNSequence);
    self.ftxtRNWith.Text := FrmScriptActions.txtRNWith.Text;
end;
procedure TScriptStepConditionExecute.Reset;
begin
	self.ftxtCE1.Text := FrmScriptActions.txtCE1.Text;
    self.fudCE1.Position := FrmScriptActions.udCE1.Position;
    self.fcbxCECondition.ItemIndex := FrmScriptActions.cbxCECondition.ItemIndex;
    self.ftxtCEText.Text := FrmScriptActions.txtCEText.text;
end;
procedure TScriptStepRemoveExpression.Reset;
begin
    self.fcbxRERemoveReplace.ItemIndex := FrmScriptActions.cbxRERemoveReplace.ItemIndex;
	self.fcbreAtThe.Checked := FrmScriptActions.cbreAtThe.Checked;
    self.fcbxreBegginingEnding.ItemIndex := FrmScriptActions.cbxreBegginingEnding.ItemIndex;
    self.fcbreThatSTartsWIth.Checked := FrmScriptActions.cbreThatSTartsWIth.Checked;
    self.fcbxrePhraseCharacters.ItemIndex := FrmScriptActions.cbxrePhraseCharacters.ItemIndex;
    self.ftxtrePhraseCharacters.Text := FrmScriptActions.txtrePhraseCharacters.Text;
    self.fcbxreHasInTheMiddle.Checked := FrmScriptActions.cbreHasInTheMiddle.Checked;
    self.fcbxrePhraseCharacters2.ItemIndex := FrmScriptActions.cbxrePhraseCharacters2.ItemIndex;
    self.ftxtrePhraseCharacters2.Text := FrmScriptActions.txtrePhraseCharacters2.Text;
    self.fcbreAndEndsWith.Checked := FrmScriptActions.cbreAndEndsWith.Checked;
    self.fcbxrePhraseCharacters3.ItemIndex := FrmScriptActions.cbxrePhraseCharacters3.ItemIndex;
    self.ftxtrePhraseCharacters3.Text := FrmScriptActions.txtrePhraseCharacters3.Text;
    self.flblReplaceExpressionWith.Visible := FrmScriptActions.lblReplaceExpressionWith.Visible;
    self.ftxtREWith.visible := FrmScriptActions.txtREWith.visible;
    self.ftxtREWith.Parent.visible := FrmScriptActions.txtREWith.visible;

    self.fcbreAndStore.Checked := FrmScriptActions.cbreAndStore.Checked;
    self.fcbreRemoveNumber.Checked := FrmScriptActions.cbreRemoveNumber.Checked;
    self.ftxtRE1.Text := FrmScriptActions.txtRE1.Text;
    self.fudRE1.Position := FrmScriptActions.udRE1.Position;
end;
procedure TScriptStepRemoveWords.Reset;
begin
	self.ftxtRemovePhrase.Text := FrmScriptActions.txtRemovePhrase.Text;
    self.fcbxRemovePhrase.ItemIndex := FrmScriptActions.cbxRemovePhrase.itemindex;
    self.fcbxRemovePhrasePart.ItemIndex := FrmScriptActions.cbxRemovePhrasePart.ItemIndex;
    self.fcbxRemovePhrase.OnChange(self.fcbxRemovePhrase);
end;
procedure TScriptStepRemoveCharacters.Reset;
begin
	self.ftxtR2X.Text := FrmScriptActions.txtR2X.Text;
    self.fudR2X.Position := FrmScriptActions.udr2x.Position;
    self.fcbxRemoveCharPosition.ItemIndex := FrmScriptActions.cbxRemoveCharPosition.ItemIndex;
    self.fcbxRemoveCharactersPart.ItemIndex := FrmScriptActions.cbxRemoveCharactersPart.ItemIndex;
    self.fcbAndSTore.Checked := FrmScriptActions.cbAndStore.Checked;
end;
procedure TScriptStepRemoveSubstring.Reset;
begin
    self.fcbxRemoveType.ItemIndex := FrmScriptActions.cbxRemoveType.ItemIndex;
    self.ftxtR3X.Text := FrmScriptActions.txtR3X.Text;
	self.ftxtR3Y.Text := FrmScriptActions.txtR3Y.Text;
	self.ftxtR3Z.Text := FrmScriptActions.txtR3Z.Text;
    self.fudR3X.Position := FrmScriptActions.udR3X.Position;
	self.fudR3Y.Position := FrmScriptActions.udR3Y.Position;
	self.fudR3Z.Position := FrmScriptActions.udR3Z.Position;
    self.fcbAndMove.Checked := FrmScriptActions.cbAndMove.Checked;
    self.fcbAndStoreRM.Checked := FrmScriptActions.cbAndStoreRM.Checked;

    self.fcbxRemoveType.OnChange(self.fcbxRemoveType);
end;
procedure TScriptStepRemoveBeforeAfter.Reset;
begin
	self.fcbxBeforeAfter.ItemIndex := FrmScriptActions.cbxBeforeAfter.ItemIndex;
    self.ftxtRemove2Phrase.Text := FrmScriptActions.txtRemove2Phrase.Text;
    self.fcbAndStoreRM2.Checked := FrmScriptActions.cbAndStoreRM2.Checked;
end;
procedure TScriptStepRemoveWordNumber.Reset;
begin
	self.ftxtRW1.Text := FrmScriptActions.txtrw1.Text;
    self.fudRW1.Position := FrmScriptActions.udrw1.Position;
    self.fcbxrwCharactersType.ItemIndex := FrmScriptActions.cbxrwCharactersType.ItemIndex;
    self.ftxtrwCharacters.Text := FrmScriptActions.txtrwCharacters.Text;
    self.fcbrwStoreTemp.Checked := FrmScriptActions.cbrwStoreTemp.Checked;
end;
procedure TScriptSTepSelectAndMove.reset;
begin

end;
procedure TScriptSTepSwap.reset;
begin
    self.ftxtSPhrase.Text := FrmScriptActions.txtSPhrase.text;
    self.fcbSTrim.Checked := FrmScriptActions.cbSTrim.Checked;
end;
procedure TScriptStepFindAndEdit.reset;
begin
    self.ftxtFEPhrase1.Text := FrmScriptActions.txtFEPhrase1.Text;
    self.fcbFEFromThe.Checked := FrmScriptActions.cbFEFromThe.Checked;
    self.fcbxFEBeginningEnding.ItemIndex := FrmScriptActions.cbxFEBeginningEnding.ItemIndex;
    self.ftxtFEX1.Text := FrmScriptActions.txtFEX1.Text;
    self.fudFEX1.Position := FrmScriptActions.udFEX1.position;
    self.ftxtFEY1.Text := FrmScriptActions.txtFEY1.Text;
    self.fudFEY1.Position := FrmScriptActions.udFEY1.Position;
    Self.ftxtFEPhrase2.Text := FrmScriptActions.txtFEPhrase2.Text;
    self.fcbFEInsert.Checked := FrmScriptActions.cbFEInsert.Checked;
    self.ftxtFEPhrase3.Text := FrmScriptActions.txtFEPhrase3.Text;
    self.fcbxFEBeforeAFter.ItemIndex := FrmScriptActions.cbxFEBeforeAFter.ItemIndex;
    self.ftxtFEPhrase4.Text := FrmScriptActions.txtFEPhrase4.Text;
end;

function TScriptStepAddBeginningEnding.GetShortSentence : string;
begin
    result := 'Add to Beginning/Ending'
end;
function TScriptStepAddCounter.GetShortSentence: string;
begin
    result := 'Add Counter';
end;
function TScriptStepAddPosition.GetShortSentence : string;
begin
    result := 'Add by Position';
end;
function TScriptStepAddBeforeAfter.GetShortSentence : string;
begin
    result := 'Add Before/After';
end;
function TScriptStepReplaceNameExtension.GetShortSentence : string;
begin
    result := 'Replace/Find and Replace';
end;
function TScriptStepReplaceCharacters.GetShortSentence : string;
begin

end;
function TScriptStepReplaceNumbers.GetShortSentence : string;
begin
    result := 'Replace Numbers';
end;
function TScriptStepConditionExecute.GetShortSentence : string;
begin
    result := 'Skip/Execute Steps';
end;
function TScriptStepRemoveWords.GetShortSentence: string;
begin
    result := 'Remove Words/Numbers';
end;
function TScriptStepRemoveSubstring.GetShortSentence : string;
begin
    result := 'Remove by Position';
end;
function TScriptStepRemoveExpression.GetShortSentence : string;
begin
    result := 'Remove/Replace Expression';
end;
function TScriptStepRemoveCharacters.GetShortSentence : string;
begin
    result := 'Remove Beginning/Ending';
end;
function TScriptStepRemoveBeforeAfter.GetShortSentence : string;
begin
    result := 'Remove/Replace Before/After/Between';
end;
function TScriptStepRemoveWordNumber.GetShortSentence : string;
begin
    result := 'Remove Words/Numbers';
end;
function TScriptStepSplitAndMove.GetShortSentence : string;
begin
    result := 'Split';
end;
function TScriptSTepSelectAndMove.GetShortSentence : string;
begin
    result := 'Select and Move';
end;
function TScriptStepSwap.GetShortSentence : string;
begin
    result := 'Swap';
end;
function TScriptStepFindAndEdit.GetShortSentence : string;
begin
    result := 'Find and Edit';
end;

procedure TScriptStep.AppendSentence(ssc : TScriptStepControl; wc : TControl);
    function GetText(wc : TControl) : string;
    begin
        if wc is TLabel then
            result := TLabel(wc).caption
        else if wc is TCheckbox then
            result := TCheckbox(wc).Caption
        else if wc is TMyRichEdit then begin
            result := TMyRichEdit(wc).DisplayedText;
            result := StringReplace(result,TMyRichEdit(wc).VarDelimiter,'"',[rfReplaceAll,rfIgnoreCase]);
        end else if wc is TcustomEdit then
            result := '''' + TcustomEdit(wc).text + ''''
        else if wc is TComboBox  then
            result := TComboBox(wc).text
        else if wc is TRadioButton then begin
            if TRadioButton(wc).checked then begin
                result := TRadioButton(wc).Caption;
            end;
        end;
    end;

var lbl : TLabel;
    t : string;
const
    FONT_SIZE = 7;
    FONT_HEIGHT = 15;
begin
    lbl := TLabel.Create(ssc.pnlMid);
    frmConfig.GetScriptFont(lbl.font);
//    lbl.Font.Height := FONT_HEIGHT;
//    lbl.Font.Name := 'Times New Roman';
    lbl.Parent := TWinControl(ssc.pnlMid);
    lbl.AlignWithMargins := true;
    lbl.Margins.Left := 0;
    lbl.Margins.Right := 3;
//    lbl.Margins.Bottom := 2; // TODO : test a way to center text in title bar
//    lbl.Margins.Top := 0;


    if ssc.ScriptStep.Enabled then begin
        lbl.Font.Color := clWindowText;
    end else begin
        lbl.Font.Color := DimColor(clGrayText, GREY_DIM);
    end;
    lbl.Transparent := true;


    t := GetText(wc);

    lbl.Caption := t;
    if wc is TcustomEdit then begin
        lbl.Font.Style := lbl.font.style + [fsBold]
    end else if wc is TComboBox then begin
        lbl.Font.Style := lbl.font.style + [fsUnderline];
    end;
    lbl.Tag := ssc.Tag;
    lbl.OnClick := FrmScriptActions.LabelClickEvent;

    FrmScriptActions.DragDropHook(lbl);

    lbl.EllipsisPosition := epEndEllipsis;
    lbl.Constraints.MaxWidth := 140;

    //result := result + t + ' ';
end;

// TScriptStepControl
function TScriptStepControl.GetHint : string;
var i : integer;
begin
    result := '';
    for i := 0 to self.pnlMid.ControlCount-1 do begin
        result := result +  TLabel(self.pnlMid.Controls[i]).Caption + ' ';
    end;
end;
function TScriptStepControl.getIsExpanded : boolean;
begin
    result := self.pnlBtm.Visible;
end;
function TScriptStepControl.getHeight : integer;
begin
    if getisExpanded then
        result := outerpnl.Height
    else
        result := pnlTop.Height;
end;

procedure TScriptStepAddCounter.GetSentence(ssc : TScriptStepControl);
var ssac : TScriptStepAddCounter;
begin
    ssac := TScriptStepAddCounter(ssc.ScriptStep);
    AppendSentence(ssc, ssac.lblAddCounter);
    AppendSentence(ssc, ssac.txtAZeros);
    AppendSentence(ssc, ssac.lblZeros);
    AppendSentence(ssc, ssac.cbxBeginningEnding);
    AppendSentence(ssc, ssac.lblAOfTheName);
end;
procedure TScriptStepAddBeginningEnding.GetSentence(ssc : TScriptStepControl);
var
    ssbe : TScriptStepAddBeginningEnding;
    ssa : TScriptStepAdd;
begin
    ssa := TScriptStepAdd(ssc.ScriptStep);
    AppendSentence(ssc, ssa.lblAdd);
    AppendSentence(ssc, ssa.txtAdd);
    AppendSentence(ssc, ssa.lblToThe);
    AppendSentence(ssc, ssa.cbxAddPart);

    ssbe := TScriptStepAddBeginningEnding(ssc.scriptstep);
    //AppendSentence(ssc, ssbe.lblAtThe);
    AppendSentence(ssc, ssbe.cbxAddAtThe);
    if ssbe.cbOnCollision.checked then begin
        AppendSentence(ssc, ssbe.cbOnCollision);
    end;
end;
procedure TScriptStepAddPosition.GetSentence(ssc : TScriptStepControl);
var ssa : TScriptStepAdd;
    ssp : TScriptStepAddPosition;
begin
    ssa := TScriptStepAdd(ssc.scriptstep);
    AppendSentence(ssc, ssa.lblAdd);
    AppendSentence(ssc, ssa.txtAdd);
    AppendSentence(ssc, ssa.lblToThe);
    AppendSentence(ssc, ssa.cbxAddPart);

    ssp := TScriptStepAddPosition(ssc.scriptstep);
    AppendSentence(ssc, ssp.lblAtCharacterPosition);
    AppendSentence(ssc, ssp.txtAddPosition);
end;
procedure TScriptStepAddBeforeAfter.GetSentence(ssc : TScriptStepControl);
var
    ssa : TScriptStepAdd;
    ssba : TScriptStepAddBeforeAfter;
begin
    ssa := TScriptStepAdd(ssc.scriptstep);
    AppendSentence(ssc, ssa.lblAdd);
    AppendSentence(ssc, ssa.txtAdd);
    AppendSentence(ssc, ssa.lblToThe);
    AppendSentence(ssc, ssa.cbxAddPart);



    ssba := TScriptStepAddBeforeAfter(ssc.scriptstep);
    AppendSentence(ssc, ssba.lblPlaceIt);
    AppendSentence(ssc, ssba.cbxAddPlace);
    AppendSentence(ssc, ssba.txtAfter);
end;
procedure TScriptStepReplaceNameExtension.GetSentence(ssc : TScriptStepControl);
var ssr : TScriptStepReplace;
begin
    ssr := TScriptStepReplace(ssc.ScriptStep);
    with ssr as TScriptStepReplaceNameExtension do begin
        //Append(ssne.lblReplaceThe);
        AppendSentence(ssc, cbxReplacePart);
        // Phrase
        if cbxReplacePart.ItemIndex = 2 then begin
            AppendSentence(ssc, txtReplacePhrase);
        end;
        AppendSentence(ssc, lblWith);
        AppendSentence(ssc, txtReplaceWith);
        if cbReplaceOnly.checked  then begin
            AppendSentence(ssc, cbReplaceOnly);
            AppendSentence(ssc, txtRP2X);
            AppendSentence(ssc, lblThru);
            AppendSentence(ssc, txtRP2Y);
        end;
    end;
end;
procedure TScriptStepReplaceCharacters.GetSentence(ssc : TScriptStepControl);
begin

end;
procedure TScriptStepReplaceNumbers.GetSentence(ssc : TScriptStepControl);
var
    ssr : TScriptStepReplace;
begin
    ssr := TScriptStepReplace(ssc.ScriptStep);

    with ssr as TScriptStepReplaceNumbers do begin
        AppendSentence(ssc, lblReplaceNumbers);
        AppendSentence(ssc, cbxNumbers);
        AppendSentence(ssc, lblOfTheName);
        AppendSentence(ssc, cbxRNSequence);
        if cbxRNSequence.ItemIndex = REPLACENUMBER_PHRASE then begin
            AppendSentence(ssc, txtRNWith);
        end;
    end;
end;
procedure TScriptStepConditionExecute.GetSentence(ssc : TScriptStepControl);
var sce : TScriptStepConditionExecute;
begin
    sce := TScriptStepConditionExecute(ssc.ScriptStep);
    with sce do begin
        AppendSentence(ssc, cbxCESkipRemove);
        AppendSentence(ssc, lblCESkipTheNext);
        AppendSentence(ssc, txtCE1);
        AppendSentence(ssc, lblCESteps);
        AppendSentence(ssc, lblCEIf);
        AppendSentence(ssc, cbxCECondition);
        if cbxCECondition.ItemIndex <> 0 then
            AppendSentence(ssc, txtCEText);
    end;
end;
procedure TScriptStepRemoveWords.GetSentence(ssc : TScriptStepControl);
var
    ssrm : TScriptStepRemove;
begin
    ssrm := TScriptStepRemove(ssc.ScriptStep);
    with ssrm as TScriptStepRemoveWords do begin
        AppendSentence(ssc, lblRemoveWords);
        AppendSentence(ssc, cbxRemovePhrase);
        if cbxRemovePhrase.ItemIndex <> 2 then begin
            AppendSentence(ssc, txtRemovePhrase);
        end;
        AppendSentence(ssc, lblFromThe);
        AppendSentence(ssc, cbxRemovePhrasePart);
    end;
end;
procedure TScriptStepRemoveSubstring.GetSentence(ssc : TScriptStepControl);
var
    ssrm : TScriptStepRemove;
begin
    ssrm := TScriptStepRemove(ssc.ScriptStep);

    with ssrm as TScriptStepRemoveSubstring do begin
        AppendSentence(ssc, lblRemoveSubstring);
        AppendSentence(ssc, cbxRemoveType);
        AppendSentence(ssc, lblChars);
        AppendSentence(ssc, txtR3X);
        AppendSentence(ssc, lblCharThru);
        AppendSentence(ssc, txtR3Y);
        if cbAndMove.checked then begin
            AppendSentence(ssc, cbAndMove);
            AppendSentence(ssc, txtR3Z);
        end;
        if cbAndStoreRM.checked  then begin
            AppendSentence(ssc, cbAndStoreRM);
        end;
    end;
end;


procedure TScriptStepRemoveExpression.GetSentence(ssc : TScriptStepControl);
var
    ssrm : TScriptStepRemove;
    l : TLabel;
    procedure Append(s : string);
    begin
            l := TLabel.Create(ssc.pnlMid);
            l.Caption := s;
            AppendSentence(ssc, l);
            l.Free;
    end;
    procedure AddExpCombo(cbx : TComboBox);
    begin
            case cbx.ItemIndex of
            REMOVEEXPRESSION_PHRASE:;
            REMOVEEXPRESSION_CHARACTERS:
                Append('letters');
            REMOVEEXPRESSION_CHARACTERS_EXCEPT:
                Append('not letters');
            else
                AppendSentence(ssc, fcbxrePhraseCharacters);
            end;
    end;
begin
    ssrm := TScriptStepRemove(ssc.ScriptStep);

    with ssrm as  TScriptStepRemoveExpression do begin
        AppendSentence(ssc, fcbxRERemoveReplace);
        AppendSentence(ssc, flblreRemoveTheExpression);

        if cbxreBegginingEnding.ItemIndex <> REMOVEEXPRESSION_ANYWHERE then  begin
            AppendSentence(ssc, cbxreBegginingEnding);
            AppendSentence(ssc, flblreOfTheName);
        end else begin
            //Append('that');
        end;

        if fcbreThatSTartsWIth.checked then begin
            //AppendSentence(ssc, fcbreThatSTartsWIth );
            Append('[');
            AddExpCombo(fcbxrePhraseCharacters);
            AppendSentence(ssc, ftxtrePhraseCharacters);
            Append(']');
        end;
        if fcbxreHasInTheMiddle.checked then begin
            //AppendSentence(ssc, fcbxreHasInTheMiddle);
            Append('[');
            AddExpCombo(fcbxrePhraseCharacters2);
            AppendSentence(ssc, ftxtrePhraseCharacters2);
            Append(']');
        end;
        if fcbreAndEndsWith.checked then begin
            //AppendSentence(ssc, fcbreAndEndsWith);
            Append('[');
            AddExpCombo(fcbxrePhraseCharacters3);
            AppendSentence(ssc, ftxtrePhraseCharacters3);
            Append(']');
        end;

        case fcbxRERemoveReplace.ItemIndex of
        REMOVEEXPRESSION_REMOVE: begin
            if fcbxREParts.ItemIndex <> REMOVEEXPRESSION_PARTS_ALL then begin
                AppendSentence(ssc, flblRERemoveThePart);
                AppendSentence(ssc, fcbxREParts);
            end;

            if (fcbreAndStore.checked) then begin
                AppendSentence(ssc, fcbreAndStore);
            end;
        end;
        REMOVEEXPRESSION_REPLACE: begin
            AppendSentence(ssc, flblReplaceExpressionWith);
            AppendSentence(ssc, ftxtREWith);

        end;
        end;
    end;
end;
procedure TScriptStepRemoveCharacters.GetSentence(ssc : TScriptStepControl);
var
    ssrm : TScriptStepRemove;
begin
    ssrm := TScriptStepRemove(ssc.ScriptStep);
    with ssrm as TScriptStepRemoveCharacters do begin
        AppendSentence(ssc, lblRemoveChars);
        AppendSentence(ssc, txtR2X);
        AppendSentence(ssc, lblCharsFromThe);
       // AppendSentence(ssc, lblOfThe);
        AppendSentence(ssc, cbxRemoveCharactersPart);
        AppendSentence(ssc, cbxRemoveCharPosition);
        if cbAndStore.checked  then begin
            AppendSentence(ssc, cbAndStore);
        end;
        if cbRCReplace.Checked then begin
            AppendSentence(ssc, cbRCReplace);
            AppendSentence(ssc, txtRCReplace);
        end;
    end;
end;
procedure TScriptStepRemoveBeforeAfter.GetSentence(ssc : TScriptStepControl);
var
    ssrm : TScriptStepRemove;
begin
    ssrm := TScriptStepRemove(ssc.ScriptStep);
    with (ssrm as TScriptStepRemoveBeforeAfter) do begin
        AppendSentence(ssc, lblRemoveEverything);
        AppendSentence(ssc, cbxBeforeAfter);
        AppendSentence(ssc, lblThePhrase);
        AppendSentence(ssc, txtRemove2Phrase);

        if cbxBeforeAfter.ItemIndex = REMOVEBEFOREAFTER_BETWEEN then begin
            AppendSentence(ssc, lblAndThePhrase);
            AppendSentence(ssc, txtRemove2Phrase2);

        end;


        if cbAndStoreRM2.checked  then begin
            AppendSentence(ssc, cbAndStoreRM2);
        end;

        if cbRBReplace.Checked then begin
            AppendSentence(ssc, cbRBReplace);
            AppendSentence(ssc, txtRBReplace);
        end;
     end;

end;
procedure TScriptStepRemoveWordNumber.GetSentence(ssc : TScriptStepControl);
begin

end;
procedure TScriptStepSplitAndMove.GetSentence(ssc : TScriptStepControl);
var ss : TScriptStepSplitAndMove;
    procedure Append(s : string);
    var l : TLabel;
    begin
            l := TLabel.Create(ssc.pnlMid);
            l.Caption := s;
            AppendSentence(ssc, l);
            l.Free;
    end;
begin
    ss := TScriptStepSplitAndMove(ssc.ScriptStep);
//    AppendSentence(ssc, ss.lblSplitTheName);
    Append('Split filemask');
    AppendSentence(ssc, ss.txtFilemask);
//    AppendSentence(ssc, ss.lblMoveTheParts);
    Append('Combine as');
    AppendSentence(ssc, ss.txtMove);
end;
procedure TScriptStepSelectAndMove.GetSentence(ssc : TScriptStepControl);
begin
    AppendSentence(ssc, lblSMSelect);
    if Self.rbSMBeginningEnding.Checked then begin
        AppendSentence(ssc, cbxSMBeginningEnding);
        AppendSentence(ssc, lblSMCharacters);
        AppendSentence(ssc, txtSMZ1);
    end else if self.rbSMBeforeAfter.Checked then begin
        AppendSentence(ssc, self.cbxSMBeforeAfter);
        AppendSentence(ssc, Self.txtSMPhrase);
    end else if self.rbSMPosition.Checked then begin
        AppendSentence(ssc, rbSMPosition);
        AppendSentence(ssc, txtSMX1);
        AppendSentence(ssc, lblSMThru);
        AppendSentence(ssc, txtSMY1);
    end;

    AppendSentence(ssc, lblSMMoveTo);
    if Self.rbSMBeginningEnding2.Checked then begin
        AppendSentence(ssc, cbxSMBeginningEnding2);
    end else if self.rbSMBeforeAfter2.Checked then begin
        AppendSentence(ssc, self.cbxSMBeforeAfter2);
        AppendSentence(ssc, Self.txtSMPhrase2);
    end else if self.rbSMPosition2.Checked then begin
        AppendSentence(ssc, rbSMPosition2);
        AppendSentence(ssc, txtSMX2);
    end;

end;
procedure TScriptStepSwap.GetSentence(ssc : TScriptStepControl);
    procedure Append(s : string);
    var l : TLabel;
    begin
            l := TLabel.Create(ssc.pnlMid);
            l.Caption := s;
            AppendSentence(ssc, l);
            l.Free;
    end;
begin
    Append('Swap Around');
    AppendSentence(ssc, self.ftxtSPhrase);
    AppendSentence(ssc, self.fcbSTrim);
end;
procedure TScriptStepFindAndEdit.GetSentence(ssc : TScriptStepControl);
    procedure Append(s : string);
    var l : TLabel;
    begin
            l := TLabel.Create(ssc.pnlMid);
            l.Caption := s;
            AppendSentence(ssc, l);
            l.Free;
    end;
begin
    Append('Find');
    AppendSentence(ssc, self.ftxtFEPhrase1);
    if self.cbFEFromThe.Checked then begin
        AppendSentence(ssc, self.cbFEFromThe);
        AppendSentence(ssc,Self.fcbxFEBeginningEnding);
        AppendSentence(ssc,self.flblFEReplacePosition);
        AppendSentence(ssc,self.ftxtFEX1);
        AppendSentence(ssc,Self.flblFELength);
        AppendSentence(ssc,self.ftxtFEY1);
        AppendSentence(ssc,Self.flblFEWith);
        AppendSentence(ssc,self.ftxtFEPhrase2);
    end;
    if self.cbFEInsert.Checked then begin
        AppendSentence(ssc, self.cbFEInsert);
        AppendSentence(ssc,self.ftxtFEPhrase3);
        AppendSentence(ssc,self.fcbxFEBeforeAFter);
        AppendSentence(ssc,self.ftxtFEPhrase4);
    end;
end;
{$endregion}
//

initialization
begin
    screen.Cursors[crHandPoint] := LoadCursor(0, IDC_HAND);

    addColor := RGB(159,183,164);
    RemoveColor := RGB(142,132,149);
    ReplaceColor := RGB(147,167,185);
    OtherColor := RGB(133,133,133);
end;
end.
